/*
** usage:
**    def2mod [options] files
**
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>
#include <unistd.h>
#include <glib.h>
#ifndef __WIN32__
#define __USE_GNU
#include <dlfcn.h>
#endif
#include <gtk/gtk.h>
#ifdef NEED_GNOMESUPPORT_H
#include <gnome.h>
#include <popt.h>
#endif
#include <gtk_drv_api.h>

#define MAX_LINE_BUF        1024
#define MAX_PARSE_ARGS      32
#define MAX_PARSE_ENUMS     64
#define MAX_ARRAY_DECL      6

#define ANY_LENGTH          0   /* array length declaration [] or * */
#define NO_LENGTH          -1   /* pointer * declaration */

#define MAX_TOKEN_LOOKAHEAD 100

#ifdef __WIN32__
#define strncasecmp(a, b, c) strnicmp((a), (b), (c))
#endif

/* #define DEBUG */

char* opt_module   = NULL;
char* opt_outdir   = NULL;
char* opt_c_outdir = NULL;
char* opt_e_outdir = NULL;
char* opt_h_outdir = NULL;
char* opt_lib      = NULL;
int opt_preload;
int opt_verbose;

void* dl_lib;

#ifdef NEED_GNOMESUPPORT_H
struct poptOption pop_options[] = {
    { "module",      'M', POPT_ARG_STRING, &opt_module, 0, NULL, 
	  "Override the default module name" },
    { "lib",         'L', POPT_ARG_STRING, &opt_lib,    0, NULL, 
	  "The GTK library shared object file" },
    { "outdir",      'o', POPT_ARG_STRING, &opt_outdir, 0, NULL,
	  "Output directory (default .)" },
    { "c_outdir",    'C', POPT_ARG_STRING, &opt_c_outdir, 0,  NULL,
	  "Overide -o for  lib<mod>.c files" },
    { "e_outdir",    'E', POPT_ARG_STRING, &opt_e_outdir, 0,  NULL, 
	  "Override -o for <mod>.inc files" },
    { "h_outdir",    'H', POPT_ARG_STRING, &opt_h_outdir, 0, NULL, 
	  "Override -o for lib<mod>.hrl files" },
    { "verbose",     'v', POPT_ARG_NONE,   &opt_verbose,  0, NULL, 
	  "Emit verbose output"},
    { "preload",     'p', POPT_ARG_NONE,   &opt_preload,  0, NULL, 
	  "Not dynamically loaded" },
    { NULL, 0, 0, NULL, 0, NULL, NULL }
};
#else
struct option long_options[] = {
    { "help",        0, NULL, 'h' },
    { "module",      1, NULL, 'M' },
    { "lib",         1, NULL, 'L' },
    { "outdir",      1, NULL, 'o' },
    { "c_outdir",    1, NULL, 'C' },
    { "e_outdir",    1, NULL, 'E' },
    { "h_outdir",    1, NULL, 'H' },
    { "verbose",     0, NULL, 'v' },
    { "version",     0, NULL, 'V' },
    { "preload",     0, NULL, 'p' },
    { 0, 0, 0, 0}};
#endif

char prog[1024];
char outdir[FILENAME_MAX];     /* general out directory    */
char c_outdir[FILENAME_MAX];   /* generated lib%.c files   */
char e_outdir[FILENAME_MAX];   /* generated .inc files     */
char h_outdir[FILENAME_MAX];   /* generated lib%.hrl files */


char* erl_keywords[] = {
    "after",
    "begin",
    "case",
    "catch",
    "end",
    "fun",
    "if",
    "let",
    "of",
    "query",
    "receive",
    "when",
    "bnot",
    "not",
    "div",
    "rem",
    "band",
    "and",
    "bor",
    "bxor",
    "bsl",
    "bsr",
    "or",
    "xor",
    NULL
};


void usage()
{
    fprintf(stderr,
"usage: %s [options] file ...\n"
" OPTIONS:\n"
"     -M <name>, --module=<name>    Override the default module name\n"
"     -p, --preload                 Not dynamically loaded\n"
"     -V, --version                 Emit version and exit\n"
"     -v, --verbose                 Emit verbose output\n"
"     -o <dir>,  --outdir=<dir>     Output directory (default .)\n"
"     -C <dir>,  --c_outdir=<dir>   Overide -o for  lib<mod>.c files\n"
"     -H <dir>,  --h_outdir=<dir>   Override -o for lib<mod>.hrl files\n"
"     -E <dir>,  --e_outdir=<dir>   Override -o for <mod>.inc files\n"
"     -h, --help                    This message\n", prog);
    exit(1);
}

void silly_warning(void* arg)
{
    /* nothing much */
}

/*
** input syntax:
**
** line-commect = '//'
** comment = '/' '*' <char>*  '*' '/'
** syntax = (<decl> ';')*
**
** <decl> = <func-decl>        |
**          <constructor-decl> |
**          <acess-decl>       |
**          <module-decl>      |
**          <struct-decl>      |
**          <typedef-decl>     |
**          <enum-decl>
**
** module-decl      = 'module' <id>
**
** constructor-decl = 'new' <func-decl>
**
** access-decl      = [&|*]<type>-><id> [ '[' <arg-list> ']' ]
**
** func-decl        = <type> <id> '(' <arg-list> ')' [ = <id> ]
**
** struct-decl      = 'struct' <id> '{' [private';'] (<type> <id>;)+ '}'
**
** typedef-decl     = 'typedef' <type> <id>
**
** enum-decl        = 'enum' <id> '{' <id>['='<int>] + '}'
**
** <arg-list>       = [ <type> [<id>] (, <type> [id])* ]
**
** <type> = ['in'|'out'|'inout'] ('void' | 'gint' | 'gfloat' | <id>) '*'*
**
** <id>   = _*[a-zA-Z0-9][a-zA-Z0-9_]*
** ----------------------------------------------------------------------
*/

typedef enum {
    T_EOF,
    T_ERROR,
    T_STAR,     /* '*'  */
    T_PLUS,     /* '+'  */
    T_AMP,      /* '&'  */
    T_LPAREN,   /* '('  */
    T_RPAREN,   /* ')'  */
    T_LBRACK,   /* '['  */
    T_RBRACK,   /* ']'  */
    T_LCBRACK,  /* '{'  */
    T_RCBRACK,  /* '}'  */
    T_COMMA,    /* ','  */
    T_SEMI,     /* ';'  */
    T_COLON,    /* ':'  */
    T_EQ,       /* '='  */
    T_ARROW,    /* '->' */
    T_ELLIPS,   /* ... */
    T_HASH,     /* # */
    T_CHAR,     /* non token character */
    T_NEW,      /* 'NEW' */
    T_MODULE,   /* 'MODULE' */
    T_IN,       /* 'IN' */
    T_OUT,      /* 'OUT' */
    T_INOUT,    /* 'INOUT' */
    T_CONST,    /* 'const' */
    T_UNSIGNED, /* 'unsigned' */
    T_SIGNED,   /* 'signed' */
    T_STRUCT,   /* 'struct' */
    T_TYPEDEF,  /* 'typedef' */
    T_ENUM,     /* 'enum' */
    T_ID,       /* identifier */
    T_STRING,   /* double quoted string */
    T_INTEGER,  /* integer constant */
    T_PRIVATE,  /* 'PRIVATE' */
    T_LAST
} token_t;

typedef struct {
    token_t t;
    int    line;
    int    tlen;
    char   tbuf[MAX_ID_BUF];
} token_entry_t;

typedef struct {
    FILE* f;
    char* filename;
    int  lineno;
    char linebuf[MAX_LINE_BUF];
    token_entry_t* t_top;   /* buffered sofar */
    token_entry_t* t_cur;   /* current for get_token */
    token_entry_t* t_end;   /* end of token buffer */
    token_entry_t t_buf[MAX_TOKEN_LOOKAHEAD];
} scan_t;

/*************** STATIC STRUCTURES **************/
typedef struct {
    char* directive;
    char* arg;
    char* val;  /* for defined */
} parse_cpp_t;

typedef struct {
    type_qualifier_t qualifier;
    char* id;
    int   indirection;             /* = number of '*'  */
    int   asize[MAX_ARRAY_DECL];   /* element size of arrays (vectors) */
} parse_type_t;

typedef struct {
    arg_status_t status;
    parse_type_t type;
} parse_arg_type_t;

typedef struct {
    parse_arg_type_t arg_type;
    char*            id;        /* optional argument id */
} parse_arg_t;

typedef struct {
    int         nargs;
    parse_arg_t arg[MAX_PARSE_ARGS];
} parse_arglist_t;

typedef struct {
    int             is_constructor;
    parse_type_t    return_type;
    char*           id;
    parse_arglist_t arglist;
    char*           cid;
} parse_funcdecl_t;

typedef struct {
    char* id;
    int             private;
    parse_arglist_t elems;
} parse_structdecl_t;

typedef struct {
    char*           id;
} parse_moduledecl_t;

typedef struct {
    parse_type_t     return_type;
    char*            type;
    char*            field;
    char             rmod;   /* & | * | blank */
    parse_arglist_t  arglist;
} parse_accessdecl_t;

typedef struct {
    parse_type_t     type;
    char*            id;
} parse_typedefdecl_t;

typedef struct {
    char* id;
    int   val_is_set;
    int   val;
} parse_enum_t;

typedef struct {
    int  nenums;
    parse_enum_t enm[MAX_PARSE_ENUMS];
} parse_enumlist_t;    

typedef struct {
    char* name;    /* type name */
    parse_enumlist_t enums;
} parse_enumdecl_t;

typedef enum {
    DECL_FUNC,
    DECL_ACCESS,
    DECL_MODULE,
    DECL_STRUCT,
    DECL_TYPE,
    DECL_ENUM,
    DECL_CPP
} decl_type_t;

typedef struct {
    int line;
    decl_type_t type;
    union {
	parse_cpp_t         p;
	parse_funcdecl_t    f;
	parse_accessdecl_t  a;
	parse_moduledecl_t  m;
	parse_structdecl_t  s;
	parse_typedefdecl_t t;
	parse_enumdecl_t    e;
    } u;
} parse_decl_t;

    
typedef enum {
    FUNCTION,
    CONSTRUCTOR,
    ACCESS_GET,
    ACCESS_PUT
} f_call_type_t;

/*************** DYNAMIC STRUCTURES **************/

typedef struct _type_t {
    int               indirection;            /* number of stars */
    int               asize[MAX_ARRAY_DECL];  /* element size of arrays (vectors) */
    arg_status_t      status;                 /* IN | OUT | INOUT | ELLIPS */
    type_qualifier_t  qualifier;              /* SIGNED UNSIGNED ... */
    char*             type_name;              /* base type name */
    GtkType           type;                   /* the gtk type "number" */
    char*             arg_name;               /* argument name */
} type_t;

typedef struct {
    char* directive;
    char* arg;
    char* val;
} cpp_t;

typedef struct {
    char*   name;
} marshall_t;

typedef struct _typedef_t {
    char*     name;          /* name of the new type */
    type_t*   type;          /* The "new" type */
    int       ext;           /* external type (not defined here) */
    type_t*   type_def;      /* definition of the new type */
} typedef_t;

typedef struct {
    f_call_type_t type;
    int     line;             /* line number in the defintion file */
    char*   fun_name;         /* The Erl function name */
    char*   c_name;           /* The C function name */
    int     c_extern;          /* 1 if the c function must be extern decl */
    char*   field_name;       /* valid when type == ACCESS */
    type_t* return_type;
    int     nargs;
    GSList* args;
} function_t;

typedef struct {
    char*   name;
    type_t* type;
    int     declare;
    int     private;
    int     nelems;
    GSList* elems;
} struct_t;

typedef struct {
    char* id;
    int   val_is_set;
    int   val;
} enum_val_t;

typedef struct {
    char*   name;
    type_t* type;
    int     nelems;
    GSList* elems;
} enum_t;


typedef struct {
    char* name;        /* module name */
    GSList* clist;     /* '#include' list */
    GSList* tlist;     /* 'typedef'  list */ 
    GSList* elist;     /* 'enum'     list */
    GSList* slist;     /* 'struct'   list */
    GSList* flist;     /* 'function' list */
} module_t;

/* global definitions (for all modules in one file) */

module_t* globals   = NULL;
GSList*   modules   = NULL;    /* module list */
int       gerror    = 0;

char* type_id(type_t* ap);
char* type_suffix(GtkType type);

extern void  gtk_drv_type_init(void);


/*********** LIST UTILS ************/

#define FOREACH(list, type, var, body) do { \
  GSList* __iter; \
  for (__iter = (list); __iter != NULL; __iter = __iter->next) { \
      type var = (type) (__iter->data); \
      body \
  } \
} while(0)

#define IS_LAST (__iter->next == NULL)

typedef void (*func0_t)(void*);
typedef void (*func1_t)(void*, void*);
typedef void (*func2_t)(void*, void*, void*);
typedef void (*func1_flag_t)(void*, int, void*);
typedef void (*func2_flag_t)(void*, int, void*, void*);

void slist_free(GSList* list)
{
    while(list) {
	GSList* tmp = g_slist_next(list);
	g_slist_free(list);
	list = tmp;
    }
}

void foreach(GSList *list, GFunc func)
{
    while (list) { 
	(*((func0_t)func))(list->data);
	list = list->next;
    }
}

void foreach1(GSList *list, GFunc func, void* arg1)
{
    while (list) { 
	(*((func1_t)func))(list->data, arg1);
	list = list->next;
    }
}

void foreach2(GSList *list, GFunc func, void* arg1, void* arg2)
{
    while (list) { 
	(*((func2_t)func))(list->data, arg1, arg2);
	list = list->next;
    }
}

void foreach_first1(GSList *list, GFunc func, void* arg1)
{
    if (list) {
	(*((func1_flag_t)func))(list->data, 1, arg1);
	list = list->next;
    }
    while (list) { 
	(*((func1_flag_t)func))(list->data, 0, arg1);
	list = list->next;
    }
}

void foreach_last1(GSList *list, GFunc func, void* arg1)
{
    while (list) { 
	GSList* next = list->next;
	(*((func1_flag_t)func))(list->data, next==NULL, arg1);
	list = next;
    }
}

void foreach_last2(GSList *list, GFunc func, void* arg1, void* arg2)
{
    while (list) { 
	GSList* next = list->next;
	(*((func2_flag_t)func))(list->data, next==NULL, arg1, arg2);
	list = next;
    }
}



GtkType lookup_type(type_t* tp)
{
    return gdrv_type_from_name(tp->type_name, tp->indirection, 
			       tp->status, tp->qualifier);
}

/*********** DYNAMIC STUCTURE ALLOC/FREE ************/

type_t* new_internal_type(char* type_name)
{
    type_t* p    = (type_t*) malloc(sizeof(type_t));

    p->indirection = 0;
    p->qualifier   = Q_NONE;
    p->status      = ARG_IN;
    p->type_name   = strdup(type_name);
    p->arg_name    = strdup("");
    p->type        = 0;    
    return p;
}

type_t* new_type(parse_type_t* type)
{
    type_t* p    = (type_t*) malloc(sizeof(type_t));
    p->indirection = type->indirection;
    memcpy(p->asize, type->asize, type->indirection*sizeof(int));
    p->qualifier   = type->qualifier;
    p->status      = ARG_IN;            /* default for return types */
    p->type_name   = strdup(type->id); /* type name must always exist */
    p->arg_name    = strdup("");
    p->type        = 0;
    return p;
}

type_t* new_void_type()
{
    type_t* p    = (type_t*) malloc(sizeof(type_t));
    p->indirection = 0;
    p->status      = ARG_IN;
    p->type_name   = strdup("void");   /* type name must always exist */
    p->arg_name    = strdup("");
    p->type        = 0;
    return p;
}

type_t* new_arg(parse_arg_t* arg, int i)
{
    type_t* p = new_type(&arg->arg_type.type);

    p->status       = arg->arg_type.status;
    free(p->arg_name);
    if (arg->id == NULL) {
	char buf[8];
	sprintf(buf, "a%d", i);
	p->arg_name = strdup(buf);
    }
    else
	p->arg_name = strdup(arg->id);
    return p;
}

void free_type(type_t* arg)
{
    free(arg->arg_name);
    free(arg->type_name);
    free(arg);
}

marshall_t* new_marshaller(char* name)
{
    marshall_t* p = (marshall_t*) malloc(sizeof(marshall_t));
    p->name = strdup(name);
    return p;
}

void free_marshaller(marshall_t* p)
{
    free(p->name);
    free(p);
}

int cmp_marshaller(marshall_t* p, char* name)
{
    return strcmp(p->name, name);
}


typedef_t* new_typedef(parse_typedefdecl_t* t)
{
    typedef_t* p = (typedef_t*) malloc(sizeof(typedef_t));
    p->name     = strdup(t->id);
    p->type     = new_internal_type(t->id);
    p->ext      = 0;
    p->type_def = new_type(&t->type);
    return p;
}

void free_typedef(typedef_t* p)
{
    free(p->name);
    free_type(p->type);
    free_type(p->type_def);
    free(p);
}

cpp_t* new_cpp(parse_cpp_t* src)
{
    cpp_t* p = (cpp_t*) malloc(sizeof(cpp_t));
    p->directive = strdup(src->directive);
    p->arg = strdup(src->arg);
    p->val = strdup(src->val);
    return p;
}

void free_cpp(cpp_t* p)
{
    free(p->directive);
    free(p->arg);
    free(p->val);
    free(p);
}

struct_t* new_struct(char* name, GSList* elems)
{
    struct_t* p = (struct_t*) malloc(sizeof(struct_t));

    p->name     = strdup(name);
    p->type     = new_internal_type(name);
    p->elems    = elems;
    p->nelems   = g_slist_length(elems);
    p->private  = 0;
    p->declare  = 0;
    return p;
}

void free_struct(struct_t* p)
{
    free(p->name);
    free_type(p->type);
    foreach(p->elems, (GFunc) free_type);
    slist_free(p->elems);
    free(p);
}

enum_t* new_enum(parse_enumdecl_t* src)
{
    enum_t* p = (enum_t*) malloc(sizeof(enum_t));
    int i;

    p->name   = strdup(src->name);
    p->type   = new_internal_type(src->name);
    p->nelems = src->enums.nenums;
    p->elems  = NULL;
    for (i = 0; i < p->nelems; i++) {
	enum_val_t* val = (enum_val_t*) malloc(sizeof(enum_val_t));
	val->id = strdup(src->enums.enm[i].id);
	val->val_is_set = src->enums.enm[i].val_is_set;
	val->val        = src->enums.enm[i].val;
	p->elems = g_slist_append(p->elems, val);
    }
    return p;
}

void free_enum_val(enum_val_t* p)
{
    free(p->id);
    free(p);
}

void free_enum(enum_t* p)
{
    free(p->name);
    free_type(p->type);
    foreach(p->elems, (GFunc) free_enum_val);
    slist_free(p->elems);
    free(p);
}

function_t* new_function(f_call_type_t ftype,
			 char* mod_name, 
			 char* fun_name,
			 char* decl_name,
			 char* c_name,
			 char* field_name,
			 type_t* return_type,
			 GSList* args)
{
    function_t* p;
    int len;

    p = (function_t*) malloc(sizeof(function_t));
    p->type = ftype;

    len =  strlen(mod_name);

    if (strncmp(mod_name, decl_name, len) == 0) {
	if (decl_name[len] == '_')
	    p->fun_name = strdup(decl_name + len + 1);
	else {
	    p->fun_name = strdup(decl_name + len);
	    if (isupper(p->fun_name[0]))
		p->fun_name[0] = tolower(p->fun_name[0]);
	}
    }
    else
	p->fun_name = strdup(decl_name);
    if (c_name != NULL) {
	p->c_extern = 0; /* was 1 (need to to do #include instead) */
	p->c_name = strdup(c_name);
    }
    else {
	p->c_extern = 0;
	p->c_name = strdup(decl_name);
    }
    if (field_name != NULL)
	p->field_name = strdup(field_name);
    else
	p->field_name = strdup("");
    p->return_type = return_type;

    p->args = args;
    if ((p->nargs = g_slist_length(p->args)) == 1) {
	type_t* arg = (type_t*) args->data;
	if ((strcmp(arg->type_name, "void") == 0) &&
	    (arg->indirection == 0)) {
	    free_type(arg);
	    slist_free(args);
	    p->args = NULL;
	    p->nargs = 0;
	}
    }
    return p;
}

void free_function(function_t* p)
{
    free(p->fun_name);
    free(p->c_name);
    free(p->field_name);
    foreach(p->args, (GFunc)free_type);
    slist_free(p->args);
    free(p);
}


module_t* new_module(char* name)
{
    module_t* p = (module_t*) malloc(sizeof(module_t));

    p->name   = strdup(name);
    p->clist  = NULL;
    p->tlist  = NULL;
    p->elist  = NULL;
    p->slist  = NULL;
    p->flist  = NULL;
    return p;
}

void free_module(module_t* p)
{
    foreach(p->clist,  (GFunc) free_cpp);
    foreach(p->tlist,  (GFunc) free_typedef);
    foreach(p->elist,  (GFunc) free_enum);
    foreach(p->slist,  (GFunc) free_struct);
    foreach(p->flist,  (GFunc) free_function);

    slist_free(p->clist);
    slist_free(p->tlist);
    slist_free(p->elist);
    slist_free(p->slist);
    slist_free(p->flist);

    free(p);
}
    

void load_lib(char* file)
{
    if ((dl_lib = dlopen(file, RTLD_LAZY)) == NULL) {
	fprintf(stderr, "could not open library %s, %s\n",
		file, dlerror());
	return;
    }
}

GtkType load_gtk_type(char* type_name)
{
    char name[256];
    char* err;
    void* sym = NULL;

    gdrv_typename_to_cname(name, type_name);
    strcat(name, "_get_type");

#ifndef __WIN32__
    sym = dlsym(RTLD_DEFAULT, name);
    if ((err = dlerror()) != NULL) {
	sym = dlsym(dl_lib, name);
	if ((err = dlerror()) != NULL) {
	    fprintf(stderr, "warning: could not resolve: %s (%s)\n",name,err);
	    return 0;
	}
    }
#else
    /* FIXME */
#endif
    if (sym == NULL) {
	fprintf(stderr, "warning: symbol %s == NULL\n", name);
	return 0;
    }
    return (*((GtkType(*)(void))sym))();
}



GSList* new_arglist(parse_arglist_t* arglist, int ibase)
{
    GSList* list = NULL;
    int i;

    for (i = 0; i < arglist->nargs; i++) {
	type_t* arg = new_arg(&arglist->arg[i], i+ibase);
	list = g_slist_append(list, arg);
    }
    return list;
}



/*
** read next token.
** skip blanks and comments
*/
token_t read_token(FILE* f, int* line, char* yytext, int* yylen)
{
    char* yyt = yytext;
    int len = *yylen;
    token_t t = T_ERROR;
    int c;

    while(len) {
	c = fgetc(f);
	while(isspace(c)) {
	    if (c == '\n') {
		(*line)++;
	    }
	    c = fgetc(f);
	}
	if (c == EOF)
	    return T_EOF;
	*yytext++ = c;
	len--;
	switch(c) {
	case '*': t = T_STAR; goto done;
	case '+': t = T_PLUS; goto done;
	case '&': t = T_AMP;  goto done;
	case '(': t = T_LPAREN; goto done;
	case ')': t = T_RPAREN; goto done;
	case '[': t = T_LBRACK; goto done;
	case ']': t = T_RBRACK; goto done;
	case '{': t = T_LCBRACK; goto done;
	case '}': t = T_RCBRACK; goto done;
	case ',': t = T_COMMA; goto done;
	case ';': t = T_SEMI; goto done;
	case '=': t = T_EQ; goto done;
	case ':': t = T_COLON; goto done;
	case '#': t = T_HASH; goto done;

	case '.':
	    if ((c = fgetc(f)) == '.') {
		if ((c = fgetc(f)) == '.') {
		    *yytext++ = '.';
		    *yytext++ = '.';
		    t = T_ELLIPS;
		    goto done;
		}
		ungetc('.', f);
		if (c != EOF)
		    ungetc(c, f);
	    }
	    else if (c != EOF)
		ungetc(c, f);
	    goto done;
	case '/':
	    if ((c = fgetc(f)) != EOF) {
		if (c == '/') {
		    yytext--;
		    len++;
		    while(((c = fgetc(f)) != '\n') && (c != EOF))
			;
		    if (c == '\n')
			(*line)++;
		    continue;
		}
		if (c == '*') {
		    yytext--;
		    len++;
		    while(1) {
			c = fgetc(f);
			if (c == EOF)
			    return T_EOF;
			else if (c == '\n')
			    (*line)++;
			else if (c == '*') {
			    if ((c = fgetc(f)) == '/')
				break;
			    if (c == EOF)
				return T_EOF;
			    ungetc(c, f);
			}
		    }
		    continue;
		}
		else {
		    ungetc(c, f);
		    t = T_CHAR;
		}
	    }
	    goto done;

	case '-':
	    if (len) {
		if ((c = fgetc(f)) != EOF) {
		    *yytext++ = c;
		    len--;
		}
		if (c == '>')
		    t = T_ARROW;
		else if (isdigit(c)) {
		    c = fgetc(f);
		    while(isdigit(c)) {
			*yytext++ = c;
			len--;
			c = fgetc(f);	
		    }
		    if (c != EOF)
			ungetc(c, f);
		    t = T_INTEGER;
		}
	    }
	    goto done;

	case '_':
	    if (len) {
		if ((c = fgetc(f)) != EOF) {
		    *yytext++ = c;
		    len--;
		}
		while(len && (c == '_')) {
		    if ((c = fgetc(f)) != EOF) {
			*yytext++ = c;
			len--;
		    }
		}
	    }

	case '"':
	    c = fgetc(f);
	    *yytext++ = c;
	    while((c != EOF) && (c != '"')) {
		c = fgetc(f);
		*yytext++ = c;
	    }
	    if (c == '"') {
		t = T_STRING;
		goto done;
	    }
	    return T_EOF;

	case '<':
	    /* DIRTY HACK */
	    c = fgetc(f);
	    *yytext++ = c;
	    while((c != EOF) && (c != '>')) {
		c = fgetc(f);
		*yytext++ = c;
	    }
	    if (c == '>') {
		t = T_STRING;
		goto done;
	    }
	    return T_EOF; 

	default:
	    if (isdigit(c)) {
		c = fgetc(f);
		while(isdigit(c)) {
		    *yytext++ = c;
		    len--;
		    c = fgetc(f);	
		}
		if (c != EOF)
		    ungetc(c, f);
		t = T_INTEGER;
	    }
	    else if (isalnum(c)) {
		c = fgetc(f);
		while(isalnum(c) || (c == '_')) {
		    *yytext++ = c;
		    len--;
		    c = fgetc(f);
		}
		if (c != EOF)
		    ungetc(c, f);
		t = T_ID;
	    }
	    goto done;
	}
    }

 done:
    *yytext++ = '\0';
    if (t == T_ID) {
	if (strcmp(yyt, "new") == 0)
	    t =T_NEW;
	else if (strcmp(yyt, "module") == 0)
	    t = T_MODULE;
	else if (strcmp(yyt, "in") == 0)
	    t = T_IN;
	else if (strcmp(yyt, "out") == 0)
	    t = T_OUT;
	else if (strcmp(yyt, "inout") == 0)
	    t = T_INOUT;
	else if (strcmp(yyt, "const") == 0)
	    t = T_CONST;
	else if (strcmp(yyt, "unsigned") == 0)
	    t = T_UNSIGNED;
	else if (strcmp(yyt, "signed") == 0)
	    t = T_SIGNED;
	else if (strcmp(yyt, "struct") == 0)
	    t = T_STRUCT;
	else if (strcmp(yyt, "private") == 0)
	    t = T_PRIVATE;
	else if (strcmp(yyt, "typedef") == 0)
	    t = T_TYPEDEF;
	else if (strcmp(yyt, "enum") == 0)
	    t = T_ENUM;
    }
    *yylen = yytext - yyt;
    return t;
}


void print_error(scan_t* scan, char* message)
{
    token_entry_t* tp;
    fprintf(stderr, "%s:%d: %s\n", scan->filename, scan->lineno, message);
    for (tp = scan->t_buf; tp != scan->t_top; tp++)
	fprintf(stderr, "%s ", tp->tbuf);
    fprintf(stderr, "\n");
}


/* read next token from stream into head */
token_t get_token(scan_t* scan, char** text)
{
    token_entry_t* cur = scan->t_cur;

    if (cur == NULL)
	cur = scan->t_buf;
    else
	cur++;

    if (cur == scan->t_top) {
	token_entry_t* tp;
	if (scan->t_top == scan->t_end) {
	    print_error(scan, "token lookahead buffer exceeded");
	    if (text != NULL)
		*text = "token lookahead buffer exceeded";
	    return T_ERROR;
	}
	tp = scan->t_top;
	tp->tlen = 255;   /* reset max length */
	tp->t = read_token(scan->f, &scan->lineno, tp->tbuf, &tp->tlen);
	tp->line = scan->lineno;
	scan->t_top++;
    }
    scan->t_cur = cur;
    if (text != NULL)
	*text = cur->tbuf;

    if (opt_verbose > 1) {
	fprintf(stderr, "TOKEN pos:%d line:%d: %d \"%s\"\n", 
		(scan->t_cur - scan->t_buf),
		cur->line, cur->t, cur->tbuf);
    }
    return cur->t;
}

int get_line(scan_t* scan)
{
    return (scan->t_cur == NULL) ? -1 : scan->t_cur->line;
}

void unget_token(scan_t* scan)
{ 
    if ((scan->t_cur == NULL) || (scan->t_cur->t == T_ERROR))
	;
    else if (scan->t_cur == scan->t_buf)
	scan->t_cur = NULL;
    else
	scan->t_cur--;
}


token_t curr_token(scan_t* scan, char** text)
{
    token_entry_t* cur = scan->t_cur;

    if (cur == NULL) {
	print_error(scan, "token reset");
	if (text != NULL)
	    *text = "token reset";
	return T_ERROR;
    }
    else if (text != NULL)
	*text = cur->tbuf;
    return cur->t;
}


void reset_scan(scan_t* scan)
{
    scan->t_cur = NULL;
    scan->t_top = scan->t_buf;
}

void sync_scan(scan_t* scan, token_t token)
{
    token_t t;
    do {
	char tmpbuf[MAX_ID_BUF];
	int len = MAX_ID_BUF;
	t = read_token(scan->f, &scan->lineno, tmpbuf, &len);
    } while((t != T_EOF) || (t != token));
    reset_scan(scan);
}

/* expect token */
int expect_token(scan_t* scan, token_t t)
{
    return (t == get_token(scan, NULL));
}

void init_scan(scan_t* scan, FILE* f, char* filename)
{
    scan->f = f;
    scan->filename = filename;
    scan->lineno = 1;
    scan->t_end = scan->t_buf + MAX_TOKEN_LOOKAHEAD;
    reset_scan(scan);
}

void print_type(parse_type_t* type)
{
    int i;
    printf("%s", type->id);
    for (i = 0; i < type->indirection; i++) {
	if (type->asize[i] == NO_LENGTH)
	    printf("*");
	else if (type->asize[i] == ANY_LENGTH)
	    printf("[]");
	else
	    printf("[%d]", type->asize[i]);
    }
}

void print_arglist(parse_arglist_t* arglist,
		   char* sep1, char* sep2, char* sep3)
{
    int i;

    for (i = 0; i < arglist->nargs; i++) {
	if (i > 0)
	    printf("%s", sep1);
	else
	    printf("%s", sep2);
	switch(arglist->arg[i].arg_type.status) {
	case ARG_ELLIPS: printf("..."); break;
	case ARG_IN:     /* printf("in "); */ break;
	case ARG_OUT:    printf("out "); break;
	case ARG_INOUT:  printf("inout "); break;
	}
	print_type(&arglist->arg[i].arg_type.type);
	if (arglist->arg[i].id != NULL)
	    printf(" %s%s", arglist->arg[i].id, sep3);
    }
}

/* decl = <func_decl> | <access_decl> | <module_decl> */
void print_decl(parse_decl_t* decl) 
{
    switch(decl->type) {
    case DECL_FUNC: {
	parse_funcdecl_t* function = &decl->u.f;
	if (function->is_constructor)
	    printf("new ");
	print_type(&function->return_type);
	printf(" %s(", function->id);
	print_arglist(&function->arglist, ", ", "", "");
	printf(")");
	if (function->cid != NULL)
	    printf(" = %s", function->cid);
	break;
    }
    case DECL_ACCESS: {
	parse_accessdecl_t* ap = &decl->u.a;
	print_type(&ap->return_type);
	printf(" %s->%s", ap->type, ap->field);
	if (ap->arglist.nargs > 0) {
	    printf("[");
	    print_arglist(&ap->arglist, ", ", "", "");
	    printf("]");
	}
	break;
    }
    case DECL_MODULE: {
	parse_moduledecl_t* module = &decl->u.m;
	printf("module %s", module->id);
	break;
    }

    case DECL_STRUCT: {
	parse_structdecl_t* sp = &decl->u.s;
	printf("struct %s {", sp->id);
	print_arglist(&sp->elems, "\n  ", "  ", ";\n");
	printf("}");
	break;
    }

    case DECL_TYPE: {
	parse_typedefdecl_t* td = &decl->u.t;
	printf("typedef ");
	print_type(&td->type);
	printf(" %s", td->id);
	break;
    }

    case DECL_ENUM: {
	parse_enumdecl_t* e = &decl->u.e;
	int i;

	printf("enum %s {\n", e->name);
	for (i = 0; i < e->enums.nenums; i++) {
	    int c = (i == e->enums.nenums-1) ? ' ' : ',';
	    if (e->enums.enm[i].val_is_set)
		printf("%s = %d%c\n",
		       e->enums.enm[i].id, e->enums.enm[i].val, c);
	    else
		printf("%s%c\n", e->enums.enm[i].id, c);
	}
	printf("}");
	break;
    }

    case DECL_CPP: {
	parse_cpp_t* p = &decl->u.p;

	printf("#%s %s\n", p->directive, p->arg);
	return;
    }
    }
    printf(";\n");
}

/* parse ( '[' [<int>] ']' )*  */
int parse_array_decl(scan_t* scan, parse_type_t* type)
{
    while(1) {
	token_t tok = get_token(scan, NULL);
	if (tok == T_LBRACK) {
	    char* val;
	    if ((tok = get_token(scan, &val)) == T_INTEGER) {
		int len;
		if ((len = atoi(val)) > 0) {
		    type->asize[type->indirection] = len;
		    if (get_token(scan, NULL) != T_RBRACK) {
			unget_token(scan);
			return 0;
		    }
		}
		else {
		    print_error(scan, "bad array length");
		    return 0;
		}
	    }
	    else if (tok == T_RBRACK)
		type->asize[type->indirection] = ANY_LENGTH;
	    else {
		unget_token(scan);
		unget_token(scan);
		return 0;
	    }
	}
	else if (tok == T_STAR)
	    type->asize[type->indirection] = NO_LENGTH;
	else {
	    unget_token(scan);
	    break;
	}
	type->indirection++;
    }
    return 1;
}


/* type = ['const'|'unsigned'|'signed']* <id>'*'*  */
int parse_type(scan_t* scan, parse_type_t* type)
{
    type->qualifier = 0;

 qloop:
    switch(get_token(scan, NULL)) {
    case T_CONST:
	type->qualifier |= Q_CONST;
	goto qloop;
    case T_UNSIGNED:
	type->qualifier |= Q_UNSIGNED;
	goto qloop;
    case T_SIGNED:
	type->qualifier |= Q_SIGNED;
	goto qloop;
    default:
	unget_token(scan);
	break;
    }
    if (get_token(scan, &type->id) != T_ID) {
	unget_token(scan);
	return 0;
    }
    type->indirection = 0;
    return parse_array_decl(scan, type);
}

/* arg_type = [IN|INOUT|OUT] <type> */
int parse_arg_type(scan_t* scan, parse_arg_type_t* type)
{
    token_t tok = get_token(scan, NULL);

    switch (tok) {
    case T_IN:
	type->status = ARG_IN;
	break;
    case T_OUT:  
	type->status = ARG_OUT; 
	break;
    case T_INOUT: 
	type->status = ARG_INOUT; 
	break;
    default:
	type->status = ARG_IN;
	unget_token(scan);
	break;
    }
    return parse_type(scan, &type->type);
}

/* arg = <arg_type> [<id> <array-decl> ] */
int parse_arg(scan_t* scan, parse_arg_t* arg)
{
    if (parse_arg_type(scan, &arg->arg_type)) {
	if (get_token(scan, &arg->id) != T_ID) {
	    arg->id = NULL;
	    unget_token(scan);
	    return 1;
	}
	return parse_array_decl(scan, &arg->arg_type.type);
    }
    return 0;
}


/* arg_list =  [ <arg> (',' <arg>)* ] */
int parse_arg_list(scan_t* scan, parse_arglist_t* arglist)
{
    int i = 0;

    while((i < MAX_PARSE_ARGS) && parse_arg(scan, &arglist->arg[i])) {
	if (get_token(scan, NULL) != T_COMMA) {
	    unget_token(scan);
	    arglist->nargs = i+1;
	    return 1;
	}
	i++;
    }
    if (i < MAX_PARSE_ARGS) {
	if (get_token(scan, NULL) == T_ELLIPS) {
	    arglist->arg[i].arg_type.status = ARG_ELLIPS;
	    arglist->arg[i].arg_type.type.qualifier = Q_NONE;
	    arglist->arg[i].arg_type.type.id = "...";
	    arglist->arg[i].arg_type.type.indirection = 0;
	    arglist->arg[i].id = "";
	    arglist->nargs = i+1;
	    return 1;
	}
	unget_token(scan);
	return 0;
    }
    else
	print_error(scan, "to many arguments");
    return 0;
}

/* elem_list =  (<type> (<id>)+ ';')+  */
int parse_elem_list(scan_t* scan, parse_arglist_t* arglist, int* private)
{
    int i = 0;

    *private = 0;
    while(i < MAX_PARSE_ARGS) {
	parse_arg_type_t arg_type;
	token_t t;

	if ((t = get_token(scan, NULL)) == T_RCBRACK) {
	    if (i == 0) {
		print_error(scan, "empty struct not allowed");
		return 0;
	    }
	    unget_token(scan);
	    arglist->nargs = i;
	    return 1;
	}
	else if (t == T_PRIVATE) {
	    *private = 1;
	    if (get_token(scan, NULL) != T_SEMI)
		return 0;
	    continue;
	}
	unget_token(scan);
	

	if (!parse_arg_type(scan, &arg_type))
	    return 0;
	do {
	    if (get_token(scan, &arglist->arg[i].id) != T_ID)
		return 0;
	    arglist->arg[i].arg_type = arg_type;
	    i++;
	} while((get_token(scan, NULL) == T_COMMA) && (i < MAX_PARSE_ARGS));

	unget_token(scan);
	if (get_token(scan, NULL) != T_SEMI)
	    return 0;
    }
    print_error(scan, "to many elements in struct");
    return 0;
}

/* struct_decl = 'struct' '{' (<type> <id> ';')+ '}' */
int parse_struct_decl(scan_t* scan, parse_structdecl_t* sp)
{
    if (get_token(scan, NULL) != T_STRUCT) return 0;
    if (get_token(scan, &sp->id) != T_ID)   return 0;
    if (get_token(scan, NULL) != T_LCBRACK) return 0;
    if (!parse_elem_list(scan, &sp->elems, &sp->private)) return 0;
    if (get_token(scan, NULL) != T_RCBRACK) return 0;
    return 1;
}


/* func_decl = ['new'] <type> <id> '(' <arg_list> ')' [ '=' <id> ] */
int parse_func_decl(scan_t* scan, parse_funcdecl_t* func)
{
    if (get_token(scan, NULL) == T_NEW)
	func->is_constructor = 1;
    else {
	func->is_constructor = 0;
	unget_token(scan);
    }
    if (!parse_type(scan, &func->return_type)) return 0;
    if (get_token(scan, &func->id) != T_ID)   return 0;
    if (get_token(scan, NULL) != T_LPAREN)     return 0;
    if (!parse_arg_list(scan, &func->arglist)) return 0;
    if (get_token(scan, NULL) != T_RPAREN)     return 0;
    if (get_token(scan, NULL) == T_EQ) {
	if (get_token(scan, &func->cid) != T_ID) return 0;
    }
    else {
	func->cid = NULL;
	unget_token(scan);
    }
    return 1;
}

/* module_decl = 'module' <id> */
int parse_module_decl(scan_t* scan, parse_moduledecl_t* module)
{
    if (get_token(scan, NULL) != T_MODULE) return 0;
    if (get_token(scan, &module->id) != T_ID) return 0;
    return 1;
}

/* access_decl = <type> [&|*]<id>'->'<id> [ '[' <arg-list> ']' ] */
int parse_access_decl(scan_t* scan, parse_accessdecl_t* a)
{
    token_t t;
    char* ftype;

    if (!parse_type(scan, &a->return_type))  return 0;
    t = get_token(scan, &ftype);
    a->rmod = ' ';
    if (t == T_ID) 
	a->type = ftype;
    else {
	if (t == T_AMP)
	    a->rmod = '&';
	else if (t == T_STAR)
	    a->rmod = '*';
	else
	    return 0;
	if (get_token(scan, &a->type) != T_ID) return 0;
    }
    if (get_token(scan, NULL) != T_ARROW)     return 0;
    if (get_token(scan, &a->field) != T_ID)  return 0;
    if (get_token(scan, NULL) == T_LBRACK) {
	if (!parse_arg_list(scan, &a->arglist))   return 0;
	if (get_token(scan, NULL) != T_RBRACK)    return 0;
    }
    else {
	a->arglist.nargs = 0;
	unget_token(scan);
    }
    return 1;
}

int parse_typedef_decl(scan_t* scan, parse_typedefdecl_t* t)
{
    if (get_token(scan, NULL) != T_TYPEDEF) return 0;
    if (!parse_type(scan, &t->type))       return 0;
    if (get_token(scan, &t->id) != T_ID)   return 0;
    return 1;
}

/* elem_list =  (<id>['='<integer>] ',')+  */
int parse_enum_list(scan_t* scan, parse_enumlist_t* enumlist)
{
    int i = 0;

    while(i < MAX_PARSE_ENUMS) {
	char* num;
	token_t t;

	if (get_token(scan, &enumlist->enm[i].id) != T_ID)
	    return 0;
	if ((t = get_token(scan, NULL)) == T_EQ) {
	    if (get_token(scan, &num) == T_INTEGER) {
		enumlist->enm[i].val_is_set = 1;
		enumlist->enm[i].val = atoi(num);
	    }
	    else
		return 0;
	    t = get_token(scan,NULL);
	}
	else
	    enumlist->enm[i].val_is_set = 0;

	if (t == T_RCBRACK) {
	    unget_token(scan);
	    enumlist->nenums = i;
	    return 1;
	}
	else if (t != T_COMMA)
	    return 0;
	i++;
    }
    print_error(scan, "to many values in enum");
    return 0;
}

int parse_enum_decl(scan_t* scan, parse_enumdecl_t* e)
{
    if (get_token(scan, NULL) != T_ENUM) return 0;
    if (get_token(scan, &e->name) != T_ID) return 0;
    if (get_token(scan, NULL) != T_LCBRACK) return 0;
    if (!parse_enum_list(scan, &e->enums)) return 0;
    if (get_token(scan, NULL) != T_RCBRACK) return 0;
    return 1;
}

int parse_cpp(scan_t* scan, parse_cpp_t* p)
{
    int line;

    if (get_token(scan, NULL) != T_HASH) return 0;

    line = get_line(scan);
    if (get_token(scan, &p->directive) != T_ID) return 0;

    if (strcmp(p->directive, "include") == 0) {
	if (get_token(scan, &p->arg) != T_STRING) return 0;
	p->val = "";
	return 1;
    }
    else if (strcmp(p->directive, "define") == 0) {
	token_t t;

	if (get_token(scan, &p->arg) != T_ID) return 0;
	t = get_token(scan, &p->val);
	if ((t != T_INTEGER) && (t != T_STRING) && (t != T_ID)) {
	    if (get_line(scan) != line) {
		p->val = "";
		return 1;
	    }
	    return 0;
	}
	return 1;
    }
    return 0;
}


int parse_decl(scan_t* scan, parse_decl_t* decl)
{
    token_entry_t* save;

    save = scan->t_cur;
    decl->line = scan->lineno;
    if (parse_func_decl(scan, &decl->u.f)) {
	decl->type = DECL_FUNC;
	return 1;
    }
    scan->t_cur = save;
    if (parse_access_decl(scan, &decl->u.a)) {
	decl->type = DECL_ACCESS;
	return 1;
    }
    scan->t_cur = save;
    if (parse_module_decl(scan, &decl->u.m)) {
	decl->type = DECL_MODULE;
	return 1;
    }
    scan->t_cur = save;
    if (parse_struct_decl(scan, &decl->u.s)) {
	decl->type = DECL_STRUCT;
	return 1;
    }
    scan->t_cur = save;
    if (parse_typedef_decl(scan, &decl->u.t)) {
	decl->type = DECL_TYPE;
	return 1;
    }
    scan->t_cur = save;
    if (parse_enum_decl(scan, &decl->u.e)) {
	decl->type = DECL_ENUM;
	return 1;
    }
    scan->t_cur = save;
    if (parse_cpp(scan, &decl->u.p)) {
	decl->type = DECL_CPP;
	return 1;
    }
    scan->t_cur = save;

    if (get_token(scan, NULL) == T_EOF)
	return 0;
    scan->t_cur = save;
    print_error(scan, "syntax error");
    return 0;
}


/*
** Parse a new def file
*/
void parse_def(FILE* f, char* filename)
{
    scan_t scanner;
    parse_decl_t decl;
    function_t* call;

    module_t* module = globals;

    init_scan(&scanner, f, filename);

    while(parse_decl(&scanner, &decl)) {
	if ((decl.type != DECL_CPP) &&
	    (get_token(&scanner, NULL) != T_SEMI)) {
	    if (curr_token(&scanner, NULL) != T_EOF) {
		print_error(&scanner, "syntax error");
		gerror++;
		sync_scan(&scanner, T_SEMI);
	    }
	    else
		break;
	}

	if (opt_verbose)
	    print_decl(&decl);

	switch (decl.type) {
	case DECL_FUNC: {
	    parse_funcdecl_t* function = &decl.u.f;
	    f_call_type_t ftype;
	    if (function->is_constructor)
		ftype = CONSTRUCTOR;
	    else
		ftype = FUNCTION;
		
	    call = new_function(ftype, 
				module->name,
				NULL,
				function->id,
				function->cid,
				NULL,
				new_type(&function->return_type),
				new_arglist(&function->arglist, 0));
	    module->flist = g_slist_append(module->flist, call);
	    break;
	}

	case DECL_ACCESS: {
	    parse_accessdecl_t* ap = &decl.u.a;
	    char tname[MAX_ID_BUF];
	    char cname[MAX_ID_BUF*3+1];
	    char fname[MAX_ID_BUF+2];
	    parse_arg_t arg0;
	    parse_arg_t arg_last;
	    GSList* args;
	    type_t* arg;

	    arg0.id = NULL;	/* auto generate */
	    arg0.arg_type.status = ARG_IN;
	    arg0.arg_type.type.qualifier = Q_NONE;
	    arg0.arg_type.type.id = ap->type;
	    arg0.arg_type.type.indirection = 1; /* MUST BE Pointer */
	    arg0.arg_type.type.asize[0] = NO_LENGTH;
	    arg = new_arg(&arg0, 0);
	    args = new_arglist(&ap->arglist, 1);
	    args = g_slist_prepend(args, arg);

	    gdrv_typename_to_cname(tname, ap->type);
	    if (ap->rmod == ' ')
		strcpy(fname, ap->field);
	    else
		sprintf(fname, "%c%s", ap->rmod, ap->field);
	    sprintf(cname, "%s_get_%s", tname, ap->field);
	    call = new_function(ACCESS_GET, 
				module->name,
				NULL,
				cname,
				NULL,
				fname,
				new_type(&ap->return_type),
				args);
	    module->flist = g_slist_append(module->flist, call);

	    /* MUST recreate since we must append element */
	    arg  = new_arg(&arg0, 0);
	    args = new_arglist(&ap->arglist, 1);
	    args = g_slist_prepend(args, arg);

	    arg_last.id = NULL;
	    arg_last.arg_type.status = ARG_IN;
	    arg_last.arg_type.type = ap->return_type;
	    
	    arg = new_arg(&arg_last, g_slist_length(args));
	    args = g_slist_append(args, arg);

	    sprintf(cname, "%s_put_%s", tname, ap->field);
		
	    call = new_function(ACCESS_PUT, 
				module->name,
				NULL,
				cname,
				NULL,
				fname,
				new_void_type(),
				args);
	    module->flist = g_slist_append(module->flist, call);
	    break;
	}

	case DECL_MODULE: {
	    parse_moduledecl_t* mp = &decl.u.m;

	    module = new_module(mp->id);
	    modules = g_slist_append(modules, module);
	    break;
	}

	case DECL_STRUCT: {
	    parse_structdecl_t* sp = &decl.u.s;
	    struct_t* slp;
	    GSList* args;

	    args = new_arglist(&sp->elems, 0);
	    slp = new_struct(sp->id, args);
	    slp->private = sp->private;

	    if ((slp->type->type = lookup_type(slp->type)) == 0) {
		e_type_info_t tab[2];
		slp->declare = 1;
		tab[0].type_name = slp->name;
		tab[0].type_id   = &slp->type->type;
		tab[0].parent    = GTK_TYPE_BOXED;
		tab[0].values    = NULL;
		tab[1].type_name = NULL;
		gdrv_type_load(tab);
		
	    }
	    module->slist = g_slist_append(module->slist, slp);
	    break;
	}

	case DECL_TYPE: {
	    typedef_t* p = new_typedef(&decl.u.t);

	    if ((p->type_def->type = lookup_type(p->type_def)) == 0) {
		fprintf(stderr, "%s:%d: invalid type name %s\n",
			scanner.filename, decl.line, p->type_def->type_name);
		gerror++;
	    }
	    else {
		/* For Gtk/Gobjects we must do some special treatment,
		 * try to load the function that initialize the GtkType
		 * i.e for GtkX the function is called 
		 * GtkType gtk_x_get_type(void)
		 * if that function exist then we should not emit any type
		 */
		if (p->type_def->type == GTK_TYPE_OBJECT) {
		    if ((p->type->type = load_gtk_type(p->name)) == 0)
			gerror++;
		    else
			p->ext = 1;
		}
		else if ((p->type->type = lookup_type(p->type)) == 0) {
		    e_type_info_t tab[2];
		    tab[0].type_name  = p->type->type_name;
		    tab[0].type_id    = &p->type->type;
		    tab[0].parent     = p->type_def->type;
		    tab[0].values     = NULL;
		    tab[1].type_name  = NULL;
		    gdrv_type_load(tab);
		}
	    }
	    module->tlist = g_slist_append(module->tlist, p);
	    break;
	}

	case DECL_CPP: {
	    cpp_t* p = new_cpp(&decl.u.p);
	    module->clist = g_slist_append(module->clist, p);
	    break;
	}

	case DECL_ENUM: {
	    enum_t* en = new_enum(&decl.u.e);
	    GtkEnumValue env[MAX_PARSE_ENUMS+1];
	    int i;
	    e_type_info_t tab[2];

	    /* load names (with dummy values) */
	    for (i = 0; i < en->nelems; i++) {
		env[i].value = i;  /* not the correct value!!! */
		env[i].value_name = decl.u.e.enums.enm[i].id;
		env[i].value_nick = NULL;
	    }
	    env[en->nelems].value = 0;
	    env[en->nelems].value_name = NULL;
	    env[en->nelems].value_nick = NULL;

	    tab[0].type_name = en->name;
	    tab[0].type_id   = &en->type->type;
	    tab[0].parent    = GTK_TYPE_ENUM;
	    tab[0].values    = env;
	    tab[1].type_name = NULL;

	    gdrv_type_load(tab);

	    module->elist = g_slist_append(module->elist, en);
	    break;
	}


	}
	
	reset_scan(&scanner);

    }
}




char* marshall_letter(GtkType type)
{
    switch(GTK_FUNDAMENTAL_TYPE(type)) {
    case GTK_TYPE_NONE:
	return "v";
    case GTK_TYPE_BOOL:
	return "b";
    case GTK_TYPE_CHAR:
    case GTK_TYPE_UCHAR:
    case GTK_TYPE_FLAGS:
    case GTK_TYPE_ENUM:
    case GTK_TYPE_INT:
    case GTK_TYPE_UINT:
    case GTK_TYPE_LONG:
    case GTK_TYPE_ULONG:
	return "i";
    case GTK_TYPE_FLOAT:
	return "f";
    case GTK_TYPE_DOUBLE:
	return "d";
    case GTK_TYPE_STRING:
    case GTK_TYPE_BOXED:
    case GTK_TYPE_POINTER:
    case G_TYPE_OBJECT:
#if GTK_VERSION >= GTK_VERS(2,0,0)
    case G_TYPE_INTERFACE:
#endif
	return "p";
    default:
	if (IS_ARRAY(type))
	    return "p";
	return "?";
    }
}


char* marshall_type(GtkType type)
{
    switch(GTK_FUNDAMENTAL_TYPE(type)) {
    case GTK_TYPE_NONE:
	return "void";
    case GTK_TYPE_BOOL:
	return "gboolean";
    case GTK_TYPE_CHAR:
    case GTK_TYPE_UCHAR:
    case GTK_TYPE_FLAGS:
    case GTK_TYPE_ENUM:
    case GTK_TYPE_INT:
    case GTK_TYPE_UINT:
    case GTK_TYPE_LONG:
    case GTK_TYPE_ULONG:
	return "gint";
    case GTK_TYPE_FLOAT:
	return "gfloat";
    case GTK_TYPE_DOUBLE:
	return "gdouble";
    case GTK_TYPE_STRING:
    case GTK_TYPE_BOXED:
    case GTK_TYPE_POINTER:
    case G_TYPE_OBJECT:
#if GTK_VERSION >= GTK_VERS(2,0,0)
    case G_TYPE_INTERFACE:
#endif
	return "gpointer";
    default:
	if (IS_ARRAY(type))
	    return "gpointer";
	return "?";
    }
}

char* marshall_id(GtkType type)
{
    switch(GTK_FUNDAMENTAL_TYPE(type)) {
    case GTK_TYPE_NONE:
	return "NONE"; 
    case GTK_TYPE_BOOL:
	return "BOOL";
    case GTK_TYPE_CHAR:
    case GTK_TYPE_UCHAR:
    case GTK_TYPE_FLAGS:
    case GTK_TYPE_ENUM:
    case GTK_TYPE_INT:
    case GTK_TYPE_UINT:
    case GTK_TYPE_LONG:
    case GTK_TYPE_ULONG:
	return "INT";
    case GTK_TYPE_FLOAT:
	return "FLOAT";
    case GTK_TYPE_DOUBLE:
	return "DOUBLE";
    case GTK_TYPE_STRING:
    case GTK_TYPE_BOXED:
    case GTK_TYPE_POINTER:
    case G_TYPE_OBJECT:
#if GTK_VERSION >= GTK_VERS(2,0,0)
    case G_TYPE_INTERFACE:
#endif
	return "POINTER";
    default:
	if (IS_ARRAY(type))
	    return "POINTER";
	return "?";
    }
}

/* return the marshaller function name */
char* marshaller(function_t* p)
{
    static char str[1024];
    GSList* args = p->args;

    strcpy(str, "MF_");
    strcat(str, marshall_letter(p->return_type->type));
    strcat(str, "_");
    if (args == NULL)
	strcat(str, "v");
    else {
	while(args) {
	    type_t* ap = (type_t*) args->data;
	    switch(ap->status) {
	    case ARG_IN:     strcat(str, "I"); break;
	    case ARG_OUT:    strcat(str, "O"); break;
	    case ARG_INOUT:  strcat(str, "X"); break;
	    default: strcat(str, "?"); break;
	    }
	    strcat(str, marshall_letter(ap->type));
	    args = args->next;
	}
    }
    return str;
}

int is_stack_allocated(GtkType type)
{
    return  (type == GDK_TYPE_COLOR) ||
	    (type == GDK_TYPE_POINT) ||
	    (type == GDK_TYPE_RECTANGLE) ||
	    (type == GDK_TYPE_SEGMENT) ||
		IS_ARRAY(type);
}

void emit_marshaller(function_t* p, GSList** mlist, FILE* f)
{
    int i;
    char* name     = marshaller(p);
    char* ret_loc  = marshall_id(p->return_type->type);
    char* ret_type = marshall_type(p->return_type->type); 
    GSList* args;


    if (g_slist_find_custom(*mlist, name,(GCompareFunc)cmp_marshaller) != NULL)
	return;
    *mlist = g_slist_append(*mlist, new_marshaller(name));

    if (p->return_type->type == GTK_TYPE_NONE) {
	fprintf(f, "static void %s(FFunction func, gpointer data,"
		" GtkArg* args)\n"
		"{\n"
		"  (*(", name);
    }
    else {
	fprintf(f, "static void %s(FFunction func, gpointer data,"
		" GtkArg* args)\n"
		"{\n"
		"  %s* r = GTK_RETLOC_%s(args[%d]);\n", 
		name, ret_type, ret_loc, p->nargs);
	fprintf(f, "  *r = (*(");
    }
    /* make function cast */
    fprintf(f, "%s (*)(", ret_type);
    for (args = p->args; args != NULL; args = args->next) {
	type_t* ap = (type_t*) args->data;
	if (ap->status & ARG_OUT)
	    fprintf(f, "%s*,", marshall_type(ap->type));
	else
	    fprintf(f, "%s,", marshall_type(ap->type));
    }
    fprintf(f, "gpointer))func)(");

    for (i = 0, args = p->args; args != NULL; args = args->next, i++) {
	type_t* ap = (type_t*) args->data;	
	if ((ap->status & ARG_OUT) && is_stack_allocated(ap->type))
	    fprintf(f, "GTK_VALUE_%s(args[%d]),", 
		    marshall_id(ap->type),
		    i);
	else if (ap->status & ARG_OUT)
	    fprintf(f, "&GTK_VALUE_%s(args[%d]),", 
		    marshall_id(ap->type),
		    i);
	else
	    fprintf(f, "GTK_VALUE_%s(args[%d]),", 
		    marshall_id(ap->type),
		    i);
    }
    fprintf(f, "data);\n");
    fprintf(f, "}\n\n");
}


char* return_type(type_t* r)
{
    if (r->type == GTK_TYPE_STRING)
	return "gchar*";
    else
	return r->type_name;
}


char* arg_type(type_t* r)
{
    if (r->type == GTK_TYPE_STRING)
	return "gchar*";
    else
	return r->type_name;
}


char* type_suffix(GtkType type)
{
    switch (GTK_FUNDAMENTAL_TYPE(type)) {
    case G_TYPE_OBJECT:
#if GTK_VERSION >= GTK_VERS(2,0,0)
    case G_TYPE_INTERFACE:
#endif
    case GTK_TYPE_BOXED:
	return "*";
    default:
	return "";
    }
}



void emit_arg_flag(type_t* ap, int last, FILE* f)
{
    fprintf(f, "%s%s %s%s", 
	    arg_type(ap),
	    type_suffix(ap->type),
	    ap->arg_name,
	    last ? "" : ",");
}

void emit_array_field(type_t* ap, FILE* f)
{
    fprintf(f, "[%s]", ap->arg_name);
}

void emit_array_field_last(type_t* ap, int last, char* access, FILE* f)
{
    if (last)
	fprintf(f, " = %s%s", access, ap->arg_name);
    else 
	fprintf(f, "[%s]", ap->arg_name);
}

void emit_get_method(function_t* p, FILE* f)
{
    type_t* arg0;

    fprintf(f, "static %s%s %s(",
	    return_type(p->return_type),
	    type_suffix(p->return_type->type),
	    p->c_name);
    foreach_last1(p->args, (GFunc)emit_arg_flag, f);
    fprintf(f, ")\n{\n");

    arg0 = (type_t*) p->args->data;
    if ((p->field_name[0] == '&') || (p->field_name[0] == '*')) {
	fprintf(f, "  return %c%s->%s",
		p->field_name[0],
		arg0->arg_name, 
		p->field_name+1);
    }
    else
	fprintf(f, "  return %s->%s",
		arg0->arg_name,
		p->field_name);
    /* HACK (for array fields) */
    foreach1(p->args->next, (GFunc)emit_array_field, f);
    fprintf(f, ";\n}\n\n");
}

void emit_put_method(function_t* p, FILE* f)
{
    char* access;
    type_t* arg0;

    fprintf(f, "static %s%s %s(",
	    return_type(p->return_type),
	    type_suffix(p->return_type->type),
	    p->c_name);
    foreach_last1(p->args, (GFunc)emit_arg_flag, f);
    fprintf(f, ")\n{\n");
    arg0 = (type_t*) p->args->data;
    if (p->field_name[0] == '&') {
	access = "*";
	fprintf(f, "  %s->%s",
		arg0->arg_name, 
		p->field_name+1);
    }
    else if (p->field_name[0] == '*') {
	access = "";
	fprintf(f, "  %s->%s",
		arg0->arg_name, 
		p->field_name+1);
    }
    else {
	access = "";
	fprintf(f, "  %s->%s",
		arg0->arg_name,
		p->field_name);
    }
    /* HACK (for array fields) */
    foreach_last2(p->args->next, (GFunc) emit_array_field_last, access, f);
    fprintf(f, ";\n}\n\n");
}


char* field_type(type_t* ap, int* xtype, char** xcast)
{
    GtkType type = ap->type;

    *xtype  = 0;
    *xcast  = NULL;

    switch(GTK_FUNDAMENTAL_TYPE(type)) {
    case GTK_TYPE_NONE:
	return "????";
    case GTK_TYPE_BOOL:
	return "boolean";
    case GTK_TYPE_CHAR:
	return "char";
    case GTK_TYPE_UCHAR:
	return "uchar";

    case GTK_TYPE_INT:
	if (strcmp(ap->type_name, "gint16") == 0)
	    return "short";
	else
	    return "int";

    case GTK_TYPE_LONG:
	return "long";

    case GTK_TYPE_ENUM:
	*xcast = "int";
	return "int";

    case GTK_TYPE_FLAGS:
	return "uint";
    case GTK_TYPE_ULONG:
	if (strcmp(ap->type_name, "guint32") == 0)
	    *xcast = "unsigned long";
	return "ulong";

    case GTK_TYPE_UINT:
	if (strcmp(ap->type_name, "guint16") == 0)
	    return "ushort";
	else if (strcmp(ap->type_name, "gushort") == 0)
	    return "ushort";
	return "uint";

    case GTK_TYPE_FLOAT:
	return "float";

    case GTK_TYPE_DOUBLE:
	return "double";

    case GTK_TYPE_STRING:
	return "string";

    case GTK_TYPE_BOXED:
	*xtype = 1;
	*xcast = "gpointer";
	return "boxed";

    case GTK_TYPE_POINTER:
	*xtype = 1;
	*xcast = "gpointer";
	return "pointer";

    case G_TYPE_OBJECT:
#if GTK_VERSION >= GTK_VERS(2,0,0)
    case G_TYPE_INTERFACE:
#endif
	*xcast = "GtkObject*";
	return "object";
    default:
	return "?";
    }
}


void emit_struct(struct_t* p, char* mod_name, FILE* f)
{
    char tnm[MAX_ID_BUF];
    char anm[MAX_ID_BUF];
    GSList* elem;

    gdrv_typename_to_cname(tnm, p->name);
    gdrv_typename_to_basename(anm, mod_name, p->name);

    /* encode function */
    fprintf(f, "static int SEF_%s(gdrv_mesg_t** mpp, %s* %s)\n", 
	    tnm, p->name, anm);
    fprintf(f, "{\n");
    fprintf(f, "  gdrv_encode_tuple_header(mpp, %d);\n",
	    1+p->private+p->nelems);
    fprintf(f, "  gdrv_encode_atom(mpp, \"%s\");\n", tnm);

    if (p->private)
	/* generate an extra slot for straight pointer */
	fprintf(f, "  gdrv_encode_ulong(mpp, (unsigned long) %s);\n", anm);

    for (elem = p->elems; elem != NULL; elem = elem->next) {
	type_t* ap = (type_t*) elem->data;
	int xtype;
	char* xcast;
	char* ftype;
	char xcastbuf[MAX_ID_BUF];
	char xtypebuf[MAX_ID_BUF];

	ftype = field_type(ap, &xtype,&xcast);

	if (xcast != NULL)
	    sprintf(xcastbuf, "(%s)", xcast);
	else
	    xcastbuf[0] = '\0';

	if (xtype)
	    sprintf(xtypebuf, ", %s", type_id(ap));
	else
	    xtypebuf[0] = '\0';
	
	fprintf(f, "  gdrv_encode_%s(mpp, %s(%s->%s)%s);\n",
		ftype, xcastbuf, anm, ap->arg_name, xtypebuf);
    }
    fprintf(f, "  return 0;\n}\n\n");

    /* decode function */
    if (p->private) {
	fprintf(f, "static int SDF_%s(gdrv_mesg_t* mp, gpointer* ptr)\n{\n", 
		tnm);
	/* declare a dummy */
	fprintf(f, "  %s stc_%s;\n",        p->name, anm);
	fprintf(f, "  %s* %s = &stc_%s;\n", p->name, anm, anm);
	fprintf(f, "  if (gdrv_decode_ulong(mp, (unsigned long*)ptr) < 0) return -1;\n");
    }
    else {
	fprintf(f, "static int SDF_%s(gdrv_mesg_t* mp, %s* %s)\n{\n", 
		tnm, p->name, anm);
    }

    for (elem = p->elems; elem != NULL; elem = elem->next) {
	type_t* ap = (type_t*) elem->data;
	int xtype;
	char* ftype;
	char* xcast;
	char xcastbuf[MAX_ID_BUF];
	char xtypebuf[MAX_ID_BUF];
	
	ftype = field_type(ap, &xtype, &xcast);

	if (xcast != NULL)
	    sprintf(xcastbuf, "(%s*)", xcast);
	else
	    xcastbuf[0] = '\0';

	if (xtype)
	    sprintf(xtypebuf, ", %s", type_id(ap));
	else
	    xtypebuf[0] = '\0';

	fprintf(f, "  if (gdrv_decode_%s(mp, %s(&(%s->%s))%s)<0) return -1;\n",
		ftype, xcastbuf, anm, ap->arg_name, xtypebuf);
    }
    fprintf(f, "  return 0;\n}\n\n");
}

/* generate the GTK_TYPE stuff */
char* type_id(type_t* ap)
{
    return gdrv_typename_to_typeidname(ap->type_name, 
				       ap->indirection,
				       ap->qualifier);
}


char* type_status(type_t* ap)
{
    switch(ap->status) {
    case ARG_IN:     return "ARG_IN";
    case ARG_OUT:    return "ARG_OUT";
    case ARG_INOUT:  return "ARG_INOUT";
    case ARG_ELLIPS: return "ARG_ELLIPS";
    default: return "????";
    }
}

char* erl_arg_name(char* name)
{
    static char buffer[MAX_ID_LEN];

    buffer[0] = toupper(name[0]);
    strcpy(buffer+1, name+1);
    return buffer;
}


void erl_arg(FILE* f, type_t* ap)
{
    char* arg_nm = erl_arg_name(ap->arg_name);

    switch(GTK_FUNDAMENTAL_TYPE(ap->type)) {
    case G_TYPE_OBJECT:
#if GTK_VERSION >= GTK_VERS(2,0,0)
    case G_TYPE_INTERFACE:
#endif
	fprintf(f, "%s", arg_nm);
	break;
    case GTK_TYPE_ENUM:
	fprintf(f, "{'%s', enum('%s',%s)}", 
		gdrv_type_to_erlname(ap->type), 
		gdrv_type_to_erlname(ap->type),
		arg_nm);
	break;
    default:
	fprintf(f, "{'%s',%s}", 
		gdrv_type_to_erlname(ap->type),
		arg_nm);
    }
}


char* erl_atom_name(char* name)
{
    static char buffer[MAX_ID_LEN];
    char** kptr = erl_keywords;

    while(*kptr != NULL) {
	if (strcmp(*kptr, name) == 0) {
	    sprintf(buffer, "'%s'", name);	    
	    return buffer;
	}
	kptr++;
    }
    return name;
}


void emit_export_args(type_t* t, FILE* f)
{
    fprintf(f, ",\n                 %s,%s",
	    type_id(t), type_status(t));
}

void emit_module_init(module_t* m, FILE* f)
{
    GSList* list;

    if (opt_preload)
	fprintf(f, "void %s_module_init()\n{\n", m->name);
    else
	fprintf(f, "G_MODULE_EXPORT void module_init()\n{\n");

    for (list = m->tlist; list != NULL; list = list->next) {
	typedef_t* p = (typedef_t*) list->data;
	if (p->ext) {
	    fprintf(f, "  gdrv_init_type(%s);\n", 
		    gdrv_typename_to_typeidname(p->name, 0, Q_NONE));
	}
    }

    fprintf(f, "  gdrv_type_load(type_info);\n\n");

    for (list = m->flist; list != NULL; list = list->next) {
	function_t* p = (function_t*) list->data;
	fprintf(f,
		"  gdrv_function_export(\"%s\",\"%s\",\n"
		"                 (FFunction) %s,\n"
		"                 (FMarshaller) %s,\n"
		"                 %s,\n"
		"                 %d",
		m->name,
		p->fun_name,
		p->c_name,
		marshaller(p),
		type_id(p->return_type),
		p->nargs);
	foreach1(p->args, (GFunc)emit_export_args, f);
	fprintf(f, ");\n");
    }

    for (list = m->slist; list != NULL; list = list->next) {
	struct_t* p = (struct_t*) list->data;
	char tnm[MAX_ID_BUF];
	gdrv_typename_to_cname(tnm, p->name);
	fprintf(f,
		"  gdrv_struct_export(\"%s\", \"%s\",\n"
		"                 %s, %s,\n"
		"                 sizeof(%s), %d,\n"
		"                 (FCoder) SEF_%s,\n"
		"                 (FCoder) SDF_%s);\n",
		m->name,
		tnm,
		gdrv_typename_to_typeidname(p->name,0,Q_NONE),
		(p->private ? "TRUE" : "FALSE"),
		p->name, p->nelems,
		tnm, tnm);
    }
    fprintf(f, "}\n");
}


void emit_module_exit(module_t* m, FILE* f)
{
    GSList* list;

    /* This is the dynamic entry point */
    if (opt_preload)
	fprintf(f, "void %s_module_exit()\n{\n", m->name);
    else
	fprintf(f, "G_MODULE_EXPORT void module_exit()\n{\n");

    for (list = m->flist; list != NULL; list = list->next) {
	function_t* p = (function_t*) list->data;
	fprintf(f,
		"  gdrv_function_unexport(\"%s\",\"%s\");\n",
		m->name,
		p->fun_name);
    }

    for (list = m->slist; list != NULL; list = list->next) {
	struct_t* p = (struct_t*) list->data;    
	char tnm[MAX_ID_BUF];
	gdrv_typename_to_cname(tnm, p->name);
	fprintf(f,
		"  gdrv_struct_unexport(\"%s\", \"%s\");\n",
		m->name, tnm);
    }
    fprintf(f, "}\n");
}

/* WARNING static string ! */
char* upper_str(char* src)
{
    static char dst[MAX_ID_LEN];
    char* ptr = dst;
    int    len = MAX_ID_LEN-1;

    while(len && *src != '\0') {
	*ptr++ = toupper(*src);
	src++;
	len--;
    }
    *ptr = '\0';
    return dst;
}

char* record_default_value(GtkType t)
{
    switch(GTK_FUNDAMENTAL_TYPE(t)) {
    case GTK_TYPE_BOOL:
	return "false";
    case GTK_TYPE_CHAR:
    case GTK_TYPE_UCHAR:
    case GTK_TYPE_FLAGS:
    case GTK_TYPE_ENUM:
    case GTK_TYPE_INT:
    case GTK_TYPE_UINT:
    case GTK_TYPE_LONG:
    case GTK_TYPE_ULONG:
	return "0";
    case GTK_TYPE_FLOAT:
    case GTK_TYPE_DOUBLE:
	return "0.0";
    case GTK_TYPE_STRING:
	return "\"\"";
    default:
	return NULL;
    }
}

void emit_typedefs(GSList* tlist, GSList* elist, GSList* slist, FILE* f)
{
    int n;

    FOREACH(slist, struct_t*, p,
	    {
		if ((p->type->type != 0) && p->declare) {
		    fprintf(f, "static GtkType %s = 0;\n", 
			    gdrv_ext_to_typeidname(p->name));
		}
	    });

    FOREACH(tlist, typedef_t*,  p, 
	    {
		if (!p->ext) {
		    fprintf(f, "static GtkType %s = 0;\n", 
			    gdrv_ext_to_typeidname(p->name));
		}
	    });

    n = 0;
    FOREACH(elist, enum_t*, p,
	    {
		fprintf(f, "static GtkType T%04d = 0;\n", n);
		silly_warning(p);
		n++;
	    });
    fprintf(f, "\n");

    n = 0;
    FOREACH(elist, enum_t*,  p,
	    {
		fprintf(f, "static GtkEnumValue E%04d[] = {\n", n);
		FOREACH(p->elems, enum_val_t*, v, 
			{
			    fprintf(f, "  { %s, \"%s\", NULL},\n",
				    v->id, v->id);
			});
		fprintf(f, "  { 0, NULL, NULL}};\n");
		n++;
	    });
    fprintf(f, "\n");


    fprintf(f, "static e_type_info_t type_info[] = {\n");
    FOREACH(slist, struct_t*, p,
	    {
		if (p->type->type != 0 && p->declare) {
		    fprintf(f, "  { \"%s\", &%s, GTK_TYPE_BOXED, NULL},\n",
			    p->name, 
			    gdrv_ext_to_typeidname(p->name));
		}
	    });
    FOREACH(tlist, typedef_t*,  p, 
	    {
		if (!p->ext) {
		    char* tn = (char*) gtk_type_name(p->type_def->type);
		    char* ti = gdrv_typename_to_typeidname
			(tn, p->type_def->indirection,
			 p->type_def->qualifier);
		    fprintf(f, "  { \"%s\", &%s, %s, NULL},\n",
			    p->name,
			    gdrv_ext_to_typeidname(p->name),
			    ti);
		    n++;
		}
	    });
    n = 0;
    FOREACH(elist, enum_t*,  p, 
	    {
		fprintf(f, "  { \"%s\", &T%04d, GTK_TYPE_ENUM, E%04d},\n",
			p->name, n, n);
		n++;
	    });


    fprintf(f, "  { NULL }\n");
    fprintf(f, "};\n\n");
}

void emit_access(function_t* p, FILE* f)
{
    if (p->type == ACCESS_GET)
	emit_get_method(p, f);
    else if (p->type == ACCESS_PUT)
	emit_put_method(p, f);
}

void emit_cpp(cpp_t* p, FILE* f)
{
    if (strlen(p->val) > 0)
	fprintf(f, "#%s %s %s\n\n", p->directive, p->arg, p->val);
    else
	fprintf(f, "#%s %s\n\n", p->directive, p->arg);	
}

void emit_extern(function_t* p, FILE* f)
{
    if (p->c_extern)
	fprintf(f, "extern int %s();\n", p->c_name);
}

void emit_module(module_t* module)
{
    char fname[BUFSIZ];
    FILE* erl_out      = NULL;
    FILE* c_out        = NULL;
    GSList* marshaller = NULL;

    sprintf(fname, "%s/lib%s.c", c_outdir, module->name);

    /* open C output */
    if ((c_out = fopen(fname, "w")) == NULL) {
	fprintf(stderr, "%s: could not open %s\n", prog, fname);
	perror("fopen");
	gerror++;
	return;
    }

    /* print C header stuff */
    fprintf(c_out, "/* Generated by %s (do not edit) */\n", prog);
    fprintf(c_out, "#include <gtk/gtk.h>\n");
    fprintf(c_out, "#include <gmodule.h>\n");
    fprintf(c_out, "#include <gtk_drv_api.h>\n\n");

    /* emit includes/defines etc. */
    foreach1(module->clist, (GFunc)emit_cpp, c_out);
    
    /* emit extern declarations */
    foreach1(module->flist, (GFunc)emit_extern, c_out);

    /* generate type definitions & enums */
    emit_typedefs(module->tlist, module->elist, module->slist, c_out);

    /* generate marshaller functions */
    foreach2(module->flist, (GFunc)emit_marshaller, &marshaller, c_out);

    /* generate access functions */
    foreach1(module->flist, (GFunc)emit_access, c_out);

    /* generate structure access functions */
    foreach2(module->slist, (GFunc)emit_struct, module->name, c_out);

    /* generate module_init function */
    emit_module_init(module, c_out);

    /* generate module_exit function */
    emit_module_exit(module, c_out);

    fclose(c_out);

    /* open Erlang output .inc */
    sprintf(fname, "%s/%s.inc", e_outdir, module->name);
    if ((erl_out = fopen(fname, "w")) == NULL) {
	fprintf(stderr, "%s: could not open %s\n", prog, fname);
	perror("fopen");
	gerror++;
	return;
    }

    /* generate Erlang stuff */
    fprintf(erl_out, "%% Generated by %s (do not edit)\n", prog);

    FOREACH(module->flist, function_t*, p, {
	fprintf(erl_out, "%s(", erl_atom_name(p->fun_name));
	if (p->nargs > 0) {
	    int first = 1;
	    FOREACH(p->args, type_t*, ap, {
		if (ap->status & ARG_IN) {
		    if (!first)
			fputc(',', erl_out);
		    fprintf(erl_out, "%s", erl_arg_name(ap->arg_name));
		    first = 0;
		}
	    });
	}
	fprintf(erl_out, ") ");	
	if (p->nargs > 0) {
	    int first = 1;
	    FOREACH(p->args, type_t*, ap, {
		if (ap->status & ARG_IN) {
		    int ix = ap->indirection;
		    int sz;
		    if ((ix > 0) && ((sz = ap->asize[ix-1]) >= 0)) {
			if (first)
			    fprintf(erl_out, "when ");
			else 
			    fputc(',', erl_out);
			fprintf(erl_out, "tuple(%s), ",
				erl_arg_name(ap->arg_name));
			if (sz > 0)
			    fprintf(erl_out, "size(%s) == %d",
				erl_arg_name(ap->arg_name), sz);
			first = 0;
		    }
		}
	    });
	}
	fprintf(erl_out, " ->\n");
	if (p->type == CONSTRUCTOR)
	    fprintf(erl_out, " fnew(%s, %s, [", 
		    module->name,  erl_atom_name(p->fun_name));
	else
	    fprintf(erl_out, " fcall(%s, %s, [", 
		    module->name,  erl_atom_name(p->fun_name));
	if (p->nargs > 0) {
	    int first = 1;
	    FOREACH(p->args, type_t*, ap, {
		if (ap->status & ARG_IN) {
		    if (!first)
			fputc(',', erl_out);
		    erl_arg(erl_out, ap);
		    first = 0;
		}
	    });
	}
	fprintf(erl_out, "]).\n\n");
    });

    fclose(erl_out);

    /* Generate the .hrl file (records and defs) */
    sprintf(fname, "%s/lib%s.hrl", h_outdir, module->name);
    if ((erl_out = fopen(fname, "w")) == NULL) {
	fprintf(stderr, "%s: could not open %s\n", prog, fname);
	perror("fopen");
	gerror++;
	return;
    }

    /* generate Erlang stuff */
    fprintf(erl_out, "%%\n");
    fprintf(erl_out, "%% Generated by %s (do not edit)\n", prog);
    fprintf(erl_out, "%%\n");
    fprintf(erl_out, "-ifndef(LIB%s_HRL).\n", upper_str(module->name));
    fprintf(erl_out, "-define(LIB%s_HRL, true).\n\n", upper_str(module->name));

    FOREACH(module->slist, struct_t*, sp, {
	char tnm[MAX_ID_BUF];
	gdrv_typename_to_cname(tnm, sp->name);

	fprintf(erl_out, "-record(%s,\n", erl_atom_name(tnm));
	fprintf(erl_out, "       {\n");
	if (sp->private)
	    fprintf(erl_out, "         private,\n");
	FOREACH(sp->elems, type_t*, ap, {
	    char* val = record_default_value(ap->type);
	    if (val == NULL)
		fprintf(erl_out, "         %s", ap->arg_name);
	    else
		fprintf(erl_out, "         %s = %s", ap->arg_name, val);
	    if (!IS_LAST)
		fprintf(erl_out, ",\n");
	});
	fprintf(erl_out, "\n       }).\n\n");
    });
    fprintf(erl_out, "-endif.\n");
    fclose(erl_out);
}

void emit_module_and_global(module_t* module)
{
    module_t mod;
    GSList* clist = g_slist_copy(globals->clist);
    GSList* tlist = g_slist_copy(globals->tlist);
    GSList* elist = g_slist_copy(globals->elist);
    GSList* slist = g_slist_copy(globals->slist);
    GSList* flist = g_slist_copy(globals->flist);

    /* merge module with global */
    mod.name = module->name;
    
    mod.clist  = g_slist_concat(clist, module->clist);
    mod.tlist  = g_slist_concat(tlist, module->tlist);
    mod.elist  = g_slist_concat(elist, module->elist);
    mod.slist  = g_slist_concat(slist, module->slist);
    mod.flist  = g_slist_concat(flist, module->flist);

    emit_module(&mod);
    /* destory lists */
}


void emit_def()
{
    if (modules == NULL) {
	emit_module(globals);
    }
    else {
	g_slist_foreach(modules, (GFunc)emit_module_and_global, NULL);
    }
}

/*
** Calculate all types
**
*/
void resolve_type(type_t* t)
{
    if ((t->type = lookup_type(t)) == 0) {
	fprintf(stderr, "%s: invalid type name %s\n", prog, t->type_name);
	gerror++;
    }
}

void resolve_function_type(function_t* p)
{
    resolve_type(p->return_type);
    foreach(p->args, (GFunc) resolve_type);
}

void resolve_struct_type(struct_t* p)
{
    foreach(p->elems, (GFunc)resolve_type);
}

void resolve_module_type(module_t* m)
{
    foreach(m->flist, (GFunc)resolve_function_type);
    foreach(m->slist, (GFunc)resolve_struct_type);
}

/*
** base name of a file
*/
char* basename(char* path, char* suffix, char* buf, int len)
{
    char* base;
    int n;
    int sn;

    if ((base = strrchr(path, '/')) != NULL)
	base++;
    else
	base = path;
    n = strlen(base);
    if ((suffix == NULL) || ((sn = strlen(suffix)) == 0) || (sn > n))
	n = len;
    else if (strcmp(suffix, base+(n-sn)) == 0) {
	n -= sn;  /* remove suffix */
	if (n > len)
	    n = len;
    }
    else if (strcmp(suffix, ".*") == 0) {
	char* ptr;
	if ((ptr = strrchr(base, '.')) != NULL)
	    n = ptr - base;
    }
    strncpy(buf, base, n);
    return buf;
}

/* return 1 iff suffix is the suffix of name */
int suffix(char* name, char* suffix)
{
    int n = strlen(name);
    int m = strlen(suffix);

    if (n < m)
	return 0;
    return (strcmp(name + n - m, suffix) == 0);
}

/* trim a directory path (i.e remove trailing slash) */
char* trim_slash(char* dir)
{
    int n = strlen(dir);
    while((n > 0) && (dir[n-1] == '/')) {
	dir[n-1] = '\0';
	n--;
    }
    return dir;
}

/* check if a directory exist and is writable */

int check_dir(char* dir)
{
    struct stat stbuf;

    if (stat(dir, &stbuf) < 0) {
	fprintf(stderr, "%s: can not stat %s\n", prog, dir);
	return -1;
    }
    if (!S_ISDIR(stbuf.st_mode)) {
	fprintf(stderr, "%s: %s is not a directory\n", prog, dir);
	return -1;
    }
    if (access(dir, W_OK) < 0) {
	fprintf(stderr, "%s: %s is not writeable\n", prog, dir);
	return -1;	
    }
    return 0;
}



#ifdef __WIN32__
int
WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
		    PSTR szCmdLine, int iCmdShow)
{
    int argc = __argc;
    char **argv = __argv;
#else
int
main(int argc, char **argv)
{
#endif
    int n;
    char** args;

#ifdef NEED_GNOMESUPPORT_H
    poptContext ctx;
#endif
    basename(argv[0], "", prog, sizeof(prog));

    strcpy(outdir, ".");

    modules = NULL;

#ifdef NEED_GNOMESUPPORT_H
    gnome_init_with_popt_table("def2mod", ERLGTK_VSN, argc, argv,
			 pop_options, 0, &ctx);
    args = (char**) poptGetArgs(ctx);
#else
    while(1) {
	int c;
	int option_index = 0;
	c = getopt_long (argc, argv, 
			 "vVpM:o:H:C:E:L:",
			 long_options, &option_index);	
	if (c == -1)
	    break;
	switch(c) {
	case 0:
	    printf ("option %s", long_options[option_index].name);
	    if (optarg)
		printf (" with arg %s", optarg);
	    printf ("\n");
	    break;
	case 'M':
	    if (optarg == NULL)
		usage();
	    opt_module = optarg;
	    break;
	case 'o':
	    if (optarg == NULL)
		usage();
	    opt_outdir = optarg;
	    break;
	case 'C':
	    if (optarg == NULL)
		usage();
	    opt_c_outdir = optarg;
	    break;
	case 'E':
	    if (optarg == NULL)
		usage();
	    opt_e_outdir = optarg;
	    break;	
	case 'H':
	    if (optarg == NULL)
		usage();
	    opt_h_outdir = optarg;
	    break;	
	case 'L':
	    if (optarg == NULL)
		usage();
	    opt_lib = optarg;
	    break;
	case 'v':
	    opt_verbose++;
	    break;
	case 'V':
	    printf("def2mod %s\n", ERLGTK_VSN);
	    exit(0);
	case 'p':
	    opt_preload = 1;
	    break;
	case '?':
	    usage();
	}
    }
    args = &argv[optind];
    gtk_init(&argc, &argv);
#endif

    gtk_drv_type_init();

    /* strip trailing / from outdir and normalize */
    if (opt_outdir != NULL) {
	strcpy(outdir, opt_outdir);
	trim_slash(outdir);
    }
    n = 0;
    if (opt_c_outdir == NULL)
	strcpy(c_outdir, outdir);
    else {
	n++;
	strcpy(c_outdir, opt_c_outdir);
	trim_slash(c_outdir);
	if (check_dir(c_outdir) < 0)
	    usage();
    }
    if (opt_e_outdir == NULL)
	strcpy(e_outdir, outdir);
    else {
	n++;
	strcpy(e_outdir, opt_e_outdir);
	trim_slash(e_outdir);
	if (check_dir(e_outdir) < 0)
	    usage();
    }
    if (opt_h_outdir  == NULL)
	strcpy(h_outdir, outdir);
    else {
	n++;
	strcpy(h_outdir, opt_h_outdir);
	trim_slash(h_outdir);
	if (check_dir(h_outdir) < 0)
	    usage();
    }

    if (opt_lib != NULL)
	load_lib(opt_lib);


    if (n != 3) { /* outdir used */
	if (check_dir(outdir) < 0)
	    usage();
    }

    if ((args == NULL) || (*args == NULL)) {
	if (opt_module == NULL) {
	    fprintf(stderr, "%s: need to set the module name\n", prog);
	    usage();
	}

	globals = new_module(opt_module);

	gerror = 0;

	parse_def(stdin, "*stdin*");
	if (gerror) exit(1);
	resolve_module_type(globals);
	foreach(modules, (GFunc) resolve_module_type);
	if (gerror) exit(1);
	emit_def(); 
	if (gerror) exit(1);		
    }
    else {
	while(*args != NULL) {
	    FILE* in;
	    char module_name[MAX_ID_BUF];

	    if ((in = fopen(*args, "r")) == NULL) {
		fprintf(stderr, "%s: can not open file %s\n", prog, *args);
		exit(1);
	    }

	    if (opt_module == NULL)
		basename(*args, ".*", module_name, sizeof(module_name));

	    globals = new_module(module_name);
	    gerror = 0;

	    parse_def(in, *args);
	    if (gerror) exit(1);
	    resolve_module_type(globals);
	    foreach(modules, (GFunc) resolve_module_type);
	    if (gerror) exit(1);
	    emit_def();
	    if (gerror) exit(1);

	    free_module(globals);

	    fclose(in);
	    args++;
	}
    }
    exit(0);
}

