/*
** Read and write messages from erlang
*/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <errno.h>
#include <ei.h>
#include <erl_interface.h>
#ifdef __WIN32__
#include <windows.h>
#endif
#include "e_mesg.h"


#define MESG_HDR_SZ	(sizeof(gdrv_mesg_t)-ALIGN_SZ)
#define MESG_BUF_SZ     256	/* must be dividable with ALIGN_SZ!!! */
#define MESG_FIXED_SZ	(MESG_HDR_SZ + MESG_BUF_SZ)
#define MESG_NO_FIXED	4	/* number of fixed messages */

/* Glib mutex stuff */
#ifdef USE_THREADS
GThreadFunctions thread_functions;
#define THREAD_INIT  g_thread_init(&thread_functions)
#define LOCK_T       GMutex*
#define LockInit(m)  *(m) = thread_functions.mutex_new()
#define Lock(mp)     thread_functions.mutex_lock(*(mp))
#define Unlock(mp)   thread_functions.mutex_unlock(*(mp))
#else
#define THREAD_INIT
#define LOCK_T  int
#define LockInit(m)
#define Lock(m)
#define Unlock(m)
#endif

#ifdef USE_THREADS
static LOCK_T        message_lock;
#endif
static gdrv_mesg_t*  message_pool;
static char          mesg_buffer[MESG_NO_FIXED*MESG_FIXED_SZ];

extern int debugf(FILE*, char*, ...);
extern FILE* logf;

void gdrv_mesg_init()
{
    char* ptr = mesg_buffer;
    gdrv_mesg_t* mesg;
    int i;

    THREAD_INIT;

    LockInit(&message_lock);
    message_pool = NULL;
    for (i = 0; i < MESG_NO_FIXED; i++) {
	mesg = (gdrv_mesg_t*) ptr;
	ptr += MESG_FIXED_SZ;
	mesg->next = message_pool;
	mesg->alen = MESG_BUF_SZ; /* dont touch */
	mesg->dynamic = 0;        /* dont touch */
	mesg->len = 0;
	mesg->ix  = 0;
	message_pool = mesg;
    }
}

gdrv_mesg_t* gdrv_mesg_new(gint len)
{
    gdrv_mesg_t* mesg = NULL;

    if (len <= MESG_BUF_SZ) {
	Lock(&message_lock);
	if (message_pool != NULL) {
	    mesg = message_pool;
	    message_pool = mesg->next;
	}
	Unlock(&message_lock);
    }
    if (mesg == NULL) {
	mesg = (gdrv_mesg_t*) g_malloc(MESG_HDR_SZ+len);
	if (mesg == NULL)
	    return NULL;
	mesg->dynamic = 1;
	mesg->alen = len;
    }
    mesg->len = len;
    mesg->ix  = 0;
    return mesg;
}

void gdrv_mesg_free(gdrv_mesg_t* mesg)
{
    if (mesg->dynamic)
	g_free((gpointer) mesg);
    else {
	Lock(&message_lock);
	mesg->next = message_pool;
	message_pool = mesg;
	Unlock(&message_lock);
    }
}

gdrv_mesg_t* gdrv_mesg_realloc(gdrv_mesg_t* mesg, gint len)
{
    if (len > mesg->alen) {
	if (mesg->dynamic || (len > MESG_BUF_SZ)) {
	    gdrv_mesg_t* m = gdrv_mesg_new(len);
	    memcpy(m->data, mesg->data, mesg->len);
	    m->ix = mesg->ix;
	    gdrv_mesg_free(mesg);
	    return m;
	}
    }
    mesg->len = len;
    return mesg;
}

static int e_read(GIOChannel* source, char* buf, int len, GIOError* err)
{
    int i, got=0;
    GIOError r;

    if (len == 0) {
	*err = G_IO_ERROR_NONE;
	return 0;
    }

    do {
	r = g_io_channel_read(source, buf+got, len-got, &i);
	if (r != G_IO_ERROR_NONE) {
	    *err = r;
	    return -1;
	}
	if (i == 0)
	    break;
	got += i;
    } while (got < len);

    *err = G_IO_ERROR_NONE;
    return got;
}

static int e_write(GIOChannel* source, char* buf, int len, GIOError* err)
{
    int i,done=0;
    GIOError r;

    if (len == 0) {
	*err = G_IO_ERROR_NONE;
	return 0;
    }

    do {
	r = g_io_channel_write(source, buf+done, len-done, &i);
	if (r != G_IO_ERROR_NONE) {
	    *err = r;
	    return -1;
	}
	done += i;
    } while (done < len);
    *err = G_IO_ERROR_NONE;
    return len;
}

gdrv_mesg_t* gdrv_mesg_recv(GIOChannel* source, GIOError* err)
{
    gdrv_mesg_t* mesg;
    int n;
    int len;
    guint8 buf[4];

    if ((n = e_read(source, (char*)buf, 4, err)) < 4)
	return NULL;
    n = GET_INT32(buf);
    mesg = gdrv_mesg_new(n);
    if ((len = e_read(source, mesg->data, n, err)) < n) {
	gdrv_mesg_free(mesg);
	return NULL;
    }
    return mesg;
}

int gdrv_mesg_send(GIOChannel* dest, gdrv_mesg_t* mesg, GIOError* err)
{
    int len = mesg->ix;
    int n;

    PUT_INT32(mesg->hdr, len);
    len += 4;
    if ((n = e_write(dest, mesg->hdr, len, err)) != len)
	return -1;
    return len;
}

#define MSG_EXPAND(m, n) do { \
  if ( ((*(m))->ix + (n)) > (*(m))->alen) { \
     *(m) = gdrv_mesg_realloc(*(m), (*(m))->ix + (n) + 256); \
   } \
 } while(0)

/* ENCODE */
int gdrv_encode_version(gdrv_mesg_t** mpp)
{
    MSG_EXPAND(mpp, 2);
    return ei_encode_version((*mpp)->data, &(*mpp)->ix);
}

int gdrv_encode_tuple_header(gdrv_mesg_t** mpp, int x)
{
    MSG_EXPAND(mpp, 5);
    return ei_encode_tuple_header((*mpp)->data, &(*mpp)->ix, x);
}

int gdrv_encode_list_header(gdrv_mesg_t** mpp, int x)
{
    MSG_EXPAND(mpp, 5);
    return ei_encode_list_header((*mpp)->data, &(*mpp)->ix, x);
}

int gdrv_encode_empty_list(gdrv_mesg_t** mpp)
{
    MSG_EXPAND(mpp, 1);
    return ei_encode_empty_list((*mpp)->data, &(*mpp)->ix);
}

int gdrv_encode_atom(gdrv_mesg_t** mpp, char* x)
{
    int n = strlen(x);
    MSG_EXPAND(mpp, 3+n);
    return ei_encode_atom((*mpp)->data, &(*mpp)->ix, x);
}

int gdrv_encode_char(gdrv_mesg_t** mpp, char x)
{
    MSG_EXPAND(mpp, 2);
    return ei_encode_char((*mpp)->data, &(*mpp)->ix, x);
}

int gdrv_encode_uchar(gdrv_mesg_t** mpp, unsigned char x)
{
    MSG_EXPAND(mpp, 2);
    return ei_encode_char((*mpp)->data, &(*mpp)->ix, (char) x);
}

int gdrv_encode_short(gdrv_mesg_t** mpp, short x)
{
    long xl = x;
    MSG_EXPAND(mpp, 7);
    return ei_encode_long((*mpp)->data, &(*mpp)->ix, xl);
}

int gdrv_encode_ushort(gdrv_mesg_t** mpp, unsigned short x)
{
    unsigned long xl = x;
    MSG_EXPAND(mpp, 7);
    return ei_encode_ulong((*mpp)->data, &(*mpp)->ix, xl);
}

int gdrv_encode_int(gdrv_mesg_t** mpp, int x)
{
    long xl = x;
    MSG_EXPAND(mpp, 7);
    return ei_encode_long((*mpp)->data, &(*mpp)->ix, xl);
}

int gdrv_encode_uint(gdrv_mesg_t** mpp, unsigned int x)
{
    unsigned long xl = x;
    MSG_EXPAND(mpp, 7);
    return ei_encode_ulong((*mpp)->data, &(*mpp)->ix, xl);
}

int gdrv_encode_ulong(gdrv_mesg_t** mpp, unsigned long x)
{
    MSG_EXPAND(mpp, 7);
    return ei_encode_ulong((*mpp)->data, &(*mpp)->ix, x);
}

int gdrv_encode_long(gdrv_mesg_t** mpp, long x)
{
    MSG_EXPAND(mpp, 7);
    return ei_encode_ulong((*mpp)->data, &(*mpp)->ix, x);
}

int gdrv_encode_double(gdrv_mesg_t** mpp, double x)
{
    MSG_EXPAND(mpp, 32);
    return ei_encode_double((*mpp)->data, &(*mpp)->ix, x);
}

int gdrv_encode_float(gdrv_mesg_t** mpp, float x)
{
    MSG_EXPAND(mpp, 32);
    return ei_encode_double((*mpp)->data, &(*mpp)->ix, 
			    (double) x);
}


int gdrv_encode_term(gdrv_mesg_t** mpp, ETERM* x)
{
    int n = erl_term_len(x);
    MSG_EXPAND(mpp, n);
    return ei_encode_term((*mpp)->data, &(*mpp)->ix, x);
}

int gdrv_encode_string(gdrv_mesg_t** mpp, char* x)
{
    if (x == NULL)
	return gdrv_encode_empty_list(mpp);
    else {
	int n = 0;
	if (ei_encode_string(NULL, &n, x) < 0)
	    return -1;
	MSG_EXPAND(mpp, n);
	return ei_encode_string((*mpp)->data, &(*mpp)->ix, x); 
    }
}

int gdrv_encode_boolean(gdrv_mesg_t** mpp, int x)
{
    int n = 0;
    if (ei_encode_boolean(NULL, &n, x) < 0)
	return -1;
    MSG_EXPAND(mpp, n);
    return ei_encode_boolean((*mpp)->data, &(*mpp)->ix, x);
}

int gdrv_encode_binary(gdrv_mesg_t** mpp, char* x, int n)
{
    MSG_EXPAND(mpp, 5+n);
    return ei_encode_binary((*mpp)->data, &(*mpp)->ix, x, n);    
}

/* DECODE */

/* decode type/len info but do not advance pointer */
int gdrv_decode_type(gdrv_mesg_t* mp, int* type, int* len)
{
    int ix = mp->ix;

    return ei_get_type(mp->data, &ix, type, len);
}

int gdrv_decode_version(gdrv_mesg_t* mp, int* vers)
{
    return ei_decode_version(mp->data, &mp->ix, vers);
}

int gdrv_decode_tuple_header(gdrv_mesg_t* mp, int* x)
{
    return ei_decode_tuple_header(mp->data, &mp->ix, x);
}

int gdrv_decode_list_header(gdrv_mesg_t* mp, int* x)
{
    return ei_decode_list_header(mp->data, &mp->ix, x);
}

int gdrv_decode_empty_list(gdrv_mesg_t* mp)
{
    int type;
    int len;

    ei_get_type(mp->data, &mp->ix, &type, &len);
    if (type == ERL_NIL_EXT) {
	mp->ix++;
	return 0;
    }
    return -1;
}

int gdrv_decode_atom(gdrv_mesg_t* mp, char* x, int max_len)
{
    /* FIXME: check max_len */
    return ei_decode_atom(mp->data, &mp->ix, x);
}

int gdrv_decode_char(gdrv_mesg_t* mp, char* x)
{
    return ei_decode_char(mp->data, &mp->ix, x);
}

int gdrv_decode_uchar(gdrv_mesg_t* mp, unsigned char* x)
{
    return ei_decode_char(mp->data, &mp->ix, (char*)x);
}

int gdrv_decode_short(gdrv_mesg_t* mp, short* x)
{
    long xl;
    if (ei_decode_long(mp->data, &mp->ix, &xl) < 0)
	return -1;
    *x = (short) xl;
    return 0;
}

int gdrv_decode_ushort(gdrv_mesg_t* mp, unsigned short* x)
{
    unsigned long xl;
    if (ei_decode_ulong(mp->data, &mp->ix, &xl) < 0)
	return -1;
    *x = (unsigned short) xl;
    return 0;
}

int gdrv_decode_int(gdrv_mesg_t* mp, int* x)
{
    long xl;
    if (ei_decode_long(mp->data, &mp->ix, &xl) < 0)
	return -1;
    *x = (int) xl;
    return 0;
}

int gdrv_decode_uint(gdrv_mesg_t* mp, unsigned int* x)
{
    unsigned long xl;
    if (ei_decode_ulong(mp->data, &mp->ix, &xl) < 0)
	return -1;
    *x = (unsigned int) xl;
    return 0;
}

int gdrv_decode_ulong(gdrv_mesg_t* mp, unsigned long* x)
{
    return ei_decode_ulong(mp->data, &mp->ix, x);
}

int gdrv_decode_long(gdrv_mesg_t* mp, long* x)
{
    return ei_decode_long(mp->data, &mp->ix, x);
}

int gdrv_decode_double(gdrv_mesg_t* mp, double* x)
{
    return ei_decode_double(mp->data, &mp->ix, x);
}

int gdrv_decode_float(gdrv_mesg_t* mp, float* x)
{
    double fl;

    if (ei_decode_double(mp->data, &mp->ix, &fl) < 0)
	return -1;
    *x = (float) fl;
    return 0;
}

int gdrv_decode_boolean(gdrv_mesg_t* mp, int* x)
{
    return ei_decode_boolean(mp->data, &mp->ix, x);
}

int gdrv_decode_term(gdrv_mesg_t* mp, ETERM** x)
{
    return ei_decode_term(mp->data, &mp->ix, (void*) x);
}

/* accept atom lists, and string */
int gdrv_decode_strn(gdrv_mesg_t* mp, char* x, int max_len)
{
    int type;
    int len;

    ei_get_type(mp->data, &mp->ix, &type, &len);
    if (len >= max_len)
	return -1;
    if (type == ERL_ATOM_EXT)
	return ei_decode_atom(mp->data, &mp->ix, x);
    else
	return ei_decode_string(mp->data, &mp->ix, x);
}


int gdrv_decode_binary(gdrv_mesg_t* mp, void* x, long* len)
{
    return ei_decode_binary(mp->data, &mp->ix, x, len);
}


