/*
** gtk driver
** author: Tony Rogvall <tony@bluetail.com>
*/
#include <sys/types.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <memory.h>
#include <gtk/gtk.h>
#ifdef NEED_GNOMESUPPORT_H
#include <gnome.h>
#endif
#include <gmodule.h>

#ifdef __WIN32__
#include <windows.h>
#define CLOSESOCKET(s) closesocket(s)
#else
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/socket.h>
#define CLOSESOCKET(s) close(s)
#define SOCKET_ERROR   -1
#define SOCKET         int
#define INVALID_SOCKET -1
extern int close(int);
#endif

#include <ei.h>
#include <erl_interface.h>
/* this #define is missing in new erl_interface (3.4): */
#ifndef ERL_MAX_COUNT
#define ERL_MAX_COUNT     0xffffff
#endif
#include <gtk_drv_api.h>

#include "gtk_drv_arg.h"
#include "e_mesg.h"



#ifdef PRELOADED
extern void gdk_module_init();
extern void gtk_module_init();
#endif

extern void  gtk_drv_type_init(void);

static int encode_fundamental(gdrv_mesg_t** mpp, gchar* ptr, GtkType type);
static int decode_gdk_point(gdrv_mesg_t* mp, GdkPoint* p);
static int decode_gdk_segment(gdrv_mesg_t* mp, GdkSegment* seg);
int gdrv_encode_arg(gdrv_mesg_t** mpp, GtkArg* arg);

#ifdef DEBUG
#define DEBUG_TO_STDERR

FILE* logf = NULL;

int debug(char* fmt, ...)
{
  va_list va;
  char buffer[1024];
  int n;

  va_start(va, fmt);
  n = vsprintf(buffer, fmt, va);
  va_end(va);
  if (n < 0)
    n = strlen(buffer);
  fwrite(buffer, n, sizeof(char), logf);
  return n;
}

#define DEBUGF(code) debug code

#else

#define DEBUGF(code)

#endif


#define MAX_ARG_STACK  0x8000  /* max # of long words (128K) */

#define ETERM_REF(e) do { \
  if (ERL_COUNT(e) < ERL_MAX_COUNT) \
      ERL_COUNT(e)++; \
 } while(0)

#define E_GTK_TYPE_ENUM_FIND_VALUE  0x0002
#define E_GTK_TYPE_FLAGS_FIND_VALUE 0x0003
#define E_GTK_ARG_GET_TYPE          0x0004
#define E_GTK_SIGNAL_EMIT           0x0009
#define E_GTK_SIGNAL_CONNECT        0x000A
#define E_GTK_CHILD_ARG_GET_TYPE    0x000D
#define E_GTK_ASYNC_SIGNAL_EMIT     0x000E
#define E_GTK_REPLY                 0x000F
#define E_GTK_FUNCTION_CALL         0x0010
#define E_GTK_TYPE_IS_A             0x0011
#define E_GTK_NEW_CALL              0x0012
#define E_GTK_ASYNC_CALL            0x0013
#define E_GTK_MODULE_DIR            0x0014
#define E_GTK_LOAD_MODULE           0x0015
#define E_GTK_UNLOAD_MODULE         0x0016
#define E_GTK_FIND_FUNCTION         0x0017

#define NAME_SIZE 256

/*
** Closure needed for signal emission to erlang
*/
typedef struct _e_closure_t {
    ETERM*     fun;           /* Erlang function */
    ETERM*     data;          /* Data - Last argument */
    ETERM*     pid;           /* Reciver - widget owner */
    GtkObject* object;        /* The connected object */
    int        object_signal; /* This is an object signal */
    int        msg_size;      /* external size of fun+data+pid */
    GtkType    return_type;   /* Signal return type */
    guint      n_params;      /* Number of arguments */
    GtkType    param_type[1]; /* Type of arguments */
} e_closure_t;

/*
** reply structure for callbacks & events
** reply will be a boolean
*/
typedef struct _e_reply_t {
    struct _e_reply_t* next;
    unsigned long id;
    gint value;
    gboolean ready;
} e_reply_t;


typedef struct _e_module_t {
    struct _e_module_t* next;
    gchar* mod_name;   /* short name, long name found in g_module_name(mod) */
    void (*initfn)();    /* init function */
    void (*exitfn)();    /* exit function */
    GModule* mod;
} e_module_t;


typedef struct _e_export_t {
    const char* mod_name;
    const char* fun_name;
    GQuark quark;            /* for the name mod:fun */
    FFunction function;
    FMarshaller marshaller;
    GtkType return_type;
    int arity;
    e_arg_t arg_type[1];
} e_export_t;


typedef struct _e_struct_t {
    const char* mod_name;
    const char* struct_name;
    GtkType     type;        /* xxx_TYPE_yyy */
    int         private;     /* 1 iff first has straight pointer */
    size_t      size;        /* sizeof(C-struct) */
    int         arity;       /* number of elements */
    GQuark quark;            /* for the name mod.struct */
    FCoder encode;           /* encoder */
    FCoder decode;           /* decoder */
} e_struct_t;


static GHashTable *exp_hash_table = NULL;
static GHashTable *stc_hash_table = NULL;
static e_module_t* module_list = NULL;

/*
** FIXME: make module_dir a path list instead 
** Dir1:Dir2:Dir...
*/
static char module_dir[BUFSIZ];

/*
** e_pmap is a pointer map beteen erlang and gtk 
** the Erlang side will always see the pointer to the epmap entry.
** The epmap entry is never free'd so that a pointer to the epmap
** entry is always valid (but the content may be NULL)
*/

typedef struct _pmap_t {
    struct _pmap_t* next;
    GtkType   type;        /* type of pointer */
    gpointer  ptr;         /* real object pointer */
} pmap_t;


static pmap_t* gtk_object_map(GtkObject* obj);

GIOChannel* erl_fd_in;
GIOChannel* erl_fd_out;

static unsigned long e_reply_id = 0;

static e_reply_t* e_reply_stack = NULL;

/* the e_ptr_free forms a circular list of epmap_t entries
** new elements are taken from the "end" in a queue like fashion
** (a stack could yield more race conditions!?)
** The pmap entry is added to the GtkObject's user data field
** with the name "pmap"
*/ 
static pmap_t*     pmap_free  = NULL;
static GHashTable* pmap_table = NULL;

/*
** Temporary storage for argument data
** and vectors
*/

static struct {
    char* e_name;  /* event name */
    char* r_name;  /* record name */
    int   r_len;   /* record length (arity) */
} erl_ev[32] = {
    { "GDK_DELETE",            "gdk_event_any",        4},
    { "GDK_DESTROY",           "gdk_event_any",        4},
    { "GDK_EXPOSE",            "gdk_event_expose",     6},
    { "GDK_MOTION_NOTIFY",     "gdk_event_motion",     16},
    { "GDK_BUTTON_PRESS",      "gdk_event_button",     16},
    { "GDK_2BUTTON_PRESS",     "gdk_event_button",     16},
    { "GDK_3BUTTON_PRESS",     "gdk_event_button",     16},
    { "GDK_BUTTON_RELEASE",    "gdk_event_button",     16},
    { "GDK_KEY_PRESS",         "gdk_event_key",        9},
    { "GDK_KEY_RELEASE",       "gdk_event_key",        9},
    { "GDK_ENTER_NOTIFY",      "gdk_event_crossing",   14},
    { "GDK_LEAVE_NOTIFY",      "gdk_event_crossing",   14},
    { "GDK_FOCUS_CHANGE",      "gdk_event_focus",      5},
    { "GDK_CONFIGURE",         "gdk_event_configure",  8},
    { "GDK_MAP",               "gdk_event_any",        4},
    { "GDK_UNMAP",             "gdk_event_any",        4},
    { "GDK_PROPERTY_NOTIFY",   "gdk_event_property",   7},
    { "GDK_SELECTION_CLEAR",   "gdk_event_selection",  9},
    { "GDK_SELECTION_REQUEST", "gdk_event_selection",  9},
    { "GDK_SELECTION_NOTIFY",  "gdk_event_selection",  9},
    { "GDK_PROXIMITY_IN",      "gdk_event_proximity",  7},
    { "GDK_PROXIMITY_OUT",     "gdk_event_proximity",  7},
    { "GDK_DRAG_ENTER",        "gdk_event_dnd",        8},
    { "GDK_DRAG_LEAVE",        "gdk_event_dnd",        8},
    { "GDK_DRAG_MOTION",       "gdk_event_dnd",        8},
    { "GDK_DRAG_STATUS",       "gdk_event_dnd",        8},
    { "GDK_DROP_START",        "gdk_event_dnd",        8},
    { "GDK_DROP_FINISHED",     "gdk_event_dnd",        8},
    { "GDK_CLIENT_EVENT",      "gdk_event_client",     7}, 
    { "GDK_VISIBILITY_NOTIFY", "gdk_event_visibility", 5},
    { "GDK_NO_EXPOSE",         "gdk_event_no_expose",  4},
    { "GDK_SCROLL",            "gdk_event_scroll",     12}
    
  };



/* WARNING GTK INTERNAL STUFF (version 1.2.8) (not in 1.3 :-( ) */

struct _GtkSignal
{
  guint		      signal_id;
  GtkType	      object_type;
  gchar		     *name;
  guint		      function_offset;
  GtkSignalMarshaller marshaller;
  GtkType	      return_val;
  guint		      signal_flags : 16;
  guint		      nparams : 16;
  GtkType	     *params;
  GHookList	     *hook_list;
};

#if 0
extern struct _GtkSignal  *_gtk_private_signals;
extern guint	           _gtk_private_n_signals;

#define LOOKUP_SIGNAL_ID(signal_id)	( \
  signal_id > 0 && signal_id < _gtk_private_n_signals ? \
    (struct _GtkSignal*) _gtk_private_signals + signal_id : \
    (struct _GtkSignal*) 0 \
)

/* get the event method call 'name' */
GtkSignalFunc gtk_get_event_method(GtkObject* object, gchar* name)
{
    guint signal_id = gtk_signal_lookup(name, GTK_OBJECT_TYPE(object));
    struct _GtkSignal* sig = LOOKUP_SIGNAL_ID(signal_id);

    if (sig == NULL)
	return NULL;
    else if (sig->function_offset)
	return G_STRUCT_MEMBER (GtkSignalFunc, object->klass,
				sig->function_offset);
    else
	return NULL;
}
#endif


void gtk_x_version(guint* major, guint* minor, guint* micro)
{
    *major = gtk_major_version;
    *minor = gtk_minor_version;
    *micro = gtk_micro_version;
}


/* speical wrapper with Erlang data */
void gtk_x_object_set_user_data(GtkObject* object, ETERM* term)
{
    if (term != NULL)
	ETERM_REF(term);
    gtk_object_set_data_full(object, "user_data", 
			     (gpointer) term,
			     (GtkDestroyNotify) erl_free_term);
}

ETERM* gtk_x_object_get_user_data(GtkObject* object)
{
    ETERM* term = gtk_object_get_data(object, "user_data");
    if (term != NULL)
	ETERM_REF(term);
    return term;
}

/* speical wrapper with Erlang data */
void gtk_x_object_set_data(GtkObject* object, char* key, ETERM* term)
{
    if (term != NULL)
	ETERM_REF(term);
    gtk_object_set_data_full(object, key, 
			     (gpointer) term,
			     (GtkDestroyNotify) erl_free_term);
}

ETERM* gtk_x_object_get_data(GtkObject* object, char* key)
{
    ETERM* term = (ETERM*) gtk_object_get_data(object, key);
    if (term != NULL)
	ETERM_REF(term);
    return term;
}

GtkWidgetAuxInfo* gtk_widget_get_aux_info(GtkWidget* widget)
{
    return gtk_object_get_data(GTK_OBJECT(widget), "gtk-aux-info");
}

void gtk_widget_set_flags(GtkWidget* widget, guint flags)
{
    GTK_WIDGET_SET_FLAGS(widget, flags);
}

void gtk_widget_unset_flags(GtkWidget* widget, guint flags)
{
    GTK_WIDGET_SET_FLAGS(widget, flags);
}

guint gtk_widget_get_flags(GtkWidget* widget)
{
    return GTK_WIDGET_FLAGS(widget);
}

guint gtk_widget_get_state(GtkWidget* widget)
{
    return GTK_WIDGET_STATE(widget);
}

guint gtk_widget_get_saved_state(GtkWidget* widget)
{
    return GTK_WIDGET_SAVED_STATE(widget);
}

void gtk_object_set_flags(GtkObject* object, guint flags)
{
    GTK_OBJECT_SET_FLAGS(object, flags);
}

void gtk_object_unset_flags(GtkObject* object, guint flags)
{
    GTK_OBJECT_SET_FLAGS(object, flags);
}

guint gtk_object_get_flags(GtkObject* object)
{
    return GTK_OBJECT_FLAGS(object);
}

static void prepend_widget(GtkWidget *widget,gpointer client_data)
{
    GList **children = (GList**) client_data;
    ETERM* v[2];
    pmap_t* eptr = gtk_object_map(GTK_OBJECT(widget));
    v[0] = erl_mk_uint((unsigned int) GTK_OBJECT_TYPE(widget));
    v[1] = erl_mk_uint((unsigned int) eptr);
    *children = g_list_prepend (*children, erl_mk_tuple(v, 2));
}


ETERM* gtk_x_container_children(GtkContainer* container)
{
    GList* children = NULL;
    ETERM* ep;

    gtk_container_foreach(container, prepend_widget, (gpointer)&children);
    ep = erl_mk_empty_list();
    while(children) {
	GList* link = children;
	ep = erl_cons((ETERM*)link->data, ep);
	children = g_list_remove_link(children, link);
    }
    return ep;
}

/*
** Allocate a signal closure
*/
static e_closure_t* create_signal_closure(guint signal_id)
{
    e_closure_t* cp;
    int i;
    int n;
#if GTK_VERSION >= GTK_VERS(1,3,2)
    GSignalQuery query;

    g_signal_query(signal_id, &query);
    if (query.signal_id == 0)
	return NULL;
    n = query.n_params;
    if ((cp = (e_closure_t*) g_malloc(sizeof(e_closure_t)+
				      n*sizeof(GtkType))) == NULL)
	return NULL;
    cp->return_type = query.return_type;
    cp->n_params = n;
    for (i = 0; i < n; i++)
	cp->param_type[i] = query.param_types[i];
#else
    GtkSignalQuery* query;

    if ((query = gtk_signal_query(signal_id)) == NULL)
	return NULL;
    n = query->nparams;
    if ((cp = (e_closure_t*) g_malloc(sizeof(e_closure_t)+
				      n*sizeof(GtkType))) == NULL)
	return NULL;
    cp->return_type = query->return_val;
    cp->n_params = n;
    for (i = 0; i < n; i++)
	cp->param_type[i] = query->params[i];
    g_free(query);
#endif

    return cp;
}



static pmap_t* pmap_create(gpointer ptr, GtkType type)
{
    pmap_t* fptr = pmap_free;
    pmap_t* eptr;

    if (fptr == NULL)
	eptr = (pmap_t*) g_malloc(sizeof(pmap_t));
    else {
	eptr = fptr->next;
	if (eptr == fptr)
	    pmap_free = NULL;
	else
	    fptr->next = eptr->next;
    }
    eptr->next = NULL;
    eptr->type = type;
    eptr->ptr = ptr;
    g_hash_table_insert(pmap_table, ptr, (gpointer) eptr);

    DEBUGF(("PMAP CREATE %s: %ld -> %ld\r\n",
	    gtk_type_name(type), 
	    (unsigned long)eptr->ptr, 
	    (unsigned long)eptr));
    return eptr;
}

#if 0
static pmap_t* pmap_lookup(gpointer ptr)
{
    return (pmap_t*) g_hash_table_lookup(pmap_table, ptr);
}

static void pmap_remove(gpointer ptr)
{
    pmap_t* eptr = (pmap_t*) g_hash_table_lookup(pmap_table, ptr);
    
    if (eptr != NULL) {
	DEBUGF(("PMAP REMOVE %s: %ld -> %ld\r\n",
		gtk_type_name(eptr->type),
		(unsigned long)eptr->ptr,
		(unsigned long)eptr));
	if (pmap_free == NULL)
	    eptr->next = eptr;
	else {
	    eptr->next = pmap_free->next;
	    pmap_free->next = eptr;
	}
	pmap_free = eptr;
	g_hash_table_remove(pmap_table, ptr);
	eptr->ptr = NULL;
    }
}
#endif


static void pmap_destroy(pmap_t* eptr)
{
    DEBUGF(("PMAP DESTROY %s: %ld -> %ld\r\n",
	    gtk_type_name(eptr->type),
	    (unsigned long)eptr->ptr,
	    (unsigned long)eptr));
    if (pmap_free == NULL)
	eptr->next = eptr;
    else {
	eptr->next = pmap_free->next;
	pmap_free->next = eptr;
    }
    pmap_free = eptr;
    g_hash_table_remove(pmap_table, eptr->ptr);
    eptr->ptr = NULL;
}


/* GdkDrawble/GdkWindow/GdkPixmap/GdkBitmap/GdkRegion */
static pmap_t* dataset_map(gpointer* obj, GtkType type)
{
    pmap_t* eptr;

    DEBUGF(("DATASET_MAP \r\n"));

    if ((eptr = g_dataset_get_data(obj, "pmap")) != NULL) {
	if (eptr->type == type)
	    return eptr;
	return NULL;
    }
    eptr = pmap_create((gpointer) obj, type);
    g_dataset_set_data_full (obj, "pmap", (gpointer)eptr, 
			     (GDestroyNotify) pmap_destroy);
    return eptr;
}

static pmap_t* gtk_object_map(GtkObject* obj)
{
    pmap_t* eptr;

#if GTK_VERSION >= GTK_VERS(1,3,2)
    GQuark quark = g_quark_from_string("pmap");

    DEBUGF(("DATASET_OBJECT_MAP \r\n"));

    if ((eptr = (pmap_t*) g_object_get_qdata((GObject*)obj, quark)) != NULL)
	return eptr;
    eptr = pmap_create((gpointer) obj, GTK_OBJECT_TYPE(obj));
    g_object_set_qdata_full((GObject*)obj, quark, eptr,
			    (GtkDestroyNotify) pmap_destroy);
#else
    DEBUGF(("DATASET_OBJECT_MAP \r\n"));

    if ((eptr = (pmap_t*) gtk_object_get_data(obj, "pmap")) != NULL)
	return eptr;
    eptr = pmap_create((gpointer) obj, GTK_OBJECT_TYPE(obj));
    gtk_object_set_data_full(obj, "pmap", eptr,
			     (GtkDestroyNotify) pmap_destroy);
#endif
    return eptr;
}

/****************************************************************************
**
** Argument stack operations
**
*****************************************************************************/

typedef union {
    double        d;
    unsigned long l[2];
    char          c[8];
} stack_elem_t;

static stack_elem_t arg_stack_bot[MAX_ARG_STACK];
static stack_elem_t* arg_sp;
static stack_elem_t* arg_stack_top;

#define ARG_STACK_SAVE(location) *(location) = arg_sp
#define ARG_STACK_RESTORE(location) arg_sp = *(location)

#define ARG_STACK_ALLOC(size) \
    arg_stack_alloc(((size)+sizeof(stack_elem_t)-1)/sizeof(stack_elem_t))

static void arg_stack_init()
{
    arg_sp = arg_stack_bot;
    arg_stack_top = arg_stack_bot + MAX_ARG_STACK;
}

static void* arg_stack_alloc(int n)
{
    stack_elem_t* mem = arg_sp;
    stack_elem_t* arg_sp_next = mem + n;
    if (arg_sp_next > arg_stack_top) {
	DEBUGF(("gtk_drv: fatal error, out of stack space\r\n"));
	return NULL;
    }
    arg_sp = arg_sp_next;
    return (void*) mem;
}


static char* arg_stack_strdup(char* str)
{
    int n = strlen(str);
    char* dup = (char*) ARG_STACK_ALLOC(n + 1);
    strcpy(dup, str);
    return dup;
}


void* gdrv_stack_alloc(int size)
{
    return ARG_STACK_ALLOC(size);
}

/****************************************************************************
** Export table 
*****************************************************************************/

static GQuark exp_try_quark(const gchar* mod_name, 
			    const char qtype,
			    const gchar* name)
{
    gchar str[255+255+2];
    int len = strlen(mod_name);

    strncpy(str, mod_name, len);
    str[len] = qtype;
    strcpy(str+len+1, name);
    return g_quark_try_string(str);
}


static GQuark exp_mk_quark(const gchar* mod_name, 
			   const char qtype,
			   const gchar* name)
{
    gchar str[255+255+2];
    int len = strlen(mod_name);

    strncpy(str, mod_name, len);
    str[len] = qtype;
    strcpy(str+len+1, name);
    return g_quark_from_string(str);
}

static e_export_t* exp_lookup(const gchar* mod_name, const gchar* func_name)
{
    GQuark quark = exp_try_quark(mod_name, ':', func_name);

    if (quark)
	return (e_export_t*) g_hash_table_lookup(exp_hash_table,
						 GUINT_TO_POINTER(quark));
    return NULL;
}

static e_struct_t* struct_lookup_by_name(const gchar* struct_name)
{
    GQuark quark = exp_try_quark("", '#', struct_name);
    
    if (quark)
	return (e_struct_t*) g_hash_table_lookup(exp_hash_table,
						 GUINT_TO_POINTER(quark));
    return NULL;
}

static e_struct_t* struct_lookup_by_type(GtkType type)
{
    return (e_struct_t*) g_hash_table_lookup(stc_hash_table,
					     GUINT_TO_POINTER(type));
}

#if 0
static int arg_set_return_pointer(GtkArg* loc, GtkType rtype, GtkArg* ret)
{
    switch(GTK_FUNDAMENTAL_TYPE(rtype)) {
    case GTK_TYPE_INVALID:
    case GTK_TYPE_NONE:
	GTK_VALUE_POINTER(*loc) = NULL;
	break;
    case GTK_TYPE_CHAR:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_CHAR(*ret);
	break;
    case GTK_TYPE_UCHAR:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_UCHAR(*ret);
	break;
    case GTK_TYPE_BOOL:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_BOOL(*ret);
	break;
    case GTK_TYPE_INT:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_INT(*ret);
	break;
    case GTK_TYPE_UINT:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_UINT(*ret);
	break;
    case GTK_TYPE_LONG:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_LONG(*ret);
	break;
    case GTK_TYPE_ULONG:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_ULONG(*ret);
	break;
    case GTK_TYPE_FLOAT:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_FLOAT(*ret);
	break;
    case GTK_TYPE_DOUBLE:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_DOUBLE(*ret);
	break;
    case GTK_TYPE_STRING:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_STRING(*ret);
	break;
    case GTK_TYPE_ENUM:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_ENUM(*ret);
	break;
    case GTK_TYPE_FLAGS:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_FLAGS(*ret);
	break;
    case GTK_TYPE_BOXED:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_BOXED(*ret);
	break;
    case GTK_TYPE_POINTER:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_POINTER(*ret);
	break; 
    case G_TYPE_OBJECT:
	GTK_VALUE_POINTER(*loc) = &GTK_VALUE_OBJECT(*ret);
	break;
    default:
	return -1;
    }
    return 0;
}

static int exp_quark_call(GQuark quark, GtkArg* params, GtkArg* ret)
{
    e_export_t* exp;

    exp = (e_export_t*) g_hash_table_lookup(exp_hash_table,
					    GUINT_TO_POINTER(quark));
    if (exp == NULL)
	return -1;
    if (arg_set_return_pointer(&params[exp->arity], 
			       exp->return_type, ret) < 0)
	return -1;
    (*exp->marshaller)((GtkSignalFunc)exp->function, NULL, params);
    if (exp->return_type == GTK_TYPE_NONE)
	return 0;
    else {
	ret->type = exp->return_type;
	return 1;
    }
}

static int exp_call(const gchar* mod_name, const gchar* fun_name,
		    GtkArg* params, GtkArg* ret)
{
    GQuark quark = exp_try_quark(mod_name, ':', fun_name);
    if (quark)
	return exp_quark_call(quark, params, ret);
    return -1;
}
#endif

/*
 * initialize the Type by creating one object of the
 * type given. This is needed when eval arguments are initialized
 * (The GnomeCanvasLine etc need this call to work)
 */
void gdrv_init_type(GtkType type)
{
    if (GTK_FUNDAMENTAL_TYPE(type) == G_TYPE_OBJECT) {
	gpointer obj = gtk_type_new(type);
	if (obj != NULL)
#if GTK_VERSION < GTK_VERS(2,0,0)
	    gtk_type_free(type, obj);
#else
	    g_object_unref(obj);
#endif
    }
}

/* EXPORTED */
gint gdrv_function_exportv(const char* mod_name, const char* fun_name,
		  FFunction function,
		  FMarshaller marshaller,
		  GtkType return_type,
		  guint arity,
		  e_arg_t* arg_type)
{
    e_export_t* exp;
    GQuark quark;

    g_return_val_if_fail (mod_name != NULL, -1);
    g_return_val_if_fail (fun_name != NULL, -1);
    g_return_val_if_fail (marshaller != NULL, -1);
    if (arity)
	g_return_val_if_fail (arg_type != NULL, -1);
    if ((exp = exp_lookup(mod_name, fun_name)) != NULL) {
	g_warning ("gdrv_function_exportv(): function \"%s:%s\" already exists\n", 
		   mod_name, fun_name);
	return -1;
    }

    quark = exp_mk_quark(mod_name, ':', fun_name);

    exp = (e_export_t*) g_malloc(sizeof(e_export_t)+(arity*sizeof(e_arg_t)));
    exp->mod_name    = mod_name;
    exp->fun_name    = fun_name;
    exp->quark       = quark;
    exp->function    = function;
    exp->marshaller  = marshaller;
    exp->return_type = return_type;
    exp->arity = arity;

    if (arity > 0) {
	guint i;
	for (i = 0; i < arity; i++)
	    exp->arg_type[i] = arg_type[i];
    }
    g_hash_table_insert (exp_hash_table, GUINT_TO_POINTER (quark),
			 exp);
    return 0;
}

/* EXPORTED */
gint gdrv_function_export(const char* mod_name, const char* fun_name,
			  FFunction function,
			  FMarshaller marshaller,
			  GtkType return_type,
			  guint arity,
			  ...)
{
    e_arg_t* arg_type;
    e_arg_t  arg_type0[10];
    gint res;
    guint i;
    va_list args;

    if (arity > 0) {
	if (arity > 10)
	    arg_type = g_new (e_arg_t, arity);
	else
	    arg_type = arg_type0;
	va_start (args, arity);
	for (i = 0; i < arity; i++) {
	    arg_type[i].type = va_arg (args, GtkType);
	    arg_type[i].status = va_arg(args, arg_status_t);
	}
	va_end (args);
    }
    else
	arg_type = NULL;

    res = gdrv_function_exportv(mod_name, 
				fun_name,
				function, 
				marshaller, return_type, arity, arg_type);
    if ((arg_type != NULL) && (arg_type != arg_type0))
	g_free(arg_type);
    return res;
}

/* EXPORTED */
gint gdrv_function_unexport(const char* mod_name, const char* fun_name)
{
    e_export_t* exp;
    GQuark quark =  exp_try_quark(mod_name, ':', fun_name);
    if (quark)  {
	exp = (e_export_t*) g_hash_table_lookup(exp_hash_table,
						GUINT_TO_POINTER(quark));
	if (exp != NULL) {
	    g_hash_table_remove(exp_hash_table, GUINT_TO_POINTER(quark));
	    g_free(exp);
	}
    }
    return 0;
}


/* EXPORTED */
gint gdrv_struct_export(const char* mod_name, 
			const char* struct_name,
			GtkType     type,
			gboolean    private,
			size_t size,
			int    arity,
			FCoder encode,
			FCoder decode)
{
    e_struct_t* sp;
    GQuark quark;

    g_return_val_if_fail (mod_name != NULL, -1);
    g_return_val_if_fail (struct_name != NULL, -1);
    g_return_val_if_fail (encode != NULL, -1);    
    g_return_val_if_fail (decode != NULL, -1);    

    if ((sp = struct_lookup_by_name(struct_name)) != NULL) {
	g_warning ("gdrv_struct_export(): record \"%s.%s\" already exists\n", 
		   mod_name, struct_name);
	return -1;
    }

    quark = exp_mk_quark("", '#', struct_name);

    if ((sp = (e_struct_t*) g_malloc(sizeof(e_struct_t))) == NULL)
	return -1;
    sp->mod_name     = mod_name;
    sp->struct_name  = struct_name;
    sp->type         = type;
    sp->private      = private;
    sp->quark        = quark;
    sp->encode       = encode;
    sp->decode       = decode;
    sp->size         = size;
    sp->arity        = arity;
    g_hash_table_insert (exp_hash_table, GUINT_TO_POINTER (quark), sp);
    g_hash_table_insert (stc_hash_table, GUINT_TO_POINTER (type),  sp);
    return 0;
}

/* EXPORTED */
gint gdrv_struct_unexport(const char* mod_name, const char* struct_name)
{
    e_struct_t* sp;
    GQuark quark = exp_try_quark("", '#', struct_name);

    if (quark)  {
	sp = (e_struct_t*) g_hash_table_lookup(exp_hash_table,
					       GUINT_TO_POINTER(quark));
	if (sp != NULL) {
	    g_hash_table_remove(exp_hash_table, GUINT_TO_POINTER(quark));
	    g_hash_table_remove(stc_hash_table, GUINT_TO_POINTER(sp->type));
	    g_free(sp);
	}
    }
    return 0;
}



static void mod_insert(GModule* mod, char* name, 
		       void (*initfn)(), void (*exitfn)())
{
    e_module_t* emod = g_malloc(sizeof(e_module_t));

    emod->mod = mod;
    emod->mod_name = g_strdup(name);
    emod->initfn = initfn;
    emod->exitfn = exitfn;
    emod->next = module_list;

    (*initfn)();  /* init module */

    module_list = emod;
}

static void mod_remove(char* name)
{
    e_module_t** emodp = &module_list;

    while(*emodp != NULL) {
	e_module_t* emod = *emodp;

	if (strcmp(name, emod->mod_name) == 0) {
	    if (emod->exitfn != NULL)
		(*emod->exitfn)();
	    *emodp = emod->next;
	    (void) g_module_close(emod->mod);
	    g_free(emod->mod_name);
	    g_free(emod);
	    return;
	}
	emodp = &emod->next;
    }
}

static e_module_t* mod_find(char* name)
{
    e_module_t* emod = module_list;

    while(emod) {
	if (strcmp(emod->mod_name, name) == 0)
	    return emod;
	emod = emod->next;
    }
    return NULL;
}

/****************************************************************************
**
** DECODE FUNCTIONS
**
**
*****************************************************************************/
/*
** Decode a GtkObject pointer. The pointer is found in the pmap
** entry that is encoded as a unsigned long value
*/
static int decode_object(gdrv_mesg_t* mp, GtkObject** obj)
{
    unsigned long id;
    pmap_t* eptr;

    if (gdrv_decode_ulong(mp, &id) < 0)
	return -1;

    if ((eptr = (pmap_t*) id) == NULL) {
	*obj = NULL;

	DEBUGF(("PMAP LOOKUP: %ld -> %ld\r\n", 
		id, (long) NULL));
	return 0;
    }

    if (GTK_FUNDAMENTAL_TYPE(eptr->type) != G_TYPE_OBJECT) {
	*obj = NULL;
	DEBUGF(("PMAP LOOKUP FAILED: %ld %s not OBJECT\r\n", 
		id, gtk_type_name(eptr->type)));
	return -1;
    }
    else if (eptr->ptr == NULL) {
	*obj = NULL;
	DEBUGF(("PMAP LOOKUP FAILED: %ld object destroyed\r\n",
		id));
	return -1;
    }
    else {
	DEBUGF(("PMAP LOOKUP: %s: %ld -> %ld\r\n", 
		gtk_type_name(eptr->type), id, (long) eptr->ptr));
	*obj = (GtkObject*) eptr->ptr;
	return 0;
    }
}

/*
** Decode erlang version of object ref {ObjType, ObjId}
*/
int gdrv_decode_object(gdrv_mesg_t* mp, GtkObject** obj)
{
    int arity;
    GtkType type;

    if ((gdrv_decode_tuple_header(mp, &arity) < 0) ||
	(arity != 2)) return -1;

    if (gdrv_decode_ulong(mp, (unsigned long*)&type) < 0)
	return -1;
    switch(GTK_FUNDAMENTAL_TYPE(type)) {
    case G_TYPE_OBJECT:
	return decode_object(mp, obj);
    case GTK_TYPE_BOXED: 
	return gdrv_decode_boxed(mp, (gpointer*) obj, type);
    default:
	return -1;
    }
}


int gdrv_decode_pointer(gdrv_mesg_t* mp, gpointer* ptr, GtkType itype)
{
    unsigned long id;

    if (itype == GTK_TYPE_ARG) {
	if (gdrv_decode_arg(mp, (GtkArg*) ptr) < 0)
	    return -1;
    }
    else if (itype == GTK_TYPE_ETERM) {
	ETERM* term;
	if (gdrv_decode_term(mp, (void*) &term) < 0)
	    return -1;
	*ptr = (gpointer) term;
    }
    else if (itype == GTK_TYPE_GLIST) {
	int len;
	GList* list = NULL;
	if (gdrv_decode_list_header(mp, &len) < 0)
	    return -1;
	DEBUGF(("decode GList: length=%d\n", len));
	if (len > 0) {
	    while(len--) {
		GtkObject* obj;
		if (gdrv_decode_object(mp, &obj) < 0) {
		    g_list_free(list);
		    return -1;
		}
		list = g_list_append(list, (gpointer) obj);
	    }
	    if ((gdrv_decode_list_header(mp,&len) < 0) || (len != 0)) {
		g_list_free(list);
		return -1;
	    }
	}
	*ptr = (gpointer) list;
    }
    else if (itype == GTK_TYPE_GSLIST) {
	int len;
	GSList* list = NULL;
	if (gdrv_decode_list_header(mp, &len) < 0) 
	    return -1;
	DEBUGF(("decode GSList: length=%d\r\n", len));
	if (len > 0) {
	    while(len--) {
		GtkObject* obj;
		if (gdrv_decode_object(mp, &obj) < 0) {
		    DEBUGF(("decode GSList: fail at length=%d\r\n", len));
		    g_slist_free(list);
		    return -1;
		}
		list = g_slist_append(list, (gpointer) obj);
	    }
	    if ((gdrv_decode_list_header(mp,&len) < 0) || (len != 0)) {
		g_slist_free(list);
		return -1;
	    }
	}
	*ptr = (gpointer) list;
    }
    else if (itype == GDK_TYPE_ARRAY_OF_POINT) {
	if (decode_gdk_point(mp, (GdkPoint*)ptr) < 0)
	    return -1;
	return 0;
    }
    else if (itype == GDK_TYPE_ARRAY_OF_SEGMENT) {
	if (decode_gdk_segment(mp, (GdkSegment*)ptr) < 0)
	    return -1;
	return 0;
    }
    else if (itype == GTK_TYPE_ARRAY_OF_ARG) {
	if (gdrv_decode_arg(mp, (GtkArg*)ptr) < 0)
	    return -1;
    }
    else {
	if (gdrv_decode_ulong(mp, &id) < 0)
	    return -1;
	*ptr = (gpointer) id;
    }
    return 0;
}

static int decode_gdk_point(gdrv_mesg_t* mp, GdkPoint* p)
{
    int arity;
    char record_name[NAME_SIZE];
    long x, y;

    if ((gdrv_decode_tuple_header(mp, &arity) < 0) ||
	(arity != 3)) return -1;
    if ((gdrv_decode_atom(mp, record_name, NAME_SIZE) < 0) || 
	(strcmp(record_name, "gdk_point") != 0)) return -1;
    if (gdrv_decode_long(mp, &x) < 0) return -1;
    if (gdrv_decode_long(mp, &y) < 0) return -1;
    p->x = x;
    p->y = y;
    return 0;
}

static int decode_gdk_segment(gdrv_mesg_t* mp, GdkSegment* seg)
{
    int arity;
    char record_name[NAME_SIZE];
    long x1, y1, x2, y2;

    if ((gdrv_decode_tuple_header(mp, &arity) < 0) ||
	(arity != 5)) return -1;
    if ((gdrv_decode_atom(mp, record_name, NAME_SIZE) < 0) || 
	(strcmp(record_name, "gdk_segment") != 0)) return -1;
    if (gdrv_decode_long(mp, &x1) < 0) return -1;
    if (gdrv_decode_long(mp, &y1) < 0) return -1;
    if (gdrv_decode_long(mp, &x2) < 0) return -1;
    if (gdrv_decode_long(mp, &y2) < 0) return -1;
    seg->x1 = x1;
    seg->y1 = y1;
    seg->x2 = x2;
    seg->y2 = y2;
    return 0;
}


int gdrv_decode_boxed(gdrv_mesg_t* mp, gpointer* p, GtkType itype)
{
    unsigned long id;
    pmap_t* eptr;

    if (gdrv_decode_ulong(mp, &id) < 0) {
	/* special case for record encoded stuff */
	char struct_name[NAME_SIZE];
	int  arity;
	e_struct_t* stc;

	if ((gdrv_decode_tuple_header(mp, &arity) < 0) || (arity < 1))
	    return -1;
	if (gdrv_decode_atom(mp, struct_name, NAME_SIZE) < 0)
	    return -1;
	if ((stc = struct_lookup_by_name(struct_name)) == NULL)
	    return -1;
	if (stc->private) {
	    if (stc->arity != arity-2) return -1;
	}
	else if (stc->arity != arity-1)
	    return -1;
	if (stc->private) {
	    if (stc->decode(mp, p) < 0)
		return -1;
	}
	else {
	    void* mem;

	    if ((mem = ARG_STACK_ALLOC(stc->size)) == NULL)
		return -1;
	    if (stc->decode(mp, mem) < 0)
		return -1;
	    *p = mem;
	}
	return 0;
    }

    if (id == 0) {
	*p = NULL;
	return 0;
    }

    if ((itype == GDK_TYPE_WINDOW) ||
	(itype == GDK_TYPE_PIXMAP) ||
	(itype == GDK_TYPE_BITMAP) ||
	(itype == GDK_TYPE_DRAWABLE)) {
	eptr = (pmap_t*) id;

	if ((itype != GDK_TYPE_DRAWABLE) && (eptr->type != itype)) {
	    *p = NULL;
	    DEBUGF(("PMAP LOOKUP FAILED: %ld %s type %s\r\n", 
		    id, gtk_type_name(eptr->type), gtk_type_name(itype)));
	    return -1;
	}
	else if (eptr->ptr == NULL) {
	    *p = NULL;
	    DEBUGF(("PMAP LOOKUP FAILED: %ld object destroyed\r\n",
		    id));
	    return -1;
	}
	else {
	    DEBUGF(("PMAP LOOKUP: %s: %ld -> %ld\r\n", 
		    gtk_type_name(eptr->type), id, (long) eptr->ptr));
	    *p = eptr->ptr;
	    return 0;
	}
    }
    else {
	/* GTK_TYPE_STYLE, GTK_TYPE_RC_STYLE, GTK_TYPE_ACCEL_GROUP,
	 * GDK_TYPE_COLORMAP,GDK_TYPE_VISUAL, GDK_TYPE_IMAGE, 
	 * GDK_TYPE_CURSOR, GDK_TYPE_GC,  
	 * ...
	 */
	DEBUGF(("DECODE BOXED: %s direct pointer %d\r\n", 
		gtk_type_name(itype), id));
	*p = (gpointer) id;
	return 0;
    }
    return -1;
}

#if 0
/* boxed or object, depends on 1.2 or 2.0 */
static int decode_by_type(gdrv_mesg_t* mp, gpointer* p, GtkType itype)
{
    if (GTK_FUNDAMENTAL_TYPE(itype) == G_TYPE_OBJECT)
	return gdrv_decode_object(mp, (GtkObject**) p);
    else
	return gdrv_decode_boxed(mp, p, itype);
}

static int decode_gdk_window(gdrv_mesg_t* mp, GdkWindow** window)
{
    return decode_by_type(mp, (gpointer*) window, GDK_TYPE_WINDOW);
}


static int decode_gdk_event(gdrv_mesg_t* mp, GdkEvent* event)
{
    int arity;
    char r_name[NAME_SIZE];
    char e_name[NAME_SIZE];

    if ((gdrv_decode_tuple_header(mp, &arity) < 0) || (arity < 4))
	goto badarg;
    if (gdrv_decode_atom(mp, r_name, NAME_SIZE) < 0) goto badarg;
    if (gdrv_decode_atom(mp, e_name, NAME_SIZE) < 0) goto badarg; /* type */
    if (decode_gdk_window(mp, &event->any.window) < 0)
	goto badarg;
    if (gdrv_decode_boolean(mp, (int*)&event->any.send_event) < 0)
	goto badarg;
    if (strncmp(e_name, "GDK_", 4) != 0) 
	goto badarg;
    switch(e_name[4]) {
    case 'D':  /* DELETE | DESTROY | DRAG_ENTER | DRAG_LEAVE | DRAG_MOTION |
		  DRAG_STATUS | DROP_START | DROP_FINISHED */
    case 'E':  /* EXPOSE | ENTER_NOTIFY */
    case 'M':  /* MOTION_NOTIFY | MAP */
    case 'B':  /* BUTTON_PRESS | BUTTON_RELEASE */
    case '2':  /* 2BUTTON_PRESS */
    case '3':  /* 3BUTTON_PRESS */
    case 'K':  /* KEY_PRESS | KEY_RELEASE */
    case 'L':  /* LEAVE_NOTIFY */
    case 'F':  /* FOCUS_CHANGE */
    case 'C':  /* CONFIGURE | CLIENT_EVENT */
    case 'U':  /* UNMAP */
    case 'P':  /* PROPERTY_NOTIFY | PROXIMITY_IN | PROXIMITY_OUT */
    case 'S':  /* SELECTION_CLEAR | SELECTION_REQUEST | SELECTION_NOTIFY */
    case 'V': /* VISIBILITY_NOTIFY */
    case 'N': /* NO_EXPOSE */
      ;
    }

 badarg:
    return -1;
}
#endif

/* allocate string on argument heap for the string */
int gdrv_decode_string(gdrv_mesg_t* mp, char** x)
{
    int type;
    int len;
    gchar* ptr = NULL;

    ei_get_type(mp->data, &mp->ix, &type, &len);
    
    /* accept the integer 0 as NULL */
    if ((len == 0) && (type == ERL_SMALL_INTEGER_EXT)) {
	long v;
	if ((gdrv_decode_long(mp, &v) < 0) || (v != 0))
	    return -1;
	*x = NULL;
	return 0;
    }

    ptr = (gchar*) ARG_STACK_ALLOC(len+1);
    if (gdrv_decode_strn(mp, (char*) ptr, len+1) < 0)
	return -1;
    *x = ptr;
    return 0;
}

static int sizeof_type(GtkType type)
{
    switch (GTK_FUNDAMENTAL_TYPE(type)) {
    case GTK_TYPE_NONE:
	return 0;
    case GTK_TYPE_CHAR:
	return sizeof(gchar);
    case GTK_TYPE_UCHAR:
	return sizeof(guchar);
    case GTK_TYPE_BOOL:
	return sizeof(gboolean);
    case GTK_TYPE_INT:
	return sizeof(gint);
    case GTK_TYPE_UINT:
	return sizeof(guint);
    case GTK_TYPE_ENUM:	
	return sizeof(gint);
    case GTK_TYPE_FLAGS:
	return sizeof(guint);
    case GTK_TYPE_LONG:
	return sizeof(glong);
    case GTK_TYPE_ULONG:
	return sizeof(gulong);
    case GTK_TYPE_FLOAT:
	return sizeof(gfloat);
    case GTK_TYPE_DOUBLE:
	return sizeof(gdouble);
    case GTK_TYPE_STRING:
	return sizeof(gchar*);
    case G_TYPE_OBJECT:
	return sizeof(GtkObject*);
    case GTK_TYPE_POINTER:
	if (type == GDK_TYPE_ARRAY_OF_POINT)
	    return sizeof(GdkPoint);
	else if (type == GDK_TYPE_ARRAY_OF_SEGMENT)
	    return sizeof(GdkSegment);
	else if (type == GTK_TYPE_ARRAY_OF_ARG)
	    return sizeof(GtkArg);
	else if (type == GTK_TYPE_ARG)
	    return sizeof(GtkArg);
	else
	    return sizeof(gpointer);
    case GTK_TYPE_BOXED: 
	if (type == GDK_TYPE_POINT)
	    return sizeof(GdkPoint);
	else if (type == GDK_TYPE_SEGMENT)
	    return sizeof(GdkSegment);
	return sizeof(gpointer);
    default:
	return -1;
    }
}

static int decode_gtk_type(gdrv_mesg_t* mp, GtkType* type)
{
    if (gdrv_decode_long(mp, (long*) type) < 0) {
	char tname[NAME_SIZE];
	if (gdrv_decode_atom(mp, tname, NAME_SIZE) < 0)
	    return -1;
	if (strcmp(tname, "gstring") == 0)
	    *type = GTK_TYPE_STRING;
	else if ((*type = gtk_type_from_name(tname)) == 0)
	    return -1;
    }
    return 0;
}


static int decode_fundamental(gdrv_mesg_t* mp, gchar* ptr, GtkType type)
{
    long lval;
    unsigned long ulval;
    double fval;
    int ival;
    char cval;
    GtkType tval;
    GtkType ftype = GTK_FUNDAMENTAL_TYPE(type);
    
    DEBUGF(("decode_fundamental: type=%d, fundamental=%d\r\n",
	    type, ftype));

    switch(ftype) {
    case GTK_TYPE_CHAR:
	if (gdrv_decode_char(mp, &cval) >= 0)
	    *ptr = (gchar) cval;
	else
	    return -1;
	return 0;
    case GTK_TYPE_UCHAR:
	if (gdrv_decode_char(mp, &cval) >= 0)
	    *((guchar*)ptr) = (guchar)cval;	
	else
	    return -1;
	return 0;
    case GTK_TYPE_BOOL:
	if (gdrv_decode_boolean(mp, &ival) >= 0)
	    *((gboolean*)ptr) = ival;
	else if (gdrv_decode_long(mp, &lval) >= 0)
	    *((gboolean*)ptr) = (lval != 0);
	else
	    return -1;
	return 0;
    case GTK_TYPE_INT:
	if (gdrv_decode_long(mp, &lval) >= 0)
	    *((gint*)ptr) = (gint) lval;
	else if (gdrv_decode_boolean(mp, &ival) >= 0)
	    *((gint*)ptr) = (gint) ival;
	else if (gdrv_decode_double(mp, &fval) >= 0)
	    *((gint*)ptr) = (gint) fval;
	else
	    return -1;
	return 0;
    case GTK_TYPE_UINT:
	if (gdrv_decode_ulong(mp, &ulval) >= 0)
	    *((guint*)ptr) = (guint) ulval;
	else if (gdrv_decode_boolean(mp, &ival) >= 0)
	    *((guint*)ptr) = (guint) ival;
	else if (gdrv_decode_double(mp, &fval) >= 0)
	    *((guint*)ptr) = (guint) fval;
	else
	    return -1;
	return 0;
    case GTK_TYPE_ENUM:	
	if (gdrv_decode_long(mp, &lval) >= 0)
	    *((gint*)ptr) = (gint) lval;
	else
	    return -1;
	return 0;

    case GTK_TYPE_FLAGS:
	if (gdrv_decode_ulong(mp, &ulval) >= 0)
	    *((guint*) ptr) = (guint) ulval;
	else
	    return -1;
	return 0;

    case GTK_TYPE_LONG:
	if (type == GTK_TYPE_TYPE) {
	    if (decode_gtk_type(mp,&tval) >= 0)
		*((glong*)ptr) = (glong) tval;
	    else
		return -1;
	}
	else if (gdrv_decode_long(mp, &lval) >= 0)
	    *((glong*)ptr) = (glong) lval;
	else
	    return -1;
	return 0;

   case GTK_TYPE_ULONG:
	if (gdrv_decode_ulong(mp, &ulval) >= 0)
	    *((gulong*)ptr) = (gulong) ulval;
	else
	    return -1;
	return 0;

    case GTK_TYPE_FLOAT:
	if (gdrv_decode_double(mp, &fval) >= 0)
	    *((gfloat*)ptr) = (gfloat) fval;
	else if (gdrv_decode_long(mp, &lval) >= 0)
	    *((gfloat*)ptr) = (gfloat) lval;
	else
	    return -1;
	return 0;
	
    case GTK_TYPE_DOUBLE:
	if (gdrv_decode_double(mp, &fval) >= 0)
	    *((gdouble*)ptr) = (gdouble) fval;
	else if (gdrv_decode_long(mp, &lval) >= 0)
	    *((gdouble*)ptr) = (gdouble) lval;
	else
	    return -1;
	return 0;

    case GTK_TYPE_STRING: {
	gchar* str;

	if (gdrv_decode_string(mp, &str) < 0)
	    return -1;
	*((gchar**)ptr) = str;
	return 0;
    }

    case G_TYPE_OBJECT: {
	GtkObject* obj;
	
	if (decode_object(mp, &obj) < 0)
	    return -1;
	*((GtkObject**)ptr) = obj;
	return 0;
    }

    case GTK_TYPE_BOXED: {
	gpointer p;
	if (gdrv_decode_boxed(mp, &p, type) < 0)
	    return -1;
	*((gpointer*) ptr) = p;
	return 0;
    }

    case GTK_TYPE_POINTER: {
	if (gdrv_decode_pointer(mp, (gpointer*) ptr, type) < 0)
	    return -1;
	return 0;
    }

    default:
	return -1;
    }
    return -1;
}


/*
** decode arguments
** input: {atom name, integer type, term value}
** output: GtkArg
*/

int gdrv_decode_arg(gdrv_mesg_t* mp, GtkArg* ap)
{
    int     arity;
    GtkType type;
    int     code;
    char    name[NAME_SIZE];

    if ((gdrv_decode_tuple_header(mp, &arity) < 0) || (arity != 3))
	return -1;

    if (gdrv_decode_atom(mp, name, NAME_SIZE) < 0)  return -1;

    if (decode_gtk_type(mp, &ap->type) < 0) return -1;

    DEBUGF(("decode_arg: name=%s, type=%d\r\n", name, ap->type));

    code = 0;
    type = ap->type;

    if (IS_ARRAY(type)) {
	gchar* array;

	if (type == GTK_TYPE_ARRAY_OF_VOID) {
	    int ext_type;
	    int len;
	    long llen;

	    if ((gdrv_decode_type(mp, &ext_type, &len) < 0) ||
		(ext_type != ERL_BINARY_EXT))
		return -1;
	    if ((array = (gchar*) ARG_STACK_ALLOC(len)) == NULL)
		return -1;
	    llen = len;
	    if (gdrv_decode_binary(mp, (void*) array, &llen) < 0)
		return -1;
	}
	else {
	    gchar* ptr;
	    GtkType etype = array_element_type(type);
	    int i;
	    int elem_size = sizeof_type(etype);

	    if ((code = gdrv_decode_tuple_header(mp, &arity)) < 0)
		return -1;

	    DEBUGF(("array length %d, element size: %d\r\n", 
		    arity, elem_size));

	    if ((array = (gchar*) ARG_STACK_ALLOC((arity+1)*elem_size))==NULL)
		return -1;
	    ptr = array;
	    for (i = 0; i < arity; i++) {
		if (decode_fundamental(mp, ptr, etype) < 0)
		    return -1;
		ptr += elem_size;
	    }
	    memset(ptr, 0, elem_size); /* null terminate as a feature */
	}
	ap->d.pointer_data = (gpointer) array;
    }
    else {
	switch(GTK_FUNDAMENTAL_TYPE(type)) {
	case GTK_TYPE_CHAR:
	    code = decode_fundamental(mp, (char*)&ap->d.char_data, type);
	    break;
	case GTK_TYPE_UCHAR:
	    code = decode_fundamental(mp,  (char*)&ap->d.uchar_data, type);
	    break;
	case GTK_TYPE_BOOL:
	    code = decode_fundamental(mp,  (char*)&ap->d.bool_data, type);
	    break;
	case GTK_TYPE_INT:
	    code = decode_fundamental(mp,  (char*)&ap->d.int_data, type);
	    break;
	case GTK_TYPE_UINT:
	    code = decode_fundamental(mp,  (char*)&ap->d.uint_data, type);
	    break;
	case GTK_TYPE_ENUM:	
	    code = decode_fundamental(mp, (char*)&ap->d.int_data, type);
	    break;
	case GTK_TYPE_FLAGS:
	    code = decode_fundamental(mp, (char*)&ap->d.uint_data, type);
	    break;
	case GTK_TYPE_LONG:
	    code = decode_fundamental(mp, (char*)&ap->d.long_data, type);
	    break;
	case GTK_TYPE_ULONG:
	    code = decode_fundamental(mp, (char*)&ap->d.ulong_data, type);
	    break;
	case GTK_TYPE_FLOAT:
	    code = decode_fundamental(mp, (char*)&ap->d.float_data, type);
	    break;
	case GTK_TYPE_DOUBLE:
	    code = decode_fundamental(mp, (char*)&ap->d.double_data, type);
	    break;
	case GTK_TYPE_STRING:
	    code = decode_fundamental(mp, (char*)&ap->d.string_data, type);
	    break;
	case G_TYPE_OBJECT:
	    code = decode_fundamental(mp, (char*)&ap->d.object_data, type);
	    break;
	case GTK_TYPE_POINTER:
	    code = decode_fundamental(mp, (char*)&ap->d.pointer_data, type);
	    break;
	case GTK_TYPE_BOXED: 
	    code = decode_fundamental(mp, (char*)&ap->d.pointer_data, type);
	    break;
	default:
	    return -1;
	}
    }
    if (code < 0)
	return -1;
    else
	ap->name = arg_stack_strdup(name);
    return 0;
}

/****************************************************************************
**
** ENCODE FUNCTIONS
** encode_gdk_window
** encode_gdk_rectangle
** encode_gdk_segment
** encode_gdk_color
** encode_gdk_point
** encode_gdk_visibility_state
** encode_gdk_crossing_mode
** encode_gdk_notify_type
** encode_gdk_drag_protocol
** encode_gdk_drag_context
** encode_gdk_atom
** encode_gdk_event
** encode_gdk_font
** encode_boxed
** gdrv_encode_arg
*****************************************************************************/


/*
** Encode a GtkObject into form {Type, ObjId}
*/
int gdrv_encode_object(gdrv_mesg_t** mpp, GtkObject* obj)
{
    gdrv_encode_tuple_header(mpp, 2);

    if (obj == NULL) {
	gdrv_encode_long(mpp, GTK_TYPE_INVALID);
	return gdrv_encode_ulong(mpp, 0);
    }
    else {
	pmap_t* eptr = gtk_object_map(obj);
	gdrv_encode_long(mpp, GTK_OBJECT_TYPE(obj));
	return gdrv_encode_ulong(mpp, (unsigned long) eptr);
    }
}


int gdrv_encode_pointer(gdrv_mesg_t** mpp, gpointer ptr, GtkType type)
{
    if (type == GTK_TYPE_ETERM) {
	ETERM* e = (ETERM*) ptr;
	if (e == NULL)
	    return gdrv_encode_atom(mpp, "undefined");
	else {
	    int code = gdrv_encode_term(mpp, e);
	    erl_free_term(e);
	    return code;
	}
    }
    else if (type == GTK_TYPE_ARG) {
	GtkArg* ap = (GtkArg*) ptr;
	if (gdrv_encode_tuple_header(mpp, 3) < 0)
	    return -1;
	if (gdrv_encode_atom(mpp, ap->name) < 0)
	    return -1;
	if (gdrv_encode_long(mpp, ap->type) < 0)
	    return -1;
	if (gdrv_encode_arg(mpp, ap) < 0)
	    return -1;
	return 0;
    }
    else if ((type == GTK_TYPE_ARRAY) || (IS_ARRAY(type))) {
	/* encode data as a tuple {E1,E2,...En} or a binary */
	GtkArray* arr  = (GtkArray*) ptr;
	int elem_sz    = sizeof_type(arr->elem_type);
	char* vp       = (char*) arr->vector;
	int code;
	int i;

	DEBUGF(("encode array: sz=%d sizeof(elem)\r\n", arr->size, elem_sz));
	if (arr->elem_type == GTK_TYPE_NONE) {
	    if ((code = gdrv_encode_binary(mpp, vp, arr->size)) < 0)
		return code;
	}
	else {
	    if ((code = gdrv_encode_tuple_header(mpp, arr->size)) < 0)
		return code;
	    for (i = 0; i < arr->size; i++) {
		code = encode_fundamental(mpp, vp, arr->elem_type);
		if (code < 0) return code;
		vp += elem_sz;
	    }
	}
	return 0;
    }
    else if (type == GTK_TYPE_GLIST) {
	GList* list;
	int len;
	int code;
	
	for (list=(GList*)ptr, len=0; list!=NULL; list=list->next, len++)
	    ;
	if (len > 0) {
	    code = gdrv_encode_list_header(mpp, len);
	    for (list=(GList*)ptr; list!=NULL; list=list->next) {
		if (GTK_IS_OBJECT(list->data))
		    gdrv_encode_object(mpp, (GtkObject*) list->data);
		else
		    gdrv_encode_ulong(mpp, (unsigned long)list->data);
	    }
	}
	return gdrv_encode_empty_list(mpp);
    }
    else if (type == GTK_TYPE_GSLIST) {
	GSList* list;
	int len;
	int code;
	
	for (list=(GSList*)ptr, len=0; list!=NULL; list=list->next, len++)
	    ;
	if (len > 0) {
	    code = gdrv_encode_list_header(mpp, len);
	    for (list=(GSList*)ptr; list!=NULL; list=list->next) {
		if (GTK_IS_OBJECT(list->data))
		    gdrv_encode_object(mpp, (GtkObject*) list->data);
		else
		    gdrv_encode_ulong(mpp, (unsigned long)list->data);
	    }
	}
	return gdrv_encode_empty_list(mpp);
    }
    else
	return gdrv_encode_ulong(mpp, (unsigned long) ptr);
}

static int encode_gdk_rectangle(gdrv_mesg_t** mpp, GdkRectangle* rect)
{
    gdrv_encode_tuple_header(mpp, 5);
    gdrv_encode_atom(mpp, "gdk_rectangle");
    gdrv_encode_long(mpp, rect->x);
    gdrv_encode_long(mpp, rect->y);
    gdrv_encode_long(mpp, rect->width);
    gdrv_encode_long(mpp, rect->height);
    return 0;
}

static int encode_gdk_visibility_state(gdrv_mesg_t** mpp, 
				       GdkVisibilityState state)
{
    switch(state) {
    case GDK_VISIBILITY_UNOBSCURED:
	return gdrv_encode_atom(mpp, "GDK_VISIBILITY_UNOBSCURED");
    case GDK_VISIBILITY_PARTIAL:
	return gdrv_encode_atom(mpp, "GDK_VISIBILITY_PARTIAL");
    case GDK_VISIBILITY_FULLY_OBSCURED:
	return gdrv_encode_atom(mpp, "GDK_VISIBILITY_FULLY_OBSCURED");
    default:
	return gdrv_encode_atom(mpp, "undefined");
    }
}

static int encode_gdk_crossing_mode(gdrv_mesg_t** mpp,
				       GdkCrossingMode mode)
{
    switch(mode) {
    case GDK_CROSSING_NORMAL:
	return gdrv_encode_atom(mpp, "GDK_CROSSING_NORMAL");
    case GDK_CROSSING_GRAB:
	return gdrv_encode_atom(mpp, "GDK_CROSSING_GRAB");
    case GDK_CROSSING_UNGRAB:
	return gdrv_encode_atom(mpp, "GDK_CROSSING_UNGRAB");
    default:
	return gdrv_encode_atom(mpp, "undefined");
    }
}

#if GTK_VERSION >= GTK_VERS(1,3,1)
static int encode_gdk_scroll_direction(gdrv_mesg_t** mpp, 
				       GdkScrollDirection dir)
{
    switch(dir) {
    case GDK_SCROLL_UP:
	return gdrv_encode_atom(mpp, "GDK_SCROLL_UP");
    case GDK_SCROLL_DOWN:
	return gdrv_encode_atom(mpp, "GDK_SCROLL_DOWN");
    case GDK_SCROLL_LEFT:
	return gdrv_encode_atom(mpp, "GDK_SCROLL_LEFT");
    case GDK_SCROLL_RIGHT:
	return gdrv_encode_atom(mpp, "GDK_SCROLL_RIGHT");
    default:
	return gdrv_encode_atom(mpp, "undefined");
    }
}
#endif

static int encode_gdk_notify_type(gdrv_mesg_t** mpp, GdkNotifyType type)
{
    switch(type) {
    case GDK_NOTIFY_ANCESTOR:
	return gdrv_encode_atom(mpp, "GDK_NOTIFY_ANCESTOR");
    case GDK_NOTIFY_VIRTUAL:
	return gdrv_encode_atom(mpp, "GDK_NOTIFY_VIRTUAL");
    case GDK_NOTIFY_INFERIOR:
	return gdrv_encode_atom(mpp, "GDK_NOTIFY_INFERIOR");
    case GDK_NOTIFY_NONLINEAR:
	return gdrv_encode_atom(mpp, "GDK_NOTIFY_NONLINEAR");
    case GDK_NOTIFY_NONLINEAR_VIRTUAL:
	return gdrv_encode_atom(mpp, "GDK_NOTIFY_NONLINEAR_VIRTUAL");
    case GDK_NOTIFY_UNKNOWN:
	return gdrv_encode_atom(mpp, "GDK_NOTIFY_UNKNOWN");
    default:
	return gdrv_encode_atom(mpp, "undefined");
    }
}

static int encode_gdk_drag_protocol(gdrv_mesg_t** mpp,
				    GdkDragProtocol proto)
{
    switch(proto) {
    case GDK_DRAG_PROTO_MOTIF:
	return gdrv_encode_atom(mpp, "GDK_DRAG_PROTO_MOTIF");	
    case GDK_DRAG_PROTO_XDND:
	return gdrv_encode_atom(mpp, "GDK_DRAG_PROTO_XDND");	
    case GDK_DRAG_PROTO_ROOTWIN:
	return gdrv_encode_atom(mpp, "GDK_DRAG_PROTO_ROOTWIN");	
    case GDK_DRAG_PROTO_NONE:
	return gdrv_encode_atom(mpp, "GDK_DRAG_PROTO_NONE");	
    default:
	return gdrv_encode_atom(mpp, "undefined");
    }
}


/* boxed or object, depends on 1.2 or 2.0 */
static int encode_by_type(gdrv_mesg_t** mpp,gpointer* ptr,GtkType type)
{
    if (GTK_FUNDAMENTAL_TYPE(type) == G_TYPE_OBJECT)
	return gdrv_encode_object(mpp, (GtkObject*) ptr);
    else
	return gdrv_encode_boxed(mpp, ptr, type);
}

static int encode_gdk_window(gdrv_mesg_t** mpp, GdkWindow* window)
{
    return encode_by_type(mpp, (gpointer) window, GDK_TYPE_WINDOW);
}

static int encode_gdk_drag_context(gdrv_mesg_t** mpp,
				   GdkDragContext* context)
{
    if (context == NULL)
	gdrv_encode_atom(mpp, "undefined");
    else {
	gdrv_encode_tuple_header(mpp, 10);
	gdrv_encode_atom(mpp, "gdk_drag_context");
	encode_gdk_drag_protocol(mpp, context->protocol);
	gdrv_encode_boolean(mpp, context->is_source);
	encode_gdk_window(mpp, context->source_window);
	encode_gdk_window(mpp, context->dest_window);
	gdrv_encode_empty_list(mpp); /* FIXME: targets */
	gdrv_encode_ulong(mpp, context->actions);
	gdrv_encode_ulong(mpp, context->suggested_action);
	gdrv_encode_ulong(mpp, context->action);
	gdrv_encode_ulong(mpp, context->start_time);
    }
    return 0;
}
    
static int encode_gdk_atom(gdrv_mesg_t** mpp, GdkAtom atom)
{
    char* name = gdk_atom_name(atom);

    if (name == NULL)
	return gdrv_encode_atom(mpp, "undefined");
    else {
	int res = gdrv_encode_atom(mpp, name);
	g_free(name);
	return res;
    }
}

static int encode_gtk_selection_data(gdrv_mesg_t** mpp,
				     GtkSelectionData* selection)
{
    gdrv_encode_tuple_header(mpp, 6);
    gdrv_encode_atom(mpp, "gtk_selection_data");
    encode_gdk_atom(mpp, selection->selection);
    encode_gdk_atom(mpp, selection->target);
    encode_gdk_atom(mpp, selection->type);
    gdrv_encode_long(mpp, selection->format);
    gdrv_encode_binary(mpp, selection->data, selection->length);
    return 0;
}

static int encode_gdk_event(gdrv_mesg_t** mpp, GdkEvent* event)
{
    if ((event->type < 0) || (event->type > 31))
	return gdrv_encode_atom(mpp, "undefined");

    gdrv_encode_tuple_header(mpp, erl_ev[event->type].r_len);
    gdrv_encode_atom(mpp, erl_ev[event->type].r_name); /* record name */
    gdrv_encode_atom(mpp, erl_ev[event->type].e_name); /* type */
    encode_gdk_window(mpp, event->any.window);
    gdrv_encode_boolean(mpp, event->any.send_event);   /* send_event */

    /* add fields to form a record for all event types .... */
    switch(event->type) {
    case GDK_DELETE:
    case GDK_DESTROY:
    case GDK_MAP:
    case GDK_UNMAP:
	break;

    case GDK_EXPOSE:
	encode_gdk_rectangle(mpp, &event->expose.area);
	gdrv_encode_long(mpp, event->expose.count);
	break;

    case GDK_MOTION_NOTIFY:  {
	gdouble pressure;
	gdouble xtilt;
	gdouble ytilt;

#if GTK_VERSION >= GTK_VERS(1,3,1)
	gdk_device_get_axis(event->motion.device, event->motion.axes,
			    GDK_AXIS_PRESSURE, &pressure);
	gdk_device_get_axis(event->motion.device, event->motion.axes,
			    GDK_AXIS_XTILT, &xtilt);
	gdk_device_get_axis(event->motion.device, event->motion.axes,
			    GDK_AXIS_YTILT, &ytilt);
#else
	pressure = event->motion.pressure;
	xtilt    = event->motion.xtilt;
	ytilt    = event->motion.ytilt;
#endif
	gdrv_encode_ulong(mpp, event->motion.time);
	gdrv_encode_double(mpp, event->motion.x);
	gdrv_encode_double(mpp, event->motion.y);
	gdrv_encode_double(mpp, pressure);
	gdrv_encode_double(mpp, xtilt);
	gdrv_encode_double(mpp, ytilt);
	gdrv_encode_ulong(mpp, event->motion.state);
	gdrv_encode_boolean(mpp, event->motion.is_hint);
	/* FIXME event->motion.source */
	gdrv_encode_atom(mpp, "undefined");

#if GTK_VERSION >= GTK_VERS(1,3,1)
	gdrv_encode_string(mpp, event->motion.device->name);
#else
	gdrv_encode_ulong(mpp, event->motion.deviceid);
#endif
	gdrv_encode_double(mpp, event->motion.x_root);
	gdrv_encode_double(mpp, event->motion.y_root);
	break;
    }
	
    case GDK_BUTTON_PRESS: 
    case GDK_2BUTTON_PRESS:
    case GDK_3BUTTON_PRESS:
    case GDK_BUTTON_RELEASE: {
	gdouble pressure;
	gdouble xtilt;
	gdouble ytilt;

#if GTK_VERSION >= GTK_VERS(1,3,1)
	gdk_device_get_axis(event->button.device, event->button.axes,
			    GDK_AXIS_PRESSURE, &pressure);
	gdk_device_get_axis(event->button.device, event->button.axes,
			    GDK_AXIS_XTILT, &xtilt);
	gdk_device_get_axis(event->button.device, event->button.axes,
			    GDK_AXIS_YTILT, &ytilt);
#else
	pressure = event->button.pressure;
	xtilt    = event->button.xtilt;
	ytilt    = event->button.ytilt;
#endif
	gdrv_encode_ulong(mpp, event->button.time);
	gdrv_encode_double(mpp, event->button.x);
	gdrv_encode_double(mpp, event->button.y);
	gdrv_encode_double(mpp, pressure);
	gdrv_encode_double(mpp, xtilt);
	gdrv_encode_double(mpp, ytilt);
	gdrv_encode_ulong(mpp, event->button.state);
	gdrv_encode_ulong(mpp, event->button.button);
	/* FIXME event->motion.source */
	gdrv_encode_atom(mpp, "undefined");

#if GTK_VERSION >= GTK_VERS(1,3,1)
	gdrv_encode_string(mpp, event->button.device->name);
#else
	gdrv_encode_ulong(mpp, event->button.deviceid);
#endif
	gdrv_encode_double(mpp, event->button.x_root);
	gdrv_encode_double(mpp, event->button.y_root);
	break;
    }

    case GDK_KEY_PRESS: 
    case GDK_KEY_RELEASE: 
	gdrv_encode_ulong(mpp,  event->key.time);
	gdrv_encode_ulong(mpp,  event->key.state);
	gdrv_encode_ulong(mpp,  event->key.keyval);
	gdrv_encode_long(mpp,   event->key.length);
	gdrv_encode_string(mpp, event->key.string);
	break;

    case GDK_ENTER_NOTIFY:
    case GDK_LEAVE_NOTIFY:
	encode_gdk_window(mpp, event->crossing.subwindow);
	gdrv_encode_ulong(mpp,   event->crossing.time);
	gdrv_encode_double(mpp,  event->crossing.x);
	gdrv_encode_double(mpp,  event->crossing.y);	
	gdrv_encode_double(mpp,  event->crossing.x_root);
	gdrv_encode_double(mpp,  event->crossing.y_root);
	encode_gdk_crossing_mode(mpp,   event->crossing.mode);	
	encode_gdk_notify_type(mpp,   event->crossing.detail);
	gdrv_encode_boolean(mpp, event->crossing.focus);
	gdrv_encode_ulong(mpp,   event->crossing.state);	
	break;

    case GDK_FOCUS_CHANGE:
	gdrv_encode_long(mpp, event->focus_change.in);
	break;

    case GDK_CONFIGURE: 
	gdrv_encode_long(mpp, event->configure.x);
	gdrv_encode_long(mpp, event->configure.y);
	gdrv_encode_long(mpp, event->configure.width);
	gdrv_encode_long(mpp, event->configure.height);
	break;

    case GDK_PROPERTY_NOTIFY:
	encode_gdk_atom(mpp, event->property.atom);
	gdrv_encode_ulong(mpp, event->property.time);
	gdrv_encode_ulong(mpp, event->property.state);
	break;

    case GDK_SELECTION_CLEAR: 
    case GDK_SELECTION_REQUEST:  
    case GDK_SELECTION_NOTIFY:  
	encode_gdk_atom(mpp, event->selection.selection);
	encode_gdk_atom(mpp, event->selection.target);
	encode_gdk_atom(mpp, event->selection.property);
	gdrv_encode_ulong(mpp, event->selection.requestor);
	gdrv_encode_ulong(mpp, event->selection.time);
	break;

    case GDK_PROXIMITY_IN: 
    case GDK_PROXIMITY_OUT:
	gdrv_encode_ulong(mpp, event->proximity.time);
	/* FIXME event->proximity.source */
	gdrv_encode_atom(mpp, "undefined");
#if GTK_VERSION >= GTK_VERS(1,3,1)
	gdrv_encode_string(mpp, event->proximity.device->name);
#else
	gdrv_encode_ulong(mpp, event->proximity.deviceid);
#endif
	break;
	
    case GDK_DRAG_ENTER:
    case GDK_DRAG_LEAVE: 
    case GDK_DRAG_MOTION:
    case GDK_DRAG_STATUS:
    case GDK_DROP_START:
    case GDK_DROP_FINISHED:
	/* FIXME: event->dnd.context */
	encode_gdk_drag_context(mpp, event->dnd.context);
	gdrv_encode_ulong(mpp, event->dnd.time);
	gdrv_encode_long(mpp, event->dnd.x_root);
	gdrv_encode_long(mpp, event->dnd.y_root);
	break;
	
    case GDK_CLIENT_EVENT:
	encode_gdk_atom(mpp, event->client.message_type);
	gdrv_encode_ulong(mpp, event->client.data_format);
	gdrv_encode_binary(mpp, event->client.data.b, 20);
	break;

    case GDK_VISIBILITY_NOTIFY:  
	encode_gdk_visibility_state(mpp, event->visibility.state);
	break;
	
    case GDK_NO_EXPOSE: 
	break;
    case GDK_NOTHING:
	break;
#if GTK_VERSION >= GTK_VERS(1,3,1)
    case GDK_SCROLL:
	gdrv_encode_ulong(mpp,  event->scroll.time);
	gdrv_encode_double(mpp, event->scroll.x);
	gdrv_encode_double(mpp, event->scroll.y);
	gdrv_encode_ulong(mpp,  event->scroll.state);
	encode_gdk_scroll_direction(mpp,  event->scroll.direction);
	gdrv_encode_string(mpp, event->button.device->name);
	gdrv_encode_double(mpp, event->button.x_root);
	gdrv_encode_double(mpp, event->button.y_root);
	break;
#endif
    default:
	break;
    }
    return 0;
}

/*
** Encode a pointer by inserting it into the
** pointer map if needed
*/
int gdrv_encode_boxed(gdrv_mesg_t** mpp, gpointer ptr, GtkType type)
{
    e_struct_t* sp;

    if (ptr == NULL) {
	return gdrv_encode_ulong(mpp, 0);
    }

    if (type == GDK_TYPE_EVENT) {
	return encode_gdk_event(mpp, (GdkEvent*) ptr);
    }
    else if ((type == GDK_TYPE_WINDOW)||(type == GDK_TYPE_PIXMAP) ||
	     (type == GDK_TYPE_BITMAP)||(type == GDK_TYPE_DRAWABLE)) {
	pmap_t* eptr = dataset_map(ptr, type);
	DEBUGF(("ENCODE BOXED: GtkWindow %ld\r\n", 
		(unsigned long) eptr->ptr));
	return gdrv_encode_ulong(mpp, (unsigned long) eptr);
    }
    else if ((sp = struct_lookup_by_type(type)) != NULL) {
	return sp->encode(mpp, ptr);
    }
    else if (type == GDK_TYPE_DRAG_CONTEXT) {
	return encode_gdk_drag_context(mpp, (GdkDragContext*) ptr);
    }
    else if (type == GTK_TYPE_SELECTION_DATA) {
	return encode_gtk_selection_data(mpp, (GtkSelectionData*) ptr);
    }
    else {
	/* GTK_TYPE_STYLE, GTK_TYPE_RC_STYLE, GTK_TYPE_ACCEL_GROUP,
	 * GDK_TYPE_COLORMAP,GDK_TYPE_VISUAL, GDK_TYPE_IMAGE, 
	 * GDK_TYPE_CURSOR, GDK_TYPE_GC,  
	 * ...
	 */
	/* FIXME: we must handle style destroy somehow (fixed in gtk2.0) */
	DEBUGF(("ENCODE BOXED: %s %ld\r\n", gtk_type_name(type), 
		(unsigned long) ptr));
	return gdrv_encode_ulong(mpp, (unsigned long) ptr);
    }
    return -1;
}

static int encode_fundamental(gdrv_mesg_t** mpp, gchar* ptr, GtkType type)
{
    int code;
    GtkType ftype = GTK_FUNDAMENTAL_TYPE(type);

    switch(ftype) {
    case GTK_TYPE_CHAR:
	code = gdrv_encode_char(mpp, *(char*)ptr);
	break;
    case GTK_TYPE_UCHAR:
	code = gdrv_encode_char(mpp, *(unsigned char*)ptr);
	break;
    case GTK_TYPE_BOOL:
	code = gdrv_encode_boolean(mpp, *(int*)ptr);
	break;
    case GTK_TYPE_INT:
    case GTK_TYPE_ENUM:	
	code = gdrv_encode_long(mpp, (long) *(int*)ptr);
	break;
    case GTK_TYPE_UINT:
    case GTK_TYPE_FLAGS:
	code = gdrv_encode_ulong(mpp, (unsigned long) *(unsigned int*)ptr);
	break;
    case GTK_TYPE_LONG:	
	if (type == GTK_TYPE_TYPE) {
	    char* name = (char*) gtk_type_name(*(long*)ptr);
	    if (name != NULL)
		return gdrv_encode_atom(mpp, name);
	}
	code = gdrv_encode_long(mpp, *(long*)ptr);
	break;
   case GTK_TYPE_ULONG:
	code = gdrv_encode_ulong(mpp, (unsigned long) *(unsigned long*)ptr);
	break;
    case GTK_TYPE_FLOAT:
	code = gdrv_encode_double(mpp, (double)*(float*)ptr);
	break;
    case GTK_TYPE_DOUBLE:
	code = gdrv_encode_double(mpp, *(double*)ptr);
	break;
    case GTK_TYPE_STRING: {
	char* str = *((char**) ptr);
	if (str == NULL)
	    code = gdrv_encode_empty_list(mpp);
	else
	    code = gdrv_encode_string(mpp, str);
	break;	
    }
    case G_TYPE_OBJECT:
	code = gdrv_encode_object(mpp, *(GtkObject**)ptr);
	break;

    case GTK_TYPE_POINTER:
	code = gdrv_encode_pointer(mpp,*(gpointer*)ptr, type);
	break;

    case GTK_TYPE_BOXED:
	code = gdrv_encode_boxed(mpp, *(gpointer*)ptr, type);
	break;	

    default:
	return -1;
    }
    return code;
}

/*
** input:  GtkArg*
** output: encoded value
*/
int gdrv_encode_arg(gdrv_mesg_t** mpp, GtkArg* arg)
{
    long lval;
    unsigned long ulval;
    int ival;
    int code;
    char cval;

    switch(GTK_FUNDAMENTAL_TYPE(arg->type)) {
    case GTK_TYPE_CHAR:
	cval = (char) arg->d.char_data;
	code = gdrv_encode_char(mpp, cval);
	break;

    case GTK_TYPE_UCHAR:
	cval = (char) arg->d.uchar_data;
	code = gdrv_encode_char(mpp, cval);
	break;

    case GTK_TYPE_BOOL:
	ival = (int) arg->d.bool_data;
	code = gdrv_encode_boolean(mpp, ival);
	break;

    case GTK_TYPE_INT:
    case GTK_TYPE_ENUM:	
	lval = (long) arg->d.int_data;
	code = gdrv_encode_long(mpp, lval);	
	break;

    case GTK_TYPE_UINT:
    case GTK_TYPE_FLAGS:
	ulval = (unsigned long) arg->d.uint_data;
	code = gdrv_encode_ulong(mpp, ulval);
	break;

    case GTK_TYPE_LONG:
	lval = (long) arg->d.long_data;
	code = gdrv_encode_long(mpp, lval);
	break;

   case GTK_TYPE_ULONG:
	ulval = (unsigned long) arg->d.ulong_data;
	code = gdrv_encode_ulong(mpp, ulval);
	break;

    case GTK_TYPE_FLOAT:
	code = gdrv_encode_double(mpp, arg->d.float_data);
	break;
	
    case GTK_TYPE_DOUBLE:
	code = gdrv_encode_double(mpp, arg->d.double_data);
	break;
	
    case GTK_TYPE_STRING:
	if (arg->d.string_data == NULL)
	    code = gdrv_encode_empty_list(mpp);
	else
	    code = gdrv_encode_string(mpp, arg->d.string_data);
	break;

    case G_TYPE_OBJECT:
	code = gdrv_encode_object(mpp,(GtkObject*)arg->d.object_data);
	break;

    case GTK_TYPE_POINTER:
	code = gdrv_encode_pointer(mpp,arg->d.pointer_data, arg->type);
	break;

    case GTK_TYPE_BOXED:
	code = gdrv_encode_boxed(mpp,arg->d.pointer_data, arg->type);
	break;

    default:
	goto badarg;
    }
    return code;

 badarg:
    return -1;
}

static gint e_wait_reply(unsigned long id)
{
    e_reply_t reply;

    reply.id    = id;
    reply.ready = FALSE;

    /* push on reply stack */
    reply.next = e_reply_stack;
    e_reply_stack = &reply;

    DEBUGF(("LEVEL=%d\r\n", gtk_main_level()));

    gtk_main();

    DEBUGF(("LEVEL=%d\r\n", gtk_main_level()));

    /* pop reply stack */
    e_reply_stack = reply.next;

    /* pop next level if ready */
    if ((e_reply_stack != NULL) && e_reply_stack->ready)
	gtk_main_quit();

    DEBUGF(("VALUE = %d\r\n", reply.value));
    return reply.value;
}

/*
** Wrapper for gtk_entry_key_press, gtk_text_key_press
** Purpose, queue key_press events until the key event is 
** processed by the Erlang handler if triggered.
*/
typedef struct _key_press_entry 
{
    struct _key_press_entry* next;
    GtkWidget* widget;
    GdkEventKey* event;
} key_press_entry;
    

static gint (*entry_key_press)(GtkWidget *widget, GdkEventKey *event);
#if GTK_VERSION < GTK_VERS(2,0,0)
static gint (*text_key_press)(GtkWidget *widget, GdkEventKey *event);
#endif
static gint key_press_in_progress;
static key_press_entry* key_press_first;
static key_press_entry* key_press_last;

static gint gtk_drv_call_key(GtkWidget *widget, GdkEventKey* event)
{
    gint ret = FALSE;

    if (GTK_IS_ENTRY(widget)) {
	key_press_in_progress++;
	ret = (*entry_key_press)(widget, event);
	key_press_in_progress--;
    }
#if GTK_VERSION < GTK_VERS(2,0,0)
    else if (GTK_IS_TEXT(widget)) {
	key_press_in_progress++;
	ret = (*text_key_press)(widget, event);
	key_press_in_progress--;
    }
#endif
    return ret;
}


static gint gtk_drv_key_press(GtkWidget *widget, GdkEventKey* event)
{
    key_press_entry* ent;
    gint ret = TRUE;

    if (key_press_in_progress) {
	ent = (key_press_entry*) g_malloc(sizeof(key_press_entry));
	ent->widget = widget;
	ent->event = (GdkEventKey*) gdk_event_copy((GdkEvent*) event);
	ent->next  = NULL;

	if (key_press_first == NULL)
	    key_press_first = ent;
	else
	    key_press_last->next = ent;
	key_press_last = ent;
    }
    else {
	if (key_press_first == NULL)
	    ret = gtk_drv_call_key(widget, event);
	while ((ent = key_press_first) != NULL) {
	    if ((key_press_first = ent->next) == NULL)
		key_press_last = NULL;
	    ret = gtk_drv_call_key(ent->widget, ent->event);
	    gdk_event_free((GdkEvent*)ent->event);
	    g_free(ent);
	}
    }
    return ret;
}

/* Init and patch entry and text key_press events */
static void gtk_drv_key_press_init()
{
    GtkWidgetClass* widget_class;

    key_press_in_progress = 0;
    key_press_first = NULL;
    key_press_last = NULL;

    widget_class = (GtkWidgetClass*) gtk_type_class(GTK_TYPE_ENTRY);
    entry_key_press = widget_class->key_press_event;
    widget_class->key_press_event = gtk_drv_key_press;
#if GTK_VERSION < GTK_VERS(2,0,0)
    widget_class = (GtkWidgetClass*) gtk_type_class(GTK_TYPE_TEXT);    
    text_key_press = widget_class->key_press_event;
    widget_class->key_press_event = gtk_drv_key_press;
#endif
}


/* wrapper for GtkEditable:insert text signal 
 * send signal to erlang and expect a position' as return 
 *       args will be [_,New_text',New_Text_Length, *position,_]
 */
static void handle_insert_text (GtkEditable *editable,
				const gchar *new_text,
				gint         new_text_length,
				gint        *position,
				e_closure_t* closure)
{
    int size;
    gdrv_mesg_t* mesg = NULL;
    int ev_size;
    GIOError err;
    int ret;
    unsigned long id = e_reply_id++;
    gchar small_buf[65];
    gchar* buf_ptr;

    DEBUGF(("handle_insert_text: %ld\r\n", id));

    /* calculate a upper limit for the signal */
    size = closure->msg_size + 512 + new_text_length;
    mesg = gdrv_mesg_new(size);
    ev_size = 0;

    gdrv_encode_version(&mesg);
    gdrv_encode_tuple_header(&mesg, 5);
    gdrv_encode_atom(&mesg, "signal");
    gdrv_encode_ulong(&mesg, id);
    gdrv_encode_term(&mesg, closure->pid);
    gdrv_encode_term(&mesg, closure->fun);

    gdrv_encode_list_header(&mesg, 5);

    if (closure->object_signal)
	gdrv_encode_term(&mesg, closure->data);
    else
	gdrv_encode_object(&mesg, closure->object);

    if (new_text_length >= 65)
	buf_ptr = g_malloc(new_text_length+1);
    else
	buf_ptr = small_buf;
    memcpy(buf_ptr, new_text, new_text_length);
    buf_ptr[new_text_length] = '\0';

    gdrv_encode_string(&mesg, buf_ptr);
    gdrv_encode_long(&mesg, new_text_length);
    gdrv_encode_long(&mesg, *position);

    if (closure->object_signal)
	gdrv_encode_object(&mesg, closure->object);	
    else
	gdrv_encode_term(&mesg, closure->data);

    gdrv_encode_empty_list(&mesg);

    if (buf_ptr != small_buf)
	g_free(buf_ptr);

    if (gdrv_mesg_send(erl_fd_out, mesg, &err) < 0) {
	gdrv_mesg_free(mesg);
	DEBUGF(("write error: errno = %d\r\n", err));
	gtk_main_quit();
	return;
    }
    gdrv_mesg_free(mesg);
    ret = e_wait_reply(id);
    *position = ret;
    return;
}



/* send signal as: {signal, ReplyId, Pid, Fun, [Object,...,Data]}
** push wait for E_GTK_REPLY ID Value
** while waiting for an E_GTK_REPLY
** gtk_main is called on return the 
** e_reply structure should hold the return value
*/
static gint handle_signal(e_closure_t* closure, ...)
{
    int size;
    gdrv_mesg_t* mesg = NULL;
    int ev_size;
    GIOError err;
    unsigned long id = e_reply_id++;
    va_list vargs;
    int n = closure->n_params;

    DEBUGF(("handle_signal: %ld args=%d\r\n", id, n));

    /* calculate a upper limit for the signal */
    size = closure->msg_size + 512;
    mesg = gdrv_mesg_new(size);
    ev_size = 0;

    gdrv_encode_version(&mesg);
    gdrv_encode_tuple_header(&mesg, 5);
    gdrv_encode_atom(&mesg, "signal");
    gdrv_encode_ulong(&mesg, id);
    gdrv_encode_term(&mesg, closure->pid);
    gdrv_encode_term(&mesg, closure->fun);

    gdrv_encode_list_header(&mesg, 2+n);

    if (closure->object_signal)
	gdrv_encode_term(&mesg, closure->data);
    else
	gdrv_encode_object(&mesg, closure->object);

    {
	char* error;
	int i;

	va_start (vargs, closure);
	for (i = 0; i < n; i++) {
	    GtkArg arg;
	    arg.type = closure->param_type[i];
	    GTK_ARG_COLLECT_VALUE((&arg),vargs,error);
	    if (error) {
		g_printerr("e_handle_signal collect error %s\n", error);
		goto error;
	    }
	    if (gdrv_encode_arg(&mesg, &arg) < 0) {
		g_printerr("e_handle_signal encode error\n");
		goto error;
	    }
	}
	va_end (vargs);
    }

    if (closure->object_signal)
	gdrv_encode_object(&mesg, closure->object);	
    else
	gdrv_encode_term(&mesg, closure->data);

    gdrv_encode_empty_list(&mesg);

    if (gdrv_mesg_send(erl_fd_out, mesg, &err) < 0) {
	gdrv_mesg_free(mesg);
	DEBUGF(("write error: errno = %d\r\n", err));
	gtk_main_quit();
	return FALSE;
    }
    gdrv_mesg_free(mesg);

    return e_wait_reply(id);

 error:
    if (mesg != NULL)
	gdrv_mesg_free(mesg);
    return FALSE;
}


/* Called when the signal object is destroyed */
static void closure_destroy(e_closure_t* closure)
{
    erl_free_term(closure->fun);
    erl_free_term(closure->pid);
    erl_free_term(closure->data);
    g_free(closure);
}

/* Called when an object is destroyed 
** send signal as: {destroy, Object} 
*/
static void e_destroy_widget(GtkObject* obj, gpointer* data)
{
    GIOError err;
    gdrv_mesg_t* mesg;

    mesg = gdrv_mesg_new(40);

    gdrv_encode_version(&mesg);
    gdrv_encode_tuple_header(&mesg, 2);
    gdrv_encode_atom(&mesg, "destroy");
    gdrv_encode_object(&mesg, obj);

    if (gdrv_mesg_send(erl_fd_out, mesg, &err) < 0) {
	gdrv_mesg_free(mesg);
	DEBUGF(("write error: errno = %d\n", err));
	gtk_main_quit();
	return;
    }
    gdrv_mesg_free(mesg);
}

/* input: {ObjectId, SignalName, [{name,type,value}]} */
/* output: {ok,ok} | {ok,Value} | {error, Reason} */
static int signal_emit(gdrv_mesg_t* mp, gdrv_mesg_t** mpp)
{
    int i;
    int nargs;
    int nvsize;
    int arity;
    GtkObject* object;
    GtkArg* av;
    char name[NAME_SIZE];

    if ((gdrv_decode_tuple_header(mp, &arity) < 0) ||
	(arity != 3))
	return -1;
    if (decode_object(mp, &object) < 0) 
	return -1;
    if (gdrv_decode_atom(mp, name, NAME_SIZE) < 0)
	return -1;
    if (gdrv_decode_list_header(mp, &nargs) < 0) 
	return -1;
    nvsize = nargs+1; /* extra space for possible return value */    
    av = (GtkArg*) ARG_STACK_ALLOC(sizeof(GtkArg)*nvsize);
    for (i = 0; i < nargs; i++) {
	if ((gdrv_decode_arg(mp, &av[i])) < 0)
	    return -1;
    }
    gtk_signal_emitv_by_name(object, name, av);
    if (mpp != NULL) {
	/* Can we handle reply from signal??? */
	gdrv_encode_atom(mpp, "ok");
    }
    return 0;
}

/* function_get
** input:  {Mod, Function}
** output: {ok, Quark} | {error, Reason}
*/
static int function_lookup(gdrv_mesg_t* mp, gdrv_mesg_t** mpp)
{
    int arity;
    GQuark quark = 0;
    char mod_name[NAME_SIZE];
    char fun_name[NAME_SIZE];

    if ((gdrv_decode_tuple_header(mp, &arity) < 0) || (arity != 2))
	return -1;
    if (gdrv_decode_atom(mp, mod_name, NAME_SIZE) < 0)
	return -1;	
    if (gdrv_decode_atom(mp, fun_name, NAME_SIZE) < 0)
	return -1;    
    quark = exp_try_quark(mod_name, ':', fun_name);
    if (quark) {
	gdrv_encode_ulong(mpp, (unsigned long) quark);
	return 0;
    }
    return -1;
}


/*
** input:  {Mod, Function, [{name,type,value}]}
**       | {Quark, [{name,type,value}]}
**
** output: Value | {Value,Out1,...,Outn} | Bool,Object
**  void functions return ok as Value
*/
static int function_call(gboolean constructor,
			 gdrv_mesg_t* mp, gdrv_mesg_t** mpp)
{
    int i;
    int i_args;
    int o_args;
    int nvsize;
    int arity;
    GtkArg* av;
    GtkArg  ret;
    GQuark quark = 0;
    e_export_t* exp;
    char mod_name[NAME_SIZE];
    char fun_name[NAME_SIZE];

    if (gdrv_decode_tuple_header(mp, &arity) < 0)
	return -1;
    if (arity == 3) {
	if (gdrv_decode_atom(mp, mod_name, NAME_SIZE) < 0)
	    return -1;	
	if (gdrv_decode_atom(mp, fun_name, NAME_SIZE) < 0)
	    return -1;
	if ((quark = exp_try_quark(mod_name, ':', fun_name)) == 0)
	    return -1;
    }
    else if (arity == 2) {
	if (gdrv_decode_long(mp, (long*)&quark) < 0)
	    return -1;
    }

    exp = (e_export_t*) g_hash_table_lookup(exp_hash_table,
					    GUINT_TO_POINTER(quark));
    if (exp == NULL) /* no such functions */
	return -1;

    DEBUGF(("function_call: %s:%s\r\n", exp->mod_name, exp->fun_name));

    if ((gdrv_decode_list_header(mp, &i_args) < 0) ||
	(i_args > exp->arity))  /* to many args */
	return -1;
    o_args = 0;
    nvsize = exp->arity + 1;

    av = (GtkArg*) ARG_STACK_ALLOC(sizeof(GtkArg)*nvsize);

    for (i = 0; i < exp->arity; i++) {
	if (exp->arg_type[i].status & ARG_IN) {
	    if (i_args <= 0)  /* to few arguments */
		return -1;
	    if (gdrv_decode_arg(mp, &av[i]) < 0)
		return -1;
	    /* FIXME: check arg types (includes some casts!!) */
	    i_args--;
	}
	else if (exp->arg_type[i].status == ARG_OUT) {
	    GtkType ftype;

	    av[i].type = exp->arg_type[i].type;	    
	    ftype = GTK_FUNDAMENTAL_TYPE(av[i].type);
	    if (ftype == GTK_TYPE_BOXED) {
		e_struct_t* stc = struct_lookup_by_type(av[i].type);

		if (stc != NULL) {
		    void* mem = ARG_STACK_ALLOC(stc->size);
		    GTK_VALUE_POINTER(av[i]) = mem;
		}
		else 
		    GTK_VALUE_POINTER(av[i]) = NULL;
	    }
	    else if (IS_ARRAY(av[i].type)) {
		GtkArray* arr= (GtkArray*) ARG_STACK_ALLOC(sizeof(GtkArray));
		DEBUGF(("array output argument\r\n"));
		arr->elem_type = array_element_type(exp->arg_type[i].type);
		arr->size = 0;
		arr->vector = NULL;
		av[i].type = GTK_TYPE_ARRAY;
		GTK_VALUE_POINTER(av[i]) = (void*) arr;
	    }
	    else
		GTK_VALUE_POINTER(av[i]) = NULL;
	}
	if (exp->arg_type[i].status & ARG_OUT)
	    o_args++;
    }

    ret.type = exp->return_type;

    switch(GTK_FUNDAMENTAL_TYPE(exp->return_type)) {
    case GTK_TYPE_INVALID:
    case GTK_TYPE_NONE:
	GTK_VALUE_POINTER(av[exp->arity]) = NULL;
	break;
    case GTK_TYPE_CHAR:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_CHAR(ret);
	break;
    case GTK_TYPE_UCHAR:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_UCHAR(ret);
	break;
    case GTK_TYPE_BOOL:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_BOOL(ret);
	break;
    case GTK_TYPE_INT:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_INT(ret);
	break;
    case GTK_TYPE_UINT:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_UINT(ret);
	break;
    case GTK_TYPE_LONG:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_LONG(ret);
	break;
    case GTK_TYPE_ULONG:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_ULONG(ret);
	break;
    case GTK_TYPE_FLOAT:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_FLOAT(ret);
	break;
    case GTK_TYPE_DOUBLE:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_DOUBLE(ret);
	break;
    case GTK_TYPE_STRING:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_STRING(ret);
	break;
    case GTK_TYPE_ENUM:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_ENUM(ret);
	break;
    case GTK_TYPE_FLAGS:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_FLAGS(ret);
	break;
    case GTK_TYPE_BOXED:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_BOXED(ret);
	break;
    case GTK_TYPE_POINTER:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_POINTER(ret);
	break;
    case G_TYPE_OBJECT:
	GTK_VALUE_POINTER(av[exp->arity]) = &GTK_VALUE_OBJECT(ret);
	break;
    default:
	return -1;
    }

    (*exp->marshaller)((GtkSignalFunc)exp->function, NULL, av);

    if (constructor) {
	/* FIXME: allow constructor to have out args */
	if ((GTK_FUNDAMENTAL_TYPE(ret.type) == G_TYPE_OBJECT)) {
	    GtkObject* object = GTK_VALUE_POINTER(ret);
	    if (object != NULL) {
		gtk_signal_connect(object, 
				   "destroy", 
				   GTK_SIGNAL_FUNC(e_destroy_widget),
				   NULL);
		gdrv_encode_boolean(mpp, GTK_IS_WINDOW(object));
		gdrv_encode_object(mpp, object);
		return 1;
	    }
	    return -1;
	}
	else {
	    /* assume contructor syntax on a regular data structure,
	     * GtkStyle, GtkAccelGroup .. (Gtk-1.2.x)
	     * GdkImage ..
	     */
	    gdrv_encode_boolean(mpp, FALSE),
	    gdrv_encode_arg(mpp, &ret);
	    return 1;
	}
    }
    else {
	if (o_args > 0)
	    gdrv_encode_tuple_header(mpp, o_args+1);
	
	if (exp->return_type == GTK_TYPE_NONE)
	    gdrv_encode_atom(mpp, "ok");
	else
	    gdrv_encode_arg(mpp, &ret);

	if (o_args > 0) {
	    for (i = 0; i < exp->arity; i++) {
		if (exp->arg_type[i].status & ARG_OUT)
		    gdrv_encode_arg(mpp, &av[i]);
	    }
	}
	return 1;
    }
}


/* dispatch gtk call:
** message structure is {InternalFunc, CallID, Data}
*/
static gdrv_mesg_t* dispatch(gdrv_mesg_t* i_mesg)
{
    int ver;
    int arity;
    long e_op;
    long callid;
    gdrv_mesg_t* o_mesg;
    stack_elem_t* save_sp;
    char* strerr = "";

    o_mesg = gdrv_mesg_new(256);

    ARG_STACK_SAVE(&save_sp);

    if (gdrv_decode_version(i_mesg, &ver) < 0) {
	g_printerr("gtk_drv: bad binary data (version)\n");
	return NULL;
    }

    if ((gdrv_decode_tuple_header(i_mesg, &arity) < 0) ||
	(arity != 3)) {
	g_printerr("gtk_drv: bad binary data (tuple)\n");
	return NULL;
    }

    if (gdrv_decode_long(i_mesg, &e_op) < 0) {
	g_printerr("gtk_drv: bad command (function index)\n");
	return NULL;
    }
    if (gdrv_decode_long(i_mesg, &callid) < 0) {
	g_printerr("gtk_drv: bad command (call id)\n");
	return NULL;
    }

    gdrv_encode_version(&o_mesg);
    gdrv_encode_tuple_header(&o_mesg, 3);
    gdrv_encode_atom(&o_mesg, "ok");
    gdrv_encode_long(&o_mesg, callid);

    switch(e_op) {

    case E_GTK_TYPE_ENUM_FIND_VALUE: {
	/* args: {(long) enum-type, (string) name} */
	/* return: integer */
	long type;
	char name[NAME_SIZE];
	GtkEnumValue* v;

	if ((gdrv_decode_tuple_header(i_mesg, &arity) < 0) ||
	    (arity != 2))
	    goto badarg;

	if (gdrv_decode_long(i_mesg, &type) < 0)
	    goto badarg;
	if (gdrv_decode_atom(i_mesg, name, NAME_SIZE) < 0)
	    goto badarg;

	if ((v = gtk_type_enum_find_value((GtkType)type, name)) == NULL)
	    goto badarg;
	gdrv_encode_long(&o_mesg, (long) v->value);
	goto reply;
    }

    case E_GTK_TYPE_FLAGS_FIND_VALUE: {
	/* args: {(long) flag-type, (string) name} */
	/* return: integer */
	long type;
	char name[NAME_SIZE];
	GtkFlagValue* v;

	if ((gdrv_decode_tuple_header(i_mesg, &arity) < 0) ||
	    (arity != 2))
	    goto badarg;

	if (gdrv_decode_long(i_mesg, &type) < 0)
	    goto badarg;
	if (gdrv_decode_atom(i_mesg, name, NAME_SIZE) < 0)
	    goto badarg;

	if ((v = gtk_type_flags_find_value((GtkType)type, name)) == NULL)
	    goto badarg;
	gdrv_encode_ulong(&o_mesg, (unsigned long) v->value);
	goto reply;
    }

    case E_GTK_ARG_GET_TYPE: {
	/* args: { object-type, arg-name } */
	/* return: arg-type */
	GtkType type;
	char name[NAME_SIZE];
#if GTK_VERSION >= GTK_VERS(2,0,0)
	GParamSpec* pspec;
#else
	GtkArgInfo* info;
	gchar* error;
#endif
	if ((gdrv_decode_tuple_header(i_mesg, &arity) < 0) ||
	    (arity != 2))
	    goto badarg;

        DEBUGF(("ARG_GET_TYPE\n"));
	if (decode_gtk_type(i_mesg, &type) < 0)
	    goto badarg;
        DEBUGF(("ARG_GET_TYPE: type=%d\r\n", type));
	if (gdrv_decode_atom(i_mesg, name, NAME_SIZE) < 0)
	    goto badarg;
        DEBUGF(("ARG_GET_TYPE: name=%s\r\n", name));
#if GTK_VERSION >= GTK_VERS(2,0,0)
	if ((pspec = g_object_class_find_property(
						  G_OBJECT_GET_CLASS(type),
						  name)) == NULL)
	    goto badarg;
	gdrv_encode_long(&o_mesg, (long) pspec->value_type);
	goto reply;
#else
	if ((error = gtk_object_arg_get_info(type,
					     name,
					     &info)) != NULL) {
	    DEBUGF(("ARG_GET_TYPE: error=%s\r\n", error));
	    g_free(error);
	    goto badarg;
	}
	gdrv_encode_long(&o_mesg, (long) info->type);
	goto reply;
#endif
    }

    case E_GTK_CHILD_ARG_GET_TYPE: {
	/* args: {object-type, arg-name} */
	/* return: arg-type */
	GtkType type;
	char name[NAME_SIZE];
#if GTK_VERSION >= GTK_VERS(2,0,0)
	GParamSpec* pspec;
#else
	GtkArgInfo* info;
	gchar* error;
#endif
	if ((gdrv_decode_tuple_header(i_mesg, &arity) < 0) ||
	    (arity != 2))
	    goto badarg;

        DEBUGF(("CHILD_ARG_GET_TYPE\n"));
	if (decode_gtk_type(i_mesg, &type) < 0)
	    goto badarg;
        DEBUGF(("CHILD_ARG_GET_TYPE: type=%d\r\n", type));
	if (gdrv_decode_atom(i_mesg, name, NAME_SIZE) < 0)
	    goto badarg;
        DEBUGF(("CHILD_ARG_GET_TYPE: name=%s\r\n", name));
#if GTK_VERSION >= GTK_VERS(2,0,0)
	if ((pspec = 
	     gtk_container_class_find_child_property(G_OBJECT_GET_CLASS(type),
						     name)) == NULL)
	    goto badarg;
	gdrv_encode_long(&o_mesg, (long) pspec->value_type);
	goto reply;
#else
	if ((error = gtk_container_child_arg_get_info(type,
						      name,
						      &info)) != NULL) {
	    DEBUGF(("CHILD_ARG_GET_TYPE: error=%s\r\n", error));
	    g_free(error);
	    goto badarg;
	}
	gdrv_encode_long(&o_mesg, (long) info->type);
	goto reply;
#endif
    }

    case E_GTK_ASYNC_SIGNAL_EMIT: 
	signal_emit(i_mesg, NULL);
	goto noreply;

    case E_GTK_SIGNAL_EMIT: {
	if (signal_emit(i_mesg, &o_mesg) < 0)
	    goto badarg;
	goto reply;
    }

    case E_GTK_SIGNAL_CONNECT: {
	/* input: {ObjectId, SignalName, After, Object, Fun, Pid, Data} */
	/* output: {ok,HandleID} | {error, reason} */
	int after;
	int object_signal;
	GtkObject* object;
	char name[NAME_SIZE];
	guint hid;
	guint signal_id;

	if ((gdrv_decode_tuple_header(i_mesg, &arity) < 0) ||
	    (arity != 7))
	    goto badarg;
	if (decode_object(i_mesg, &object) < 0) goto badarg;
	if (gdrv_decode_atom(i_mesg, name, NAME_SIZE) < 0)   goto badarg;
	if (gdrv_decode_boolean(i_mesg, &after) < 0) goto badarg;
	if (gdrv_decode_boolean(i_mesg, &object_signal) < 0) goto badarg;
	signal_id = gtk_signal_lookup(name, GTK_OBJECT_TYPE(object));
	if (!signal_id)
	    goto badarg;

	switch(i_mesg->data[i_mesg->ix]) {
#ifdef ERL_FUN_EXT
	case ERL_FUN_EXT:
#endif
#ifdef ERL_NEW_FUN_EXT
	case ERL_NEW_FUN_EXT:
#endif
	case ERL_BINARY_EXT:  /* accept "wrapped" functions */
	case ERL_SMALL_TUPLE_EXT:
	case ERL_LARGE_TUPLE_EXT: {
	    ETERM* data = NULL;
	    ETERM* pid  = NULL;
	    ETERM* fun  = NULL;
	    e_closure_t* closure;

	    if (gdrv_decode_term(i_mesg, (void*) &fun) < 0)
		goto badarg;
	    if (gdrv_decode_term(i_mesg, (void*) &pid) < 0) {
		erl_free_term(fun);
		goto badarg;
	    }
	    if (gdrv_decode_term(i_mesg, (void*) &data) < 0) {
		erl_free_term(fun);
		erl_free_term(pid);
		goto badarg;
	    }

	    /* we need to query the signal for number of arguments */
	    if ((closure = create_signal_closure(signal_id)) == NULL) {
		erl_free_term(fun);
		erl_free_term(pid);
		goto badarg;
	    }
	    closure->pid  = pid;
	    closure->data = data;
	    closure->fun  = fun;
	    closure->object = object;
	    closure->object_signal = object_signal;
	    closure->msg_size = erl_term_len(pid) + 
		erl_term_len(data) + erl_term_len(fun);

	    if (GTK_IS_EDITABLE(object) && (strcmp(name, "insert_text") == 0)){
		hid = gtk_signal_connect_full(object,
					      name,
					      (GtkSignalFunc) handle_insert_text,
					      (GtkCallbackMarshal) NULL,
					      (gpointer) closure,
					      (GtkDestroyNotify)closure_destroy,
					      (gint) FALSE,
					      (gint) after);
	    }
	    else {
		/* gtk_signal_connect_full will lookup up the signal again! */
		hid = gtk_signal_connect_full(object,
					      name,
					      (GtkSignalFunc) handle_signal,
					      (GtkCallbackMarshal) NULL,
					      (gpointer) closure,
					      (GtkDestroyNotify)closure_destroy,
					      (gint) TRUE,
					      (gint) after);
	    }
	    break;
	}


	default:
	    goto badarg;

	}
	gdrv_encode_ulong(&o_mesg, (unsigned long) hid);
	goto reply;
    }

    case E_GTK_REPLY: {
	/* input: {ReplyId, true | false| integer } */
	unsigned long id;
	gint value;
	long lval;
	e_reply_t* ptr;

	if ((gdrv_decode_tuple_header(i_mesg, &arity) < 0) ||
	    (arity != 2))
	    goto noreply;
	if (gdrv_decode_ulong(i_mesg, &id) < 0)
	    goto noreply;
	if (gdrv_decode_boolean(i_mesg, &value) < 0) {
	    if (gdrv_decode_long(i_mesg, &lval) < 0)
		goto noreply;
	    value = lval;
	}
	if ((ptr = e_reply_stack) != NULL) {
	    while(ptr != NULL) {
		if (ptr->id == id) {
		    ptr->value = value;
		    ptr->ready = TRUE;
		    break;
		}
		ptr = ptr->next;
	    }
	    if (e_reply_stack->id == id)
		gtk_main_quit();
	}
	goto noreply;
    }
	
    case E_GTK_FUNCTION_CALL: {
	/* input: {Function_name, [{name,type,value}...]} */
	/* output: ok | {ok,Value} | {error,Reason} */
	if (function_call(FALSE, i_mesg, &o_mesg) < 0)
	    goto badarg;
	goto reply;
    }

    case E_GTK_ASYNC_CALL: {
	/* input: {Function_name, [{name,type,value}...]} */
	/* output: ok | {ok,Value} | {error,Reason} */
	function_call(FALSE, i_mesg, &o_mesg);
	goto noreply;
    }
	
    case E_GTK_NEW_CALL: {
	/* input: {Function_name, [{name,type,value}...]} 
	 * output: {created,ID,TopLevel,Value} | {error,ID,Reason} 
	 * Where TopLevel=true|false
	 */
	o_mesg->ix = 0;
	gdrv_encode_version(&o_mesg);
	gdrv_encode_tuple_header(&o_mesg, 4);
	gdrv_encode_atom(&o_mesg, "created");
	gdrv_encode_long(&o_mesg, callid);

	if (function_call(TRUE, i_mesg, &o_mesg) < 0)
	    goto badarg;
	goto reply;
    }

    case E_GTK_MODULE_DIR: {
        /* Input: Directory
         * Ouput: ok | {error, Reason}
         */
        if (gdrv_decode_strn(i_mesg, module_dir, 1024) < 0)
            goto badarg;
        gdrv_encode_atom(&o_mesg, "ok");
        goto reply;
    }

    case E_GTK_LOAD_MODULE: {
        /* Input: ModuleName 
         * Ouput: ok | {error, Reason}
         */
        char mod_name[256];
        GModule* mod;
        void (*initf)();
        void (*exitf)();
        char* path;

        if (gdrv_decode_strn(i_mesg, mod_name, 256) < 0)
            goto badarg;
        if (!g_module_supported()) {
            strerr = "moduled not supported";
            goto badstrarg;
        }

        DEBUGF(("LOAD_MODULE: %s\r\n", mod_name));

	if (mod_find(mod_name) != NULL) {
	    DEBUGF(("LOAD_MODULE: already loaded\r\n"));
	    gdrv_encode_atom(&o_mesg, "ok");
	    goto reply;
	}

        path = g_module_build_path(module_dir, mod_name);
        mod = g_module_open(path, 0);
        g_free(path);

        if (mod == NULL) {
            strerr = (char*) g_module_error();
            DEBUGF(("LOAD_MODULE: error %s\r\n", strerr));
            goto badstrarg;
        }

        if (!g_module_symbol(mod, "module_init", (gpointer) &initf)) {
            g_module_close(mod);
            strerr = "module_init not found";
            goto badstrarg;
        }
	if (!g_module_symbol(mod, "module_exit", (gpointer) &exitf))
	    exitf = NULL;

	mod_insert(mod, mod_name, initf, exitf);
	
        gdrv_encode_atom(&o_mesg, "ok");
        goto reply;
    }

    case E_GTK_UNLOAD_MODULE: {
        /*
         * Input: ModuleName
         * Ouput: ok | {error,Reason}
         */
        char mod_name[256];

        if (gdrv_decode_strn(i_mesg, mod_name, 256) < 0)
            goto badarg;
        DEBUGF(("UNLOAD_MODULE: %s\r\n", mod_name));

	mod_remove(mod_name);
        gdrv_encode_atom(&o_mesg, "ok");  
        goto reply;
    }


    case E_GTK_FIND_FUNCTION: {
        if (function_lookup(i_mesg, &o_mesg) < 0)
            goto badarg;
        goto reply;
    }

    default:
	goto badarg;
    }


 badarg:
    ARG_STACK_RESTORE(&save_sp);
    o_mesg->ix = 0;
    gdrv_encode_version(&o_mesg);
    gdrv_encode_tuple_header(&o_mesg, 3);
    gdrv_encode_atom(&o_mesg, "error");
    gdrv_encode_long(&o_mesg, callid);
    gdrv_encode_atom(&o_mesg, "badarg");
    return o_mesg;

badstrarg:
    ARG_STACK_RESTORE(&save_sp);
    o_mesg->ix = 0;
    gdrv_encode_version(&o_mesg);
    gdrv_encode_tuple_header(&o_mesg, 3);
    gdrv_encode_atom(&o_mesg, "error");
    gdrv_encode_long(&o_mesg, callid);
    gdrv_encode_string(&o_mesg, strerr);
    return o_mesg;

 reply:
    ARG_STACK_RESTORE(&save_sp);
    return o_mesg;

 noreply:
    ARG_STACK_RESTORE(&save_sp);
    gdrv_mesg_free(o_mesg);
    return NULL;
}


/* read a command and dispatch */
static gboolean input(GIOChannel   *source,
		      GIOCondition  condition,
		      gpointer      data)
{
    gdrv_mesg_t* request;
    gdrv_mesg_t* reply;
    GIOError err;

    DEBUGF(("input condition = %x\r\n", condition));

    if (condition & (G_IO_HUP | G_IO_ERR)) {
	gtk_main_quit();
    }
    else if (condition & G_IO_IN) {
	if ((request = gdrv_mesg_recv(source, &err)) == NULL) {
	    DEBUGF(("read error: errno = %d\n", err));
	    gtk_main_quit();
	}
	else {
	    reply = dispatch(request);
	    if (reply != NULL) {
		if (gdrv_mesg_send(erl_fd_out, reply, &err) < 0) {
		    DEBUGF(("write error: errno = %d\n", err));
		    gtk_main_quit();
		}
		gdrv_mesg_free(reply);
	    }
	    gdrv_mesg_free(request);
	}
    }
    return TRUE;
}


static SOCKET do_connect(unsigned short port)
{
    struct sockaddr_in a;
    SOCKET s = INVALID_SOCKET;
    int len = sizeof(a);
    int on = 1;

#ifdef __WIN32__
    WORD wVersionRequested;
    WSADATA wsaData;

    wVersionRequested = MAKEWORD(2,0);
    if (WSAStartup(wVersionRequested, &wsaData) != 0)
      goto error;
    if ((LOBYTE(wsaData.wVersion) != 2) || (HIBYTE(wsaData.wVersion) != 0))
      goto error;
#endif
    if ((s = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
      goto error;
    memset((char*) &a, 0, sizeof(a));
    a.sin_family = AF_INET;
    a.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    a.sin_port = htons(port);
    if (connect(s, (struct sockaddr*) &a, len) == SOCKET_ERROR)
	goto error;
    /* The TCP_NODELAY option is needed */ 
    setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char*) &on, sizeof(int)); 
    return s;

 error:
    if (s != INVALID_SOCKET)
	CLOSESOCKET(s);
#ifdef __WIN32__
    WSACleanup();
#endif    
    return INVALID_SOCKET;
}


static void debug_init()
{
#ifdef DEBUG

#if defined(__WIN32__) || defined(DEBUG_TO_GTK_LOG)
    logf = fopen("gtk.log", "w");
    setvbuf(logf, NULL, _IONBF, 0);
#elif defined(DEBUG_TO_STDERR)
    logf = stderr;
#endif

#endif
}

guint watch_io(GIOChannel* channel, GIOCondition condition)
{
#if GTK_VERSION >= GTK_VERS(2,0,0)
    GSource *source;
    guint id;

    source = g_io_create_watch (channel, condition);

    g_source_set_callback(source, (GSourceFunc)input, NULL, NULL);

    g_source_set_can_recurse(source, TRUE); /* This is needed */

    id = g_source_attach (source, NULL);
    g_source_unref (source);
    return id;
#else
    return g_io_add_watch_full (erl_fd_in,
				G_PRIORITY_DEFAULT, 
				G_IO_IN | G_IO_HUP | G_IO_ERR,
				input,
				NULL, NULL);
#endif 
}


#ifdef __WIN32__
int
WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
		    PSTR szCmdLine, int iCmdShow)
{
    int argc = __argc;
    char **argv = __argv;
#else
int
main(int argc, char **argv)
{
#endif
    SOCKET socket;
    GIOError result;

    debug_init();

    if (argc >= 2) {
	unsigned short port = atoi(argv[1]);
	if ((socket = do_connect(port)) == INVALID_SOCKET) {
	    DEBUGF(("could not connect to port %d\r\n", port));
	    exit(1);
	}
	DEBUGF(("gtk connected to localhost:%d\r\n", port));
    }
    else 
	socket = INVALID_SOCKET;

    erl_init(NULL, 0);
    gdrv_mesg_init();
#ifdef NEED_GNOMESUPPORT_H
    gnome_init("gtk_drv", ERLGTK_VSN, argc, argv);
#else
    gtk_init(&argc, &argv);
#endif
    arg_stack_init();

    pmap_table = g_hash_table_new(g_direct_hash, g_direct_equal);
    pmap_free  = NULL;

    exp_hash_table = g_hash_table_new (g_direct_hash, g_direct_equal);
    stc_hash_table = g_hash_table_new (g_direct_hash, g_direct_equal);

    gtk_drv_type_init();

    gtk_drv_key_press_init();     /* patch key_press_event */

#ifdef PRELOADED
    DEBUGF(("loading preloaded modules\r\n"));
    /* Load preloaded modules */
    mod_insert(NULL, "gdk", gdk_module_init, NULL);
    mod_insert(NULL, "gtk", gtk_module_init, NULL);
#endif

    DEBUGF(("gtk adding input\r\n"));

#ifdef __WIN32__
    if (socket == INVALID_SOCKET)
	erl_fd_in = g_io_channel_win32_new_fd (0);
    else
	erl_fd_in = g_io_channel_win32_new_stream_socket (socket);
#else
    if (socket == INVALID_SOCKET)
	erl_fd_in = g_io_channel_unix_new (0);	
    else
	erl_fd_in = g_io_channel_unix_new (socket);
#endif
    result = watch_io(erl_fd_in,  G_IO_IN | G_IO_HUP | G_IO_ERR);
    erl_fd_out = erl_fd_in;
    gtk_main();
    exit(0);
}


