%%% File    : clist.erl
%%% Author  : Tony Rogvall <tony@bit.hemma.se>
%%% Purpose : clist example
%%% Created : 17 Dec 2000 by Tony Rogvall <tony@bit.hemma.se>

-module(clist).
-author('tony@bit.hemma.se').

-compile(export_all).

%% example-start clist clist.c 

-include("../include/gtk.hrl").

%% User clicked the "Add List" button. 
button_add_clicked( Button, Data ) ->
    %% Something silly to add to the list. 4 rows of 2 columns each */
    Drink = [ { "Milk",    "3 Oz" },
	      { "Water",   "6 l" },
	      { "Carrots", "2" },
	      { "Snakes",  "55" } ],

    %% Here we do the actual adding of the text. It's done once for
    %% each row.
    %%
    lists:foreach(
      fun(Row) ->
	      gtk:clist_append(Data, Row)
      end, Drink).


%% User clicked the "Clear List" button. 
button_clear_clicked(Button, Data ) ->
    %% Clear the list using gtk_clist_clear. This is much faster than
    %% calling gtk_clist_remove once for each row.
    %%
    gtk:clist_clear(Data).


%% The user clicked the "Hide/Show titles" button. 
button_hide_show_clicked(Button, Data) ->
    %% Just a flag to remember the status. 0 = currently visible 
    Flag = case get(flag) of
	       undefined -> 0;
	       N -> N
	   end,
    if Flag == 0 ->
	    %% Hide the titles and set the flag to 1 */
	    gtk:clist_column_titles_hide(Data),
	    put(flag, 1);
       Flag == 1 ->
	    %% Show the titles and reset flag to 0 
	    gtk:clist_column_titles_show(Data),
	    put(flag, 0)
    end.


%% If we come here, then the user has selected a row in the list. 
selection_made( Clist, Row, Column, Event, Data) ->
    %% Get the text that is stored in the selected row and column
    %% which was clicked in. We will receive it as a pointer in the
    %% argument text.
    %%
    {_, Text} = gtk:clist_get_text(?GTK_CLIST(Clist), Row, Column), %% , &text),
    %% Just prints some information about the selected row 
    io:format("You selected row ~w. More specifically you clicked in "
	      "column ~w, and the text in this cell is ~s\n\n",
	      [Row, Column, Text]).

run() ->
    gtk:start(),
    spawn_link(?MODULE, main, []).

main( ) ->
    Titles = { "Ingredients", "Amount" },

    Window=gtk:window_new('GTK_WINDOW_TOPLEVEL'),
    gtk:widget_set_usize(?GTK_WIDGET(Window), 300, 150),

    gtk:window_set_title(?GTK_WINDOW(Window), "GtkCList Example"),
    gtk:signal_connect(?GTK_OBJECT(Window),
		       'destroy',
		       fun(_,_) -> gtk:main_quit() end, 
		       ?NULL),
    
    Vbox=gtk:vbox_new(false, 5),
    gtk:container_set_border_width(?GTK_CONTAINER(Vbox), 5),
    gtk:container_add(?GTK_CONTAINER(Window), Vbox),
    gtk:widget_show(Vbox),
    
    %% Create a scrolled window to pack the CList widget into 
    Scrolled_window = gtk:scrolled_window_new (?GTK_NULL, ?GTK_NULL),
    gtk:scrolled_window_set_policy (?GTK_SCROLLED_WINDOW (Scrolled_window),
                                    'GTK_POLICY_AUTOMATIC', 
				    'GTK_POLICY_ALWAYS'),

    gtk:box_pack_start(?GTK_BOX(Vbox), Scrolled_window, true, true, 0),
    gtk:widget_show (Scrolled_window),

    %% Create the CList. For this example we use 2 columns 
    Clist = gtk:clist_new_with_titles( 2, Titles),

    %% When a selection is made, we want to know about it. The callback
    %% used is selection_made, and its code can be found further down */
    gtk:signal_connect(?GTK_OBJECT(Clist), 'select_row',
		       fun selection_made/5, ?NULL),


    %% It isn't necessary to shadow the border, but it looks nice :) 
    gtk:clist_set_shadow_type (?GTK_CLIST(Clist), 'GTK_SHADOW_OUT'),

    %% What however is important, is that we set the column widths as
    %% they will never be right otherwise. Note that the columns are
    %% numbered from 0 and up (to 1 in this case).
    %%
    gtk:clist_set_column_width (?GTK_CLIST(Clist), 0, 150),

    %% Add the CList widget to the vertical box and show it. 
    gtk:container_add(?GTK_CONTAINER(Scrolled_window), Clist),
    gtk:widget_show(Clist),

    %% Create the buttons and add them to the window. See the button
    %% tutorial for more examples and comments on this.
    %%
    Hbox = gtk:hbox_new(false, 0),
    gtk:box_pack_start(?GTK_BOX(Vbox), Hbox, false, true, 0),
    gtk:widget_show(Hbox),

    Button_add = gtk:button_new_with_label("Add List"),
    Button_clear = gtk:button_new_with_label("Clear List"),
    Button_hide_show = gtk:button_new_with_label("Hide/Show titles"),

    gtk:box_pack_start(?GTK_BOX(Hbox), Button_add, true, true, 0),
    gtk:box_pack_start(?GTK_BOX(Hbox), Button_clear, true, true, 0),
    gtk:box_pack_start(?GTK_BOX(Hbox), Button_hide_show, true, true, 0),

    %% Connect our callbacks to the three buttons 
    gtk:signal_connect(?GTK_OBJECT(Button_add), 'clicked',
		       fun button_add_clicked/2, Clist),
    gtk:signal_connect(?GTK_OBJECT(Button_clear), 'clicked',
		       fun button_clear_clicked/2, 
		       Clist),
    gtk:signal_connect(?GTK_OBJECT(Button_hide_show), 'clicked',
		       fun button_hide_show_clicked/2,
		       Clist),

    gtk:widget_show(Button_add),
    gtk:widget_show(Button_clear),
    gtk:widget_show(Button_hide_show),

    %% The interface is completely set up so we show the window and
    %% enter the gtk_main loop.
    %%
    gtk:widget_show(Window),
    gtk:main().

%% example-end 
