%%% File    : progressbar.erl
%%% Author  : Tony Rogvall <tony@bit.hemma.se>
%%% Purpose : progressbar example
%%% Created : 10 Dec 2000 by Tony Rogvall <tony@bit.hemma.se>

-module(progressbar).
-author('tony@bit.hemma.se').

-compile(export_all).

%% example-start progressbar progressbar.c */

-include("../include/gtk.hrl").

-record(progress_data, 
	{
	  window,
	  pbar,
	  timer
	 }).

%% Update the value of the progress bar so that we get
%% some movement 
progress_timeout( Data ) ->
    %% Calculate the value of the progress bar using the
    %% value range set in the adjustment object 
    New_val0 = gtk:progress_get_value( ?GTK_PROGRESS(Data) ) + 1,

    Adj = gtk:progress_get_adjustment(?GTK_PROGRESS (Data)),
    Upper = gtk:adjustment_get_upper(Adj),
    New_val = if New_val0 > Upper -> 
		      gtk:adjustment_get_lower(Adj);
		 true ->
		      New_val0
	      end,
    %% Set the new value 
    gtk:progress_set_value (?GTK_PROGRESS (Data), New_val),

    %% As this is a timeout function, return TRUE so that it
    %% continues to get called 
    true.


%% Callback that toggles the text display within the progress
%% bar trough 
toggle_show_text(Widget, Pdata) ->
    gtk:progress_set_show_text(
      ?GTK_PROGRESS (Pdata#progress_data.pbar),
      gtk:toggle_button_get_active(?GTK_TOGGLE_BUTTON (Widget))).


%% Callback that toggles the activity mode of the progress
%% bar 
toggle_activity_mode( Widget, Pdata ) ->
    gtk:progress_set_activity_mode (
      ?GTK_PROGRESS (Pdata#progress_data.pbar),
      gtk:toggle_button_get_active(?GTK_TOGGLE_BUTTON(Widget))).


%% Callback that toggles the continuous mode of the progress
%% bar
set_continuous_mode( Widget,  Pdata ) ->
    gtk:progress_bar_set_bar_style(
      ?GTK_PROGRESS_BAR(Pdata#progress_data.pbar),
      'GTK_PROGRESS_CONTINUOUS').


%% Callback that toggles the discrete mode of the progress
%% bar 
set_discrete_mode( Widget, Pdata ) ->
    gtk:progress_bar_set_bar_style
      (?GTK_PROGRESS_BAR (Pdata#progress_data.pbar),
       'GTK_PROGRESS_DISCRETE').

 
%% Clean up allocated memory and remove the timer 
destroy_progress( Widget, _, Pdata) ->
    gtk:timeout_remove (Pdata#progress_data.timer),
    gtk:main_quit().

run() ->
    gtk:start(),
    spawn_link(?MODULE, main, []).


main() ->
    Window = gtk:window_new ('GTK_WINDOW_TOPLEVEL'),
    gtk:window_set_policy (?GTK_WINDOW (Window), false, false, true),


    gtk:window_set_title (?GTK_WINDOW (Window), "GtkProgressBar"),
    gtk:container_set_border_width (?GTK_CONTAINER (Window), 0),

    Vbox = gtk:vbox_new (false, 5),
    gtk:container_set_border_width (?GTK_CONTAINER (Vbox), 10),
    gtk:container_add (?GTK_CONTAINER(Window), Vbox),
    gtk:widget_show(Vbox),
  
    %% Create a centering alignment object 
    Align = gtk:alignment_new (0.5, 0.5, 0, 0),
    gtk:box_pack_start (?GTK_BOX (Vbox), Align, false, false, 5),
    gtk:widget_show(Align),

    %% Create a Adjusment object to hold the range of the
    %% progress bar 
    Adj = gtk:adjustment_new(0, 1, 150, 0, 0, 0),

    %% Create the GtkProgressBar using the adjustment 
    Pbar = gtk:progress_bar_new_with_adjustment (Adj),

    %% gtk:object_unref(Adj), %% do not need it after this

    %% Set the format of the string that can be displayed in the
    %% trough of the progress bar:
    %% %p - percentage
    %% %v - value
    %% %l - lower range value
    %% %u - upper range value 
    gtk:progress_set_format_string (?GTK_PROGRESS (Pbar),
	                            "%v from [%l-%u] (=%p%%)"),
    gtk:container_add (?GTK_CONTAINER (Align), Pbar),
    gtk:widget_show(Pbar),

    %% Add a timer callback to update the value of the progress bar 
    Timer = gtk:timeout_add (100, fun progress_timeout/1, Pbar),

    Pdata = #progress_data { window = Window, pbar = Pbar, timer = Timer},
    %% Moved the signal connect to here
    gtk:signal_connect (?GTK_OBJECT (Window), 'delete_event',
	                fun destroy_progress/3,
                        Pdata),

    Separator1 = gtk:hseparator_new (),
    gtk:box_pack_start (?GTK_BOX (Vbox), Separator1, false, false, 0),
    gtk:widget_show(Separator1),

    %% rows, columns, homogeneous 
    Table = gtk:table_new (2, 3, false),
    gtk:box_pack_start (?GTK_BOX (Vbox), Table, false, true, 0),
    gtk:widget_show(Table),

    %% Add a check button to select displaying of the trough text 
    Check1 = gtk:check_button_new_with_label ("Show text"),
    gtk:table_attach (?GTK_TABLE (Table), Check1, 0, 1, 0, 1,
                      ?GTK_EXPAND bor ?GTK_FILL, ?GTK_EXPAND bor ?GTK_FILL,
                      5, 5),
    gtk:signal_connect (?GTK_OBJECT (Check1), 'clicked',
                        fun toggle_show_text/2,
                        Pdata),
    gtk:widget_show(Check1),

    %% Add a check button to toggle activity mode 
    Check2 = gtk:check_button_new_with_label ("Activity mode"),
    gtk:table_attach (?GTK_TABLE (Table), Check2, 0, 1, 1, 2,
                      ?GTK_EXPAND bor ?GTK_FILL, ?GTK_EXPAND bor ?GTK_FILL,
                      5, 5),
    gtk:signal_connect (?GTK_OBJECT (Check2), 'clicked',
                        fun toggle_activity_mode/2,
                        Pdata),
    gtk:widget_show(Check2),

    Separator2 = gtk:vseparator_new (),
    gtk:table_attach (?GTK_TABLE (Table), Separator2, 1, 2, 0, 2,
                      ?GTK_EXPAND bor ?GTK_FILL, ?GTK_EXPAND bor ?GTK_FILL,
                      5, 5),
    gtk:widget_show(Separator2),

    %% Add a radio button to select continuous display mode 
    Button1 = gtk:radio_button_new_with_label ([], "Continuous"),
    gtk:table_attach (?GTK_TABLE (Table), Button1, 2, 3, 0, 1,
                      ?GTK_EXPAND bor ?GTK_FILL, ?GTK_EXPAND bor ?GTK_FILL,
                      5, 5),
    gtk:signal_connect (?GTK_OBJECT (Button1), 'clicked',
			fun set_continuous_mode/2,
                        Pdata),
    gtk:widget_show (Button1),

    %% Add a radio button to select discrete display mode 
    Button2 = gtk:radio_button_new_with_label(
		gtk:radio_button_group (?GTK_RADIO_BUTTON (Button1)),
		"Discrete"),
    gtk:table_attach (?GTK_TABLE (Table), Button2, 2, 3, 1, 2,
                      ?GTK_EXPAND bor ?GTK_FILL, ?GTK_EXPAND bor ?GTK_FILL,
                      5, 5),
    gtk:signal_connect (?GTK_OBJECT (Button2), 'clicked',
                        fun set_discrete_mode/2,
                        Pdata),
    gtk:widget_show (Button2),

    Separator3 = gtk:hseparator_new (),
    gtk:box_pack_start (?GTK_BOX (Vbox), Separator3, false, false, 0),
    gtk:widget_show(Separator3),

    %% Add a button to exit the program 
    Button3 = gtk:button_new_with_label ("close"),
    gtk:signal_connect(?GTK_OBJECT (Button3), 'clicked',
		       fun(_,W) -> destroy_progress(W,[],Pdata) end, Window),

    gtk:box_pack_start (?GTK_BOX (Vbox), Button3, false, false, 0),

    %% This makes it so the button is the default. 
    ?GTK_WIDGET_SET_FLAGS (Button3, ?GTK_CAN_DEFAULT),

    %% This grabs this button to be the default button. Simply hitting
    %% the "Enter" key will cause this button to activate. 
    gtk:widget_grab_default (Button3),
    gtk:widget_show(Button3),

    gtk:widget_show (Window),

    gtk:main (),
    
    0.

%% example-end 
