%%% File    : glade2erl.erl
%%% Author  :  <tony@orre.bluetail.com>
%%% Description : Glade to erlang generator
%%% Created : 11 Jan 2001 by  <tony@orre.bluetail.com>

-module(glade2erl).

-compile(export_all).

-include("../include/gtk.hrl").
-include("../include/gnome.hrl").

-import(lists, [foreach/2, reverse/1, foldl/3, foldr/3, map/2]).

-record(compile, 
	{
	  basename,
	  srcdir,
	  pixdir,
	  ui_fd,
	  cb_fd,
	  main_fd
	 }).

-define(GLADE_PORT,   7681).
-define(GLADE_IFADDR, any).
-define(GLADE_HOST,   localhost).

-ifdef(debug).
-define(Debug(Fmt, As), io:format((Fmt),(As))).
-else.
-define(Debug(Fmt, As), ok).
-endif.


%%
%% Start with 
%% -s glade2erl view <file.glade>  [options]
%% -s glade2erl server             [options]
%% -s glade2erl client <file>      [options]
%% -s glade2erl compile <file>
%%

%%
%% View mode
%%
view(Args) ->
    case catch view1(Args) of
	{'EXIT', Reason} ->
	    io:format("glade2erl view: crashed ~p\n", [Reason]),
	    halt(1);
	_ ->
	    halt(0)
    end.

view1([Url]) when atom(Url) ->
    view1(atom_to_list(Url));
view1(Url) ->
    case glade_parse:url(Url) of
	{ok, XML, _} ->
	    view_xml(XML);
	Error ->
	    fatal("parse error: ~p", [Error])
    end.
%%
%% Server mode
%% options -port <port>
%%         -ifaddr <port>
%%
server(_) ->
    Port = case init:get_argument(port) of
	       {ok,[[Val0]]} ->
		   list_to_integer(Val0);
	       {ok,_} ->
		   fatal("bad argument to -port", []);
	       error ->
		   ?GLADE_PORT
	   end,
    IfAddr = case init:get_argument(ifaddr) of
		 {ok,[[Val1]]} ->
		     case inet_parse:address(Val1) of
			 {ok,IP} -> {ok,IP};
			 _ -> fatal("bad argument to -ifaddr")
		     end;
		 {ok,_} ->
		     fatal("bad argument to -ifaddr");
		 error ->
		     ?GLADE_IFADDR
	     end,
    case gen_tcp:listen(Port, [{ifaddr,IfAddr},
			       {packet,4}, 
			       {mode,binary},
			       {active,false},
			       {reuseaddr, true}]) of
	{ok, L} ->
	    server_loop(L);
	Error ->
	    fatal("can not start server: ~p\n", [Error])
    end.

server_loop(L) ->
    case gen_tcp:accept(L) of
	{ok, S} ->
	    case gen_tcp:recv(S, 0, 10000) of
		{ok,Data} ->
		    server_data(S, Data),
		    gen_tcp:close(S),
		    server_loop(L);
		Error ->
		    io:format("warning: recv error ~p\n", [Error]),
		    gen_tcp:close(S),
		    server_loop(L)
	    end;
	Error ->
	    io:format("warning: accept error ~p\n", [Error]),
	    server_loop(L)
    end.

server_data(S, Data) ->    
    case catch glade_parse:string(Data) of
	{'EXIT', Reason} ->
	    gen_tcp:send(S, io_lib:format("parse error: ~p", [Reason]));
	{ok, XML, _} ->
	    gtk:start(),
	    spawn(?MODULE, view_main, [self(),XML]),
	    gen_tcp:send(S, "ok")
    end.

%%
%% Client mode
%% options -port port
%%         -host host
%%
client([File]) when atom(File) ->
    client(atom_to_list(File));
client(File) ->
    case file:read_file(File) of
	{ok,Data} ->
	    Port = case init:get_argument(port) of
		       {ok,[[Val0]]} ->
			   list_to_integer(Val0);
		       {ok,_} ->
			   fatal("bad argument to -port", []);
		       error ->
			   ?GLADE_PORT
		   end,
	    Host = case init:get_argument(host) of
		       {ok,[[Val1]]} ->  
			   Val1;
		       {ok,_} ->
			   fatal("bad argument to -host", []);
		       error ->
			   ?GLADE_HOST
		   end,
	    case gen_tcp:connect(Host, Port, [{packet,4},{active,false}]) of
		{ok, S} ->
		    gen_tcp:send(S, Data),
		    case gen_tcp:recv(S, 0, 30000) of
			{ok,"ok"} ->
			    gen_tcp:close(S),
			    halt(0);
			{ok,Error} ->
			    fatal("~s", [Error])
		    end;
		Error ->
		    fatal("connect error: ~p", [Error])
	    end;
	Error ->
	    fatal("file error: ~p", [Error])
    end.
		
%%
%% Glade browser
%%
browser([Url]) when atom(Url) ->
    browser(atom_to_list(Url));
browser(Url) ->
    case glade_parse:url(Url) of
	{ok, XML, _} ->
	    view_xml(XML);
	Error ->
	    fatal("parse error: ~p", [Error])
    end.


%%
%% Compile mode
%% options -o <dir>
%%         -nocb
%%         -nomain
%%         -noui
%%	
%%
compile(Args) ->
    case catch compile1(Args) of
	{'EXIT', Reason} ->
	    io:format("glade2erl compile: crashed ~p\n", [Reason]),
	    halt(1);
	_ ->
	    halt(0)
    end.
    

compile1([File]) when atom(File) ->
    compile1(atom_to_list(File));
compile1(File) ->
    case glade_parse:file(File) of
	{ok, XML, _} ->
	    Project = project_xml(XML),
	    C0 = compile_setup(File, Project),
	    C1 = compile_open(C0),
	    gen_headers(C1),
	    Res = gen_ui(C1, XML),
	    compile_close(C0),
	    halt(0);
	Error ->
	    fatal("parse error: ~p", [Error])
    end.

compile_setup(File, Project) ->
    Base = filename:basename(File, ".glade"),
    SrcDir = attr("source_directory",gstring, Project,filename:dirname(File)),
    PixDir = attr("pixmaps_directory",gstring,Project, SrcDir),
    #compile { basename = Base,
	       srcdir   = SrcDir,
	       pixdir   = PixDir }.


compile_open(C0) ->    
    C1 = case init:get_argument(o) of
	     {ok,[Dir]} ->
		 C0#compile { srcdir = Dir };
	     error ->
		 C0
	 end,
    C2 = case init:get_argument(nocb) of
	     error ->
		 CBFile = C1#compile.basename ++ "_cb.erl",
		 CB = filename:join(C1#compile.srcdir, CBFile),
		 case file:open(CB, [write]) of
		     {ok,Fd1} ->
			 C1#compile { cb_fd = Fd1 };
		     Error1 ->
			 fatal("could not open ~s", [CB])
		 end;
	     {ok,_} ->
		 C1
	 end,
    C3 = case init:get_argument(noui) of
	     error ->
		 UIFile = C2#compile.basename ++ "_ui.erl",
		 UI = filename:join(C2#compile.srcdir, UIFile),
		 case file:open(UI, [write]) of
		     {ok,Fd2} ->
			 C2#compile { ui_fd = Fd2 };
		     Error2 ->
			 fatal("could not open ~s", [UI])
		 end;
	     {ok,_} ->
		 C2
	 end,
    C4 = case init:get_argument(nomain) of
	     error ->
		 MainFile = C3#compile.basename ++ "_main.erl",
		 Main = filename:join(C3#compile.srcdir, MainFile),
		 case file:open(Main, [write]) of
		     {ok,Fd3} ->
			 C3#compile { main_fd = Fd3 };
		     Error3 ->
			 fatal("could not open ~s", [Main])
		 end;
	     {ok,_} ->
		 C3
	 end,
    C4.

compile_close(C) ->
    if C#compile.ui_fd == undefined -> ok;
       true -> file:close(C#compile.ui_fd)
    end,
    if C#compile.cb_fd == undefined -> ok;
       true -> file:close(C#compile.cb_fd)
    end,
    if C#compile.main_fd == undefined -> ok;
       true -> file:close(C#compile.main_fd)
    end,
    C#compile { ui_fd = undefined, cb_fd = undefined, main_fd = undefined}.


%%
%% create a GtkPixmap from url!
%%
pixmap_new(Widget,Url) ->
    case glade_url:parse(Url) of
	{scheme,"file",File} ->    pixmap_file(Widget,File);
	{abs_path, Path} ->        pixmap_file(Widget,Path);
	{rel_path, Path} ->        pixmap_file(Widget,Path);
	{absoluteURI,http,Host,Port,Path} ->
	    pixmap_http(Widget,Host, Port, Path);
	{error, not_supported} ->
	    pixmap_broken(Widget)
    end.


pixmap_file(Widget,Path) ->
    case file:read_file(Path) of
	{ok, Bin} -> pixmap_data(Widget,Bin);
	{error,_} -> pixmap_broken(Widget)
    end.

pixmap_http(Widget,Host, Port, Path) ->
    case glade_http_cli:load(Host,Port,Path) of
	{ok, Data} -> pixmap_data(Widget,Data);
	{error,_} -> pixmap_broken(Widget)
    end.


pixmap_data(Widget,Bin) when binary(Bin) ->
    XpmData = xpm_parse(binary_to_list(Bin)),
    Colormap  = gtk:widget_get_colormap(Widget),    
    {Pixmap,Mask} = gdk:pixmap_colormap_create_from_xpm_d(?NULL,
							  Colormap,
							  ?NULL,
							  XpmData),
    gtk:pixmap_new(Pixmap,Mask);
pixmap_data(Widget,Data) when list(Data) ->
    pixmap_data(Widget,list_to_binary(Data)).


pixmap_broken(Widget) ->
    Colormap  = gtk:widget_get_colormap(Widget),
    {Pixmap,Mask} = gdk:pixmap_colormap_create_from_xpm_d(?NULL,
							  Colormap,
							  ?NULL,
							  broken_xpm()),
    gtk:pixmap_new(Pixmap,Mask).

xpm_parse(Data) ->
    list_to_tuple(
      lists:foldr(
      fun("/*"++ _, Lines) -> Lines;
	 ("static "++_, Lines) -> Lines;
	 ("};"++_, Lines) -> Lines;
	 ("", Lines) -> Lines;
	 ([$\" | Cs], Lines) ->
              case reverse(Cs) of
                  [$\" | Cs1] ->
                      [ reverse([0 | Cs1]) | Lines];
                  [$,, $\" | Cs1] ->
                      [ reverse([0 | Cs1]) | Lines]
              end
       end, 
       [],   string:tokens(Data, "\n\r"))).


broken_xpm() ->
 {
"    20    22        5            1\0",
". c #ffffff\0",
"# c #ccffff\0",
"a c #999999\0",
"b c #333333\0",
"c c #000000\0",
"bbbbbbbbbbbbb#######\0",
"b..........b.b######\0",
"b..........b..b#####\0",
"b..........b...b####\0",
"b..........b....b###\0",
"b..........cbcbcbc##\0",
"b..........ccccccc##\0",
"b............aaaac##\0",
"b.............aaac##\0",
"b...............ac##\0",
"b.b...........b.ac##\0",
"bbcb...b...b.bcbac##\0",
"cc#cb.bcb.bcbc#cbc##\0",
"c###cbc#cbc#c###cc##\0",
"##c##c###c####c##c##\0",
"#c.c###c###c#c.c####\0",
"c...c#c.c#c.c...c###\0",
"c....c...c......ac##\0",
"b...............ac##\0",
"b...............ac##\0",
"bbaaaaaaaaaaaaaabc##\0",
"##cccccccccccccccc##\0"  }.
   
    



gen_headers(C) ->
    gen_module_header(C#compile.ui_fd, C#compile.basename ++ "_ui"),
    gen_module_header(C#compile.cb_fd, C#compile.basename ++ "_cb"),
    gen_module_header(C#compile.main_fd, C#compile.basename ++ "_main"),
    C.

gen_module_header(undefined, _) ->
    ok;
gen_module_header(Fd, Module) ->
    io:format(Fd, "%%\n%% ~s was generated by glade2erl\n", [Module]),
    {YY,MM,DD} = date(),
    {H,M,S} = time(),
    io:format(Fd, "%% date: ~s ~w ~2.2.0w:~2.2.0w:~2.2.0w ~w~n",
	      [month(MM), DD, H, M, S, YY]),
    io:format(Fd, "%%\n", []),
    io:format(Fd, "-module(~s).\n\n", [Module]),
    io:format(Fd, "-include_lib(\"erlgtk/include/gdk.hrl\").\n", []),
    io:format(Fd, "-include_lib(\"erlgtk/include/gtk.hrl\").\n", []),
    io:format(Fd, "-include_lib(\"erlgtk/include/gnome.hrl\").\n", []),
    io:format(Fd, "-compile(export_all).\n\n", []).
    

%%
%% Month conversion
%%
month(1) -> "Jan";
month(2) -> "Feb";
month(3) -> "Mar";
month(4) -> "Apr";
month(5) -> "May";
month(6) -> "Jun";
month(7) -> "Jul";
month(8) -> "Aug";
month(9) -> "Sep";
month(10) -> "Oct";
month(11) -> "Nov";
month(12) -> "Dec".

fatal(Fmt) ->
    fatal(Fmt, []).

fatal(Fmt, As) ->
    io:format("glade2erl:" ++ Fmt ++ "\n", As),
    halt(1).

%%
%% Extract project attributes
%%
project_xml({"GTK-Interface", _, Ws}) ->
    case lists:keysearch("project", 1, Ws) of
	{value, {_, _, Attr}} ->
	    reduce_attributes(Attr, []);
	_ ->
	    []
    end;
project_xml(_) ->
    [].

%%
%% The tag structure MUST be on form
%% {Tag, Params, Children}
%%
view_xml(XML) ->
    gtk:start(),
    Pid = spawn_link(?MODULE, view_main, [self(),XML]),
    receive
	{Pid,Answer} ->
	    Answer
    end.

view_main(Pid, XML) ->
    case view_main1(Pid, XML) of
	{'EXIT', Reason} ->
	    io:format("CRASH: ~p\n", [Reason]);
	Other ->
	    ok
    end.


view_main1(Pid, {"GTK-Interface", _, Ws}) ->
    put(wcount, 0),
    foreach(
      fun(W) -> 
	      {Creation,Widget,Class,As} = view_widget(W, []),
	      case ?GTK_IS_WINDOW(Widget) of
		  true ->
		      gtk:widget_show(Widget),
		      gtk:signal_connect(Widget, 'destroy',
					 fun(_,_) ->
						 Count = get(wcount),
						 if Count == 1 ->
							 Pid ! {self(),ok},
							 gtk:main_quit();
						    true ->
							 put(wcount,Count-1)
						 end
					 end, []),
		      put(wcount, get(wcount)+1);
		  false ->
		      ok
	      end
      end, Ws),
    gtk:main().


view_widget({"widget", _, Attr}, ParentList) ->
    {As,Ws} = collect_attributes(Attr, [], []),
    ?Debug("view_widget: ~p\n", [As]),
    case view_create_widget(As,ParentList) of
	{error,W, Class, ContainerType} ->
	    {error,?GTK_NULL,"",[]};

	{new,Widget,Class="Placeholder", ContainerType} ->
	    %% add child spec to place holder
	    {new,Widget,Class,[{"child",
				[{"padding","0"},
				 {"expand","True"},
				 {"fill", "True"}]} | As]};

	{Creation,Widget,Class,ContainerType} ->
	    view_extra(Widget,Class,As),
	    view_tooltip(Widget, As),
	    view_signals(Widget, As),
	    ParentList1 = [Widget|ParentList],
	    Children = foldl(fun(W,Ls) ->
				     case view_widget(W,ParentList1) of
					 {error,_,_,_} -> Ls;
					 Child -> [Child|Ls]
				     end
			     end, [], Ws),
	    view_children(ParentList1, reverse(Children), [], ContainerType),
	    case ?GTK_IS_WINDOW(Widget) of
		true ->
		    ok; %% do not show toplevel windows here
		false ->
		    gtk:widget_show(Widget)
	    end,
	    {Creation,Widget,Class,As}
    end;
view_widget({"project", _, Attr},Parent) ->
    {error,?GTK_NULL,"",[]}.


view_extra(Widget, "GtkCombo", As) ->
    view_combo_items(Widget, As);
view_extra(Widget, "GtkOptionMenu", As) ->
    view_option_menu_items(Widget, As);
view_extra(Widget, "GtkVScale", As) ->
    view_range_adjustment(Widget, As);
view_extra(Widget, "GtkHScale", As) ->
    view_range_adjustment(Widget, As);
view_extra(Widget, "GtkVScrollbar", As) ->
    view_range_adjustment(Widget, As);
view_extra(Widget, "GtkHScrollbar", As) ->
    view_range_adjustment(Widget, As);
view_extra(Widget, _, As) ->
    ok.


view_range_adjustment(Widget, As) ->
    Value = attr("value", gfloat, As, 0.0),
    Upper = attr("upper", gfloat, As, 0.0),
    Lower = attr("lower", gfloat, As, 0.0),
    Step = attr("step", gfloat, As, 0.0),
    Page = attr("page", gfloat, As, 0.0),
    Page_Size = attr("page_size", gfloat, As, 0.0),
    Adjustment = gtk:range_get_adjustment(?GTK_RANGE(Widget)),
    gtk:adjustment_put_lower(Adjustment, Lower),
    gtk:adjustment_put_upper(Adjustment, Upper),    
    gtk:adjustment_put_step_increment(Adjustment, Step),
    gtk:adjustment_put_page_increment(Adjustment, Page),
    gtk:adjustment_put_page_size(Adjustment, Page_Size),
    gtk:adjustment_put_value(Adjustment, Value),
    ok.

    
view_create_widget(As,ParentList) ->
    case lists:keysearch("class", 1, As) of
	{value, Item={_, Class}} ->
	    Name = attr("name", gstring, As, fun() -> gen_tmp() end),
	    {Spec,ContainerType} = spec(Class, Name, view),
	    Parent = case ParentList of
			 [P |_] -> P;
			 _ -> undefined
		     end,
	    case lists:keysearch("child_name", 1, As) of
		{value,{_,ChildName}} ->
		    case view_get_child(ChildName,ParentList) of
			false ->
			    W = exec_spec([], Spec, [{"parent",Parent} | As]),
			    {new,W,Class,ContainerType};
			Child ->
			    [_|Spec1] = Spec,  %% remove create part
			    W = exec_spec([Child],Spec1,
					  [{"parent",Parent}|As]),
			    {get,W,Class,ContainerType}
		    end;
		false ->
		    W = exec_spec([], Spec, [{"parent",Parent} | As]),
		    {new, W, Class, ContainerType}
	    end;
	false ->
	    io:format("error no \"class\" found in ~p\n", [As]),
	    {error, ?GTK_NULL, "", undefined}
    end.

		    
view_get_child("Dialog:vbox", [Dialog|_]) ->
    gtk:dialog_get_vbox(Dialog);
view_get_child("Dialog:action_area", [VBox,Dialog|_]) ->
    gtk:dialog_get_action_area(Dialog);
view_get_child("FileSel:ok_button", [FileSel|_]) ->
    gtk:file_selection_get_ok_button(FileSel);
view_get_child("FileSel:cancel_button", [FileSel|_]) ->
    gtk:file_selection_get_cancel_button(FileSel);
view_get_child("ColorSel:ok_button", [ColorSel|_]) ->
    gtk:color_selection_dialog_get_ok_button(ColorSel);
view_get_child("ColorSel:cancel_button", [ColorSel|_]) ->
    gtk:color_selection_dialog_get_cancel_button(ColorSel);
view_get_child("ColorSel:help_button", [ColorSel|_]) ->
    gtk:color_selection_dialog_get_help_button(ColorSel);
view_get_child("FontSel:ok_button", [FontSel|_]) ->
    gtk:font_selection_dialog_get_ok_button(FontSel);
view_get_child("FontSel:cancel_button", [FontSel|_]) ->
    gtk:font_selection_dialog_get_cancel_button(FontSel);
view_get_child("FontSel:apply_button", [FontSel|_]) ->
    gtk:font_selection_dialog_get_apply_button(FontSel);
view_get_child("GtkCombo:entry", [Combo|_]) ->
    gtk:combo_get_entry(Combo);
view_get_child("InputDialog:save_button", [Dialog|_]) ->
    gtk:input_dialog_get_save_button(Dialog);
view_get_child("InputDialog:close_button", [Dialog|_]) ->
    gtk:input_dialog_get_close_button(Dialog);

view_get_child("GnomeApp:dock", [GnomeApp|_]) ->
    gnome_app:get_dock(GnomeApp);
view_get_child("GnomeApp:appbar",[GnomeApp|_]) ->
    gnome_app:get_appbar(GnomeApp);
view_get_child("GnomeDialog:vbox",[GnomeDialog|_]) ->
    gnome_dialog:get_vbox(GnomeDialog);
view_get_child("GnomeDialog:action_area",[GnomeDialog|_]) ->
    gnome_dialog:get_action_area(GnomeDialog);
view_get_child("GnomeDruidPageStandard:vbox",[Druid|_]) ->
    gnome_druid:page_standard_get_vbox(Druid);
view_get_child("GnomeEntry:entry", [Entry|_]) ->
    gnome_entry:get_entry(Entry);
view_get_child("GnomePixmapEntry:file-entry", [PixmapEntry|_]) ->
    gnome_pixmap_entry:get_file_entry(PixmapEntry);
view_get_child(_, Parent) ->
    false.



view_tooltip(Widget, As) ->
    case attr("tooltip", gstring, As, "") of
	"" -> ok;
	Text ->
	    Tips = gtk:tooltips_new(),
	    gtk:tooltips_set_tip(Tips, Widget, Text, ?NULL)
    end.

view_signals(Widget, As) ->
    case lists:keysearch("signal", 1, As) of
	false -> ok;
	{value,{_,SAs}} ->
	    Signal = attr("name", gstring, SAs),
	    Handler = attr("handler", gstring, SAs),
	    Raw = glade_parse:xml2raw(Handler),
	    Exprs = parse_erl(Raw),
	    Data = attr("data", gstring, SAs, []),
	    gtk:signal_connect(Widget, list_to_atom(Signal),
			       {?MODULE,sig_exec}, {Exprs,Data})
    end.


view_option_menu_items(Widget, As) ->
    Items = attr("items", gstring, As, ""),
    Initial = attr("initial_choice", gint, As, 0),
    Menu = gtk:menu_new(),
    List = string:tokens(Items, "\n"),
    foreach(fun(Label) ->
		    I = gtk:menu_item_new_with_label(Label),
		    gtk:menu_append(Menu, I),
		    gtk:widget_show(I)
	    end, List),
    gtk:option_menu_set_menu(Widget, Menu),
    gtk:widget_show(Menu),
    gtk:option_menu_set_history(Widget, Initial).



view_combo_items(Widget, As) ->
    Items = attr("items", gstring, As, ""),
    case string:tokens(Items, "\n") of
	[] -> ok;
	L ->
	    LW = gtk:combo_get_list(?GTK_COMBO(Widget)),
	    foreach(fun(Item) ->
			    I = gtk:list_item_new_with_label(Item),
			    gtk:container_add(?GTK_LIST(LW),I),
			    gtk:widget_show(I)
		    end, L)
    end.



view_children(ParentList, [{get,_,_,_}|Cs], XArg, ContainerType) ->
    %% Filter get "accessed" children i.e not created
    view_children(ParentList, Cs, XArg, ContainerType);

view_children(ParentList, [{_,Child,_,As}|Cs], XArg, container) ->
    [Parent|_] = ParentList,
    gtk:container_add(?GTK_CONTAINER(Parent), Child),
    view_children(ParentList, Cs, XArg, container);
view_children(ParentList, [{_,Child,Class,As}|Cs], XArg, box) ->
    [Parent|_] = ParentList,
    XArg1 = if Class == "GtkRadioButton" ->
		    gtk:radio_button_set_group(Child, XArg),
		    gtk:radio_button_group(Child);
	       true ->
		    XArg
	    end,
    case cattr("pack",genum,As,'GTK_PACK_START') of
	'GTK_PACK_START' ->
	    gtk:box_pack_start(Parent, Child,
			       cattr("expand",gboolean,As),
			       cattr("fill",gboolean,As),
			       cattr("padding",gint,As));
	'GTK_PACK_END' ->
	    gtk:box_pack_end(Parent, Child, 
			     cattr("expand",gboolean,As),
			     cattr("fill",gboolean,As),
			     cattr("padding",gint,As))
    end,
    view_children(ParentList, Cs, XArg1, box);
view_children(ParentList, [{_,Child,_,As}|Cs], XArg, table) ->
    [Parent|_] = ParentList,    
    XOpts = 
	cflag("xexpand",As,false,?GTK_EXPAND,0) bor
	cflag("xshrink",As,false,?GTK_SHRINK,0) bor
	cflag("xfill",As,true,?GTK_FILL,0),
    YOpts = 
	cflag("yexpand",As,false,?GTK_EXPAND,0) bor
	cflag("yshrink",As,false,?GTK_SHRINK,0) bor
	cflag("yfill",As,true,?GTK_FILL,0),
    gtk:table_attach(?GTK_TABLE(Parent), Child,
		     cattr("left_attach", guint, As),
		     cattr("right_attach", guint, As),
		     cattr("top_attach", guint, As),
		     cattr("bottom_attach", guint, As),
		     XOpts, YOpts,
		     cattr("xpad",guint,As),
		     cattr("ypad",guint,As)),
    view_children(ParentList, Cs, XArg, table);
view_children(ParentList, [{_,Child1,_,As1},{_,Child2,_,As2}], XArg, paned) ->
    [Parent|_] = ParentList,        
    gtk:paned_pack1(?GTK_PANED(Parent),
		    Child1,
		    cattr("resize", gboolean, As1, true),
		    cattr("shrink", gboolean, As1, false)),
    gtk:paned_pack2(?GTK_PANED(Parent),
		    Child2,
		    cattr("resize", gboolean, As2,  true),
		    cattr("shrink", gboolean, As2, false)),
    ok;
view_children(ParentList, [{_,Child,_,As} | Cs], XArg, clist) ->
    [Parent|_] = ParentList,        
    Col = if XArg == [] -> 0;
	     true -> XArg
	  end,
    gtk:clist_set_column_widget(?GTK_CLIST(Parent), Col, Child),
    gtk:clist_set_column_width(?GTK_CLIST(Parent), Col, 80),
    view_children(ParentList, Cs, Col+1, clist);
view_children(ParentList, [{_,Child,_,As}], XArg, combo) ->
    [Parent|_] = ParentList,
    %% FIXME set entry props
    ok;
view_children(ParentList, [{_,Child,_,As} | Cs], XArg, toolbar) ->
    [Parent|_] = ParentList,    
    gtk:toolbar_append_widget(?GTK_TOOLBAR(Parent),
			      Child, ?NULL, ?NULL),
    view_children(ParentList, Cs, XArg, toolbar);
view_children(ParentList, [{_,Child,_,As} | Cs], XArg, menubar) ->
    [Parent|_] = ParentList,    
    gtk:menu_bar_append(?GTK_MENU_BAR(Parent), Child),
    view_children(ParentList, Cs, XArg, menubar);
view_children(ParentList, [{_,Child,_,As} | Cs], XArg, menu) ->
    [Parent|_] = ParentList,    
    gtk:menu_append(?GTK_MENU(Parent), Child),
    view_children(ParentList, Cs, XArg, menu);
view_children(ParentList, [{_,Child,_,As}|Cs], XArg, menuitem) ->
    [Parent|_] = ParentList,
    gtk:menu_item_set_submenu(?GTK_MENU_ITEM(Parent), Child),
    view_children(ParentList, Cs, XArg, menuitem);
view_children(ParentList, [{_,Child,_,As1},
			   {_,Label,_,As2}|Cs], XArg, notebook) ->
    [Parent|_] = ParentList,    
    gtk:notebook_append_page(?GTK_NOTEBOOK(Parent), Child, Label),
    view_children(ParentList, Cs, XArg, notebook);
view_children(ParentList, [{_,Child,_,As} | Cs], XArg, fixed) ->
    [Parent|_] = ParentList,    
    gtk:fixed_put(?GTK_FIXED(Parent), Child, 
		  attr("x", gint, As, 0),
		  attr("y", gint, As, 0)),
    view_children(ParentList, Cs, XArg, fixed);
view_children(ParentList, [], XArg, _) ->
    ok.


parse_erl(String) ->
    case erl_scan:string(String) of
	{ok, Tokens, _} ->
	    ?Debug("tokens = ~p", [Tokens]),
	    Tokens1 = case lists:last(Tokens) of
			  {dot,_} -> Tokens;
			  _ -> Tokens ++ [{dot,1}]
		      end,
	    case erl_parse:parse_exprs(Tokens1) of
		{ok, Exprs} ->
		    ?Debug("exprs = ~p", [Exprs]),
		    Exprs;
		_ ->
		    error_exprs()
	    end;
	_ ->
	    error_exprs()
    end.

error_exprs() ->
    [{call,1,
      {remote,1,{atom,1,io},{atom,1,format}},
      [{string,1,"Error\n"}]}].

%% bind "Widget" to W in environment !!!
%% bind More?
sig_exec(W, {Exprs,Data}) ->
    Bs0 = erl_eval:new_bindings(),
    Bs1 = erl_eval:add_binding('Widget', W, Bs0),
    erl_eval:exprs(Exprs, Bs1).

sig_exec(W, _, {Exprs,Data}) ->
    Bs0 = erl_eval:new_bindings(),
    Bs1 = erl_eval:add_binding('Widget', W, Bs0),
    erl_eval:exprs(Exprs, Bs1).

sig_test(W, _, _, {Exprs,Data}) ->
    Bs0 = erl_eval:new_bindings(),
    Bs1 = erl_eval:add_binding('Widget', W, Bs0),
    erl_eval:exprs(Exprs, Bs1).

sig_test(W, _, _, _, {Exprs,Data}) ->
    Bs0 = erl_eval:new_bindings(),
    Bs1 = erl_eval:add_binding('Widget', W, Bs0),
    erl_eval:exprs(Exprs, Bs1).


%%
%% code generator
%%    
gen_ui(C, {"GTK-Interface", _, Widgets}) ->
    put(tmp, 1),
    emit_main(C, 
	      "start() ->\n"
	      "  gtk:start(),\n"
	      "  spawn_link(?MODULE, main, []).\n\n", []),
    emit_main(C, "main() ->\n", []),
    foreach(fun(W) -> gen_widget(C, W, []) end, Widgets),
    emit_main(C, "  gtk:main().\n", []), 
    ok.

gen_tmp() ->
    I = get(tmp),
    put(tmp, I+1),
    "tmp" ++ integer_to_list(I).


gen_widget(C, {"widget", _, Attr}, ParentList) ->
    {As,Ws} = collect_attributes(Attr, [], []),
    {Creation,Name,Class,Spec,ContainerType} = gen_spec(As, ParentList),
    case is_window(Class) of
	error ->
	    {error,Name,Class,As};
	true ->
	    emit_ui(C, "create_~s() ->\n", [fmt_name(Name)]),
	    emit_spec(C, Name, Spec, As),
	    emit_main(C, "  ~s = ~s:create_~s(),\n", 
		      [fmt_name(Name),
		       C#compile.basename ++ "_ui",
		       fmt_name(Name)]),
	    emit_main(C, "  gtk:widget_show(~s),\n",[fmt_name(Name)]),
	    emit_extra(C,Name,Class,As),
	    emit_tooltip(C, Name, As),
	    emit_signals(C, Name, As),
	    ParentList1 = [Name | ParentList],
	    Children = foldl(fun(W,Ls) ->
				     case gen_widget(C,W,ParentList1) of
					 {error,_,_,_} -> Ls;
					 Child -> [Child|Ls]
				     end
			     end, [], Ws),
	    gen_children(C, ParentList1, reverse(Children), [], ContainerType),
	    emit_ui(C, "  ~s.\n\n", [fmt_name(Name)]),
	    {Creation,Name,Class,As};

	false ->
	    As1 = if Class == "Placeholder" ->
			  [{"child",
			    [{"padding","0"},
			     {"expand","True"},
			     {"fill", "True"}]} | As];
		     true ->
			  As
		  end,
	    Parent = case ParentList of
			 [P |_] -> P;
			 _ -> undefined
		     end,	    
	    emit_spec(C, Name, Spec, [{"parent",{var,fmt_name(Parent)}}|As1]),
	    emit_extra(C,Name,Class,As1),
	    emit_tooltip(C, Name, As1),
	    emit_signals(C, Name, As1),
	    ParentList1 = [Name | ParentList],
	    Children = foldl(fun(W,Ls) ->
				     case gen_widget(C,W,ParentList1) of
					 {error,_,_,_} -> Ls;
					 Child -> [Child|Ls]
				     end
			     end, [], Ws),
	    gen_children(C, ParentList1, reverse(Children), [], ContainerType),
	    emit_ui(C, "  gtk:widget_show(~s),\n", [fmt_name(Name)]),
	    {Creation,Name,Class,As1}
    end;
gen_widget(C, {"project", _, Attr}, _) ->
    {gen_tmp(),"", []}.



gen_children(C,ParentList, [{get,Child,_,As}|Cs], XArg, ContainterType) ->
    gen_children(C,ParentList,Cs,XArg,ContainterType);

gen_children(C,ParentList, [{_,Child,_,As}|Cs], XArg, container) ->
    [Parent|_] = ParentList,
    emit_ui(C, "  gtk:container_add(?GTK_CONTAINER(~s),~s),\n", 
	    [fmt_name(Parent), fmt_name(Child)]),
    gen_children(C, ParentList, Cs, XArg, container); 
gen_children(C,ParentList, [{_,Child,Class,As}|Cs], XArg, box) ->
    [Parent|_] = ParentList,
    XArg1 = 
	if Class == "GtkRadioButton" ->
		GrpName = if XArg == [] -> "[]";
			     true -> XArg
			  end,
		emit_ui(C,
			"  gtk:radio_button_set_group(~s,~s),\n", 
			[fmt_name(Child), GrpName]),
		Grp = fmt_name(Child) ++ "_group",
		emit_ui(C,
			"  ~s =gtk:radio_button_group(~s),\n", 
			[Grp, fmt_name(Child)]),
		Grp;
	   true -> XArg
	end,
    case cattr("pack",genum,As,'GTK_PACK_START') of
	'GTK_PACK_START' ->
	    emit_ui(C, "  gtk:box_pack_start(~s,~s,~s,~s,~s),\n",
		    [fmt_name(Parent),fmt_name(Child),
		     fmt(cattr("expand",gboolean,As)),
		     fmt(cattr("fill",gboolean,As)),
		     fmt(cattr("padding",gint,As))]);
	'GTK_PACK_END' ->
	    emit_ui(C, "  gtk:box_pack_end(~s,~s,~s,~s,~s),\n",
		    [fmt_name(Parent),fmt_name(Child),
		     fmt(cattr("expand",gboolean,As)),
		     fmt(cattr("fill",gboolean,As)),
		     fmt(cattr("padding",gint,As))])
    end,
    gen_children(C,ParentList, Cs, XArg1, box);
gen_children(C,ParentList, [{_,Child,_,As}|Cs], XArg, table) ->
    [Parent|_] = ParentList,
    XOpts = 
	cflag("xexpand",As,false,?GTK_EXPAND,0) bor
	cflag("xshrink",As,false,?GTK_SHRINK,0) bor
	cflag("xfill",As,true,?GTK_FILL,0),
    YOpts =
	cflag("yexpand",As,false,?GTK_EXPAND,0) bor
	cflag("yshrink",As,false,?GTK_SHRINK,0) bor
	cflag("yfill",As,true,?GTK_FILL,0),
    FSpec = [{?GTK_EXPAND,"?GTK_EXPAND"},
	     {?GTK_SHRINK,"?GTK_SHRINK"},
	     {?GTK_FILL, "?GTK_FILL"}],
    emit_ui(C, "  gtk:table_attach(?GTK_TABLE(~s),~s,"
	    "~s,~s,~s,~s,~s,~s,~s,~s),\n",
	    [fmt_name(Parent),fmt_name(Child),
	     fmt(cattr("left_attach", guint, As)),
	     fmt(cattr("right_attach", guint, As)),
	     fmt(cattr("top_attach", guint, As)),
	     fmt(cattr("bottom_attach", guint, As)),
	     fmt_flags(XOpts,FSpec), fmt_flags(YOpts,FSpec),
	     fmt(cattr("xpad",guint,As)),
	     fmt(cattr("ypad",guint,As))]),
    gen_children(C,ParentList, Cs, XArg, table);
gen_children(C,ParentList, [{_,Child1,_,As1},{_,Child2,_,As2}], XArg, paned) ->
    [Parent|_] = ParentList,    
    emit_ui(C,"  gtk:paned_pack1(?GTK_PANED(~s),~s,~s,~s),\n",
	    [fmt_name(Parent),fmt_name(Child1),
	     fmt(cattr("resize", gboolean, As1, true)),
	     fmt(cattr("shrink", gboolean, As1, false))]),
    emit_ui(C, "  gtk:paned_pack2(?GTK_PANED(~s),~s,~s,~s),\n",
	    [fmt_name(Parent),fmt_name(Child2),
	     fmt(cattr("resize", gboolean, As2,  true)),
	     fmt(cattr("shrink", gboolean, As2, false))]),
    ok;
gen_children(C, ParentList, [{_,Child,_,As} | Cs], XArg, clist) ->
    [Parent|_] = ParentList,
    Col = if XArg == [] -> 0;
	     true -> XArg
	  end,
    emit_ui(C, "  gtk:clist_set_column_widget(?GTK_CLIST(~s),~w,~s),\n",
	    [fmt_name(Parent), Col, fmt_name(Child)]),
    gen_children(C, ParentList, Cs, Col+1, clist);

gen_children(C,ParentList,[{_,Child,_,As},{_,Label,_,LAs}|Cs],XArg,notebook) ->
    [Parent|_] = ParentList,    
    emit_ui(C, "  gtk:notebook_append_page(?GTK_NOTEBOOK(~s),~s,~s),\n",
	    [fmt_name(Parent), fmt_name(Child), fmt_name(Label)]),
    gen_children(C,ParentList, Cs, XArg, notebook);

gen_children(C, ParentList, [{_,Child,_,As}], XArg, combo) ->
    [Parent|_] = ParentList,    
    %% FIXME set entry props
    ok;
gen_children(C, ParentList, [{_,Child,_,As} | Cs], XArg, toolbar) ->
    [Parent|_] = ParentList,    
    emit_ui(C,"  gtk:toolbar_append_widget(?GTK_TOOLBAR(~s),~s,?NULL,?NULL),\n",[fmt_name(Parent), fmt_name(Child)]),
    gen_children(C,ParentList, Cs, XArg, toolbar);
gen_children(C, ParentList, [{_,Child,_,As} | Cs], XArg, menubar) ->
    [Parent|_] = ParentList,    
    emit_ui(C, "  gtk:menu_bar_append(?GTK_MENU_BAR(~s), ~s),\n",
	    [fmt_name(Parent), fmt_name(Child)]),
    gen_children(C, ParentList, Cs, XArg, menubar);
gen_children(C,ParentList, [{_,Child,_,As}|Cs], XArg, fixed) ->
    [Parent|_] = ParentList,    
    emit_ui(C, "  gtk:fixed_put(?GTK_FIXED(~s),~s,~w,~w),\n",
	    [fmt_name(Parent), fmt_name(Child), 
	     attr("x", gint, As, 0),
	     attr("y", gint, As, 0)]),
    gen_children(C, ParentList, Cs, XArg, fixed);

gen_children(C,ParentList, [], XArg, _) ->
    ok.


fmt_flags(0, Spec) -> "0";
fmt_flags(Flags, Spec) -> 
    fmt_flags(Flags, Spec, "").

fmt_flags(Flags, [{Bit,Name}|Spec], Acc) ->
    if Flags band Bit == 0 ->
	    fmt_flags(Flags, Spec, Acc);
       true ->
	    if Acc == "" ->
		    fmt_flags(Flags, Spec, Name);
	       true ->
		    fmt_flags(Flags, Spec, Acc ++ " bor " ++ Name)
	    end
    end;
fmt_flags(Flags, [], Acc) -> Acc.
	
    
gen_spec(As, ParentList) ->
    case lists:keysearch("class", 1, As) of
	{value, Item={_, Class}} ->
	    Name = attr("name", gstring, As, fun() -> gen_tmp() end),
	    {Spec,ContainerType} = spec(Class, Name, emit),
	    case lists:keysearch("child_name", 1, As) of
		{value,{_,ChildName}} ->
		    ?Debug("gen_spec: childname ~s\n", [ChildName]),
		    case gen_get_child(ChildName, ParentList) of
			false ->
			    {new,Name,Class,Spec,ContainerType};
			ChildSpec ->
			    [_|SpecTail] = Spec,
			    {get,Name,Class,[ChildSpec|SpecTail],ContainerType}
		    end;
		false ->
		    {new,Name,Class,Spec,ContainerType}
	    end;
	false ->
	    io:format("error no \"class\" found in ~p\n", [As]),
	    {error,gen_tmp(), "", [], undefined}
    end.


gen_get_child("Dialog:vbox", [Dialog|_]) ->
    {create,{gtk,dialog_get_vbox},
     [{"vbox",{var,fmt_name(Dialog)}}]};
gen_get_child("Dialog:action_area", [VBox,Dialog|_]) ->
    {create,{gtk,dialog_get_action_area},
      [{"action-area",{var,fmt_name(Dialog)}}]};
gen_get_child("FileSel:ok_button", [FileSel|_]) ->
    {create,{gtk,file_selection_get_ok_button},
     [{"ok-button",{var,fmt_name(FileSel)}}]};
gen_get_child("FileSel:cancel_button", [FileSel|_]) ->
    {create,{gtk,file_selection_get_cancel_button},
     [{"cancel-button",{var,fmt_name(FileSel)}}]};
gen_get_child("ColorSel:ok_button", [ColorSel|_]) ->
    {create,{gtk,color_selection_dialog_get_ok_button},
     [{"ok-button",{var,fmt_name(ColorSel)}}]};
gen_get_child("ColorSel:cancel_button", [ColorSel|_]) ->
    {create,{gtk,color_selection_dialog_get_cancel_button},
      [{"cancel_button",{var,fmt_name(ColorSel)}}]};
gen_get_child("ColorSel:help_button", [ColorSel|_]) ->
    {create,{gtk,color_selection_dialog_get_help_button},
     [{"help_button",{var,fmt_name(ColorSel)}}]};
gen_get_child("FontSel:ok_button", [FontSel|_]) ->
    {create,{gtk,font_selection_dialog_get_ok_button},
      [{"ok_button",{var,fmt_name(FontSel)}}]};
gen_get_child("FontSel:cancel_button", [FontSel|_]) ->
    {create,{gtk,font_selection_dialog_get_cancel_button},
     [{"cancel_button",{var,fmt_name(FontSel)}}]};
gen_get_child("FontSel:apply_button", [FontSel|_]) ->
    {create,{gtk,font_selection_dialog_get_apply_button},
      [{"apply_button",{var,fmt_name(FontSel)}}]};
gen_get_child("GtkCombo:entry", [Combo|_]) ->
    {create,{gtk,combo_get_entry},
     [{"entry",{var,fmt_name(Combo)}}]};
gen_get_child("InputDialog:close_button", [Dialog|_]) ->
    {create,{gtk,input_dialog_get_close_button},
     [{"close_button",{var,fmt_name(Dialog)}}]};
gen_get_child("InputDialog:save_button", [Dialog|_]) ->
    {create,{gtk,input_dialog_get_save_button},
     [{"save_button",{var,fmt_name(Dialog)}}]};
gen_get_child(_, ParentList) ->
    false.


emit_extra(C, Name, "GtkCombo", As) ->
    emit_combo_items(C, Name, As);
emit_extra(C, Name, "GtkOptionMenu", As) ->
    emit_option_menu_items(C, Name, As);
emit_extra(C, Name, "GtkVScale", As) ->
    emit_range_adjustment(C, Name, As);
emit_extra(C, Name, "GtkHScale", As) ->
    emit_range_adjustment(C, Name, As);
emit_extra(C, Name, "GtkVScrollbar", As) ->
    emit_range_adjustment(C, Name, As);
emit_extra(C, Name, "GtkHScrollbar", As) ->
    emit_range_adjustment(C, Name, As);
emit_extra(C, Name, Class, As) ->
    ok.


emit_range_adjustment(C, Name, As) ->
    Value = attr("value", gfloat, As, 0.0),
    Upper = attr("upper", gfloat, As, 0.0),
    Lower = attr("lower", gfloat, As, 0.0),
    Step = attr("step", gfloat, As, 0.0),
    Page = attr("page", gfloat, As, 0.0),
    Page_Size = attr("page_size", gfloat, As, 0.0),
    Adjustment = fmt_name(Name) ++ "_adjustment",
    emit_ui(C, "  ~s = gtk:range_get_adjustment(?GTK_RANGE(~s)),\n",
	    [Adjustment, fmt_name(Name)]),
    emit_ui(C, "  gtk:adjustment_put_lower(~s,~p),\n", [Adjustment, Lower]),
    emit_ui(C, "  gtk:adjustment_put_upper(~s,~p),\n", [Adjustment, Upper]), 
    emit_ui(C, "  gtk:adjustment_put_step_increment(~s,~p),\n",
	     [Adjustment, Step]),
    emit_ui(C, "  gtk:adjustment_put_page_increment(~s,~p),\n",
	     [Adjustment, Page]),
    emit_ui(C, "  gtk:adjustment_put_page_size(~s,~p),\n",
	     [Adjustment, Page_Size]),
    emit_ui(C, "  gtk:adjustment_set_value(~s,~p),\n",
	     [Adjustment, Value]),
    ok.
    

emit_combo_items(C, Name, As) ->
    Items = attr("items", gstring, As, ""),
    case string:tokens(Items, "\n") of
	[] -> ok;
	L ->
	    emit_ui(C,"  ~s_list = gtk:combo_get_list"
		    "(?GTK_COMBO(~s)),\n",
		    [fmt_name(Name), fmt_name(Name)]),
	    foldl(
	      fun(Item,I) ->
		      emit_ui(C,"  ~s_~w = gtk:list_item_new_with_label"
			      "(\"~s\"),\n",
			      [fmt_name(Name),I,Item]),
		      emit_ui(C, "  gtk:container_add"
			      "(?GTK_LIST(~s_list),~s_~w),\n",
			      [fmt_name(Name),fmt_name(Name),I]),
		      emit_ui(C, "  gtk:widget_show(~s_~w),\n",
			      [fmt_name(Name),I]),
		      I+1
	      end, 0, L)
    end.


emit_option_menu_items(C, Name, As) ->
    Items = attr("items", gstring, As, ""),
    Initial = attr("initial_choice", gint, As, 0),
    case string:tokens(Items, "\n") of
	[] -> ok;
	List ->
	    Menu = fmt_name(Name) ++ "_menu",
	    emit_ui(C, "  ~s = gtk:menu_new(),\n", [Menu]),
	    foldl(
	      fun(Label,I) ->
		      Item = Menu ++ "_item" ++ integer_to_list(I),
		      emit_ui(C,"  ~s = gtk:menu_item_new_with_label"
			      "(\"~s\"),\n", 
			      [Item, Label]),
		      emit_ui(C,"  gtk:menu_append(~s, ~s),\n",
			      [Menu, Item]),
		      emit_ui(C,"  gtk:widget_show(~s),\n",
			      [Item]),
		      I+1
	      end, 0, List),
	    emit_ui(C,"  gtk:option_menu_set_menu(~s, ~s),\n",
		    [Name, Menu]),
	    emit_ui(C,"  gtk:widget_show(~s),\n", [Menu]),
	    emit_ui(C,"  gtk:option_menu_set_history(~s,~w),\n",
		    [Name, Initial])
    end.

    

emit_signals(C, Name, As) ->
    case lists:keysearch("signal", 1, As) of
	false -> ok;
	{value,{_,SAs}} ->
	    Signal = attr("name", gstring, SAs),
	    Handler = attr("handler", gstring, SAs),
	    Raw = glade_parse:xml2raw(Handler),
	    Exprs = parse_erl(Raw),
	    Data = attr("data", gstring, SAs, []),
	    emit_ui(C, "  gtk:signal_connect(~s,'~s',~s,~p),\n",
		    [fmt_name(Name),Signal,fmt_handler(C,Signal,Exprs),Data])
    end.

emit_tooltip(C, Name, As) ->
    case attr("tooltip", gstring, As, "") of
	"" -> ok;
	Text ->
	    emit_ui(C, "  ~s_tips = gtk:tooltips_new(),\n",
		    [fmt_name(Name)]),
	    emit_ui(C, "  gtk:tooltips_set_tip(~s_tips,~s,\"~s\",?NULL),\n",
		    [fmt_name(Name), fmt_name(Name), Text])
    end.    


fmt_handler(C,Signal,Exprs) ->
    case Exprs of
	[{atom,_,Fun}] ->
	    io_lib:format("{~s_cb, ~w}", [C#compile.basename, Fun]);

	[{var,_,Fun}] ->
	    io_lib:format("{~s_cb, ~w}", [C#compile.basename, Fun]);

	[{remote,_,{atom,_,Mod},{atom,_,Fun}}] ->
	    io_lib:format("{~w, ~w}", [Mod, Fun]);

	[{'fun',_,_}] ->
	    %% assume the call know the arity
	    erl_pp:exprs(Exprs);
	_ ->
	    %% ok list of expression generate a fun
	    ArgList = case lists:suffix("signal", Signal) of 
			  true ->			  
			      [{var,1,'Widget'},{var,1,'Event'},
			       {var,1,'Data'}];
			  false ->
			      [{var,1,'Widget'}, {var,1,'Data'}]
		      end,
	    Exprs1 = [{'fun',1,
		       {clauses,[{clause,1,
				  ArgList, [], Exprs}]}}],
	    erl_pp:exprs(Exprs1)
    end.
	    
    

fmt_handler(C,Data) ->
    %% if only chars digits and _ then 
    %% assume the funtion to be in xxx_cb module
    case is_identifier(Data) of
	true -> "{" ++ C#compile.basename ++ "_cb," ++ Data ++ "}";
	false -> Data
    end.

is_identifier([C|Cs]) when C >= $0, C =< $9 ->
    is_identifier(Cs);
is_identifier([C|Cs]) when C >= $a, C =< $z ->
    is_identifier(Cs);
is_identifier([C|Cs]) when C >= $A, C =< $Z ->
    is_identifier(Cs);
is_identifier([$_|Cs]) ->
    is_identifier(Cs);
is_identifier([]) -> true;
is_identifier(_) -> false.

is_window("GtkWindow") -> true;
is_window("GtkDialog") -> true;
is_window("GtkInputDialog") -> true;
is_window("GtkColorSelectionDialog") -> true;
is_window("GtkFileSelection") -> true;
is_window("GtkFontSelectionDialog") -> true;
is_window("") -> error;
is_window(_) -> false.


emit_spec(Fd, Name, Spec, As) ->
    emit_spec(Fd, [], Name, Spec, As).

emit_spec(C, Ps, Name, [H | T], As) ->
    case H of
	{create,Fun,A} ->
	    {_,Vs} = eval_args(A, As),
	    Nm = fmt_name(Name),
	    emit_ui(C, "  ~s = ~s(~s),\n", 
		      [Nm,fmt_fun(Fun),fmt_args(Vs)]),
	    emit_spec(C, [{var,Nm}], Name, T, As);
	{call,Fun,A} ->
	    {_,Vs} = eval_args(A, As),
	    emit_ui(C, "  ~s(~s),\n", [fmt_fun(Fun),fmt_args(Ps++Vs)]),
	    emit_spec(C, Ps, Name,T, As);
	{def,Fun,A} ->
	    case eval_args(A, As) of
		{default,_} -> 
		    emit_spec(C,Ps,Name,T,As);
		{value,Vs} ->
		    emit_ui(C, "  ~s(~s),\n", 
			    [fmt_fun(Fun),fmt_args(Ps++Vs)]),
		    emit_spec(C,Ps,Name,T,As)
	    end;
	{flags,SetF,UnSetF,Default,Fs} ->
	    {On,Off} = eval_flags(Fs, As, 0, 0),
	    OnBits = On band (bnot Default),
	    OffBits = Off band Default,
	    FSpec = [{?GTK_SENSITIVE,"?GTK_SENSITIVE"},
		     {?GTK_CAN_DEFAULT,"?GTK_CAN_DEFAULT"},
		     {?GTK_HAS_DEFAULT,"?GTK_HAS_DEFAULT"},
		     {?GTK_CAN_FOCUS,"?GTK_CAN_FOCUS"},
		     {?GTK_HAS_FOCUS,"?GTK_HAS_FOCUS"},
		     {?GTK_VISIBLE,"?GTK_VISIBLE"},
		     {?GTK_VISIBLE,"?GTK_VISIBLE"},
		     {?GTK_PARENT_SENSITIVE,"?GTK_PARENT_SENSITIVE"}],
	    if OnBits == 0 ->
		    ok;
	       true -> 
		    emit_ui(C, "  ~s(~s),\n",
			    [fmt_fun(SetF),
			     fmt_args(Ps++[{raw,fmt_flags(OnBits,FSpec)}])])
	    end,
	    if OffBits == 0 ->
		    ok;
	       true ->
		    emit_ui(C, "  ~s(~s),\n",
			    [fmt_fun(UnSetF),
			     fmt_args(Ps++[{raw,fmt_flags(OffBits,FSpec)}])])
	    end,
	    emit_spec(C,Ps,Name,T,As);
	{conditional,Type,Then,Else} ->
	    case eval_cond(Type,As) of
		true -> 
		    emit_spec(C,Ps,Name,Then, As),
		    emit_spec(C,Ps,Name,T,As);
		false ->
		    emit_spec(C,Ps,Name,Else, As),
		    emit_spec(C,Ps,Name,T, As)
	    end
    end;
emit_spec(C,[],Name,[],As) ->
    "";
emit_spec(C,[W],Name,[],As) ->
    W.


emit_ui(C, Fmt, As) ->
    if C#compile.ui_fd == undefined -> ok;
       true -> io:format(C#compile.ui_fd, Fmt, As)
    end.

emit_cb(C, Fmt, As) ->
    if C#compile.cb_fd == undefined -> ok;
       true -> io:format(C#compile.cb_fd, Fmt, As)
    end.

emit_main(C, Fmt, As) ->
    if C#compile.main_fd == undefined -> ok;
       true -> io:format(C#compile.main_fd, Fmt, As)
    end.
	    

fmt_fun({gtk,widget_set_flags}) -> "?GTK_WIDGET_SET_FLAGS";
fmt_fun({gtk,widget_unset_flags}) -> "?GTK_WIDGET_UNSET_FLAGS";
fmt_fun({M,F}) -> atom_to_list(M) ++ ":" ++ atom_to_list(F);
fmt_fun(F) when atom(F) -> atom_to_list(F).


fmt_args([]) -> "";
fmt_args([A1]) -> fmt(A1);
fmt_args([A1|As]) -> fmt(A1) ++ "," ++ fmt_args(As).

fmt(A) when integer(A) -> integer_to_list(A);
fmt(true) -> "true";
fmt(false) -> "false";
fmt(A) when atom(A) -> "'" ++ atom_to_list(A) ++ "'";
fmt(A) when float(A) -> lists:flatten(io_lib:format("~g", [A]));
fmt(Macro=[$?|Cs]) -> Macro;
fmt(A) when list(A) -> "\"" ++ A ++ "\"";
fmt({var,Nm})        -> Nm;
fmt({raw,Cs})        -> Cs;
fmt({'GtkObject',0}) -> "?GTK_NULL".

    

fmt_name([C|Cs]) when C >= $a, C =< $z ->
    [(C-$a) + $A | fmt_name1(Cs)];
fmt_name(Cs) when atom(Cs) ->
    fmt_name(atom_to_list(Cs));
fmt_name(Cs) ->
    fmt_name1(Cs).

fmt_name1([$-|Cs]) -> [$_|fmt_name1(Cs)];
fmt_name1([C|Cs]) -> [C|fmt_name1(Cs)];
fmt_name1([]) -> [].
    

    

    

collect_attributes([W={"widget",Param,Children}|T], As, Ws) ->
    collect_attributes(T, As, [W|Ws]);
collect_attributes([A|T], As, Ws) ->
    collect_attributes(T, [A|As], Ws);
collect_attributes([], As, Ws) ->
    { reduce_attributes(As,[]), reverse(Ws)}.

reduce_attributes([{Tag,_,[{pcdata,Data}]} | Ts], Acc) ->
    case Tag of
	"last_modification_time" ->
	    reduce_attributes(Ts, Acc);
	_ ->
	    reduce_attributes(Ts, [{Tag,Data}|Acc])
    end;
reduce_attributes([{Tag,_,[]}|Ts], Acc) ->
    reduce_attributes(Ts, [{Tag,""} | Acc]);
reduce_attributes([{Tag,_,Children}|Ts], Acc) ->
    Cs = reduce_attributes(Children, []),
    reduce_attributes(Ts, [{Tag,Cs} | Acc]);
reduce_attributes([], Acc) ->
    reverse(Acc).


%%
%% Create a build spec for creating a widget
%% Return
%%   { [{Mod,Fun,Args} | MFAs], ChildrenSpec
%% args = {Name,Type,Default}
%%
%% ChildrenSpec = none | box | container | notebook ...
%% 
%% {create,{Mod,Fun},Args}      unconditional call returning a new widget
%% {call,{Mod,Fun},Args}        unconditional call
%% {def,{Mod,Fun},Args}         optional call (ignore if default value on args)
%% {flags,SetF,UnsetF,Default,Flags}  
%%                              special call for setting flag arguments
%% {conditional,Cond,Then,Else} conditional 
%%
widget_spec(DefaultFlags) ->
    [
     {def,{gtk,widget_set_uposition},[{"x", gint, 0},
				      {"y", gint, 0}]},
     {def,{gtk,widget_set_usize}, [{"width", gint, 0},
				  {"height", gint, 0}]},
     {flags,
      {gtk,widget_set_flags}, {gtk,widget_unset_flags},
      DefaultFlags bor ?GTK_SENSITIVE bor ?GTK_PARENT_SENSITIVE,
      [{"can_default", ?GTK_CAN_DEFAULT},
       {"has_default", ?GTK_HAS_DEFAULT},
       {"can_focus", ?GTK_CAN_FOCUS},
       {"has_focus", ?GTK_HAS_FOCUS},
       {"visible", ?GTK_VISIBLE},
       {"sensitive", ?GTK_SENSITIVE}]}].

container_spec(Flags) ->
    [
     {def,{gtk,container_set_border_width},[{"border_width",guint,0}]}
     | widget_spec(Flags)].

bin_spec(Flags) ->
    container_spec(Flags).

clist_spec(Flags) ->
    [
     {def,{gtk,clist_set_selection_mode},
      [{"selection_mode",genum,'GTK_SELECTION_SINGLE'}]},
     {def,{gtk,clist_set_shadow_type},
      [{"shadow_type",genum,'GTK_SHADOW_IN'}]},
     {conditional, {value,"show_titles",false},
      [{call,{gtk,clist_column_titles_show},[]}],
      [{call,{gtk,clist_column_titles_hide},[]}]}
     | container_spec(?GTK_CAN_FOCUS bor Flags)].

box_spec(Flags) ->
    container_spec(Flags).

bbox_spec(Flags) ->
    [
     {def,{gtk,button_box_set_layout},
      [{"layout_style", genum, 'GTK_BUTTONBOX_DEFAULT_STYLE'}]},
     {def,{gtk,button_box_set_spacing},
      [{"spacing", gint, 30}]},
     {def,{gtk,button_box_set_child_size},
      [{"child_min_width", gint, 85},
       {"child_min_height", gint, 27}]},
     {def,{gtk,button_box_set_child_ipadding},
      [{"child_ipad_x", gint, 7},
       {"child_ipad_y", gint, 0}]}
     | box_spec(Flags)].
	   

button_spec(Flags) ->
    [{def,{gtk,button_set_relief},
	   [{"relief",genum, 'GTK_RELIEF_NORMAL'}]}
     | container_spec(?GTK_CAN_FOCUS bor ?GTK_RECEIVES_DEFAULT bor Flags)].

toggle_button_spec(Flags) ->
    [{def,{gtk,toggle_button_set_active},
	   [{"active",gboolean,false}]},
     {def,{gtk,toggle_button_draw_indicator},
	   [{"draw_indicator", gboolean,true}]}
     | button_spec(Flags)].

paned_spec(Flags) ->
    [{def,{gtk,paned_set_handle_size},
	   [{"handle_size",guint,10}]},
     {def,{gtk,paned_set_gutter_size},
	   [{"gutter_size",guint,6}]}
     | container_spec(Flags)].

misc_spec(Flags) ->
    [{def,{gtk,misc_set_alignment},
      [{"xalign", gfloat, 0.5},
       {"yalign", gfloat, 0.5}]},
     {def,{gtk,misc_set_padding},
      [{"xpad", gint, 0},
       {"ypad", gint, 0}]} | widget_spec(Flags)].


set_entry_spec() ->
    [
     {def,{gtk,entry_set_editable},
       [{"editable",gboolean,true}]},
      {def,{gtk,entry_set_visiblility},
       [{"text_visible",gboolean,true}]},
      {def,{gtk,entry_set_max_length},
       [{"text_max_length",guint,0}]},
      {def,{gtk,entry_set_text},[{"text",gstring,""}]}
     | widget_spec(?GTK_CAN_FOCUS)].

window_spec(Name,Flags) ->
    [ 
      {call,{gtk,window_set_title},
       [{"title",gstring,Name}]},
      {def,{gtk,window_set_modal},
       [{"modal",gboolean,false}]},
      {def,{gtk,window_set_policy},
       [{"allow_shrink",gboolean,false},
	{"allow_grow",gboolean,true},
	{"auto_shrink",gboolean,false}]} 
      | container_spec(Flags)].

range_spec(Flags) ->
    [
     {def,{gtk,range_set_update_policy},
      [{"policy", genum, 'GTK_UPDATE_CONTINOUS'}]}
    ].

scale_spec(Flags) ->
    [
     {def,{gtk,scale_set_draw_value},
      [{"draw_value", gboolean, true}]},
     {def,{gtk,scale_set_value_pos},
      [{"value_pos", genum, 'GTK_POS_TOP'}]},
     {def,{gtk,scale_set_digits},
      [{"digits", gint, 1}]} 
     | range_spec(Flags)].

menu_spec(Flags) ->
    container_spec(Flags).


spec(Class, Name, Type) ->
    case Class of
	"GtkWindow" ->
	    S = [
		 {create,{gtk,window_new},
		  [{"type",genum,'GTK_WINDOW_TOPLEVEL'}]} |
		 window_spec(Name, 0)],
	    {S, container};

	"GtkDialog" ->
	    S = [
		 {create,{gtk,dialog_new},[]} |
		 window_spec(Name, 0)],
	    {S, container};

	"GtkInputDialog" ->
	    S = [
		 {create,{gtk,input_dialog_new},[]} |
		 window_spec(Name, 0)],
	    {S, container};

	"GtkColorSelectionDialog" ->
	    S = [
		 {create,{gtk,color_selection_dialog_new},
		  [{"title",gstring,""}]} |
		 window_spec(Name, 0)],
	    {S, container};

	"GtkFileSelection" ->
	    S = [
		 {create,{gtk,file_selection_new},
		  [{"title",gstring,""}]} |
		 window_spec(Name, 0)],
	    {S, container};

	"GtkFontSelectionDialog" ->
	    S = [
		 {create,{gtk,font_selection_dialog_new},
		  [{"title",gstring,""}]} |
		 window_spec(Name, 0)],
	    {S, container};

	"GtkAlignment" ->
	    S = [
		 {create,{gtk,alignment_new},
		  [{"xalign", gfloat, 0.5},
		   {"yalign", gfloat, 0.5},
		   {"xscale", gfloat, 1},
		   {"yscale", gfloat, 1}]}
		 | bin_spec(0)],
	    {S, container};

	"GtkHScale" ->
	    S = [
		 {create,{gtk,hscale_new},
		  [{"adjustment", ?GTK_NULL}]}
		 | scale_spec(0)],
	    {S, none};

	"GtkVScale" ->
	    S = [
		 {create,{gtk,vscale_new},
		  [{"adjustment", ?GTK_NULL}]}
		 | scale_spec(0)],
	    {S, none};

	"GtkHScrollbar" ->
	    S = [
		 {create,{gtk,hscrollbar_new},
		  [{"adjustment", ?GTK_NULL}]}
		 | range_spec(0)],
	    {S, none};

	"GtkVScrollbar" ->
	    S = [
		 {create,{gtk,vscrollbar_new},
		  [{"adjustment", ?GTK_NULL}]}
		 | range_spec(0)],
	    {S, none};
	    
	"GtkScrolledWindow" ->
	    S = [
		 {create,{gtk,scrolled_window_new},
		  [{"hadjustment", ?GTK_NULL},
		   {"vadjustment", ?GTK_NULL}]},
		 {def,{gtk,scrolled_window_set_policy},
		  [{"hscrollbar_policy", genum, 'GTK_POLICY_ALWAYS'},
		   {"vscrollbar_policy", genum, 'GTK_POLICY_ALWAYS'}]}
		 | container_spec(0)],
	    {S, container};


	"GtkViewport" ->
	    S = [
		 {create,{gtk,viewport_new},
		  [{"hadjustment", ?GTK_NULL},
		   {"vadjustment", ?GTK_NULL}]},
		 {def,{gtk,viewport_set_shadow_type},
		  [{"shadow_type", genum, 'GTK_SHADOW_IN'}]}
		| container_spec(0) ],
	    {S, container};

	    
	"GtkVBox" ->
	    S = [
		 {create,{gtk,vbox_new},
		  [{"homogeneous",gboolean,false},
		   {"spacing",gint,0}]}
		 | box_spec(0) ],
	    {S, box};

	"GtkHBox" ->
	    S = [
		 {create,{gtk,hbox_new},
		  [{"homogeneous",gboolean,false},
		   {"spacing",gint,0}]} 
		 | box_spec(0)],
	    {S, box};

	"GtkVButtonBox" ->
	    S = [
		 {create,{gtk,vbutton_box_new},[]}
		 | bbox_spec(0) ],
	    {S, container};

	"GtkHButtonBox" ->
	    S = [
		 {create,{gtk,hbutton_box_new},[]}
		 | bbox_spec(0)],
	    {S, container};

	"GtkTable" ->
	    S = [
		 {create,{gtk,table_new},
		  [{"rows",guint,2},{"columns",guint,2},
		   {"homogeneous",gboolean,true}]},
		 {def,{gtk,table_set_row_spacings},
		  [{"row_spacing",guint,0}]},
		 {def,{gtk,table_set_col_spacings},
		       [{"column_spacing",guint,0}]}
		 | container_spec(0)],
	    {S, table};
		 
	"GtkLabel" ->
	    S = [
		 {create,{gtk,label_new},
		  [{"label",gstring,Name}]},
		 {def,{gtk,label_set_justify},
		  [{"justify",genum,'GTK_JUSTIFY_CENTER'}]},
		 {def,{gtk,label_set_line_wrap},
		  [{"wrap",gboolean,false}]}
		 | misc_spec(0)],
	    {S, none};
	    
	"GtkFrame" ->
	    S = [
		 {create,{gtk,frame_new},[{"label",gstring,Name}]},
		 {def,{gtk,frame_set_label_align},
		       [{"label_xalign",gfloat,0.0},
			{"label_yalign",gfloat,0.0}]},
		 {def,{gtk,frame_set_shadow_type},
		  [{"shadow_type",genum,'GTK_SHADOW_ETCHED_IN'}]}
		 | container_spec(0)],
	    {S, container};

	"GtkNotebook" ->
	    S = [
		 {create,{gtk,notebook_new},
		  []},
		 {def,{gtk,notebook_set_show_tab},
		  [{"show_tabs",gboolean,true}]},
		 {def,{gtk,notebook_set_show_tab},
		  [{"show_border",gboolean,true}]},
		 {def,{gtk,notebook_set_tab_pos},
		  [{"tab_pos",genum,'GTK_POS_TOP'}]},
		 {def,{gtk,notebook_set_scrollable},
		  [{"scrollable",gboolean,false}]},
		 {def,{gtk,notebook_set_tab_hborder},
		  [{"tab_hborder",gint,2}]},
		 {def,{gtk,notebook_set_tab_vborder},
		  [{"tab_vborder",gint,2}]},
		 {def,{gtk,notebook_set_popup_enable},
		  [{"popup_enable",gboolean,false}]}
		 | container_spec(?GTK_CAN_FOCUS)],
	    {S, notebook};

	"GtkToggleButton" ->
	    S = [
		 {create,{gtk,toggle_button_new_with_label},
		  [{"label",gstring,Name}]}
		 | toggle_button_spec(0)],
	    {S, container};

	"GtkRadioButton" ->
	    S = [
		 {create,{gtk,radio_button_new_with_label},
		  [{"group", []},
		   {"label",gstring,Name}]} 
		 | toggle_button_spec(0)],
	    {S, container};

	"GtkCheckButton" ->
	    S = [
		 {create,{gtk,check_button_new_with_label},
		  [{"label",gstring,Name}]} 
		 | toggle_button_spec(0)],
	    {S, container};

	"GtkSpinButton" ->
	    S = [
		 {create,{gtk,spin_button_new},
		  [{"adjustment",?GTK_NULL},
		   {"climbe_rate", gfloat, 1.0},
		   {"digits", gint, 0}]},
		 {def,{gtk,spin_button_set_update_policy},
		  [{"update_policy", genum, 'GTK_UPDATE_ALWAYS'}]},
		 {def,{gtk,spin_button_set_numeric},
		  [{"numeric",gboolean,false}]},
		 {def,{gtk,spin_button_set_snap_to_ticks},
		  [{"snap",gboolean,false}]},
		 {def,{gtk,spin_button_set_wrap},
		  [{"wrap",gboolean,false}]},
		 {def,{gtk,spin_button_set_value},
		  [{"value",gfloat,1.0}]}
		 | widget_spec(?GTK_CAN_FOCUS)],
	    {S, none};

	"GtkProgressBar" ->
	    S = [
		 {create,{gtk,progress_bar_new},
		  []},
		 {def,{gtk,progress_bar_set_bar_style},
		  [{"bar_style", genum, 'GTK_PROGRESS_CONTINUOUS'}]},
		 {def,{gtk,progress_bar_set_bar_orientaion},
		  [{"orientation",genum,'GTK_PROGRESS_LEFT_TO_RIGHT'}]},

		 {def,{gtk,progress_set_show_text},
		  [{"show_text",gboolean,false}]},
		 {def,{gtk,progress_set_text_alignment},
		  [{"text_xalign",gfloat,0.5},
		   {"text_yalign",gfloat,0.5}]},
		 {def,{gtk,progress_set_format_string},
		  [{"format",gstring,"%P %%"}]},
		 {def,{gtk,progress_set_activity_mode},
		  [{"activity_mode",gboolean,false}]},
		 {def,{gtk,progress_set_value}
		  ,[{"value", gfloat, 0.0}]}
		 | widget_spec(0)],
	    {S, none};

	"GtkStatusbar" ->
	    S = [
		 {create,{gtk,statusbar_new},
		  []}
		 | box_spec(0)],
	    {S, box};

	"GtkButton" ->
	    S = [
		 {conditional,  {member,"stock_button"},
		  [{create,{gnome_stock,button},
		    [{"stock_button", {gmacro,Type}, "Button"}]}],
		  [{create,{gtk,button_new_with_label},
		    [{"label",gstring,Name}]}]}
		 | button_spec(0)],
	    {S, container};

	"GtkEntry" ->
	    S = [
		 {create,{gtk,entry_new_with_max_length},
		  [{"text_max_length",guint,0}]},
		 {def,{gtk,entry_set_editable},
		  [{"editable",gboolean,true}]},
		 {def,{gtk,entry_set_visiblility},
		  [{"text_visible",gboolean,true}]},
		 {def,{gtk,entry_set_text},[{"text",gstring,""}]}
		 | widget_spec(?GTK_CAN_FOCUS)],
	    {S, none};

	"GtkText" ->
	    S = [
		 {create,{gtk,text_new},
		  [{"hadj", ?GTK_NULL},{"vadj",?GTK_NULL}]},
		 {def,{gtk,text_set_editable},
		  [{"editable",gboolean,true}]},
		 {def,{gtk,text_insert},
		  [{"font",?NULL},
		   {"fore",?NULL},
		   {"back",?NULL}, 
		   {"text",gstring,""}, 
		   {"text_length",-1}]}
		 | widget_spec(?GTK_CAN_FOCUS)],
	    {S, none};

	"GtkHPaned" ->
	    S = [
		 {create,{gtk,hpaned_new},[]} |
		 paned_spec(0)],
	    {S, paned};

	"GtkVPaned" ->
	    S = [
		 {create,{gtk,vpaned_new},[]} |
		 paned_spec(0)],
	    {S, paned};

	"GtkHSeparator" ->
	    S = [
		 {create,{gtk,hseparator_new},[]}
		| widget_spec(0) ],
	    {S, none};

	"GtkVSeparator" ->
	    S = [
		 {create,{gtk,vseparator_new},[]}
		 | widget_spec(0)],
	    {S, none};

	"GtkList" ->
	    S = [
		 {create,{gtk,list_new}, []},
		 {def,{gtk,list_set_selection_mode},
		       [{"selection_mode",genum,'GTK_SELECTION_SINGLE'}]} 
		 | container_spec(0)],
	    {S, container};

	"GtkCList" ->
	    S = [
		 {create,{gtk,clist_new}, [{"columns",gint,1}]}
		 | clist_spec(0)],
	    {S, clist};

	"GtkTree" ->
	    S = [
		 {create,{gtk,tree_new}, []},
		 {def, {gtk,tree_set_selection_mode},
		  [{"selection_mode",genum,'GTK_SELECTION_SINGLE'}]},
		 {def,{gtk,tree_set_view_mode},
		  [{"view_mode",genum,'GTK_TREE_VIEW_LINE'}]},
		 {def,{gtk,tree_set_view_line},
		  [{"view_line",gboolean,true}]}
		 | container_spec(0)],
	    {S,container};

	"GtkCTree" ->
	    S = [
		 {create,{gtk,ctree_new},
		  [{"columns",gint,1},
		   {"tree_column",gint,1}]}
		 | clist_spec(0)],
	    {S,clist};

	"GtkOptionMenu" ->
	    S = [
		 {create,{gtk,option_menu_new}, []}
		 | button_spec(?GTK_CAN_FOCUS)
		],
	    {S,container};

	"GtkPixmap" ->
	    S = [
		 {create,{glade2erl,pixmap_new},
		  [{"parent",gobject,?GTK_NULL},
		   {"filename",gstring,""}]}
		 | misc_spec(0)],
	    {S, none};

	"GtkFixed" ->
	    S = [
		 {create,{gtk,fixed_new},
		  []}
		 | container_spec(0)],
	    {S, fixed};

	"GtkCombo" ->
	    S = [
		 {create,{gtk,combo_new},
		  []},
		 {def,{gtk,combo_set_value_in_list},
		  [{"value_in_list", gboolean, false},
		   {"ok_if_empty", gboolean, true}]},
		 {def,{gtk,combo_set_case_sensitive},
		  [{"case_sensetive",gboolean, false}]},
		 {def,{gtk,combo_set_use_arrows},
		  [{"use_arrows", gboolean, true}]},
		 {def,{gtk,combo_set_use_arrows_always},
		  [{"use_arrows_always", gboolean, true}]}
		 | box_spec(0)],
	    {S, combo};

	"GtkMenuBar" ->
	    S = [
		 {create,{gtk,menu_bar_new},[]},
		 {def,{gtk,menu_bar_set_shadow_type},
		  [{"shadow_type", genum, 'GTK_SHADOW_OUT'}]}
		 | container_spec(0)],
	    {S, menubar};

	"GtkMenu" ->
	    S = [
		 {create,{gtk,menu_new},[]}
		 | menu_spec(0)],
	    {S, menu};

	"GtkMenuItem" ->
	    S = [
		 {create,{gtk,menu_item_new_with_label},
		  [{"label", gstring, Name}]},
		 {conditional, {value,"right_justify",false},
		  [{call,{gtk,menu_item_right_justify},[]}],
		  []} | container_spec(0)],
	    {S, menuitem};

	"GtkCheckMenuItem" ->
	    S = [
		 {create,{gtk,check_menu_item_new_with_label},
		  [{"label", gstring, "label"}]},
		 {def, {gtk,check_menu_item_set_active},
		  [{"active", gboolean, false}]},
		 {def, {gtk,check_menu_item_set_show_toggle},
		  [{"always_show_toggle", gboolean, false}]}
		 | container_spec(0)],
	    {S, menuitem};

	"GtkRadioMenuItem" ->
	    S = [
		 {create,{gtk,radio_menu_item_new_with_label},
		  [{"group", []},
		   {"label", gstring, "label"}]},
		 {def, {gtk,check_menu_item_set_active},
		  [{"active", gboolean, false}]},
		 {def, {gtk,check_menu_item_set_show_toggle},
		  [{"always_show_toggle", gboolean, false}]}
		 | container_spec(0)],
	    {S, menuitem};

	"GtkToolbar" ->
	    S = [
		 {create,{gtk,toolbar_new},
		  [{"orientaion", genum, 'GTK_ORIENTATION_HORIZONTAL'},
		   {"type", genum, 'GTK_TOOLBAR_BOTH'}]},
		 {def,{gtk,toolbar_set_space_size},
		  [{"space_size", gint, 5}]},
		 {def,{gtk,toolbar_set_space_style},
		  [{"space_style", genum, 'GTK_TOOLBAR_SPACE_EMPTY'}]},
		 {def,{gtk,toolbar_set_button_relief},
		  [{"relief", genum, 'GTK_RELIEF_NORMAL'}]},
		 {def,{gtk,toolbar_set_tooltips},
		  [{"tooltips", gboolean, true}]}
		 | container_spec(0)],
	    {S, toolbar};

	"GtkHandleBox" ->
	    S = [
		 {create,{gtk,handle_box_new},
		  []},
		 {def,{gtk,handle_box_set_shadow_type},
		  [{"shadow_type", genum, 'GTK_SHADOW_OUT'}]},
		 {def,{gtk,handle_box_set_handle_poistion},
		  [{"handle_position", genum, 'GTK_POS_LEFT'}]},
		 {def,{gtk,handle_box_set_snap_edge},
		  [{"snap_edge", genum, 'GTK_POS_TOP'}]}
		 | container_spec(0)],
	    {S, container};
		 
	"Placeholder" ->  %% glade special (create a vbox)
	    S = [{create,{gtk,vbox_new},
		  [{"homogeneous",gboolean,false},
		   {"spacing",gint,0}]}
		 | box_spec(0)],
	    {S, box};
	
	_ ->
	    io:format("Warning: can not create ~s:~s YET\n", [Name,Class]),
	    S = [{create,{gtk,label_new},
		  [{"label", Class ++ " not implemented"}]}
		 | widget_spec(0)
		],
	    {S, none}
    end.
    
%%
%% Exectute and create a widget
%%
exec_spec(Ps, [H | T], As) ->
    ?Debug("exec: ~p\n", [H]),
    case H of
	{create,Fun,A} ->
	    {_,Vs} = eval_args(A, As),
	    Widget = ex(Fun,Ps++Vs),
	    exec_spec([Widget], T, As);
	{call,Fun,A} ->
	    {_,Vs} = eval_args(A, As),
	    ex(Fun,Ps++Vs),
	    exec_spec(Ps,T, As);
	{def,Fun,A} ->
	    case eval_args(A, As) of
		{default,_} -> 
		    exec_spec(Ps,T,As);
		{value,Vs} ->
		    ex(Fun,Ps++Vs),
		    exec_spec(Ps,T,As)
	    end;
	{flags,SetF,UnSetF,Default,Fs} ->
	    {On,Off} = eval_flags(Fs, As, 0, 0),
	    OnBits = On band (bnot Default),
	    OffBits = Off band Default,
	    if OnBits == 0 ->
		    ok;
	       true -> 
		    ex(SetF, Ps++[OnBits])
	    end,
	    if OffBits == 0 ->
		    ok;
	       true ->
		    ex(UnSetF, Ps++[OffBits])
	    end,
	    exec_spec(Ps,T,As);
	{conditional,Type,Then,Else} ->
	    case eval_cond(Type,As) of
		true -> 
		    exec_spec(Ps, Then++T, As);
		false ->
		    exec_spec(Ps, Else++T, As)
	    end
    end;
exec_spec([], [], As) ->
    ?GTK_NULL;
exec_spec([Widget],[],As) ->
    Widget.

ex(Fun,As) ->
    ?Debug("ex: ~p ~p\n", [Fun, As]),
    Res = (catch apply(Fun, As)),
    case Res of 
	{'EXIT',Reason} ->
	    io:format("apply(~p,~p) = ~p\n", [Fun,As,Res]),
	    exit(Reason);
	_ -> Res
    end.

%%
%% Evaluate arguments return {default,Vs} if all arguments are defaults
%% return {value,Vs} otherwise
%%

eval_args([], _) -> %% argument-less calls are mandatory!!
    {value,[]};  
eval_args(Spec, As) ->
    eval_args(Spec, As, default, []).

eval_args([{Name,Type,Default} | Spec], As, Status, Acc) ->  
    case lists:keysearch(Name,1,As) of
	{value,{_,Val}} ->
	    case conv(Type,Val) of
		Default -> eval_args(Spec, As, Status, [Default|Acc]);
		Value   -> eval_args(Spec, As, value,  [Value|Acc])
	    end;
	false ->
	    eval_args(Spec, As, Status, [Default|Acc])
    end;
eval_args([{Name,Value} | Spec], As, Status, Acc) ->  
    eval_args(Spec, As, Status, [Value|Acc]);
eval_args([], As, Status, Acc) ->
    {Status, reverse(Acc)}.


eval_cond({'and', List}, As) -> 
    lists:all(fun(X) -> eval_cond(X,As) end, List);
eval_cond({'or', List}, As)  -> 
    lists:any(fun(X) -> eval_cond(X,As) end, List);
eval_cond({'xor', A, B}, As) -> eval_cond(A,As) xor eval_cond(B,As);
eval_cond({'not', A}, As)    -> not eval_cond(A,As);
eval_cond({member,Name}, As) ->
    case lists:keysearch(Name,1,As) of
	{value, _} -> true;
	_ -> false
    end;
eval_cond({value,Name,Default}, As) ->
    case lists:keysearch(Name,1,As) of
	{value, {_,Val}} -> conv(gboolean,Val);
	_ -> Default
    end;
eval_cond('true', As)        -> true;
eval_cond('false', As)       -> false.



eval_flags([{Name,Flag} | Fs], As, Set, Unset) ->
    case lists:keysearch(Name,1,As) of
	{value,{_,"True"}} ->
	    eval_flags(Fs, As, Set bor Flag, Unset);
	{value,{_,"False"}} ->
	    eval_flags(Fs, As, Set, Unset bor Flag);
	_ ->
	    eval_flags(Fs, As, Set, Unset)
    end;
eval_flags([], As, Set, Unset) ->
    {Set, Unset}.
	    

flag(Name, As, Default, On, Off) ->
    case attr(Name, gboolean, As, Default) of
	true -> On;
	false -> Off
    end.

attr(Name, Type, As) ->
    {value,{_,Value}} = lists:keysearch(Name, 1, As),
    conv(Type, Value).

attr(Name, Type, As, Default) ->
    case lists:keysearch(Name, 1, As) of
	false ->
	    if function(Default) ->
		    Default();
	       true ->
		    Default
	    end;
	{value,{_,Value}} -> 
	    conv(Type,Value)
    end.

%% child version of above

cattr(Name, Type, As) ->
    %% ?Debug("cattr: Name=~p, Type=~p, As=~p\n", [Name, Type, As]),
    case lists:keysearch("child", 1, As) of
	{value,{_,CAs}} ->
	    attr(Name, Type, CAs);
	false ->
	    exit({no_attr, Name})
    end.
	    
cattr(Name, Type, As, Default) ->
    case lists:keysearch("child", 1, As) of
	false -> 
	    Default;
	{value,{_,CAs}} ->
	    attr(Name, Type, CAs, Default)
    end.

cflag(Name, As, Default, On, Off) ->
    case lists:keysearch("child", 1, As) of
	false -> Default;
	{value,{_,CAs}} ->
	    flag(Name, CAs, Default, On, Off)
    end.



conv(gint, Cs) -> list_to_integer(Cs);
conv(guint, Cs) -> list_to_integer(Cs);
conv(gfloat,Cs) ->
    case catch list_to_float(Cs) of
	{'EXIT',_} -> 
	    float(list_to_integer(Cs));
	F -> F
    end;
conv(gstring,Cs) -> Cs;
conv(genum, Cs) -> list_to_atom(Cs);
conv(gboolean,"True") -> true; 
conv(gboolean,"False") -> false;
conv(gobject, Object) -> Object;
conv({gmacro,emit}, Name) -> "?" ++ Name;
conv({gmacro,view}, Name) ->
    case Name of
	"GNOME_STOCK_BUTTON_OK" -> ?GNOME_STOCK_BUTTON_OK;
	"GNOME_STOCK_BUTTON_CANCEL" -> ?GNOME_STOCK_BUTTON_CANCEL;
	"GNOME_STOCK_BUTTON_YES" -> ?GNOME_STOCK_BUTTON_YES;
	"GNOME_STOCK_BUTTON_NO" -> ?GNOME_STOCK_BUTTON_NO;
	"GNOME_STOCK_BUTTON_CLOSE" -> ?GNOME_STOCK_BUTTON_CLOSE;
	"GNOME_STOCK_BUTTON_APPLY" -> ?GNOME_STOCK_BUTTON_APPLY;
	"GNOME_STOCK_BUTTON_HELP" -> ?GNOME_STOCK_BUTTON_HELP;
	"GNOME_STOCK_BUTTON_NEXT" -> ?GNOME_STOCK_BUTTON_NEXT;
	"GNOME_STOCK_BUTTON_PREV" -> ?GNOME_STOCK_BUTTON_PREV;
	"GNOME_STOCK_BUTTON_UP" -> ?GNOME_STOCK_BUTTON_UP;
	"GNOME_STOCK_BUTTON_DOWN" -> ?GNOME_STOCK_BUTTON_DOWN;
	"GNOME_STOCK_BUTTON_FONT" -> ?GNOME_STOCK_BUTTON_FONT;
	_ -> Name
    end.


%% wrapper for the undefined case
is_a(undefined, Type) -> false;
is_a(Widget, Type) -> gtk:type_is_a(Widget, Type).

    
