%%% File    : glade_http.erl
%%% Author  : Tony Rogvall <tony@bit.hemma.se>
%%% Purpose : Basic http protocol routines
%%% Created : 13 Oct 2000 by Tony Rogvall <tony@bit.hemma.se>

-module(glade_http).
-author('tony@bit.hemma.se').

-export([recv_response/2, format_response/1,
	 recv_request/2, format_request/1,
	 recv_body/6,
	 parse_date/1, 
	 format_date/1, 
	 format_local_date/1,
	 format_current_date/0]).

-include("glade_http.hrl").
-include("glade_http_header.hrl").

recv_request(Socket, Timeout) ->
    case gen_tcp:recv(Socket, 0, Timeout) of
	{ok, {http_request, Method, Uri, Version}} ->
	    HRes = case Version of
		       {0,9} ->
			   {ok, glade_http_header:empty()};
		       _ ->
			   glade_http_header:parse(Socket, Timeout)
		   end,
	     case HRes of
		{ok,Headers} ->
		     {ok, #http_request { method = Method,
					  uri = Uri,
					  version = Version,
					  headers = Headers }};
		 Error ->
		     Error
	     end;
	Error ->
	    Error
    end.
    
format_request(Req) ->
    case Req#http_request.version of
	{0,9} ->
	    [fmt_method(Req#http_request.method)," ", 
	     glade_url:format(Req#http_request.uri),
	     "\r\n"];
	{Maj,Min} ->
	    [
	     fmt_method(Req#http_request.method)," ", 
	     glade_url:format(Req#http_request.uri)," ",
	     "HTTP/",
	     integer_to_list(Maj),$.,
	     integer_to_list(Min),"\r\n",
	     glade_http_header:format(Req#http_request.headers),
	     "\r\n"
	    ]
    end.


recv_response(Socket, Timeout) ->
    case gen_tcp:recv(Socket, 0, Timeout) of
	{ok, {http_response, Version, Status, Message}} ->
	    case glade_http_header:parse(Socket, Timeout) of
		{ok, Headers} ->
		    {ok, #http_response { version = Version,
					  status = Status,
					  message = Message,
					  headers = Headers }};
		Error ->
		    Error
	    end;
	Error ->
	    Error
    end.

format_response(Resp) ->
    case Resp#http_response.version of
	{0,9} ->  [];
	{Maj,Min} ->
	    [
	     "HTTP/",integer_to_list(Maj),$.,integer_to_list(Min)," ",
	     integer_to_list(Resp#http_response.status),$\s,
	     Resp#http_response.message, "\r\n",
	     glade_http_header:format(Resp#http_response.headers),
	     "\r\n"
	    ]
    end.

fmt_method(M) when atom(M) ->
    atom_to_list(M);
fmt_method(M) when list(M) ->
    M.

%% accumulator argument.
recv_body(Socket, Hs, BufSize, Timeout, CallBack, State) ->
    case glade_http_header:lookup(?HTTP_HDR_TRANSFER_ENCODING, Hs) of
	{_,_, Type} when Type =/= "identity"->
	    recv_chunked(Socket, BufSize, Timeout, CallBack, State);
	_ ->
	    case glade_http_header:lookup(?HTTP_HDR_CONTENT_LENGTH, Hs) of
		{_, _, Length} ->
		    recv_length(Socket, BufSize, Timeout, 
				CallBack, State, list_to_integer(Length));
		false ->
		    recv_closed(Socket, BufSize, Timeout, CallBack, State)
	    end
    end.    

recv_chunked(Socket, BufSize, Timeout, CallBack, State) ->
    inet:setopts(Socket, [{packet, line}, {mode,list}]),
    recv_chunks(Socket, BufSize, Timeout, CallBack, State).
    
recv_chunks(Socket, BufSize, Timeout, CallBack, State) ->
    case recv_chunk_length(Socket, Timeout) of
	{ok, 0} ->
	    case recv_chunk_trailer(Socket, Timeout) of
		{ok, Hs} ->
		    CallBack(eof, State);
		Error -> Error
	    end;
	{ok, Length} ->
	    inet:setopts(Socket, [{packet,0}, {mode, binary}]),
	    case recv_len(Socket, BufSize, Timeout, CallBack, State, Length) of
		{ok,State1} ->
		    inet:setopts(Socket, [{packet,line}, {mode, list}]),
		    case recv_crlf(Socket, Timeout) of
			ok ->
			    recv_chunks(Socket,BufSize,Timeout,CallBack,State1);
			Error ->
			    Error
		    end;
		Error ->
		    Error
	    end;
	Error ->
	    Error
    end.

recv_chunk_trailer(Socket, Timeout) ->
    inet:setopts(Socket, [{packet, http}]),
    %% force driver into header state
    gen_tcp:unrecv(Socket, "HTTP/1.1 200 OK\r\n"), 
    case gen_tcp:recv(Socket, 0, Timeout) of
	{ok, {http_response, _, _, _}} ->
	    glade_http_header:parse(Socket, Timeout);
	Error ->
	    Error
    end.

recv_chunk_length(Socket, Timeout) ->
    case gen_tcp:recv(Socket, 0, Timeout) of
	{ok, L} ->
	    {ok, hex_to_integer(L)};
	Error ->
	    Error
    end.


recv_crlf(Socket, Timeout) ->
    case gen_tcp:recv(Socket, 0, Timeout) of
	{ok, "\r\n"} -> ok;
	{ok, "\n"} -> ok;
	{ok, _} -> {error, bad_chunk};
	Error -> Error
    end.
	     

recv_closed(Socket, BufSize, Timeout, CallBack, State) ->
    inet:setopts(Socket, [{packet,0}, {mode, binary}]),
    recv_closed1(Socket, BufSize, Timeout, CallBack, State).


recv_closed1(Socket, BufSize, Timeout, CallBack, State) ->
    case gen_tcp:recv(Socket, 0, Timeout) of
	{ok, Data} ->
	    State1 = CallBack(Data, State),
	    recv_closed1(Socket, BufSize, Timeout, CallBack, State1);
	{error, closed} ->
	    CallBack(eof, State);
	Error ->
	    Error
    end.

recv_length(Socket, BufSize, Timeout, CallBack, State, 0) ->
    CallBack(eof, State);
recv_length(Socket, BufSize, Timeout, CallBack, State, Len) ->
    inet:setopts(Socket, [{packet,0}, {mode, binary}]),
    Res = case recv_len(Socket, BufSize, Timeout, CallBack, State, Len) of
	      {ok,State1} ->
		  CallBack(eof, State1);
	      Error ->
		  Error
	  end,
    inet:setopts(Socket, [{packet,http},{mode,list}]),
    Res.


recv_len(Socket, BufSize, Timeout, CallBack, State, 0) ->
    {ok,State};
recv_len(Socket, BufSize, Timeout, CallBack, State, Len) ->
    Sz = if Len >= BufSize ->
		 BufSize;
	    true ->
		 Len
	 end,
    case gen_tcp:recv(Socket, Sz, Timeout) of
	{ok, Data} ->
	    State1 = CallBack(Data, State),
	    recv_len(Socket, BufSize, Timeout, CallBack, State1,
		     Len - size(Data));
	Error ->
	    Error
    end.

hex_to_integer(Cs) ->
    hex_to_int(Cs, 0).

hex_to_int([C | Cs], Value) when C >= $A, C =< $F ->
    hex_to_int(Cs, Value * 16 + (C - $A + 10));
hex_to_int([C | Cs], Value) when C >= $0, C =< $9 ->
    hex_to_int(Cs, Value * 16 + (C - $0));
hex_to_int([C | Cs], Value) when C >= $a, C =< $f ->
    hex_to_int(Cs, Value * 16 + (C - $a + 10));
hex_to_int(_, Value) ->
    Value.

%%%
%% format_date_entry(..) - same date format as Apaches 
%% common_log_format uses.
%%
%%           date = [day/month/year:hour:minute:second zone] 
%%           day = 2*digit
%%           month = 3*letter
%%           year = 4*digit
%%           hour = 2*digit
%%           minute = 2*digit
%%           second = 2*digit
%%           zone = (`+' | `-') 4*digit
%%%

format_date_entry(Now) ->
    universal_time_to_string(calendar:now_to_universal_time(Now)).
    
%% Below code from misc.erl + modified.
universal_time_to_string(UTC) ->
    dat2str(
      local_time_to_date_and_time(
	calendar:universal_time_to_local_time(UTC))).
%% "Fri 1999-09-24,10:59:13+0200"
%% [15/Jun/1999:21:33:58 +0200]
dat2str([Y1,Y2, Mo, D, H, M, S | Diff]) ->
    binary_to_list(
      list_to_binary(
	io_lib:format("[~2.2.0w/~s/~w:~2.2.0w:~2.2.0w:~2.2.0w",
		      [D,int_to_month(Mo),y(Y1,Y2),H,M,S]) ++
	case Diff of
	    [Sign,Hd,Md] ->
		io_lib:format(" ~c~2.2.0w~2.2.0w]",
			      [Sign,Hd,Md]);
	    _ -> [$]]
	end)).

int_to_month(N) when N >= 1, N =< 12 ->
    enc_month(N);
int_to_month(_) ->
    "Err".

y1(Y) -> (Y bsr 8) band 255.
y2(Y) -> Y band 255.

y(Y1, Y2) -> 256 * Y1 + Y2.
    
diff(Secs) ->
    case calendar:seconds_to_daystime(Secs) of
	{0, {H, M,_}} ->
	    [$+, H, M];
	{-1, _} ->
	    {0, {H, M, _}} = calendar:seconds_to_daystime(-Secs),
	    [$-, H, M]
    end.

local_time_to_date_and_time(Local) ->
    UTC = calendar:local_time_to_universal_time(Local),
    date_and_time(Local, UTC).

date_and_time(Local, UTC) ->
    DiffSecs = calendar:datetime_to_gregorian_seconds(Local) -
	       calendar:datetime_to_gregorian_seconds(UTC),
    short_time(Local) ++ diff(DiffSecs).

short_time({{Y,M,D},{H,Mi,S}}) ->
    [y1(Y), y2(Y), M, D, H, Mi, S].

%%%
%% Misc utils.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% <tony@bluetail.com> wrote the decode/encode below.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% decode http-date (RFC 2068). (MUST be send in RFC1123 date format)
%%          HTTP-date    = rfc1123-date | rfc850-date | asctime-date
%%          rfc1123-date = wkday "," SP date1 SP time SP "GMT"
%%          rfc850-date  = weekday "," SP date2 SP time SP "GMT"
%%          asctime-date = wkday SP date3 SP time SP 4DIGIT
%%
%%          date1        = 2DIGIT SP month SP 4DIGIT
%%                         ; day month year (e.g., 02 Jun 1982)
%%          date2        = 2DIGIT "-" month "-" 2DIGIT
%%                         ; day-month-year (e.g., 02-Jun-82)
%%          date3        = month SP ( 2DIGIT | ( SP 1DIGIT ))
%%                         ; month day (e.g., Jun  2)
%%
%%          time         = 2DIGIT ":" 2DIGIT ":" 2DIGIT
%%                         ; 00:00:00 - 23:59:59
%%
%%          wkday        = "Mon" | "Tue" | "Wed"
%%                       | "Thu" | "Fri" | "Sat" | "Sun"
%%
%%
%%          weekday      = "Monday" | "Tuesday" | "Wednesday"
%%                       | "Thursday" | "Friday" | "Saturday" | "Sunday"
%%
%%          month        = "Jan" | "Feb" | "Mar" | "Apr"
%%                       | "May" | "Jun" | "Jul" | "Aug"
%%                       | "Sep" | "Oct" | "Nov" | "Dec"
%%
%% decode date or crash!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

parse_date(Line) -> dec_http_date(tolower(Line)).

dec_http_date("monday " ++ Cs) -> dec_date2(Cs);
dec_http_date("tuesday " ++ Cs) -> dec_date2(Cs);
dec_http_date("wednesday " ++ Cs) -> dec_date2(Cs);
dec_http_date("thursday " ++ Cs) -> dec_date2(Cs);
dec_http_date("friday " ++ Cs) -> dec_date2(Cs);
dec_http_date("saturday " ++ Cs) -> dec_date2(Cs);
dec_http_date("sunday " ++ Cs) -> dec_date2(Cs);
dec_http_date("mon" ++ Cs) -> dec_date13(Cs);
dec_http_date("tue" ++ Cs) -> dec_date13(Cs);
dec_http_date("wed" ++ Cs) -> dec_date13(Cs);
dec_http_date("thu" ++ Cs) -> dec_date13(Cs);
dec_http_date("fri" ++ Cs) -> dec_date13(Cs);
dec_http_date("sat" ++ Cs) -> dec_date13(Cs);
dec_http_date("sun" ++ Cs) -> dec_date13(Cs).

dec_date13([$\s|Cs]) -> dec_date3(Cs);
dec_date13([$,,$\s |Cs]) -> dec_date1(Cs).

%% date1
dec_date1([D1,D2,$\s,M1,M2,M3,$\s,Y1,Y2,Y3,Y4,$\s | Cs]) ->
    M = dec_month([M1,M2,M3]),
    D = list_to_integer([D1,D2]),
    Y = list_to_integer([Y1,Y2,Y3,Y4]),
    {Time," gmt" ++ Cs1} = dec_time(Cs),
    { {{Y,M,D},Time}, Cs1}.

%% date2
dec_date2([D1,D2,$-,M1,M2,M3,$-,Y1,Y2 | Cs]) ->
    M = dec_month([M1,M2,M3]),
    D = list_to_integer([D1,D2]),
    Y = 1900 + list_to_integer([Y1,Y2]),
    {Time," gmt" ++ Cs1} = dec_time(Cs),
    {{{Y,M,D}, Time}, Cs1}.

%% date3
dec_date3([M1,M2,M3,$\s,D1,D2,$\s| Cs]) ->
    M = dec_month([M1,M2,M3]),
    D = if D1 == $\s -> list_to_integer([D2]);
	   true -> list_to_integer([D1,D2])
	end,
    {Time,[$\s,Y1,Y2,Y3,Y4|Cs1]} = dec_time(Cs),
    Y = list_to_integer([Y1,Y2,Y3,Y4]),
    { {{Y,M,D}, Time}, Cs1 }.

%% decode lowercase month
dec_month("jan") -> 1;
dec_month("feb") -> 2;
dec_month("mar") -> 3;
dec_month("apr") -> 4;
dec_month("may") -> 5;
dec_month("jun") -> 6;
dec_month("jul") -> 7;
dec_month("aug") -> 8;
dec_month("sep") -> 9;
dec_month("oct") -> 10;
dec_month("nov") -> 11;
dec_month("dec") -> 12.

%% decode time HH:MM:SS
dec_time([H1,H2,$:,M1,M2,$:,S1,S2|Cs]) ->
    { {list_to_integer([H1,H2]), 
       list_to_integer([M1,M2]),
       list_to_integer([S1,S2]) }, Cs}.

%% encode date into rfc1123-date (must be a GMT time!!!)
format_local_date(DateTime) ->
    format_date(calendar:local_time_to_universal_time(DateTime)).

format_date({{Y,M,D},{TH,TM,TS}}) ->
    WkDay = case calendar:day_of_the_week({Y,M,D}) of
		1 -> "Mon";
		2 -> "Tue";
		3 -> "Wed";
		4 -> "Thu";
		5 -> "Fri";
		6 -> "Sat";
		7 -> "Sun"
	    end,
    lists:flatten(io_lib:format("~s, ~2..0w ~s ~4..0w "
				"~2..0w:~2..0w:~2..0w GMT",
				[WkDay, D, enc_month(M), Y, TH, TM, TS])).

format_current_date() ->
    format_date(calendar:universal_time()).

%% encode lowercase month
enc_month(1) -> "Jan";
enc_month(2) -> "Feb";
enc_month(3) -> "Mar";
enc_month(4) -> "Apr";
enc_month(5) -> "May";
enc_month(6) -> "Jun";
enc_month(7) -> "Jul";
enc_month(8) -> "Aug";
enc_month(9) -> "Sep";
enc_month(10) -> "Oct";
enc_month(11) -> "Nov";
enc_month(12) -> "Dec".


tolower([C|Cs]) when C >= $A, C =< $Z -> [((C-$A)+$a) | tolower(Cs)];
tolower([C|Cs]) -> [C|tolower(Cs)];
tolower([]) -> [].



