%%% File    : glade_http_cli.erl
%%% Author  : Tony Rogvall <tony@bit.hemma.se>
%%% Purpose : Http client
%%% Created : 12 Oct 2000 by Tony Rogvall <tony@bit.hemma.se>

-module(glade_http_cli).
-author('tony@bit.hemma.se').

-export([open_url/1, open_url/2]).
-export([open/1, open/2, open/3, close/1]).
-export([load/3, load/5]).
-export([set_headers/2, add_header/3]).
-export([set_version/3]).
-export([set_timeout/2]).
-export([send_request/2]).
-export([recv_response/1, recv_response/2, recv_body/4]).

-export([get/2]).
-export([show_resp/1]).

-include("glade_http.hrl").
-include("glade_http_header.hrl").

-record(http_client,
	{
	  socket,
	  version = {1,1},    %% client version
	  base,               %% base url to use
	  timeout = 30000,    %% timeout value
	  headers             %% http_headers
	 }).

-define(RCVBUF_SIZE, 2048).
-define(SNDBUF_SIZE, 2048).

open_url(BaseUrl) ->
    open_url(BaseUrl, []).

open_url(BaseUrl, Headers) ->
    case glade_url:parse(BaseUrl) of
	Url = {absoluteURI,_,Host,Port,_} ->
	    case open(Host, Port, Headers) of
		{ok, Cli} ->
		    {ok, Cli#http_client { base = Url }};
		Error ->
		    Error
	    end;
	{error,Reasson} ->
	    {error,Reasson};
	Other ->
	    {error, not_absolute}
    end.

open(Host) ->
    open(Host, 80).

open(Host, Port) ->
    open(Host, Port, []).
	  
open(Host, Port, Headers) ->
    HttpPort = if Port == undefined -> 80;
		  true -> Port
	       end,
    case gen_tcp:connect(Host, HttpPort,
			 [{active,false},
			  {packet,http},
			  {mode,list},
			  {nodelay,true},
			  {buffer, ?RCVBUF_SIZE},
			  {sndbuf, ?SNDBUF_SIZE}]) of
	{ok,S} ->
	    H0 = glade_http_header:new(Headers),
	    H1 = glade_http_header:add_if_not_there(?HTTP_HDR_HOST, Host, H0),
	    Cli = #http_client { socket = S,
				 headers = H1 },
	    {ok, Cli};
	Error ->
	    Error
    end.


load(Host, Port, Path) ->
    load(Host,Port,Path,
	 fun(eof,State) ->
		 State;
	    (Data,State) ->
		 [Data|State]
	 end,
	 []).


load(Host, Port, Path, BodyFun, BodyState) ->
    case open(Host,Port) of
	{ok,Cli} ->
	    case get(Cli, Path) of
		ok ->
		    Res = case recv_response(Cli, 5000) of
			      {ok,R} when R#http_response.status == 200 ->
				  recv_body(Cli,R#http_response.headers,
					    BodyFun, BodyState);
			      {ok,R} ->
				  {error, {http,
					   R#http_response.status,
					   R#http_response.message}};
			      Error ->
				  Error
			  end,
		    close(Cli),
		    Res;
		Error ->
		    close(Cli),
		    Error
	    end;
	Error -> Error
    end.


close(Cli) ->
    gen_tcp:close(Cli#http_client.socket).

set_headers(Cli, Headers) ->
    Cli#http_client { 
      headers = glade_http_headers:new(Headers) }.

add_header(Cli, Key, Value) ->
    H0 = Cli#http_client.headers,
    Cli#http_client { 
      headers = glade_http_header:add_last(Key, Value, H0)
     }.

set_version(Cli, Major, Minor) ->
    Cli#http_client { version = {Major, Minor}}.

set_timeout(Cli, Timeout) ->
    Cli#http_client { timeout = Timeout }.

send_request(Cli, Req) ->
    Fmt = glade_http:format_request(Req),
    ?Debug("REQUEST:\n~s\n", [Fmt]),
    gen_tcp:send(Cli#http_client.socket, Fmt).

recv_response(Cli) ->
    recv_response(Cli, Cli#http_client.timeout).

recv_response(Cli, Timeout) ->
    case glade_http:recv_response(Cli#http_client.socket, Timeout) of
	{ok,Resp} ->
	    ?Debug("RESPONSE:\n~s\n", [glade_http:format_response(Resp)]),
	    {ok,Resp};
	Error ->
	    ?Debug("RESPONSE:\n~p\n", [Error]),
	    Error
    end.


get(Cli, Object) ->
    Url = case Object of 
	      [$/ | _] ->  {abs_path, Object};
	      _ ->
		  case Cli#http_client.base of
		      undefined ->
			  {abs_path, filename:join("/", Object)};
		      {absoluteURI,Scheme,Host,Port,Path} ->
			  {abs_path, filename:join(Path, Object)};
		      {abs_path,Path} ->
			  {abs_path, filename:join(Path, Object)};
		      _ ->
			  {error, bad_url}
		  end
	  end,
    Req = #http_request { method = 'GET',
			  uri = Url,
			  version = Cli#http_client.version,
			  headers = Cli#http_client.headers 
			 },
    send_request(Cli, Req).

recv_body(Cli, Hs, CallBack, State) ->
    glade_http:recv_body(Cli#http_client.socket, Hs, ?RCVBUF_SIZE,
		   Cli#http_client.timeout, CallBack, State).

%%
%% Utility
%%

show_resp(Cli) ->
    case recv_response(Cli) of
	{ok, Resp} ->
	    recv_body(Cli, Resp#http_response.headers, 
		      fun(eof,_) ->
			      io:format("END\n");
			 (Data,_) ->
			      io:format("~s", [binary_to_list(Data)])
		      end,
		      []);
	Error ->
	    Error
    end.

