%%% File    : glade_http_header.erl
%%% Author  : Tony Rogvall <tony@bit.hemma.se>
%%% Purpose : HTTP Header field managment
%%% Created : 23 Oct 2000 by Tony Rogvall <tony@bit.hemma.se>

-module(glade_http_header).
-author('tony@bit.hemma.se').

-export([parse/2, format/1]).

-export([empty/0, new/1,
	 add_last/3, add_update/3, add_if_not_there/3,
	 change_if_there/3, add_replace/3,
	 lookup/2,
	 member/2,
	 member_any/2,
	 member_all/2,
	 delete/2, delete_all/2]).

-import(lists, [reverse/1]).

-include("glade_http_header.hrl").


-record(http_fields,
	{
	  bits = 0,
	  fields = []
	 }).

key_code(Key) when Key > 0, Key =< ?HTTP_BITS ->
    ?BIT_MASK(Key);
key_code(Key) when atom(Key) ->
    case Key of
	%% General-header
	'Cache-Control' ->
	    ?BIT_MASK(?HTTP_BIT_CACHE_CONTROL);
	'Connection' ->
	    ?BIT_MASK(?HTTP_BIT_CONNECTION);
	'Date' ->
	    ?BIT_MASK(?HTTP_BIT_DATE);
	'Pragma' ->
	    ?BIT_MASK(?HTTP_BIT_PRAGMA);
	'Transfer-Encoding' ->
	    ?BIT_MASK(?HTTP_BIT_TRANSFER_ENCODING);
	'Upgrade' ->
	    ?BIT_MASK(?HTTP_BIT_UPGRADE);
	'Via' -> 
	    ?BIT_MASK(?HTTP_BIT_VIA);

	%% Request header fields
	'Accept' ->
	    ?BIT_MASK(?HTTP_BIT_ACCEPT);
	'Accept-Charset' ->
	    ?BIT_MASK(?HTTP_BIT_ACCEPT_CHARSET);
	'Accept-Encoding' ->
	    ?BIT_MASK(?HTTP_BIT_ACCEPT_ENCODING);
	'Accept-Language' ->
	    ?BIT_MASK(?HTTP_BIT_ACCEPT_LANGUAGE);
	'Authorization' ->
	    ?BIT_MASK(?HTTP_BIT_AUTHORIZATION);
	'From' ->
	    ?BIT_MASK(?HTTP_BIT_FROM);
	'Host' ->
	    ?BIT_MASK(?HTTP_BIT_HOST);
	'If-Modified-Since' ->
	    ?BIT_MASK(?HTTP_BIT_IF_MODIFIED_SINCE);
	'If-Match' ->
	    ?BIT_MASK(?HTTP_BIT_IF_MATCH);
	'If-None-Match' ->
	    ?BIT_MASK(?HTTP_BIT_IF_NONE_MATCH);
	'If-Range' ->
	    ?BIT_MASK(?HTTP_BIT_IF_RANGE);
	'If-Unmodified-Since' ->
	    ?BIT_MASK(?HTTP_BIT_IF_UNMODIFIED_SINCE);
	'Max-Forwards' ->
	    ?BIT_MASK(?HTTP_BIT_MAX_FORWARDS);
	'Proxy-Authorization' ->
	    ?BIT_MASK(?HTTP_BIT_PROXY_AUTHORIZATION);
	'Range' ->
	    ?BIT_MASK(?HTTP_BIT_RANGE);
	'Referer' ->
	    ?BIT_MASK(?HTTP_BIT_REFERER);
	'User-Agent' ->
	    ?BIT_MASK(?HTTP_BIT_USER_AGENT);
	
	%% Response header fields
	'Age' ->
	    ?BIT_MASK(?HTTP_BIT_AGE);
	'Location' ->
	    ?BIT_MASK(?HTTP_BIT_LOCATION);
	'Proxy-Authenticate' ->
	    ?BIT_MASK(?HTTP_BIT_PROXY_AUTHENTICATE);
	'Public' ->
	    ?BIT_MASK(?HTTP_BIT_PUBLIC);
	'Retry-After' ->
	    ?BIT_MASK(?HTTP_BIT_RETRY_AFTER);
	'Server' ->
	    ?BIT_MASK(?HTTP_BIT_SERVER);
	'Vary' ->
	    ?BIT_MASK(?HTTP_BIT_VARY);
	'Warning' ->
	    ?BIT_MASK(?HTTP_BIT_WARNING);
	'Www-Authenticate' ->
	    ?BIT_MASK(?HTTP_BIT_WWW_AUTHENTICATE);

	%% Entity header fields
	'Allow' ->
	    ?BIT_MASK(?HTTP_BIT_ALLOW);
        'Content-Base' ->
	    ?BIT_MASK(?HTTP_BIT_CONTENT_BASE);
        'Content-Encoding' ->
	    ?BIT_MASK(?HTTP_BIT_CONTENT_ENCODING);
        'Content-Language' ->
	    ?BIT_MASK(?HTTP_BIT_CONTENT_LANGUAGE);
        'Content-Length' ->
	    ?BIT_MASK(?HTTP_BIT_CONTENT_LENGTH);
	'Content-Location' ->
	    ?BIT_MASK(?HTTP_BIT_CONTENT_LOCATION);
        'Content-Md5' ->
	    ?BIT_MASK(?HTTP_BIT_CONTENT_MD5);
        'Content-Range' ->
	    ?BIT_MASK(?HTTP_BIT_CONTENT_RANGE);
        'Content-Type' ->
	    ?BIT_MASK(?HTTP_BIT_CONTENT_TYPE);
        'ETag' ->
	    ?BIT_MASK(?HTTP_BIT_ETAG);
        'Expires' ->
	    ?BIT_MASK(?HTTP_BIT_EXPIRES);
        'Last-Modified' ->
	    ?BIT_MASK(?HTTP_BIT_LAST_MODIFIED);

	%% Other optional (recognised) header fields
	'Accept-Ranges' ->
	    ?BIT_MASK(?HTTP_BIT_ACCEPT_RANGES);
	'Set-Cookie' ->
	    ?BIT_MASK(?HTTP_BIT_SET_COOKIE);
	'Set-Cookie2' ->
	    ?BIT_MASK(?HTTP_BIT_SET_COOKIE2);
	'X-Forwarded-For' ->
	    ?BIT_MASK(?HTTP_BIT_X_FORWARDED_FOR);
	'Cookie' ->
	    ?BIT_MASK(?HTTP_BIT_COOKIE);
	'Keep-Alive' ->
	    ?BIT_MASK(?HTTP_BIT_KEEP_ALIVE);
	'Proxy-Connection' ->
	    ?BIT_MASK(?HTTP_BIT_PROXY_CONNECTION);
	_ ->
	    0
    end;
key_code(Key) ->
    0.


new(List) ->
    new1(List, [], 0).

new1([{Key,Val}|List], Acc, Bits) ->
    Bit = key_code(Key),
    KV = {Key,undefined,Val},
    new1(List, [KV|Acc], Bits bor Bit);
new1([], Acc, Bits) ->
    #http_fields { bits = Bits, fields = Acc}.
	
empty() ->
    new([]).

%% unconditionally add last
add_last(Key,Value, F) ->
    Bit = key_code(Key),
    KV  = {Key,undefined,Value},
    F#http_fields { bits = Bit bor F#http_fields.bits,
		      fields = [KV | F#http_fields.fields]}.

%% add last or update value if found
add_update(Key, Value, F) ->
    Bit = key_code(Key),
    case Bit band F#http_fields.bits of
	0 when (Bit > 0) ->
	    KV  = {Key,undefined,Value},
	    F#http_fields { bits = Bit bor F#http_fields.bits,
			    fields = [KV | F#http_fields.fields]};
	_ ->
	    add_update(Key,Value,Bit,F#http_fields.fields,[],F)
    end.

add_update(Key, Value, _, [{Key,_,Val}|Ks], Acc, F) ->
    KV = {Key,undefined, [Val,$\r,$\n,$\s,$, ,Value]},
    F#http_fields { fields = reverse(Acc) ++ [KV | Ks] };
add_update(Key, Value, Bit, [KV|Ks], Acc, F) ->
    add_update(Key,Value,Bit,Ks,[KV|Acc],F);
add_update(Key, Value, Bit, [], Acc, F) ->
    KV  = {Key,undefined,Value},
    F#http_fields { bits = Bit bor F#http_fields.bits,
		    fields = [KV | F#http_fields.fields]}.


%% add last or replace value if found
add_replace(Key, Value, F) ->
    Bit = key_code(Key),
    case Bit band F#http_fields.bits of
	0 when (Bit > 0) ->
	    KV  = {Key,undefined,Value},
	    F#http_fields { bits = Bit bor F#http_fields.bits,
			    fields = [KV | F#http_fields.fields]};
	_ ->
	    add_replace(Key,Value,Bit,F#http_fields.fields,[],F)
    end.

add_replace(Key, Value, _, [{Key,_,Val}|Ks], Acc, F) ->
    KV = {Key,undefined,Value},
    F#http_fields { fields = reverse(Acc) ++ [KV | Ks] };
add_replace(Key, Value, Bit, [KV|Ks], Acc, F) ->
    add_replace(Key,Value,Bit,Ks,[KV|Acc],F);
add_replace(Key, Value, Bit, [], Acc, F) ->
    KV  = {Key,undefined,Value},
    F#http_fields { bits = Bit bor F#http_fields.bits,
		    fields = [KV | F#http_fields.fields]}.

%%
%% Add last if header field is NOT present
%%
add_if_not_there(Key, Value, F) ->
    Bit = key_code(Key),
    case Bit band F#http_fields.bits of
	A when (A > 0) -> F;  %% its there already
	_ -> 
	    add_if_not_x(Key, Value, Bit, F#http_fields.fields, F)
    end.

add_if_not_x(Key, Value, _, [{Key,_,Val}|Ks], F) ->
    %% present return original
    F;
add_if_not_x(Key, Value, Bit, [KV|Ks], F) ->
    add_if_not_x(Key,Value,Bit, Ks, F);
add_if_not_x(Key, Value, Bit, [], F) ->
    KV  = {Key,undefined,Value},
    F#http_fields { bits = Bit bor F#http_fields.bits,
		    fields = [KV | F#http_fields.fields]}.

%% 
%% Replace header field value if present
%%

change_if_there(Key, Value, F) when atom(Key) ->
    Bit = key_code(Key),
    case Bit band F#http_fields.bits of
	0 when (Bit > 0) -> F;
	_ -> mod_if_x(Key, Value, F#http_fields.fields, [], F)
    end.

mod_if_x(Key, Value, [{Key,_,Val}|Ks], Acc, F) ->
    KV = {Key,undefined,Value},
    F#http_fields { fields = reverse(Acc) ++ [KV | Ks] };
mod_if_x(Key, Value, [KV|Ks], Acc, F) ->
    mod_if_x(Key,Value,Ks, [KV|Acc], F);
mod_if_x(Key, Value, [], Acc, F) ->
    F.

%%
%% Lookup a header field value
%%
lookup(Key, F) ->
    Bit = key_code(Key),
    case Bit band F#http_fields.bits of
	0 when (Bit > 0) -> false;
	_ ->
	    case lists:keysearch(Key, 1, F#http_fields.fields) of
		{value, KV} ->
		    KV;
		false -> false
	    end
    end.

%% Check if header is present
member(Key, F) ->
    Bit = key_code(Key),
    case Bit band F#http_fields.bits of
	0 when (Bit > 0) -> false;
	0 ->
	    case lists:keysearch(Key, 1, F#http_fields.fields) of
		{value, _} -> true;
		false -> false
	    end;
	_ -> true
    end.    

member_any([Key|Ks], F) ->
    case member(Key, F) of
	true -> true;
	false -> member_any(Ks, F)
    end;
member_any([], F) -> false.


member_all([Key|Ks], F) ->
    case member(Key, F) of
	true -> member_all(Ks, F);
	false -> false
    end;
member_all([], F) -> true.

%%
%% Delete ALL instances of a field 
%%
delete(Key, F) ->
    Bit =  key_code(Key),
    Bits = F#http_fields.bits,
    case Bit band Bits of
	0 when (Bit > 0) -> F;
	_ ->
	    Fs = delete1(Key, F#http_fields.fields),
	    F#http_fields { bits = Bits band (bnot Bit),
			    fields = Fs }
    end.


delete1(Key, [{Key,_,_} | Hs]) ->
    delete1(Key, Hs);
delete1(Key, [KV | Hs]) ->
    [KV|delete1(Key, Hs)];
delete1(Key, []) ->
    [].

%%
%% Delete multiples fields
%%
delete_all(Keys, F) ->
    delete_all(Keys,F#http_fields.bits, F#http_fields.fields,F).

delete_all([Key|Ks], Bits, Hs,F) ->
    Bit = key_code(Key),
    case Bit band Bits of
	0 when (Bit > 0) ->
	    delete_all(Ks, Bits, Hs, F);
	_ ->
	    delete_all(Ks, Bits band (bnot Bit), delete1(Key, Hs), F)
    end;
delete_all([], Bits, Hs, F) ->
    F#http_fields {  bits = Bits, fields = Hs }.
    
%%%
%% Handling of header fields.
%% --------------------------
%% Only some headers are interesting, and only those are actually parsed.
%% Each line is read and saved in a list.
%% If the line was interesting (say: it was a Via: line) it is
%% referenced. If many e.g. Via: lines are found, only the latest 
%% is referenced (XXX FIXME).
%%
%% return value is:
%%   [ Field ]
%% Field:     {atom(), string()}   [atom_to_list(atom),$:,\s,string(),\r\n]
%%            {string(),string()}  [string(),$:,\s,string(),\r\n]
%%            
%%%
parse(S, Read_timeout) ->
    %% FIXME: giv max number of header lines
    header_start(S, Read_timeout).

%%
%% Parse:   
%%   field = <key> ':' <value> ((<sp>|<tab>)<value><crlf>)* <crlf>
%%
%%   fields = <crlf> |
%%            field | fields
%%
%%   Parse at most MAX_HEADERS then error 
%%
%%   The headers are kept in reverse order for simple append oerations
%%
header_start(S, Timeout) ->
    case gen_tcp:recv(S, 0, Timeout) of
	{ok,{http_header,Bit,Key,IValue,Value}} ->
	    header_continue(S, Timeout, ?BIT_MASK(Bit), [{Key,IValue,Value}]);
	{ok, http_eoh} ->
	    {ok,empty()};
	{ok,_} ->
	    {error, 400};
	{error,{http_error,Line}} ->
	    {error, 400};
	E ->
	    {error, 502}
    end.


header_continue(S, Timeout, Bits, Acc) ->
    case gen_tcp:recv(S, 0, Timeout) of
	{ok,{http_header,Bit,Key,IValue,Value}} ->
	    header_continue(S, Timeout, ?BIT_MASK(Bit) bor Bits,
			    [{Key,IValue,Value}|Acc]);
	{ok, http_eoh} ->
	    {ok,#http_fields { bits = Bits, fields = Acc }};
	{ok,_} ->
	    {error, 400};
	{error,{http_error,Line}} ->
	    {error, 400};
	E ->
	    {error, 502}
    end.

%% Format and reverse the headers !!!
format(Fld) ->
    format(Fld#http_fields.fields, []).

format([{Key,_,Value}|Hs], Acc) ->
    if atom(Key) ->
	    format(Hs,
		   [atom_to_list(Key),$:,$\s,
		    Value,$\r,$\n|Acc]);
       true ->
	    format(Hs,
		   [Key,$:,$\s,Value,$\r,$\n | Acc])
    end;
format([], Acc) -> Acc.



