%%% File    : glade_parse.erl
%%% Author  : Tony Rogvall <tony@bit.hemma.se>
%%% Description : glade xml parser
%%% Created :  3 May 2001 by Tony Rogvall <tony@bit.hemma.se>

-module(glade_parse).

-include("glade_http.hrl").

-compile(export_all).
-import(lists, [all/2, reverse/1]).

-define(in(X,Low,Hi), Low =< X, X =< Hi).
-define(white(X), X==$ ; X==$\n ; X == $\t; X == $\r ).

-ifdef(debug).
debug(Fmt, As) -> io:format(Fmt, As).
-else.
debug(Fmt, As) -> ok.
-endif.


url(URL) ->
    case glade_url:parse(URL) of
	{absoluteURI,https,Host,Port,Path} ->
	    {error, not_supported};
	{scheme,"file",File} ->
	    file(File);
	{abs_path, Path} ->
	    file(Path);
	{rel_path, Path} ->
	    file(Path);
	{absoluteURI,http,Host,Port,Path} ->
	    glade_http_cli:load(Host,Port,Path,
				fun body/2, start_cont());
	_ ->
	    {error, not_supported}
    end.



file(F) ->
    case file:read_file(F) of
        {ok, Bin} ->
            parse_start(binary_to_list(Bin));
        Error ->
            Error
    end.


body(eof,State) -> 
    State;
body(Data0, State0) ->
    Data = if binary(Data0) -> 
		   binary_to_list(Data0);
	      true -> Data0
	   end,
    case State0 of
	{error,Reason} ->
	    {error,Reason};
	{ok,Form,Rest} ->
	    {ok,Form,Rest++Data};
	_ ->
	    case parse_more(Data, State0) of
		{more,State1} ->
		    State1;
		Other -> Other
	    end
    end.


string(Data) when binary(Data) ->
    parse_start(binary_to_list(Data));
string(Data) ->
    parse_start(Data).

%% This is a re-parser for XML streams

%% +deftype parse_result() =  {ok, FinalForm, RestString : string()}
%%                         |  {more, cont()}
%%                         |  {error, parse_error()}
%% +type parse_start(Str)             -> parse_result().
%% +type parse_more(Str, Cont)        -> parse_result().
%% +type format_error(parser_error()) -> string().
%% +type start_cont()                 -> cont().

parse_start(Str) ->
    parse_more(Str, start_cont()).

start_cont() ->
    {parse_top(), char_top()}.

parse_top() -> [].

char_top() ->  fun top/1.

parse_more(Str, {Stack, TokParser}) ->
    %% io:format("PARSE: State =~p ~p~n",[State, Str]),
    %% Call the token parser
    Res = TokParser(Str),
    %% io:format("Res=~p~n",[Res]),
    case Res of
	{item, Item, RestStr} ->
	    %% We've got an Item
	    %% RestStr are the characters left over at 
	    %% The end of the parse
	    %% io:format("Calling parser: ~p ~p ~n",[State, Item]),
	    parse_item(Stack, Item, RestStr);
	{error, X, Char} ->
	    {error, {expect, X, Char}};
	{more, F} ->
	    {more, {Stack, F}}
    end.

parse_item(Stack, {sTag, {Tag, Args}}, RestStr) ->
    %% Push new frame onto the stack the stack
    more([{Tag, Args, []}|Stack], RestStr);

parse_item(Stack, I = {pTag, {Tag, Args}}, RestStr) ->
    debug("ignoring processing tag ~p\n",[I]),
    more(Stack, RestStr);
parse_item([], P = {pcdata,_}, RestStr) ->
    debug("dropping top level pcdata~p\n",[P]),
    more([], RestStr);
parse_item([{Tag,Args,C}|L], P={pcdata, S}, RestStr) ->
    more([{Tag,Args,[P|C]}|L], RestStr);

parse_item([{Tag, Args, C}|L], {eTag, Tag}, RestStr) ->
    case deblank(reverse(C)) of
	{error, Str} ->
	    {error, {nonblank_item_found_in_tag, Tag, Str}};
	Other ->
	    pfinish([{Tag,Args,Other}|L], RestStr)
    end;

parse_item([S,{STag, Args, C}|L], {eTag, Tag}, RestStr) ->
    {error,{badendtagfound,Tag,starttagis,STag}};

parse_item(S, I, R) ->
    io:format("UUgh:Stack=~p Item=~p Rest=~p~n",[S, I, R]).

more(Stack, [])  -> {more, {Stack, char_top()}};
more(Stack, Str) -> parse_more(Str, {Stack, char_top()}).

deblank(S=[{pcdata, C}]) -> S;
deblank(X) -> deblank1(X).

deblank1([{pcdata,X}|T]) ->
    case all_blanks(X) of
	true ->
	    deblank1(T);
	false ->
	    {error, X}
    end;
deblank1([H|T]) ->
    [H|deblank1(T)];
deblank1([]) ->
    [].

pfinish([X], Rest)          -> 
    {ok, X, Rest};
pfinish([H1,{Tag,Args,L}|T], RestStr) -> 
    more([{Tag,Args,[H1|L]}|T], RestStr).

all_blanks(L) -> all(fun is_Blank/1, L).

top([$<|T]) -> get_tag(T);   
top([H|T])  -> get_pcdata(T, [H]);
top([])     -> {more, fun top/1}.

get_pcdata(S=[$<|T], L) -> {item, {pcdata, reverse(L)}, S};
get_pcdata(S=[$>|T], L) -> {item, {pcdata, reverse(L)}, S};
get_pcdata([H|T], L)    -> get_pcdata(T, [H|L]);
get_pcdata([], L)       -> {more, fun(Str) -> get_pcdata(Str, L) end}.

get_tag([H|T])  when ?white(H) -> get_tag(T);
get_tag([$/|T])                -> get_etag(T);
get_tag([$?|T])                -> get_ptag(T);
get_tag([H|T]) ->
    case is_tag_start(H) of
	true  -> get_tag(T, [H]);
	false -> error("Tag character", H)
    end;
get_tag([]) -> {more, fun get_tag/1}.

get_etag([H|T]) when ?white(H) -> get_etag(T);
get_etag([H|T]) ->
    case is_tag_start(H) of
	true  -> get_etag(T, [H]);
	false -> error("Tag character", H)
    end;
get_etag([]) -> {more, fun get_etag/1}.

get_ptag([H|T]) when ?white(H) -> get_ptag(T);
get_ptag([H|T]) ->
    case is_tag_start(H) of
	true  -> get_ptag(T, [H]);
	false -> error("Tag character", H)
    end;
get_ptag([]) -> {more, fun get_ptag/1}.



get_etag([$>|T], L) ->
    {item, {eTag, reverse(L)}, T};
get_etag([H|T], L) ->
    case is_NameChar(H) of
	true  -> get_etag(T, [H|L]);
	false -> error("character or >", H)
    end;
get_etag([], L) ->
    {more, fun(Str) -> get_etag(Str, L) end}.

get_tag(S=[H|T], L) ->
    case is_NameChar(H) of
	true  -> get_tag(T, [H|L]);
	false -> get_tag_args(S, reverse(L), [])
    end;
get_tag([], L) ->
    {more, fun(Str) -> get_tag(Str, L) end}.

get_ptag(S=[H|T], L) ->
    case is_NameChar(H) of
	true  -> get_ptag(T, [H|L]);
	false -> get_ptag_args(S, reverse(L), [])
    end;
get_ptag([], L) ->
    {more, fun(Str) -> get_ptag(Str, L) end}.


get_ptag_args([$?,$>|T], Tag, L)               ->
    {item, {pTag, {Tag, reverse(L)}}, T};
get_ptag_args([$?], Tag, L) -> 
    {more, fun(Str) -> get_ptag_args([$?|Str], Tag, L) end}; 
get_ptag_args([H|T], Tag, L) when ?white(H) -> 
    get_ptag_args(T, Tag, L);
get_ptag_args([H|T], Tag, L) ->
    case is_tag_start(H) of
	true  -> get_arg_name(T, L, [H], 
			      fun(T1,L1) -> 
				      get_ptag_args(T1,Tag,L1) 
			      end);
	false -> error("argument name or ?>", [H])
    end;
get_ptag_args([], Tag, L) ->
    {more, fun(Str) -> get_ptag_args(Str, Tag, L) end}.


get_tag_args([$>|T], Tag, L)               ->
    {item, {sTag, {Tag, reverse(L)}}, T};
get_tag_args([H|T], Tag, L) when ?white(H) -> 
    get_tag_args(T, Tag, L);
get_tag_args([H|T], Tag, L) ->
    case is_tag_start(H) of
	true  -> get_arg_name(T, L, [H],
			      fun(T1,L1) -> 
				      get_tag_args(T1,Tag,L1) 
			      end);
	false -> error("argument name or >", [H])
    end;
get_tag_args([], Tag, L) ->
    {more, fun(Str) -> get_tag_args(Str, Tag, L) end}.


get_arg_name(S=[H|T], L, Name, Cont) ->
    case is_NameChar(H) of
	true  -> get_arg_name(T, L, [H|Name], Cont);
	false -> 'get_arg_='(S, L, reverse(Name), Cont)
    end;
get_arg_name([], L, Name, Cont) ->
    {more, fun(Str) -> get_arg_name(Str, L, Name, Cont) end}.

'get_arg_='([$=|T], L, Name, Cont) ->
    get_quoted_value(T, Name, L, Cont);
'get_arg_='([H|T],  L, Name, Cont) when ?white(H) ->
    'get_arg_='(T, L, Name, Cont);
'get_arg_='([H|_], L, Name,Cont) ->
    error("white space or =", H);
'get_arg_='([], L, Name,Cont) ->
    {more, fun(Str) -> 'get_arg_='(Str, L, Name,Cont) end}.

get_quoted_value([$"|T], Name, L, Cont) ->
    get_val(T, L, Name, $", [], Cont);
get_quoted_value([$'|T], Name, L, Cont) ->
    get_val(T, L, Name, $', [], Cont);
get_quoted_value([H|T], Name, L, Cont) when ?white(H) ->
    get_quoted_value(T, Name, L, Cont);
get_quoted_value([H|_], Name, L, Cont) ->
    error("white space or \" or '", H);
get_quoted_value([], Name, L, Cont) ->
    {more, fun(Str) -> get_quoted_value(Str, Name, L, Cont) end}.

get_val([Stop|T], L, Name, Stop, Acc, Cont) ->
    L1 = [{Name, reverse(Acc)}|L],
    Cont(T, L1);
get_val([H|T], L, Name, Stop, Acc, Cont) ->
    get_val(T, L, Name, Stop, [H|Acc], Cont);
get_val([], L, Name, Stop, Acc, Cont) ->
    {more, fun(Str) -> get_val(Str, L, Name, Stop, Acc, Cont) end}.

error(Str, C) ->
    {error, Str, C}.

is_tag_start($_) -> true;
is_tag_start($:) -> true;
is_tag_start(X) -> is_Letter(X).
    
is_Letter(X) when ?in(X, $a, $z) -> true;
is_Letter(X) when ?in(X, $A, $Z) -> true;
is_Letter(_)                     -> false.

is_Digit(X)  when ?in(X, $0, $9) -> true;
is_Digit(_)                      -> false.
               
is_Name($:) -> true;
is_Name($_) -> true;
is_Name(H)  -> is_Letter(H).
 
is_MiscName($.) -> true;
is_MiscName($-) -> true;
is_MiscName($_) -> true;
is_MiscName($:) -> true;
is_MiscName(_)  -> false.
 
is_NameChar(X) ->
    case is_Letter(X) of
        true  -> true;
        false -> case is_Digit(X) of
                     true  -> true;
                     false -> is_MiscName(X)
                 end
    end.                            

is_Blank($ )  -> true;
is_Blank($\n) -> true;
is_Blank($\t) -> true;
is_Blank($\r) -> true;
is_Blank(_)   -> false.


xml2raw([$& | Cs]) ->
    {Cs1,Amp} = collect_amp(Cs,[]),
    [translate_amp(Amp) | xml2raw(Cs1)];
xml2raw([C | Cs]) ->
    [C | xml2raw(Cs)];
xml2raw([]) ->
    [].

collect_amp([$ | T], L)   -> {T, reverse(L)};
collect_amp([$\n | T], L) -> {T, reverse(L)};
collect_amp([$\r | T], L) -> {T, reverse(L)};
collect_amp([$; | T], L)  -> {T, reverse(L)};
collect_amp([H | T], L)   -> collect_amp(T, [H|L]);
collect_amp([], L)        -> {[], reverse(L)}.

translate_amp([$# | Ds]) ->
    amp_digits(Ds, 0);
translate_amp(Name) ->
    case Name of
        "lt" -> $<;
        "gt" -> $>;
        "amp" -> $&;
        "quot" -> $";
        "nbsp" -> 160;
        "iexcl" -> 161;
        "cent" -> 162;
        "pound" -> 163;
        "curren" -> 164;
        "yen" -> 165;
        "brvbar" -> 166;
        "sect" -> 167;
        "uml" -> 168;
        "copy" -> 169;
        "ordf" -> 170;
        "laquo" -> 171;
        "not" -> 172;
        "shy" -> 173;
        "reg" -> 174;
        "macr" -> 175;
        "deg" -> 176;
        "plusmn" -> 177;
        "sup2" -> 178;
        "sup3" -> 179;
        "acute" -> 180;
        "micro" -> 181;
        "para" -> 182;
        "middot" -> 183;
        "cedil" -> 184;
        "sup1" -> 185;
        "ordm" -> 186;
        "raquo" -> 187;
        "frac14" -> 188;
        "frac12" -> 189;
        "frac34" -> 190;
        "iquest" -> 191;
        "Agrave" -> 192;
        "Aacute" -> 193;
        "Acirc" -> 194;
        "Atilde" -> 195;
        "Auml" -> 196;
        "Aring" -> 197;
        "AElig" -> 198;
        "Ccedil" -> 199;
        "Egrave" -> 200;
        "Eacute" -> 201;
        "Ecirc" -> 202;
        "Euml" -> 203;
        "Igrave" -> 204;
        "Iacute" -> 205;
        "Icirc" -> 206;
        "Iuml" -> 207;
        "ETH" -> 208;
        "Ntilde" -> 209;
        "Ograve" -> 210;
        "Oacute" -> 211;
        "Ocirc"-> 212;
        "Otilde" -> 213;
        "Ouml" -> 214;
        "times" -> 215;
        "Oslash" -> 216;
        "Ugrave" -> 217;
        "Uacute" -> 218;
        "Ucirc" -> 219;
        "Uuml" -> 220;
        "Yacute" -> 221;
        "THORN" -> 222;
        "szlig" -> 223;
        "agrave" -> 224;
        "aacute" -> 225;
        "acirc" -> 226;
        "atilde" -> 227;
        "auml" -> 228;
        "aring" -> 229;
        "aelig" -> 230;
        "ccedil" -> 231;
        "egrave" -> 232;
        "eacute" -> 233;
        "ecirc" -> 234;
        "euml" -> 235;
        "igrave" -> 236;
        "iacute" -> 237;
        "icirc" -> 238;
        "iuml" -> 239;
        "eth" -> 240;
        "ntilde" -> 241;
        "ograve" -> 242;
        "oacute" -> 243;
        "ocirc" -> 244;
        "otilde" -> 245;
        "ouml" -> 246;
        "divide" -> 247;
        "oslash" -> 248;
        "ugrave" -> 249;
        "uacute" -> 250;
        "ucirc" -> 251;
        "uuml" -> 252;
        "yacute" -> 253;
        "thorn" -> 254;
        "yuml" -> 255;
        _ -> error
    end.

amp_digits([X | Xs], N) when X >= $0, X =< $9 ->
    amp_digits(Xs, N*10 + (X-$0));
amp_digits([], N) ->
    if
        N >= 0, N =< 8 -> error;
        N >= 127, N =< 159 -> error;
        N > 255 -> error;
        true -> N
    end.
