%%% File    : glade_url.erl
%%% Author  : Tony Rogvall <tony@bit.hemma.se>
%%% Purpose : Base http functionality
%%% Created : 12 Oct 2000 by Tony Rogvall <tony@bit.hemma.se>

-module(glade_url).
-author('tony@bit.hemma.se').

-export([parse/1, format/1]).

-import(lists, [reverse/1]).
%%
%% Parse url or href
%%
%% return:
%%       {absoluteURI,http|https,Host,Port|undefined,Path}
%%       {abs_path, Path}
%%       {rel_path, Path}
%%       {scheme, Scheme, Data}
%%       Data
%%
parse(Str) ->
    case strstr(Str, ":") of
	{"http", "//" ++ Str1} ->
	    parse_abs_uri(Str1, http);
	{"https", "//" ++ Str1} ->
	    parse_abs_uri(Str1, https);
	{Scheme, Data} ->
	    {scheme, Scheme, Data};
	[$/|Ps] ->
	    {abs_path, Str};
	Path ->
	    {rel_path, Path}
    end.

parse_abs_uri(Str, Scheme) ->    
    case strchr(Str, $/) of
	{HostPort, Cs1} ->
	    {Path,Cs2} = parse_path([$/|Cs1]),
	    parse_url_port(HostPort, Scheme, Path, Cs2);
	HostPort ->
	    parse_url_port(HostPort, Scheme, "/", undefined)
    end.
	    
parse_url_port(Str, Scheme, Path, Tail) ->
    case domain(Str) of
	{Domain, [$: | Cs]} when Domain =/= [] ->
	    case alphanum(Cs) of
		{"",_} ->
		    {error, bad_url};
		{Port,Cs1} ->
		    P = case catch list_to_integer(Port) of
			    N when integer(N) -> N;
			    _ -> list_to_atom(Port)
			end,
		    if Tail == undefined ->
			    {absoluteURI,Scheme,Domain,P,Path};
		       Cs1 =/= [] ->
			    {error, bad_url};
		       true ->
			    {absoluteURI,Scheme,Domain,P,Path}
		    end
	    end;
	{Domain, Cs1}  when Domain =/= [] ->
	    if Tail == undefined ->
		    {absoluteURI,Scheme,Domain,undefined,Path};
	       Cs1 =/= [] ->
		    {error, bad_url};
	       true ->
		    {absoluteURI,Scheme,Domain,undefined,Path}
	    end;
	_ ->
	    {error, bad_url}
    end.

%% split a string at SubStr return {Before, After}
%% or Str if not found
strstr(Str, SubStr) ->
    strstr(Str, SubStr, [], Str, SubStr).

strstr([C|Cs1], [C|Cs2], Head, Tail, SubStr) -> 
    strstr(Cs1, Cs2, Head, Tail, SubStr);
strstr(Cs1, [], Head, Tail, SubStr) ->
    {reverse(Head), Cs1};
strstr([], SubStr, Head, Tail, SubStr) ->  
    reverse(Head) ++ Tail;
strstr(_, _, Head, Tail, SubStr) ->
    NTail = tl(Tail),
    strstr(NTail, SubStr, [hd(Tail)|Head], NTail, SubStr).

strchr(Str, Char) ->
    strchr(Str, Char, []).

strchr([C|Cs1], C, Head) ->
    {reverse(Head), Cs1};
strchr([C|Cs1], Char, Head) ->
    strchr(Cs1, Char, [C|Head]);
strchr([], Char, Head) ->
    reverse(Head).


%% Get alphanumeric characters
alphanum(Str) ->
    alphanum(Str, []).

alphanum([C|Cs], Acc) when C >= $0, C =< $9;
			   C >= $a, C =< $z;
			   C >= $A, C =< $Z ->
    alphanum(Cs, [C|Acc]);
alphanum(Cs, Acc) ->
    {reverse(Acc), Cs}.

%% Get domain name characters
domain(Str) ->
    domain(Str, []).

domain([C|Cs], Acc) when C >= $0, C =< $9;
			 C >= $a, C =< $z;
			 C >= $A, C =< $Z;
			 C == $.; C == $- ->
    domain(Cs, [C|Acc]);
domain(Cs, Acc) ->
    {reverse(Acc), Cs}.

parse_path(Str) ->
    parse_path(Str, []).

parse_path([C|Cs], Acc) when C == $\s; C == $\t; C == $\n; C == $\r ->
    {reverse(Acc), [C|Cs]};
parse_path([C|Cs], Acc) ->
    parse_path(Cs, [C|Acc]);
parse_path([], Acc) ->
    {reverse(Acc), []}.


format({absoluteURI, Scheme, Host, Port, Path}) ->
    [ atom_to_list(Scheme),"://", Host, fmt_port(Port), Path];
format({abs_path, Path}) ->
    Path;
format({rel_path, Path}) ->
    Path;
format({scheme, Scheme, Data}) ->
    [ Scheme, $:, Data];
format(Uri) when list(Uri) ->
    Uri.
      
fmt_port(undefined) -> "";
fmt_port(Port) when integer(Port) -> [$: | integer_to_list(Port)];
fmt_port(Port) when atom(Port) -> [$: | atom_to_list(Port)].






