
module gnome_canvas;

#include <gnome.h>
#include "../gnome/gdrv_gnome_canvas.h"

typedef GtkObject GnomeCanvas;
typedef GtkObject GnomeCanvasItem;
typedef GtkObject GnomeCanvasGroup;
typedef GtkObject GnomeCanvasWidget;

typedef gpointer       GdkImlibImage;
typedef gpointer       GnomeCanvasPoints;


typedef GtkObject GnomeCanvasLine;


/* Line item for the canvas.  This is a polyline with configurable width, cap/join styles, and arrowheads.
 * If arrowheads are enabled, then three values are used to specify their shape:
 *
 *      arrow_shape_a:  Distance from tip of arrowhead to the center point.
 *      arrow_shape_b:  Distance from tip of arrowhead to trailing point, measured along the shaft.
 *      arrow_shape_c:  Distance of trailing point from outside edge of shaft.
 *
 * The following object arguments are available:
 *
 * name                 type                    read/write      description
 * ------------------------------------------------------------------------------------------
 * points               GnomeCanvasPoints*      RW              Pointer to a GnomeCanvasPoints structure.
 *                                                              This can be created by a call to
 *                                                              gnome_canvas_points_new() (in gnome-canvas-util.h). 
 *                                                              X coordinates are in the even indices of the
 *                                                              points->coords array, Y coordinates are in
 *                                                              the odd indices.
 * fill_color           string                  W               X color specification for line
 * fill_color_gdk       GdkColor*               RW              Pointer to an allocated GdkColor
 * fill_stipple         GdkBitmap*              RW              Stipple pattern for the line
 * width_pixels         uint                    R               Width of the line in pixels.  The line width
 *                                                              will not be scaled when the canvas zoom factor changes.
 * width_units          double                  R               Width of the line in canvas units.  The line width
 *                                                              will be scaled when the canvas zoom factor changes. 
 * cap_style            GdkCapStyle             RW              Cap ("endpoint") style for the line.
 * join_style           GdkJoinStyle            RW              Join ("vertex") style for the line.
 * line_style           GdkLineStyle            RW              Line dash style
 * first_arrowhead      boolean                 RW              Specifies whether to draw an arrowhead on the
 *                                                              first point of the line.
 * last_arrowhead       boolean                 RW              Specifies whether to draw an arrowhead on the
 *                                                              last point of the line.
 * smooth               boolean                 RW              Specifies whether to smooth the line using
 *                                                              parabolic splines.
 * spline_steps         uint                    RW              Specifies the number of steps to use when rendering curves.
 * arrow_shape_a        double                  RW              First arrow shape specifier.
 * arrow_shape_b        double                  RW              Second arrow shape specifier.
 * arrow_shape_c        double                  RW              Third arrow shape specifier.
 */

typedef GtkObject GnomeCanvasPolygon;

/* Polygon item for the canvas.  A polygon is a bit different from rectangles and ellipses in that
 * points inside it will always be considered "inside", even if the fill color is not set.  If you
 * want to have a hollow polygon, use a line item instead.
 *
 * The following object arguments are available:
 *
 * name                 type                    read/write      description
 * ------------------------------------------------------------------------------------------
 * points               GnomeCanvasPoints*      RW              Pointer to a GnomeCanvasPoints structure.
 *                                                              This can be created by a call to
 *                                                              gnome_canvas_points_new() (in gnome-canvas-util.h). 
 *                                                              X coordinates are in the even indices of the
 *                                                              points->coords array, Y coordinates are in
 *                                                              the odd indices.
 * fill_color           string                  W               X color specification for fill color,
 *                                                              or NULL pointer for no color (transparent).
 * fill_color_gdk       GdkColor*               RW              Allocated GdkColor for fill.
 * outline_color        string                  W               X color specification for outline color,
 *                                                              or NULL pointer for no color (transparent).
 * outline_color_gdk    GdkColor*               RW              Allocated GdkColor for outline.
 * fill_stipple         GdkBitmap*              RW              Stipple pattern for fill
 * outline_stipple      GdkBitmap*              RW              Stipple pattern for outline
 * width_pixels         uint                    RW              Width of the outline in pixels.  The outline will
 *                                                              not be scaled when the canvas zoom factor is changed.
 * width_units          double                  RW              Width of the outline in canvas units.  The outline
 *                                                              will be scaled when the canvas zoom factor is changed.
 */


typedef GtkObject GnomeCanvasRect;
typedef GtkObject GnomeCanvasEllipse;
typedef GtkObject GnomeCanvasRE;

/* Base class for rectangle and ellipse item types.  These are defined by their top-left and
 * bottom-right corners.  Rectangles and ellipses share the following arguments:
 *
 * name                 type            read/write      description
 * ------------------------------------------------------------------------------------------
 * x1                   double          RW              Leftmost coordinate of rectangle or ellipse
 * y1                   double          RW              Topmost coordinate of rectangle or ellipse
 * x2                   double          RW              Rightmost coordinate of rectangle or ellipse
 * y2                   double          RW              Bottommost coordinate of rectangle or ellipse
 * fill_color           string          W               X color specification for fill color,
 *                                                      or NULL pointer for no color (transparent)
 * fill_color_gdk       GdkColor*       RW              Allocated GdkColor for fill
 * outline_color        string          W               X color specification for outline color,
 *                                                      or NULL pointer for no color (transparent)
 * outline_color_gdk    GdkColor*       RW              Allocated GdkColor for outline
 * fill_stipple         GdkBitmap*      RW              Stipple pattern for fill
 * outline_stipple      GdkBitmap*      RW              Stipple pattern for outline
 * width_pixels         uint            RW              Width of the outline in pixels.  The outline will
 *                                                      not be scaled when the canvas zoom factor is changed.
 * width_units          double          RW              Width of the outline in canvas units.  The outline
 *                                                      will be scaled when the canvas zoom factor is changed.
 */




typedef GtkObject GnomeCanvasText;

/* Text item for the canvas.  Text items are positioned by an anchor point and an anchor direction.
 *
 * A clipping rectangle may be specified for the text.  The rectangle is anchored at the text's anchor
 * point, and is specified by clipping width and height parameters.  If the clipping rectangle is
 * enabled, it will clip the text.
 *
 * In addition, x and y offset values may be specified.  These specify an offset from the anchor
 * position.  If used in conjunction with the clipping rectangle, these could be used to implement
 * simple scrolling of the text within the clipping rectangle.
 *
 * The following object arguments are available:
 *
 * name                 type                    read/write      description
 * ------------------------------------------------------------------------------------------
 * text                 string                  RW              The string of the text label
 * x                    double                  RW              X coordinate of anchor point
 * y                    double                  RW              Y coordinate of anchor point
 * font                 string                  W               X logical font descriptor
 * fontset              string                  W               X logical fontset descriptor
 * font_gdk             GdkFont*                RW              Pointer to a GdkFont
 * anchor               GtkAnchorType           RW              Anchor side for the text
 * justification        GtkJustification        RW              Justification for multiline text
 * fill_color           string                  W               X color specification for text
 * fill_color_gdk       GdkColor*               RW              Pointer to an allocated GdkColor
 * fill_stipple         GdkBitmap*              RW              Stipple pattern for filling the text
 * clip_width           double                  RW              Width of clip rectangle
 * clip_height          double                  RW              Height of clip rectangle
 * clip                 boolean                 RW              Use clipping rectangle?
 * x_offset             double                  RW              Horizontal offset distance from anchor position
 * y_offset             double                  RW              Vertical offset distance from anchor position
 * text_width           double                  R               Used to query the width of the rendered text
 * text_height          double                  R               Used to query the rendered height of the text
 */

typedef GtkObject GnomeCanvasImage;

/* Image item for the canvas.  Images are positioned by anchoring them to a point.
 * The following arguments are available:
 *
 * name         type                    read/write      description
 * ------------------------------------------------------------------------------------------
 * image        GdkImlibImage*          RW              Pointer to a GdkImlibImage
 * pixbuf       ArtPixBuf*              W               Pointer to an ArtPixBuf (aa-mode)
 * x            double                  RW              X coordinate of anchor point
 * y            double                  RW              Y coordinate of anchor point
 * width        double                  RW              Width to scale image to, in canvas units
 * height       double                  RW              Height to scale image to, in canvas units
 * anchor       GtkAnchorType           RW              Anchor side for the image
 */



/* Create a canvas item using the standard Gtk argument mechanism.  The item is
 * automatically inserted at the top of the specified canvas group.  The last
 * argument must be a NULL pointer.
 */

/* GnomeCanvasItem *gnome_canvas_item_new (GnomeCanvasGroup *parent, GtkType type,
   const gchar *first_arg_name, ...); */

/* Same as above, with parsed args */

new GnomeCanvasItem *gnome_canvas_item_newv (GnomeCanvasGroup *parent, 
					     GtkType type,
					     guint nargs, GtkArg *args);

/* Constructors for use in derived classes and language wrappers */
/* void gnome_canvas_item_construct (GnomeCanvasItem *item, GnomeCanvasGroup *parent,
				  const gchar *first_arg_name, va_list args); */
/*
   void gnome_canvas_item_constructv (GnomeCanvasItem *item, 
                                      GnomeCanvasGroup *parent,
				      guint nargs, GtkArg *args);
*/

/* Configure an item using the standard Gtk argument mechanism.  The last
 * argument must be a NULL pointer.
 */

/* void gnome_canvas_item_set (GnomeCanvasItem *item, const gchar *first_arg_name, ...); */

/* Same as above, with parsed args */
void gnome_canvas_item_setv (GnomeCanvasItem *item, guint nargs, GtkArg *args);

/* Used only for language wrappers and the like */
/* void gnome_canvas_item_set_valist (GnomeCanvasItem *item,
				   const gchar *first_arg_name, va_list args); */

/* Move an item by the specified amount */
void gnome_canvas_item_move (GnomeCanvasItem *item, double dx, double dy);

/* Apply a relative affine transformation to the item. */
void gnome_canvas_item_affine_relative (GnomeCanvasItem *item, const double affine[6]);

/* Apply an absolute affine transformation to the item. */
void gnome_canvas_item_affine_absolute (GnomeCanvasItem *item, const double affine[6]);

/* Scale an item about a point by the specified factors */

/* new stuff ?
 void gnome_canvas_item_scale (GnomeCanvasItem *item,
			      double x, double y,
			      double scale_x, double scale_y);

void gnome_canvas_item_rotate (GnomeCanvasItem *item, double x, double y, double angle);

*/

/* Raise an item in the z-order of its parent group by the specified number of
 * positions.
 */
void gnome_canvas_item_raise (GnomeCanvasItem *item, int positions);

/* Lower an item in the z-order of its parent group by the specified number of
 * positions.
 */
void gnome_canvas_item_lower (GnomeCanvasItem *item, int positions);

/* Raise an item to the top of its parent group's z-order. */
void gnome_canvas_item_raise_to_top (GnomeCanvasItem *item);

/* Lower an item to the bottom of its parent group's z-order */
void gnome_canvas_item_lower_to_bottom (GnomeCanvasItem *item);

/* Show an item (make it visible).  If the item is already shown, it has no
 * effect.
 */
void gnome_canvas_item_show (GnomeCanvasItem *item);

/* Hide an item (make it invisible).  If the item is already invisible, it has
 * no effect.
 */
void gnome_canvas_item_hide (GnomeCanvasItem *item);

/* Grab the mouse for the specified item.  Only the events in event_mask will be
 * reported.  If cursor is non-NULL, it will be used during the duration of the
 * grab.  Time is a proper X event time parameter.  Returns the same values as
 * XGrabPointer().
 */
int gnome_canvas_item_grab (GnomeCanvasItem *item, unsigned int event_mask,
			    GdkCursor *cursor, guint32 etime);

/* Ungrabs the mouse -- the specified item must be the same that was passed to
 * gnome_canvas_item_grab().  Time is a proper X event time parameter.
 */
void gnome_canvas_item_ungrab (GnomeCanvasItem *item, guint32 etime);

/* These functions convert from a coordinate system to another.  "w" is world
 * coordinates and "i" is item coordinates.
 */
void gnome_canvas_item_w2i (GnomeCanvasItem *item, double *x, double *y);
void gnome_canvas_item_i2w (GnomeCanvasItem *item, double *x, double *y);

/* Gets the affine transform that converts from item-relative coordinates to
 * world coordinates.
 */
void gnome_canvas_item_i2w_affine (GnomeCanvasItem *item, double affine[6]);

/* Gets the affine transform that converts from item-relative coordinates to
 * canvas pixel coordinates.
 */
void gnome_canvas_item_i2c_affine (GnomeCanvasItem *item, double affine[6]);

/* Remove the item from its parent group and make the new group its parent.  The
 * item will be put on top of all the items in the new group.  The item's
 * coordinates relative to its new parent to *not* change -- this means that the
 * item could potentially move on the screen.
 * 
 * The item and the group must be in the same canvas.  An item cannot be
 * reparented to a group that is the item itself or that is an inferior of the
 * item.
 */
void gnome_canvas_item_reparent (GnomeCanvasItem *item, GnomeCanvasGroup *new_group);

/* Used to send all of the keystroke events to a specific item as well as
 * GDK_FOCUS_CHANGE events.
 */
void gnome_canvas_item_grab_focus (GnomeCanvasItem *item);

/* Fetch the bounding box of the item.  The bounding box may not be exactly
 * tight, but the canvas items will do the best they can.  The returned bounding
 * box is in the coordinate system of the item's parent.
 */
void gnome_canvas_item_get_bounds (GnomeCanvasItem *item,
				   out double x1, 
				   out double y1, 
				   out double x2, 
				   out double y2);

/* Request that the update method eventually get called.  This should be used
 * only by item implementations.
 */
void gnome_canvas_item_request_update (GnomeCanvasItem *item);

/* Creates a new canvas.  You should check that the canvas is created with the
 * proper visual and colormap.  Any visual will do unless you intend to insert
 * gdk_imlib images into it, in which case you should use the gdk_imlib visual.
 *
 * You should call gnome_canvas_set_scroll_region() soon after calling this
 * function to set the desired scrolling limits for the canvas.
 */
GtkWidget *gnome_canvas_new (void);

/* Returns the root canvas item group of the canvas */
GnomeCanvasGroup *gnome_canvas_root (GnomeCanvas *canvas);

/* Sets the limits of the scrolling region, in world coordinates */
void gnome_canvas_set_scroll_region (GnomeCanvas *canvas,
				     double x1, double y1, double x2, double y2);

/* Gets the limits of the scrolling region, in world coordinates */
void gnome_canvas_get_scroll_region (GnomeCanvas *canvas,
				     out double x1, 
				     out double y1, 
				     out double x2, 
				     out double y2);

/* Sets the number of pixels that correspond to one unit in world coordinates */
void gnome_canvas_set_pixels_per_unit (GnomeCanvas *canvas, double n);

/* Scrolls the canvas to the specified offsets, given in canvas pixel coordinates */
void gnome_canvas_scroll_to (GnomeCanvas *canvas, int cx, int cy);

/* Returns the scroll offsets of the canvas in canvas pixel coordinates.  You
 * can specify NULL for any of the values, in which case that value will not be
 * queried.
 */
void gnome_canvas_get_scroll_offsets (GnomeCanvas *canvas, 
				      out int cx, out int cy);

/* Requests that the canvas be repainted immediately instead of in the idle
 * loop.
 */
void gnome_canvas_update_now (GnomeCanvas *canvas);

/* Returns the item that is at the specified position in world coordinates, or
 * NULL if no item is there.
 */
GnomeCanvasItem *gnome_canvas_get_item_at (GnomeCanvas *canvas, double x, double y);

/* For use only by item type implementations. Request that the canvas eventually
 * redraw the specified region. The region is specified as a microtile
 * array. This function takes over responsibility for freeing the uta argument.
 */
/* void gnome_canvas_request_redraw_uta (GnomeCanvas *canvas, ArtUta *uta); */

/* For use only by item type implementations.  Request that the canvas
 * eventually redraw the specified region, specified in canvas pixel
 * coordinates.  The region contains (x1, y1) but not (x2, y2).
 */
void gnome_canvas_request_redraw (GnomeCanvas *canvas, int x1, int y1, int x2, int y2);

/* Gets the affine transform that converts world coordinates into canvas pixel
 * coordinates.
 */
void gnome_canvas_w2c_affine (GnomeCanvas *canvas, double affine[6]);

/* These functions convert from a coordinate system to another.  "w" is world
 * coordinates, "c" is canvas pixel coordinates (pixel coordinates that are
 * (0,0) for the upper-left scrolling limit and something else for the
 * lower-left scrolling limit).
 */
void gnome_canvas_w2c (GnomeCanvas *canvas, double wx, double wy, int *cx, int *cy);
void gnome_canvas_w2c_d (GnomeCanvas *canvas, double wx, double wy, double *cx, double *cy);
void gnome_canvas_c2w (GnomeCanvas *canvas, int cx, int cy, double *wx, double *wy);

/* This function takes in coordinates relative to the GTK_LAYOUT
 * (canvas)->bin_window and converts them to world coordinates.
 */
void gnome_canvas_window_to_world (GnomeCanvas *canvas,
				   double winx, double winy, double *worldx, double *worldy);

/* This is the inverse of gnome_canvas_window_to_world() */
void gnome_canvas_world_to_window (GnomeCanvas *canvas,
				   double worldx, double worldy, double *winx, double *winy);

/* Takes a string specification for a color and allocates it into the specified
 * GdkColor.  If the string is null, then it returns FALSE. Otherwise, it
 * returns TRUE.
 */
int gnome_canvas_get_color (GnomeCanvas *canvas, const char *spec, out GdkColor *color);

/* Allocates a color from the RGB value passed into this function. */
gulong gnome_canvas_get_color_pixel (GnomeCanvas *canvas,
				     guint        rgba);
     

/* Sets the stipple origin of the specified gc so that it will be aligned with
 * all the stipples used in the specified canvas.  This is intended for use only
 * by canvas item implementations.
 */
void gnome_canvas_set_stipple_origin (GnomeCanvas *canvas, GdkGC *gc);

/* Controls the dithering used when the canvase renders.
 * Only applicable to antialiased canvases - ignored by non-antialiased canvases.
 */
void gnome_canvas_set_dither (GnomeCanvas *canvas, GdkRgbDither dither);

/* Returns the dither mode of an antialiased canvas.
 * Only applicable to antialiased canvases - ignored by non-antialiased canvases.
 */
GdkRgbDither gnome_canvas_get_dither (GnomeCanvas *canvas);

/* Note that it will only loads png files */
GdkImlibImage *gnome_canvas_load_alpha (const gchar *file);
void gnome_canvas_destroy_image (GdkImlibImage *image);

/* Allocate a new GnomeCanvasPoints structure with enough space for the specified number of points */
GnomeCanvasPoints *gnome_canvas_points_new (int num_points);

/* Increate ref count */
GnomeCanvasPoints *gnome_canvas_points_ref (GnomeCanvasPoints *points);
/* #define gnome_canvas_points_unref gnome_canvas_points_free   (defined in gnome_canvas.erl) */

/* Decrease ref count and free structure if it has reached zero */
void gnome_canvas_points_free (GnomeCanvasPoints *points);

/* Given three points forming an angle, compute the coordinates of the inside and outside points of
 * the mitered corner formed by a line of a given width at that angle.
 *
 * If the angle is less than 11 degrees, then FALSE is returned and the return points are not
 * modified.  Otherwise, TRUE is returned.
 */
int gnome_canvas_get_miter_points (double x1, double y1, double x2, double y2, double x3, double y3,
				   double width,
				   out double mx1, out double my1, out double mx2, out double my2);

/* Compute the butt points of a line segment.  If project is FALSE, then the results are as follows:
 *
 *            -------------------* (bx1, by1)
 *                               |
 *   (x1, y1) *------------------* (x2, y2)
 *                               |
 *            -------------------* (bx2, by2)
 *
 * that is, the line is not projected beyond (x2, y2).  If project is TRUE, then the results are as
 * follows:
 *
 *            -------------------* (bx1, by1)
 *                      (x2, y2) |
 *   (x1, y1) *-------------*    |
 *                               |
 *            -------------------* (bx2, by2)
 */
void gnome_canvas_get_butt_points (double x1, double y1, double x2, double y2,
				   double width, int project,
				   out double bx1, out double by1, out double bx2, out double by2);

/* Calculate the distance from a polygon to a point.  The polygon's X coordinates are in the even
 * indices of the poly array, and the Y coordinates are in the odd indices.
 */
double gnome_canvas_polygon_to_point (double *poly, int num_points, double x, double y);

/* Some wrappers needed to access GnomeCanvasPoints */
void gnome_canvas_set_point(GnomeCanvasPoints* points, int index, 
			    double x, double y)
    = gdrv_gnome_canvas_set_point;

void gnome_canvas_get_point(GnomeCanvasPoints* points, int index, 
			    out double x, out double y)
    = gdrv_gnome_canvas_get_point;
