%%% File    : gtk.erl
%%% Author  :  <tony@bit.hemma.se>
%%% Purpose : Gtk interface
%%% Created : 18 Nov 1999 by  <tony@bit.hemma.se>

-module(gtk).
-author('tony@bit.hemma.se').

-compile(export_all).
%%-export([Function/Arity, ...]).

-include("../include/gtk.hrl").
-include("../include/gdk.hrl").

-define(E_GTK_TYPE_ENUM_FIND_VALUE,  16#0002).
-define(E_GTK_TYPE_FLAGS_FIND_VALUE, 16#0003).
-define(E_GTK_ARG_GET_TYPE,          16#0004).
-define(E_GTK_SIGNAL_FROM_NAME,      16#0008).
-define(E_GTK_SIGNAL_EMIT,           16#0009).
-define(E_GTK_SIGNAL_CONNECT,        16#000A).
-define(E_GTK_CHILD_ARG_GET_TYPE,    16#000D).
-define(E_GTK_ASYNC_SIGNAL_EMIT,     16#000E).
-define(E_GTK_REPLY,                 16#000F).
-define(E_GTK_FUNCTION_CALL,         16#0010).
-define(E_GTK_NEW_CALL,              16#0012).
-define(E_GTK_ASYNC_CALL,            16#0013).
-define(E_GTK_MODULE_DIR,            16#0014).
-define(E_GTK_LOAD_MODULE,           16#0015).
-define(E_GTK_UNLOAD_MODULE,         16#0016).
-define(E_GTK_FIND_FUNCTION,         16#0017).

-define(APPNAME, erlgtk).

-ifdef(erl_fun_ext).
-define(ENCODE_FUN(F), (F)).
-define(DECODE_FUN(F), (F)).
-else.
-define(ENCODE_FUN(F), 
	if integer((F)) -> (F);
	   true -> term_to_binary((F))
	end).
-define(DECODE_FUN(F), binary_to_term((F))).
-endif.

-ifdef(erl_fun_ari).
-define(HANDLE_ARGS(F, As), handle_args((F), (As))).
-else.
-define(HANDLE_ARGS(F, As), (As)).
-endif.



-ifdef(debug).
-define(DEBUG(Fmt, As), io:format((Fmt),(As))).
-else.
-define(DEBUG(Fmt, As), ok).
-endif.

-define(GC_TIMEOUT, 100).	       

-record(state,
	{
	  types,      %% type (atom) to typeid (integer) map 
	              %% 'GtkWindow' -> gtk_window_get_type()
	  enums,      %% enum name (atom) to enumvalue (integer)
	  quarks,     %% Function quark table
	  port,       %% pipe or socket to external process
	  %% Data send function
	  sendf = fun(Data) -> ok end,
	  %% port stop function
	  stopf = fun() -> ok end,
	  callid = 1, %% callid number
	  reply = [],  %% list of replies wanted {CallId, From}
	  gc_list = []
	 }).

start() ->
    start([socket]).


start(Opts) ->
    Tag = make_ref(),
    spawn(?MODULE, init, [[self()|Tag], Opts]),
    receive
	{Tag, Reply} -> 
	    Reply
    end.

stop() ->
    call(stop).

call(Call) ->
    Tag = make_ref(),
    gtk ! {call, Call, [self()|Tag]},
    call_reply(Tag).

call_reply(Tag) ->
    receive
	{signal, RID, Fun, Args} ->
	    handle_signal(RID,Fun,Args),
	    call_reply(Tag);
	{timeout, Ref, Info} ->
	    handle_timeout(Info),
	    call_reply(Tag);
	{Tag, Reply} -> 
	    Reply
    end.


cast(Cast) ->
    gtk ! {cast, Cast, [self()|undefined]}.


reply([Pid|Tag], Reply) ->
    Pid ! {Tag, Reply};
reply(none, Reply) ->  %% no reply from internal operation
    Reply.


init(From, Options) ->
    case whereis(gtk) of
	undefined ->
	    case catch register(gtk, self()) of
		true -> init1(From, Options);
		{'EXIT', _} -> 
		    gtk ! {sync_start, From}
	    end;
	Pid when pid(Pid) ->
	    gtk ! {sync_start, From}
    end.


init1(From, Options) ->
    process_flag(trap_exit, true),
    GtkTypeTab = ets:new(gtk_type,   [set,public,named_table]),
    GtkEnumTab = ets:new(gtk_enum,   [set,public,named_table]),
    GtkGlobal  = ets:new(gtk_global, [set,public,named_table]),
    GtkLink    = ets:new(gtk_link,   [bag,public,named_table]),
    GtkQuark   = ets:new(gtk_quark,  [set,public,named_table]),
    {SendF,StopF,Port}
	= case Options of
	      [port] ->
		  P = open_port({spawn, port_prog()},[{packet,4}, binary]),
		  { fun(Data) -> erlang:port_command(P, Data) end, 
		    fun() -> erlang:port_close(P) end,
		    P};
	      [socket] ->
		  {ok,L} = inet_tcp:listen(0, [{ifaddr,loopback},
					       {packet,4},
					       {nodelay,true},
					       {mode,binary}]),
		  {ok,Pn} = inet:port(L),
		  OsCmd = filename:nativename(port_prog()) ++ " " ++ 
		      integer_to_list(Pn),
		  P = open_port({spawn, OsCmd}, []),
		  unlink(P),
		  {ok,S} = inet_tcp:accept(L),
		  inet_tcp:close(L),
		  { fun(Data) -> inet_tcp:send(S, Data) end, 
		    fun() -> inet_tcp:close(S) end,
		    S }
	  end,
    set_module_dir(SendF, 0),
    ets:insert(gtk_global, {port, Port}),
    reply(From, {ok, started}),
    Res = (catch serv(#state { types = GtkTypeTab,
			       enums = GtkEnumTab,
			       port = Port,
			       sendf = SendF,
			       stopf = StopF
			      })),
    ?DEBUG("gtk server terminated: ~p\n", [Res]),
    Res.

port_prog() ->
    filename:join(bin_path(), "gtk_drv").

bin_path() ->
    case code:priv_dir(?APPNAME) of
        {error,bad_name} ->
            exit({error,bad_name});
        PrivDir ->
            filename:join([PrivDir, "bin"])
    end.    

module_path() ->
    case code:priv_dir(?APPNAME) of
        {error,bad_name} ->
            exit({error,bad_name});
        PrivDir ->
            filename:join([PrivDir, "modules"])
    end.

set_module_dir(SendF, ID) ->
    Cmd = term_to_binary({?E_GTK_MODULE_DIR,ID,module_path()}),
    SendF(Cmd),
    receive
	{tcp,S,Bin} ->
            case binary_to_term(Bin) of
                {ok,ID,ok} ->
                    ok;
                Error ->
                    set_module_dir_error(Error)
            end;
        {Port,{data,Bin}} ->
            case binary_to_term(Bin) of
                {ok,ID,ok} ->
                    ok;
                Error ->
                    set_module_dir_error(Error)
            end
    after 10000 ->
            Error = {error, timeout},
            set_module_dir_error(Error)
    end.


set_module_dir_error(Error) ->
    io:format("erlgtk: unable to set module dir ~p\n", [Error]),
    exit(Error).


serv(St) ->
    GcTimeout = if St#state.gc_list == [] -> infinity;
		   true ->
			?DEBUG("gc_list length = ~w\n", 
			    [length(St#state.gc_list)]),
			?GC_TIMEOUT
		end,
    receive
	{call, Request, From} ->
	    handle_call(Request, From,  St);

	{cast, Request, From} ->
	    handle_cast(Request, From,  St);

	{tcp,S,Bin} when S == St#state.port ->
	    srv_input(Bin, St);

	{Port,{data,Bin}} when Port == St#state.port ->
	    srv_input(Bin, St);

	{sync_start, From} ->
	    reply(From, {error, already_started}),
	    serv(St);

	{tcp_closed, S} when S == St#state.port ->
	    srv_died(dead, St);

	{'EXIT', Port, Reason} when Port == St#state.port ->
	    srv_died(dead, St);

	{'DOWN', Ref, process, Pid, Reason } ->
	    %% remove all objects that process owns
	    St1 = process_died(Pid, St),
	    serv(St1);

	Other ->
	    ?DEBUG("got: ~p\n", [Other]),
	    serv(St)
    after GcTimeout ->
	    serv_gc(St)
    end.

serv_gc(St) ->
    SendF = St#state.sendf,
    case St#state.gc_list of
	[{Type,Ptr} | List] ->
	    Cmd = {?E_GTK_ASYNC_CALL,0,
		   {gtk,object_unref,
		    [{'',Type,Ptr}]}},
	    SendF(term_to_binary(Cmd)),
	    serv(St#state { gc_list = List });
	[] ->
	    serv(St)
    end.

srv_input(Bin, St) ->
    Term =  binary_to_term(Bin),
    ?DEBUG("serv: ~p\n", [Term]),
    case Term of
	{ok,CallID,Value} ->
	    srv_reply(CallID,{ok,Value},St);

	{signal, RID, Pid, Fun, Args} ->
	    case ets:lookup(gtk_global, {links,Pid}) of
		[] ->
		    %% process has no object and can not respond
		    %% to this signal, we do it here
		    handle_cast({?E_GTK_REPLY,{RID,false}},none,St);
		_ ->
		    Pid ! {signal,RID,Fun,Args},
		    serv(St)
	    end;

	{created,CallID,IsWindow,Object} ->
	    srv_created(CallID,IsWindow,Object,St);

	{destroy, Object} ->
	    St1 = object_destroyed(Object, St),
	    serv(St1);

	{error,CallID,Reason} ->
	    srv_reply(CallID,{error,Reason},St);

	Other ->
	    ?DEBUG("got: port data: ~p\n", [Other]),
	    serv(St)
    end.

%%
%% An object was created link it to the process
%%
srv_created(CallID,IsWindow,Object,St) ->
    Reply = St#state.reply,
    case lists:keysearch(CallID,1,Reply) of
	{value,R = {_,From}} ->
	    link_object(From, IsWindow, Object), 
	    reply(From, {ok,Object}),
	    serv(St#state { reply = Reply -- [R] });
	false ->
	    %% FIXME: kill object
	    serv(St)
    end.
%%
%% gtk_drv has terminated lets die,
%% but first we must respond to all processes waiting for
%% a reply
%%
srv_died(Reason, St) ->
    lists:foreach(
      fun({CallId,From}) -> reply(From, {error,Reason}) end,
      St#state.reply).

%%
%% Reply to a request identified by CallID
%%
srv_reply(CallID,Value,St) ->
    Reply = St#state.reply,
    case lists:keysearch(CallID,1,Reply) of
	{value,R = {_,From}} ->
	    reply(From, Value),
	    serv(St#state { reply = Reply -- [R] });
	false ->
	    serv(St)
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% CALLS
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

handle_call({ID, Args}, From, St) when integer(ID) ->
    CallID = St#state.callid,
    ?DEBUG("gtk_call: (~p) ~p ~p\n", [CallID, ID, Args]),
    (St#state.sendf)(term_to_binary({ID,CallID,Args})),
    serv(St#state { callid = (CallID + 1) band 16#ffffff,
		    reply = [{CallID,From} | St#state.reply] });
handle_call(main, From=[Pid|_], St) ->
    %% keep track on main level
    case ets:lookup(gtk_global, {main,Pid}) of
	[] ->
	    ets:insert(gtk_global,{{main,Pid},1});
	_ ->
	    ets:update_counter(gtk_global, {main,Pid}, 1)
    end,
    reply(From, ok),
    serv(St);
handle_call(stop, From, St) ->
    reply(From, ok),
    srv_died(stopped, St);
handle_call(Request, From, St) ->
    reply(From, {error, badcall}),
    serv(St).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% CASTS
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
handle_cast({ID, Args}, From, St) when integer(ID) ->
    (St#state.sendf)(term_to_binary({ID,0,Args})),
    serv(St);
handle_cast(Request, From, St) ->
    serv(St).


link_object([Pid|_], IsWindow, Object) ->
    case ets:lookup(gtk_global, {links,Pid}) of
	[] ->
	    ets:insert(gtk_global, {{links,Pid},1}),
	    erlang:monitor(process, Pid);
	[{_,Count}] ->
	    ets:update_counter(gtk_global, {links,Pid}, 1),
	    ok
    end,
    ets:insert(gtk_link, {Pid,IsWindow,Object}),
    ets:insert(gtk_link, {Object,IsWindow,Pid}).


process_died(Pid, St) ->
    ?DEBUG("Process ~p died\n", [Pid]),
    ets:delete(gtk_global, {links,Pid}),
    ets:delete(gtk_global, {main,Pid}),
    SendF = St#state.sendf,
    ObjList = ets:lookup(gtk_link, Pid),
    GcList = lists:foldl(
	       fun({_Pid,true,{Type,Ptr}}, Gc) ->
		       Cmd = {?E_GTK_ASYNC_CALL,0,
			      {gtk,object_destroy,[{'',Type,Ptr}]}},
		       SendF(term_to_binary(Cmd)),
		       Gc;
		  ({_Pid,_,Obj}, Gc) ->
		       [Obj | Gc]
	       end, St#state.gc_list, ObjList),
    St#state { gc_list = GcList }.



object_destroyed(Object, St) ->
    case ets:lookup(gtk_link, Object) of
	[] ->
	    ?DEBUG("Object destroyed ~p - not found\n", [Object]),
	    St;
	[{Object,IsWindow,Pid}] ->
	    ?DEBUG("Object destroyed ~p - owner ~p\n", [Object,Pid]),
	    %% This catch is necessary since if the process died
	    %% then this slot is not here, but there is still cleaning
	    %% to do
	    catch ets:update_counter(gtk_global, {links,Pid}, -1),
	    ets:match_delete(gtk_link, {Pid,IsWindow,Object}),
	    ets:match_delete(gtk_link, {Object,IsWindow,Pid}),
	    GcList = St#state.gc_list -- [Object],
	    St#state { gc_list = GcList };
	ObjList ->
	    ?DEBUG("Error: object_destroyed object has multiple owner: "
		"~p\n", [ObjList]),
	    St
    end.


timeout_add(Timeout, Callback, Data) ->
    ID = case get(timer_id) of
	     undefined -> 0;
	     N -> N
	 end,
    put(timer_id, ID+1),
    Ref = erlang:start_timer(Timeout, self(), {ID,Timeout,Callback,Data}),
    put({timeout,ID}, Ref),
    ID.


timeout_remove(ID) ->
    case get({timeout,ID}) of
	undefined -> ok;
	Ref ->
	    case erlang:cancel_timer(Ref) of
		false ->
		    receive
			{timeout,Ref,_} ->
			    ok
		    after 0 ->
			    ok
		    end;
		_ ->
		    ok
	    end,
	    erase({timeout,ID})
    end.


input_add(Callback) ->
    ID = case get(input_id) of
	     undefined -> 0;
	     N  -> N
	 end,
    put(input_id, ID+1),
    put({input_callback,ID}, Callback),
    case get(input_handler) of
	undefined ->
	    put(input_handler, [ID]);
	Hs ->
	    put(input_handler, Hs ++ [ID])
    end,
    ID.

input_remove(ID) ->
    case get({input_callback,ID}) of
	undefined -> ok;
	_ ->
	    erase({input_callback,ID}),
	    case get(input_handler) of
		undefined -> ok;
		Hs ->
		    put(input_handler, Hs -- [ID]),
		    ok
	    end
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% MAIN
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

main() ->
    main([]).

main(St) ->
    call(main), %% register this main loop with server
    put('_state', St),  %% FIXME
    main_loop(St).

main_loop(St) ->
    receive
	{signal, RID, Fun, Args} ->
	    put('_state', St),  %% FIXME
	    handle_signal(RID,Fun,Args),
	    main_loop(St);
	{timeout, Ref, Info} ->
	    put('_state', St),  %% FIXME
	    handle_timeout(Info),
	    main_loop(St);
	quit ->
	    ets:update_counter(gtk_global, {main,self()}, -1),
	    ok;
	Message ->
	    main_loop(handle_input(Message, St))
    end.


handle_input(Message, St) ->
    handle_input(get(input_handler), Message, St).

handle_input([H | Hs], Message, St) ->
    case get({input_callback,H}) of
	undefined -> handle_input(Hs, Message, St);
	Cb ->
	    case (catch apply(Cb, [Message, St])) of
		true ->
		    St;
		{true,St1} ->
		    St1;
		false -> 
		    handle_input(Hs, Message, St);
		{false,St1} ->
		    handle_input(Hs, Message, St1);
		{'EXIT', Reason} ->
		    io:format("input handler ~w (~p) crashed: ~p\n", 
			      [H, Cb, Reason]), 
		    handle_input(Hs, Message, St);
		Val ->
		    io:format("input handler ~w (~p) bad return value: ~p\n",
			      [H, Cb, Val]),
		    handle_input(Hs, Message, St)
	    end
    end;
handle_input([], Message, St) ->
    St;
handle_input(_, Message, St) ->
    St.
		

handle_signal(RID, Fun0, Args0) ->
    Fun = ?DECODE_FUN(Fun0),
    Args = ?HANDLE_ARGS(Fun, Args0),
    Ret = (catch apply(Fun, Args)),
    case Ret of
	true ->
	    cast({?E_GTK_REPLY,{RID,true}});
	{'EXIT',Reason} ->
	    io:format("signal handler ~w (~p) crashed: ~p\n", 
		      [RID, Fun, Reason]),
	    cast({?E_GTK_REPLY,{RID,false}});
	X when integer(X) ->
	    cast({?E_GTK_REPLY,{RID,X}});	    
	_ ->
	    cast({?E_GTK_REPLY,{RID,false}})
    end.

%% when fun_info arity (R8) is defined we may use this info to adjust the
%% argument list to fit the callback, i.e remove unwanted argument
%% or fill in extra (probably buggy arguments)
handle_args({M,F}, Args) -> Args;
handle_args(F, Args) ->
    {arity,Arity} = erlang:fun_info(F, arity),
    ?DEBUG("Fun=~p, Arity=~w\n", [F, Arity]),
    adjust_args(Arity, Args).

adjust_args(0, _)      -> [];
adjust_args(I, [A|As]) -> [A   | adjust_args(I-1, As)];
adjust_args(I, [])     -> [ undefined | adjust_args(I-1,[])].

handle_timeout(Info = {ID,Timeout,Callback,Data}) ->
    case catch apply(Callback, [Data]) of
	true ->
	    Ref = erlang:start_timer(Timeout,self(),Info),
	    put({timeout,ID}, Ref);
	{'EXIT',Reason} ->
	    io:format("timeout handler ~w (~p) crashed: ~p\n", 
		      [ID, Callback, Reason]),
	    erase({timeout,ID});	    
	_ ->
	    erase({timeout,ID})
    end.


quit() ->
    self() ! quit.

main_quit() ->
    self() ! quit,
    false.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% VERSION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
version() ->
    {ok,Maj,Min,Mic} = x_version(),
    {Maj,Min,Mic}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% OBJECT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% object_new(Class, Args) ->
%%    object_newv(Class, length(Args), enc_set_args({Class,0},Args)).

%% object_set(Object, Args) ->
%%    object_setv(Object, length(Args), enc_set_args(Object,Args)).

%% object_get(Object, Args) ->
%%    object_getv(Object, length(Args), enc_get_args(Object, Args)).

object_set_user_data(Object, Data) ->
    x_object_set_user_data(Object,Data).

object_get_user_data(Object) ->
    x_object_get_user_data(Object).

object_set_data(Object, Key, Data) ->
    x_object_set_data(Object, Key, Data).

object_get_data(Object, Key) ->
    x_object_get_data(Object, Key).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WIDGET
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% widget_new(Class, Args) ->
%%    widget_newv(Class, length(Args), enc_set_args({Class,0},Args)).

%% widget_set(Object, Args) ->
%%    widget_setv(Object, length(Args), enc_set_args(Object,Args)).

%% widget_get(Object, Args) ->
%%    widget_getv(Object, length(Args), enc_get_args(Object, Args)).


signal_emit(Object, Signal) ->
    signal_emit(Object, Signal, []).

signal_emit({_,Object}, Signal, Args) when atom(Signal) ->
    ?DEBUG("signal_emit: ~p, ~p ~p\n", [Object,Signal,Args]),
    call({?E_GTK_SIGNAL_EMIT, {Object, Signal, 
					enc_signal_args(Args)}}).

signal_connect(Widget, Name, Fun) ->
    signal_connect(Widget, Name, Fun, []).

signal_connect({_,Object}, Name, Fun, Data) when atom(Name) ->
    ?DEBUG("signal_connect: ~p ~s ~p ~p\n", [Object, Name, Fun, Data]),
    call({?E_GTK_SIGNAL_CONNECT, {Object,Name,false,false,
					   ?ENCODE_FUN(Fun),
					   self(),Data}}).  

signal_connect_after(Widget, Name, Fun) ->
    signal_connect_after(Widget, Name, Fun, []).

signal_connect_after({_,Object}, Name, Fun, Data) when atom(Name) ->
    ?DEBUG("signal_connect: ~p ~s ~p ~p\n", [Object, Name, Fun, Data]),    
    call({?E_GTK_SIGNAL_CONNECT, {Object,Name,true,false,
					   ?ENCODE_FUN(Fun),
					   self(),Data}}).


signal_connect_object(Widget, Name, Fun) ->
    signal_connect_object(Widget, Name, Fun, []).    

signal_connect_object({_,Object}, Name, Fun, Data) when atom(Name) ->
    ?DEBUG("signal_connect_object: ~p ~s ~p ~p\n",
	   [Object, Name, Fun, Data]),    
    call({?E_GTK_SIGNAL_CONNECT,{Object,Name,false,true,
				 ?ENCODE_FUN(Fun),
				 self(),Data}}).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  CONTAINER
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
container_children(Container) ->
    x_container_children(Container).

container_foreach(Container, Fun) ->
    lists:map(Fun, container_children(Container)).

%% container_child_set(Container, Child, Args) ->
%%    container_child_setv(Container, Child, length(Args), 
%%			 enc_child_set_args(Container, Args)).

%% container_child_get(Container, Child, Args) ->
%%    container_child_getv(Container, Child, length(Args),
%%			 enc_child_get_args(Container, Args)).
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  UTILS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

load(Module) ->
    ?DEBUG("load: ~s\n", [Module]),
    case call({?E_GTK_LOAD_MODULE, Module}) of
        {ok, Value} ->
            Value;
        Error ->
            Error
    end.

unload(Module) ->
    ?DEBUG("unload: ~s\n", [Module]),
    case call({?E_GTK_UNLOAD_MODULE, Module}) of
        {ok, Value} ->
            Value;
        Error ->
            Error
    end.

fcall(Mod, Fun, Args) ->
    ?DEBUG("fcall: ~s:~s ~p\n", [Mod, Fun, Args]),
    Quark = fquark(Mod, Fun),
    case call({?E_GTK_FUNCTION_CALL, {Quark,enc_signal_args(Args)}}) of
        {ok, Value} ->
            Value;
        {error, Reason} ->
            exit(Reason)
    end.

fnew(Mod, Fun, Args) ->
    ?DEBUG("fcall: ~s:~s ~p\n", [Mod, Fun, Args]),
    Quark = fquark(Mod, Fun),
    case call({?E_GTK_NEW_CALL, {Quark, enc_signal_args(Args)}}) of
        {ok, Value} ->
            Value;
        {error, Reason} ->
            exit(Reason)
    end.


fquark(Mod, Fun) ->
    case find_function(Mod, Fun) of
        undefined ->
            case load(Mod) of
                ok ->
                    case find_function(Mod, Fun) of
                        undefined ->
                            exit({undef,Mod,Fun});
                        Quark -> Quark
                    end;
                Error ->
                    exit({load,Error})
            end;
        Quark -> Quark
    end.


enc_signal_args([{Type, Value} | As]) ->
    [{'', Type, Value} | enc_signal_args(As)];
enc_signal_args([]) -> [].

%%
%% Widget argument access
%%
enc_set_args({ObjectType,_}, Args) ->
    list_to_tuple(e_set_args(ObjectType, Args)).

e_set_args(ObjectType, [{Name, Value} | As]) ->
    Type = arg_get_type(ObjectType,Name),
    [{Name,Type,Value} | e_set_args(ObjectType, As)];
e_set_args(_, []) -> [].

enc_get_args({ObjectType,_}, Args) ->
    list_to_tuple(e_get_args(ObjectType, Args)).

e_get_args(ObjectType, [ArgName | As]) ->
    Type = arg_get_type(ObjectType,ArgName),
    [{ArgName,Type} | e_get_args(ObjectType, As)];
e_get_args(_, []) -> [].

%%
%% Container child access
%%
enc_child_set_args({ObjectType,_}, Args) ->
    list_to_tuple(e_child_set_args(ObjectType, Args)).

e_child_set_args(ObjectType, [{Name, Value} | As]) ->
    Type = child_arg_get_type(ObjectType,Name),
    [{Name,Type,Value} | e_child_set_args(ObjectType, As)];
e_child_set_args(_, []) -> [].

enc_child_get_args({ObjectType,_}, Args) ->
    list_to_tuple(e_child_set_args(ObjectType, Args)).
    
e_child_get_args(ObjectType, [ArgName | As]) ->
    [{ArgName,0,0} | e_child_get_args(ObjectType, As)];
e_child_get_args(_, []) -> [].


type_number(Number) when integer(Number) ->
    Number;
type_number(Name) when atom(Name) ->
    case ets:lookup(gtk_type, Name) of
	[{_, ID}] -> ID;
	[] ->
	    case type_from_name(Name) of
		0  -> 
		    exit({badarg,Name});
		ID ->
		    ets:insert(gtk_type, {Name,ID}),
		    ID
	    end
    end.


arg_get_type(ObjectType, Arg) when integer(ObjectType), atom(Arg) ->
    Key = {ObjectType,Arg},
    case ets:lookup(gtk_type, Key) of
	[{_, Type}] -> Type;
	[] ->
	    case call({?E_GTK_ARG_GET_TYPE, {ObjectType,Arg}}) of
		{ok, Type} -> 
		    ets:insert(gtk_type, {Key,Type}),
		    Type;
		{error,Reason} ->
		    exit(Reason)
	    end
    end;
arg_get_type(ObjectType, Arg) when atom(ObjectType) ->
    arg_get_type(type_number(ObjectType), Arg).


child_arg_get_type(ObjectType, Arg) when integer(ObjectType), atom(Arg) ->
    Key = {child,ObjectType,Arg},
    case ets:lookup(gtk_type, Key) of
	[{_, Type}] -> Type;
	[] ->
	    case call({?E_GTK_CHILD_ARG_GET_TYPE,{ObjectType,Arg}}) of
		{ok, Type} -> 
		    ets:insert(gtk_type, {Key,Type}),
		    Type;
		{error,Reason} ->
		    exit(Reason)
	    end
    end;
child_arg_get_type(ObjectType, Arg) when atom(ObjectType) ->
    child_arg_get_type(type_number(ObjectType), Arg).
    

is_a(?GTK_NULL, Is_a_type) -> 
    false;
is_a({Type,_}, Is_a_type) when atom(Is_a_type) ->
    ID = type_number(Is_a_type),
    Key = {is_a, Type, ID},
    case ets:lookup(gtk_type, Key) of
	[{_, Bool}] -> Bool;
	[] ->
	    case type_is_a(Type,ID) of
		Bool ->
		    ets:insert(gtk_type,{Key,Bool}),
		    Bool;
		{error,Reason} ->
		    exit(Reason)
	    end
    end.


find_function(Mod, Fun) when atom(Mod), atom(Fun) ->
    Key = {Mod,Fun},
    case ets:lookup(gtk_quark, Key) of
        [{_,Quark}] ->
            Quark;
        [] ->
            case call({?E_GTK_FIND_FUNCTION, Key}) of
                {ok, Quark} ->
                    ets:insert(gtk_quark,{Key,Quark}),
                    Quark;
                Error ->
                    undefined
            end
    end.


enum(Type, Enum) when atom(Type), integer(Enum) ->
    ?DEBUG("enum ~s ~w\n", [Type, Enum]),
    Enum;
enum(Type, Enum) when atom(Type), atom(Enum) ->
    ?DEBUG("enum ~s ~s\n", [Type, Enum]),
    case ets:lookup(gtk_enum, {Type,Enum}) of
	[{_, Value}] -> Value;
	[] ->
	    case call({?E_GTK_TYPE_ENUM_FIND_VALUE,
		       {type_number(Type),Enum}}) of
		{ok, Value} ->
		    ets:insert(gtk_enum, {{Type,Enum}, Value}),
		    Value;
		{error, Reason} ->
		    exit(Reason)
	    end
    end.

flags(Type, Enum) when atom(Type), atom(Enum) ->
    case ets:lookup(gtk_enum, {Type,Enum}) of
	[{_, Value}] -> Value;
	[] ->
	    case call({?E_GTK_TYPE_FLAGS_FIND_VALUE,
		       {type_number(Type),Enum}}) of
		{ok, Value} ->
		    ets:insert(gtk_enum, {{Type,Enum}, Value}),
		    Value;
		{error, Reason} ->
		    exit(Reason)
	    end
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  DEBUG
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% debug util to dump all objects in link table
dump_links(Pid) ->
    L = ets:lookup(gtk_link, Pid),
    lists:foreach(
      fun ({_, _, Object = {ObjType,_}}) ->
	      io:format("~w -> ~s(~w)\n", [Pid, 
					   dump_type(ObjType), 
					   Object])
      end, L).

dump_tree(Object) ->
    dump_tree(Object, "").


dump_tree(Object={ObjType,_}, Indent) ->
    io:format("~s~s (~w)\n", [Indent, dump_type(ObjType), Object]),
    Cs = case type_is_a(Object, 'GtkContainer') of
	     true -> container_children(Object);
	     false -> []
	 end,
    lists:foreach(fun(Obj) -> dump_tree(Obj,"  "++Indent) end, Cs).

dump_type(ObjType) ->
    {ok,Name} = type_name(ObjType),
    Name.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  GENERATED CALLS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-include("../priv/src/gtk.inc").

