%%% File    : gtk_lib.erl
%%% Author  : Tony Rogvall <tony@bit.hemma.se>
%%% Description : Useful? GTK stuff
%%% Created : 10 Sep 2001 by Tony Rogvall <tony@bit.hemma.se>

-module(gtk_lib).

-export([menu_bar/1, menu_bar/2]).
-export([menu/1,     menu/2]).
-export([option_menu/1]).

-import(lists, [foreach/2, reverse/1]).

-include("../include/gtk.hrl").


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Menus
%%  syntax:
%%    Item = {Label, Type, Accel, Signals, SubItems}
%%         | {Label, Type, Signals}
%%         | {Label, Type}
%%         | {Label}
%%         | separator
%% 
%%   Type    = {radio,ID} | check | item
%%   Signals = [{SignalName, SignalFun}]
%%
%%   Items = [ Item* ]
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%
%% Create an option menu
%% option_menu(MenuItems) -> GtkOptionMenu
%%

option_menu(MenuItems) ->
    OptionMenu = gtk:option_menu_new(),
    Menu = gtk:menu_new(),
    {_Items,_GIDs} = menu(Menu, ?GTK_NULL, [], MenuItems),
    gtk:option_menu_set_menu(OptionMenu, Menu),
    gtk:widget_show(Menu),
    OptionMenu.

%%
%% create a menu
%%
%% menu(Items [,AccelGroup]) -> {GtkMenu, GtkAccelGroup}
%%
menu(Items) ->
    AccelGroup = gtk:accel_group_new(),
    menu(Items, AccelGroup).

menu(Items, AccelGroup) ->
    Menu = gtk:menu_new(),
    {_Items,_GIDs} = menu(Menu, AccelGroup, [], Items),
    {Menu,AccelGroup}.

%%
%% create a menu bar
%%
%% menubar(Items [,AccelGroup]) -> {GtkMenuBar, GtkAccelGroup}
%%
menu_bar(MenuItems) ->
    AccelGroup = gtk:accel_group_new(),
    menu_bar(MenuItems, AccelGroup).


menu_bar(MenuItems, AccelGroup) ->
    Menu = gtk:menu_bar_new(),
    gtk:menu_set_accel_group(Menu, AccelGroup),
    gtk:menu_bar_set_shadow_type(Menu, 'GTK_SHADOW_OUT'),
    {Items,GIDs} = menu(Menu, ?GTK_NULL, AccelGroup, MenuItems),
    {Menu, AccelGroup}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%  INTERNAL %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_group(GID, GIDs) ->
    case lists:keysearch(GID, 1, GIDs) of
	{value, {_,Group}} -> Group;
	false -> []
    end.

set_group(GID, Group, GIDs) ->
    case lists:keysearch(GID, 1, GIDs) of
	false -> [{GID,Group} | GIDs];
	_     -> lists:keyreplace(GID, 1, GIDs, {GID,Group})
    end.

%% Create menu items
menu_item(AccelGroup,GIDs,{Label,Type,Accel,Signals,Items}) ->
    %% create the item
    {Item,GIDs1} = 
	case Type of
	    {radio,GID} ->
		Group = get_group(GID, GIDs),
		RItem = gtk:radio_menu_item_new_with_label(Group, Label),
		{RItem, set_group(GID, [RItem|Group], GIDs)};
	    check ->
		{ gtk:check_menu_item_new_with_label(Label), GIDs};
	    item ->
		{ gtk:menu_item_new_with_label(Label), GIDs}
	end,

    %% install signals
    foreach(
      fun({SignalName, SignalFun}) ->
	      gtk:signal_connect(Item, SignalName, SignalFun)
      end, Signals),

    %% install accelerator
    if Accel == []             -> true;
       AccelGroup == ?GTK_NULL -> true;
       true ->
	    case gtk:accelerator_parse(Accel) of
		{ok,0,0} ->
		    io:format("unknown accelerator \"~s\" in menu ~s\n",
			      [Accel, Label]),
		    false;
		{ok,Key,Mod} ->
		    gtk:widget_add_accelerator(Item,
					       'activate', AccelGroup,
					       Key, Mod, ?GTK_ACCEL_VISIBLE)
	    end
    end,
    %% install submenu
    if Items == [] ->
	    {Item, GIDs1};
       true ->
	    SubMenu = gtk:menu_new(),
	    {_,GIDs2} = menu(SubMenu, AccelGroup, GIDs, Items),
	    gtk:menu_item_set_submenu(Item, SubMenu),
	    gtk:widget_show(SubMenu),
	    {Item, GIDs2}
    end;
menu_item(AccelGroup, GIDs, {Label, Type, Signals}) ->
    menu_item(AccelGroup, GIDs, {Label, Type, "", Signals, []});
menu_item(AccelGroup, GIDs, {Label, Type}) ->
    menu_item(AccelGroup, GIDs, {Label, Type, "",  [], []});    
menu_item(AccelGroup, GIDs, {Label}) ->
    menu_item(AccelGroup, GIDs, {Label, item, "",  [], []});
menu_item(AccelGroup, GIDs, separator) ->
    {gtk:menu_item_new(), GIDs}.


%% Create menu return list of menu items
menu(Menu, AccelGroup, GIDs, Items) ->
    menu(Menu, AccelGroup, GIDs, Items, []).

menu(Menu, AccelGroup, GIDs, [Item|Items], Acc) ->
    {MenuItem,GIDs1} = menu_item(AccelGroup, GIDs, Item),
    gtk:menu_append(Menu, MenuItem),
    gtk:widget_show(MenuItem),
    menu(Menu, AccelGroup, GIDs1, Items, [MenuItem|Acc]);
menu(Menu, AccelGroup, GIDs, [], Acc) ->
    {reverse(Acc), GIDs}.

